
/*
 * Copyright (C) 2002-2003 Stefan Holst
 * Copyright (C) 2004-2005 Maximilian Schwerin
 *
 * This file is part of oxine a free media player.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: disc.c 3927 2009-05-25 10:44:19Z mschwerin $
 *
 */
#include "config.h"

#include <assert.h>
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <stddef.h>
#include <stdio.h>
#include <mntent.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/types.h>

#ifdef HAVE_HAL
#include <libhal.h>
#include <libhal-storage.h>
#endif

#ifdef HAVE_LINUX_CDROM_H
#include <linux/cdrom.h>
#endif

#include "disc.h"
#include "environment.h"
#include "filelist.h"
#include "heap.h"
#include "i18n.h"
#include "logger.h"
#include "meta_info.h"
#include "mutex.h"
#include "oxine.h"
#include "utils.h"

extern oxine_t *oxine;


#ifdef HAVE_HAL
static char *
hal_drive_get_title (LibHalDrive * drive)
{
    char *title = NULL;

    LibHalDriveType type = libhal_drive_get_type (drive);

    const char *dtype = NULL;
    const char *vendor = libhal_drive_get_vendor (drive);
    if (vendor && (strlen (vendor) == 0)) {
        vendor = NULL;
    }
    const char *model = libhal_drive_get_model (drive);
    if (model && (strlen (model) == 0)) {
        model = NULL;
    }

    switch (type) {
    case LIBHAL_DRIVE_TYPE_REMOVABLE_DISK:
        dtype = _("External Hard Disk");
        break;
    case LIBHAL_DRIVE_TYPE_DISK:
        if (libhal_drive_is_hotpluggable (drive)) {
            dtype = _("External Hard Disk");
        }
        else {
            dtype = _("Hard Disk");
        }
        break;
    case LIBHAL_DRIVE_TYPE_CDROM:
        dtype = _("CDROM Drive");
        break;
    case LIBHAL_DRIVE_TYPE_COMPACT_FLASH:
        dtype = _("Compact Flash Drive");
        break;
    case LIBHAL_DRIVE_TYPE_MEMORY_STICK:
        dtype = _("Memory Stick Drive");
        break;
    case LIBHAL_DRIVE_TYPE_SMART_MEDIA:
        dtype = _("Smart Media Drive");
        break;
    case LIBHAL_DRIVE_TYPE_SD_MMC:
        dtype = _("SD/MMC Drive");
        break;
    case LIBHAL_DRIVE_TYPE_CAMERA:
        dtype = _("Digital Camera");
        break;
    case LIBHAL_DRIVE_TYPE_PORTABLE_AUDIO_PLAYER:
        dtype = _("Music Player");
        break;
    case LIBHAL_DRIVE_TYPE_ZIP:
        dtype = _("Zip Drive");
        break;
    case LIBHAL_DRIVE_TYPE_JAZ:
        dtype = _("Jaz Drive");
        break;
    case LIBHAL_DRIVE_TYPE_FLASHKEY:
        dtype = _("Pen Drive");
        break;
    default:
        if (libhal_drive_is_hotpluggable (drive)) {
            dtype = _("External Drive");
        }
        else {
            dtype = _("Drive");
        }
        break;
    }

    if (vendor && model) {
        title = ho_strdup_printf ("%s %s", vendor, model);
    }
    else if (vendor) {
        title = ho_strdup_printf ("%s %s", vendor, dtype);
    }
    else {
        title = ho_strdup (dtype);
    }

    return title;
}
#endif /* HAVE_HAL */


char *
volume_get_mountpoint (const char *_device)
{
    char *mountpoint = NULL;

    char *device = file_resolve_softlink (_device);
    if (!device) {
        goto out_free;
    }

#ifdef HAVE_HAL
    if (!mountpoint) {
        LibHalVolume *volume = hal_get_volume (device);
        if (volume) {
            const char *mp = libhal_volume_get_mount_point (volume);
            mountpoint = mrl_from_filename (mp);
            libhal_volume_free (volume);
        }
    }
#endif

    if (!mountpoint) {
        FILE *fstab = fopen ("/etc/fstab", "r");
        if (!fstab) {
            goto out_free;
        }

        struct mntent *entry = getmntent (fstab);
        while (entry) {
            if (mrls_are_equal (device, entry->mnt_fsname)) {
                mountpoint = mrl_from_filename (entry->mnt_dir);
                break;
            }
            entry = getmntent (fstab);
        }
        endmntent (fstab);
    }

  out_free:
    ho_free (device);

    return mountpoint;
}


volume_type_t
volume_get_type (const char *_device)
{
    volume_type_t type = VOLUME_TYPE_UNKNOWN;

    char *device = file_resolve_softlink (_device);
    if (!device) {
        goto out_free;
    }

#ifdef HAVE_HAL
    char *udi = hal_get_udi_for_volume (device);
    if (!udi) {
        goto out_free;
    }

    if (hal_volume_test_type (udi, "volume.is_disc")) {
        if (hal_volume_test_type (udi, "volume.disc.is_blank")) {
            type = VOLUME_TYPE_BLANK;
        }
        else if (hal_volume_test_type (udi, "volume.disc.has_audio")) {
            type = VOLUME_TYPE_AUDIO_CD;
        }
        else if (hal_volume_test_type (udi, "volume.disc.is_vcd")) {
            type = VOLUME_TYPE_VIDEO_CD;
        }
        else if (hal_volume_test_type (udi, "volume.disc.is_svcd")) {
            type = VOLUME_TYPE_VIDEO_CD;
        }
        else if (hal_volume_test_type (udi, "volume.disc.is_videodvd")) {
            type = VOLUME_TYPE_VIDEO_DVD;
        }
        else {
            type = VOLUME_TYPE_CDROM;
        }
    }
    else if (hal_volume_is_mountable (device)) {
        type = VOLUME_TYPE_MOUNTABLE;
    }

    ho_free (udi);
#endif /* HAVE_HAL */

  out_free:
    ho_free (device);

    return type;
}


char *
volume_get_title (const char *_device)
{
    char *title = NULL;
    char *drive_title = NULL;
    char *volume_title = NULL;

    char *device = file_resolve_softlink (_device);
    if (!device) {
        goto out_free;
    }

    volume_type_t vol_type = volume_get_type (device);
    switch (vol_type) {
    case VOLUME_TYPE_AUDIO_CD:
        drive_title = ho_strdup (_("Audio CD"));
        {
            char *mrl = mrl_build ("cdda", device, 1);
            char *album = meta_info_get (META_INFO_ALBUM, mrl);
            char *artist = meta_info_get (META_INFO_ARTIST, mrl);

            if (album && artist) {
                volume_title = ho_strdup_printf ("%s - %s", artist, album);
            }
            else {
                volume_title = ho_strdup (_("Unknown Album"));
            }

            ho_free (mrl);
            ho_free (album);
            ho_free (artist);
        }
        break;
    case VOLUME_TYPE_VIDEO_CD:
        drive_title = ho_strdup (_("Video CD"));
        {
            char *mrl = mrl_build ("vcd", device, 0);
            volume_title = meta_info_get (META_INFO_TITLE, mrl);
            ho_free (mrl);
        }
        break;
    case VOLUME_TYPE_VIDEO_DVD:
        drive_title = ho_strdup (_("DVD"));
        {
            char *mrl = mrl_build ("dvd", device, 0);
            volume_title = meta_info_get (META_INFO_TITLE, mrl);
            ho_free (mrl);
        }
        break;
    case VOLUME_TYPE_MOUNTABLE:
    case VOLUME_TYPE_CDROM:
        break;
    case VOLUME_TYPE_BLANK:
        volume_title = ho_strdup (_("Blank Disc"));
        break;
    default:
        break;
    }

#ifdef HAVE_HAL
    LibHalDrive *drive = hal_get_drive (device);
    LibHalVolume *volume = hal_get_volume (device);
    if (!volume || !drive) {
        goto hal_out;
    }

    if (!drive_title) {
        drive_title = hal_drive_get_title (drive);
    }

    const char *label = libhal_volume_get_label (volume);
    if (!volume_title && label) {
        volume_title = ho_strdup (label);
    }

  hal_out:
    if (volume) {
        libhal_volume_free (volume);
    }
    if (drive) {
        libhal_drive_free (drive);
    }
#endif /* HAVE_HAL */

    if (!drive_title) {
        drive_title = ho_strdup (_("Drive"));
    }
    if (!volume_title) {
        volume_title = ho_strdup (_("Disc"));
    }

    title = ho_strdup_printf ("%s: %s", drive_title, volume_title);
    replace_chars (title, '_', ' ');

  out_free:
    ho_free (device);
    ho_free (drive_title);
    ho_free (volume_title);

    return title;
}


bool
volume_is_mounted (const char *_device)
{
    bool mounted = false;

    char *device = file_resolve_softlink (_device);
    if (!device) {
        goto out_free;
    }

#ifdef HAVE_HAL
    mounted = hal_volume_is_mounted (device);
#else
    FILE *f = fopen ("/proc/mounts", "r");
    if (!f) {
        error (_("Could not open '%s': %s!"),
               "/proc/mounts", strerror (errno));
        goto out_free;
    }

    char line[256];
    while (fgets (line, 256, f) != NULL) {
        if (strstr (line, device) != NULL) {
            mounted = true;
            break;
        }
    }
    fclose (f);
#endif /* !HAVE_HAL */

  out_free:
    ho_free (device);

    return mounted;
}


bool
volume_is_present (const char *_device)
{
    bool present = false;

    char *device = file_resolve_softlink (_device);
    if (!device) {
        goto out_free;
    }

#ifdef HAVE_LINUX_CDROM_H
    int fd = open (device, O_RDONLY | O_NONBLOCK);
    if (fd == -1) {
        error (_("Could not open '%s': %s!"), device, strerror (errno));
        goto out_free;
    }

    int drive_status = ioctl (fd, CDROM_DRIVE_STATUS);
    switch (drive_status) {
    case CDS_NO_INFO:
        debug ("drive reports: no info");
        break;
    case CDS_NO_DISC:
        debug ("drive reports: no disc");
        break;
    case CDS_TRAY_OPEN:
        debug ("drive reports: tray open");
        break;
    case CDS_DRIVE_NOT_READY:
        debug ("drive reports: drive not ready");
        break;
    case CDS_DISC_OK:
        debug ("drive reports: disc ok");
        present = true;
        break;
    default:
        debug ("drive reports: unknown value");
        break;
    }
    close (fd);
#endif

  out_free:
    ho_free (device);

    return present;
}


#ifdef BIN_DBUS_SEND
static bool
dbus_send (const char *udi, const char *dbus_cmd, const char *dbus_prm)
{
    bool result = false;
    char *command = NULL;

    command = ho_strdup_printf ("%s --system --print-reply "
                                "--dest=org.freedesktop.Hal %s %s %s",
                                BIN_DBUS_SEND, udi, dbus_cmd, dbus_prm);
    result = execute_shell (command, 0);
    ho_free (command);

    return result;
}
#endif


bool
volume_mount (const char *_device)
{
    bool result = false;

    if (volume_is_mounted (_device)) {
        return true;
    }

    char *device = file_resolve_softlink (_device);
    if (!device) {
        goto out_free;
    }

#if (defined (HAVE_HAL) && defined (BIN_DBUS_SEND))
    if (!result) {
        uid_t uid = getuid ();
        char *udi = hal_get_udi_for_volume (device);
        char *cmd = "org.freedesktop.Hal.Device.Volume.Mount";
        char *prm = ho_strdup_printf ("string: string: "
                                      "array:string:noexec,uid=%d", uid);
        if (udi && dbus_send (udi, cmd, prm)) {
            result = volume_is_mounted (device);
        }
        ho_free (prm);
        ho_free (udi);
    }
    if (!result) {
        char *udi = hal_get_udi_for_volume (device);
        char *cmd = "org.freedesktop.Hal.Device.Volume.Mount";
        char *prm = ho_strdup_printf ("string: string: "
                                      "array:string:noexec");
        if (udi && dbus_send (udi, cmd, prm)) {
            result = volume_is_mounted (device);
        }
        ho_free (prm);
        ho_free (udi);
    }
#endif
#ifdef BIN_PMOUNT
    if (!result) {
        char *cmd = ho_strdup_printf ("%s %s", BIN_PMOUNT, device);
        if (execute_shell (cmd, 0)) {
            result = volume_is_mounted (device);
        }
        ho_free (cmd);
    }
#endif
#ifdef BIN_MOUNT
    if (!result) {
        char *cmd = ho_strdup_printf ("%s %s", BIN_MOUNT, device);
        if (execute_shell (cmd, 0)) {
            result = volume_is_mounted (device);
        }
        ho_free (cmd);
    }
#endif

  out_free:
    if (!result) {
        error (_("Could not mount '%s': %s!"), device, _("Unknown"));
    }

    ho_free (device);

    return result;
}


bool
volume_umount (const char *_device)
{
    bool result = false;

    if (!volume_is_mounted (_device)) {
        return true;
    }

    char *device = file_resolve_softlink (_device);
    if (!device) {
        goto out_free;
    }

#if (defined (HAVE_HAL) && defined (BIN_DBUS_SEND))
    if (!result) {
        char *udi = hal_get_udi_for_volume (device);
        char *cmd = "org.freedesktop.Hal.Device.Volume.Unmount";
        char *prm = "array:string:";
        if (udi && dbus_send (udi, cmd, prm)) {
            result = !volume_is_mounted (device);
        }
        ho_free (udi);
    }
#endif
#ifdef BIN_PMOUNT
    if (!result) {
        char *cmd = ho_strdup_printf ("%s %s", BIN_PUMOUNT, device);
        if (execute_shell (cmd, 0)) {
            result = !volume_is_mounted (device);
        }
        ho_free (cmd);
    }
#endif
#ifdef BIN_MOUNT
    if (!result) {
        char *cmd = ho_strdup_printf ("%s %s", BIN_UMOUNT, device);
        if (execute_shell (cmd, 0)) {
            result = !volume_is_mounted (device);
        }
        ho_free (cmd);
    }
#endif

  out_free:
    if (!result) {
        error (_("Could not unmount '%s': %s!"), device, _("Unknown"));
    }

    ho_free (device);

    return result;
}


bool
drive_is_ejected (const char *_device)
{
    bool ejected = false;

#ifdef HAVE_LINUX_CDROM_H
    char *device = file_resolve_softlink (_device);
    if (!device) {
        goto out_free;
    }

    int fd = open (device, O_RDONLY | O_NONBLOCK);
    if (fd == -1) {
        error (_("Could not open '%s': %s!"), device, strerror (errno));
        /* If a device does not exist, we assume, that it was ejected a
         * moment ago. */
        ejected = true;
        goto out_free;
    }

    /** 
     * @bug This does not always work. Apparently ther is a bug in the SCSI
     * code of the 2.6.10 kernels that will return CDS_TRAY_OPEN if there is
     * no disc in the drive. We'll just wait and hope this gets fixed soon.
     */
    int drive_status = ioctl (fd, CDROM_DRIVE_STATUS);
    switch (drive_status) {
    case CDS_NO_INFO:
        debug ("drive reports: no info");
        break;
    case CDS_NO_DISC:
        debug ("drive reports: no disc");
        break;
    case CDS_TRAY_OPEN:
        debug ("drive reports: tray open");
        ejected = true;
        break;
    case CDS_DRIVE_NOT_READY:
        debug ("drive reports: drive not ready");
        break;
    case CDS_DISC_OK:
        debug ("drive reports: disc ok");
        break;
    default:
        debug ("drive reports: unknown value");
        break;
    }
    close (fd);

  out_free:
    ho_free (device);
#endif

    return ejected;
}


bool
drive_inject (const char *_device)
{
    char *errmsg = _("Unknown");
    bool result = false;

    char *device = file_resolve_softlink (_device);
    if (!device) {
        goto out_free;
    }

#if (defined (HAVE_HAL) && defined (BIN_DBUS_SEND))
    if (!result) {
        char *udi = hal_get_udi_for_drive (device);
        char *cmd = "org.freedesktop.Hal.Device.Storage.CloseTray";
        char *prm = "array:string:";
        if (udi && dbus_send (udi, cmd, prm)) {
            result = !volume_is_mounted (device);
        }
        ho_free (udi);
    }
#endif
#ifdef BIN_EJECT
    if (!result) {
        char *cmd = ho_strdup_printf ("%s -t %s", BIN_EJECT, device);
        if (execute_shell (cmd, 0)) {
            result = !drive_is_ejected (device);
        }
        ho_free (cmd);
    }
#endif
#ifdef HAVE_LINUX_CDROM_H
    if (!result) {
        int fd = open (device, O_RDONLY | O_NONBLOCK);
        if (fd == -1) {
            errmsg = strerror (errno);
            goto out_free;
        }
        if (ioctl (fd, CDROMCLOSETRAY) < 0) {
            errmsg = strerror (errno);
            close (fd);
        }
        else {
            result = !drive_is_ejected (device);
        }
        close (fd);
    }
#endif

  out_free:
    if (!result) {
        error (_("Could not inject '%s': %s!"), device, errmsg);
    }

    ho_free (device);

    return result;
}


bool
drive_eject (const char *_device)
{
    char *errmsg = _("Unknown");
    bool result = false;

    char *device = file_resolve_softlink (_device);
    if (!device) {
        goto out_free;
    }

#if (defined (HAVE_HAL) && defined (BIN_DBUS_SEND))
    if (!result) {
        char *udi = hal_get_udi_for_volume (device);
        char *cmd = "org.freedesktop.Hal.Device.Volume.Eject";
        char *prm = "array:string:";
        if (udi && dbus_send (udi, cmd, prm)) {
            result = drive_is_ejected (device);
        }
        ho_free (udi);
    }
#endif
#ifdef BIN_EJECT
    if (!result) {
        char *cmd = ho_strdup_printf ("%s %s", BIN_EJECT, device);
        if (execute_shell (cmd, 0)) {
            result = drive_is_ejected (device);
        }
        ho_free (cmd);
    }
#endif
#ifdef HAVE_LINUX_CDROM_H
    if (!result) {
        int fd = open (device, O_RDONLY | O_NONBLOCK);
        if (fd == -1) {
            errmsg = strerror (errno);
            goto out_free;
        }

        if (ioctl (fd, CDROMEJECT) < 0) {
            errmsg = strerror (errno);
            close (fd);
        }
        else {
            result = drive_is_ejected (device);
        }

        close (fd);
    }
#endif

  out_free:
    if (!result) {
        error (_("Could not eject '%s': %s!"), device, errmsg);
    }

    ho_free (device);

    return result;
}


char *
drive_get_title (const char *_device)
{
    char *title = NULL;
    char *device = NULL;

    device = file_resolve_softlink (_device);
    if (!device) {
        goto out_free;
    }

#ifdef HAVE_HAL
    LibHalDrive *drive = hal_get_drive (device);
    if (!drive) {
        goto out_free;
    }

    if (!title) {
        title = hal_drive_get_title (drive);
    }

    libhal_drive_free (drive);
#endif /* HAVE_HAL */

  out_free:
    ho_free (device);

    if (!title) {
        title = ho_strdup (_("Drive"));
    }

    return title;
}
