
/*
 * Copyright (C) 2006 Maximilian Schwerin
 *
 * This file is part of oxine a free media player.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: disc_fam.c 3924 2009-05-18 07:23:27Z mschwerin $
 *
 */
#include "config.h"

#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#ifdef HAVE_FAM
#include <fam.h>
#endif /* HAVE_FAM */

#include <sys/stat.h>
#include <sys/select.h>

#include "disc_fam.h"
#include "filelist.h"
#include "heap.h"
#include "i18n.h"
#include "list.h"
#include "logger.h"
#include "mutex.h"
#include "odk.h"
#include "oxine.h"
#include "utils.h"

//#define DEBUG_FAM

extern oxine_t *oxine;

typedef struct {
#ifdef HAVE_FAM
    FAMRequest request;
    bool handled_by_fam;
#endif

    filelist_t *filelist;
    bool please_remove;
} oFAMRequest_t;

#ifdef HAVE_FAM
static FAMConnection fc;

static bool thread_fam_run = false;
static pthread_t thread_fam;
#endif /* HAVE_FAM */

static bool thread_poll_run = false;
static bool thread_poll_active = false;
static pthread_t thread_poll;
static pthread_mutex_t poll_wait_mutex;
static pthread_cond_t poll_continue;

static l_list_t *request_list = NULL;
static pthread_mutex_t request_list_mutex;

#ifdef HAVE_FAM
#ifdef DEBUG_FAM
static const char *
fam_code_to_reason (int code)
{
    const char *reason[] = {
        "FAMChanged",
        "FAMDeleted",
        "FAMStartExecuting",
        "FAMStopExecuting",
        "FAMCreated",
        "FAMMoved",
        "FAMAcknowledge",
        "FAMExists",
        "FAMEndExist",
    };

    switch (code) {
    case FAMChanged:
    case FAMDeleted:
    case FAMStartExecuting:
    case FAMStopExecuting:
    case FAMCreated:
    case FAMMoved:
    case FAMAcknowledge:
    case FAMExists:
    case FAMEndExist:
        return reason[code - 1];
    default:
        return "unknown";
    }
}
#endif /* DEBUG_FAM */
#endif /* HAVE_FAM */


static void
request_list_free (void *p)
{
    oFAMRequest_t *fr = (oFAMRequest_t *) p;
    ho_setref (fr->filelist, NULL);
    ho_free (fr);
}


void
fam_monitor_add_list (filelist_t * filelist)
{
    if (!filelist) {
        return;
    }

    if (false) {
        /* do nothing */
    }
    else if (!filelist->update_allowed) {
        /* do nothing */
    }
    else if (mrl_has_protocol (filelist->mrl, "cdda")) {
        /* do nothing */
    }
#ifdef HAVE_FAM
    else if (mrl_is_local (filelist->mrl)) {
        char *filename = mrl_get_filename (filelist->mrl);

        oFAMRequest_t *fr = ho_new (oFAMRequest_t);
        ho_setref (fr->filelist, filelist);
        fr->handled_by_fam = true;

        if (mrl_is_directory (filelist->mrl)) {
            if (FAMMonitorDirectory (&fc, filename, &fr->request,
                                     fr->filelist) != 0) {
                error (_("Could not start monitor for '%s': %s!"),
                       filelist->mrl, FamErrlist[FAMErrno]);
                ho_free (fr);
            }
            else {
#ifdef DEBUG_FAM
                debug ("Started monitoring '%s'...", filelist->mrl);
#endif
                mutex_lock (&request_list_mutex);
                l_list_append (request_list, fr);
                mutex_unlock (&request_list_mutex);
            }
        }
        else {
            if (FAMMonitorFile (&fc, filename, &fr->request,
                                fr->filelist) != 0) {
                error (_("Could not start monitor for '%s': %s!"),
                       filelist->mrl, FamErrlist[FAMErrno]);
                ho_free (fr);
            }
            else {
#ifdef DEBUG_FAM
                debug ("Started monitoring '%s'...", filelist->mrl);
#endif
                mutex_lock (&request_list_mutex);
                l_list_append (request_list, fr);
                mutex_unlock (&request_list_mutex);
            }
        }

        ho_free (filename);
    }
#endif /* HAVE_FAM */
    else {
        oFAMRequest_t *fr = ho_new (oFAMRequest_t);
        ho_setref (fr->filelist, filelist);
        mutex_lock (&request_list_mutex);
        l_list_append (request_list, fr);
        mutex_unlock (&request_list_mutex);
#ifdef DEBUG_FAM
        debug ("Started monitoring '%s'...", filelist->mrl);
#endif
    }
}


void
fam_monitor_del_list (filelist_t * filelist)
{
    if (!filelist) {
        return;
    }

    mutex_lock (&request_list_mutex);
    oFAMRequest_t *fr = l_list_first (request_list);
    while (fr && (fr->filelist != filelist)) {
        fr = l_list_next (request_list, fr);
    }
    if (!fr) {
        /* do nothing */
    }
    else if (fr->please_remove) {
        /* do nothing */
    }
    else if (fr->filelist == filelist) {
#ifdef DEBUG_FAM
        debug ("Canceled monitoring '%s'...", filelist->mrl);
#endif

#ifdef HAVE_FAM
        if (!fr->handled_by_fam) {
            /* do nothing */
        }
        else if (FAMCancelMonitor (&fc, &fr->request) != 0) {
            error (_("Could not cancel monitor for '%s': %s!"),
                   filelist->mrl, FamErrlist[FAMErrno]);
        }
#endif /* HAVE_FAM */

        /* We do not remove the entry at once, as it is possible, that we're
         * checking for changes at the moment. We just mark the entry as to
         * be removed and send a signal to the polling thread to continue.
         * This is necessary, as the thread may be paused at the moment. */
        fr->please_remove = true;
        pthread_cond_signal (&poll_continue);
    }
    mutex_unlock (&request_list_mutex);
}


static void *
filelist_poll_thread (void *data)
{
    info (_("Successfully started file alteration monitor thread."));
    debug ("file alteration monitor thread: 0x%X", (int) pthread_self ());

    struct timespec ts;
    oxine_event_t event;
    oxine_event_init (&event);
    event.type = OXINE_EVENT_FILELIST_UPDATE;

    pthread_mutex_lock (&poll_wait_mutex);
    while (thread_poll_run) {
        /* If polling is active at the moment we wait for the continue
         * signal with a timeout of 10 seconds. */
        if (thread_poll_active) {
            ts.tv_sec = time (NULL) + 30;
            pthread_cond_timedwait (&poll_continue, &poll_wait_mutex, &ts);
        }
        /* If polling is paused at the moment we wait for the continue
         * signal without a timeout. */
        else {
            pthread_cond_wait (&poll_continue, &poll_wait_mutex);
        }
        /* We search the list, to remove anything that was marked as to be
         * removed. */
        {
            mutex_lock (&request_list_mutex);
            oFAMRequest_t *fr = l_list_first (request_list);
            while (fr) {
                oFAMRequest_t *nr = l_list_next (request_list, fr);
                if (fr->please_remove) {
#ifdef DEBUG_FAM
                    debug ("Removing '%s'...", fr->filelist->mrl);
#endif
                    l_list_remove (request_list, fr);
                    ho_setref (fr->filelist, NULL);
                    ho_free (fr);
                }
                fr = nr;
            }
            mutex_unlock (&request_list_mutex);
        }
        /* If polling is active at the moment we search for a filelist to
         * update. If we find one we check it for changes and if it has
         * changed send an event to the frontend telling it to update the
         * GUI accordingly. */
        if (thread_poll_active) {
            filelist_t *need_update = NULL;
            mutex_lock (&request_list_mutex);
            oFAMRequest_t *fr = l_list_first (request_list);
            while (fr && !need_update && thread_poll_run) {
                if (fr->please_remove) {
                    /* do nothing */
                }
#ifdef HAVE_FAM
                else if (fr->handled_by_fam) {
                    /* do nothing */
                }
#endif /* HAVE_FAM */
                else {
                    filelist_t *list = NULL;
                    ho_setref (list, fr->filelist);
                    mutex_unlock (&request_list_mutex);
                    if (filelist_check_for_changes (list)) {
                        ho_setref (need_update, list);
                    }
                    mutex_lock (&request_list_mutex);
                    ho_setref (list, NULL);
                }
                fr = l_list_next (request_list, fr);
            }
            mutex_unlock (&request_list_mutex);
            if (need_update) {
                ho_setref (event.data.user_data, need_update);
                odk_oxine_event_send_synchronized (oxine->odk, &event);
                ho_setref (need_update, NULL);
            }
        }
    }
    pthread_mutex_unlock (&poll_wait_mutex);

    ho_setref (event.data.user_data, NULL);
    pthread_exit (NULL);
    return NULL;
}


#ifdef HAVE_FAM
static bool
fam_code_relevant (FAMEvent *fe)
{
    switch (fe->code) {
    case FAMChanged:
    case FAMDeleted:
    case FAMCreated:
    case FAMMoved:
        return true;
    default:
        return false;
    }
}


static bool
fam_file_relevant (FAMEvent *fe)
{
    filelist_t *filelist = (filelist_t *) fe->userdata;

    if (fe->code == FAMCreated) {
        return true;
    }
    else if (mrl_is_directory (filelist->mrl)) {
        bool relevant = false;
        filelist_lock (filelist);
        fileitem_t *fileitem = filelist_first (filelist);
        while (fileitem && !relevant) {
            relevant |= str_ends_with (fileitem->mrl, fe->filename);
            fileitem = filelist_next (filelist, fileitem);
        }
        filelist_unlock (filelist);
        return relevant;
    }
    else {
        return true;
    }
}


static void *
fam_monitor_thread (void *data)
{
    info (_("Successfully started file alteration monitor thread."));
    debug ("file alteration monitor thread: 0x%X", (int) pthread_self ());

    int fd;
    FAMEvent fe;
    fd_set readfds;

    oxine_event_t event;
    oxine_event_init (&event);
    event.type = OXINE_EVENT_FILELIST_UPDATE;

    fd = FAMCONNECTION_GETFD (&fc);
    FD_ZERO (&readfds);
    FD_SET (fd, &readfds);

    while (thread_fam_run) {
        if (select (fd + 1, &readfds, NULL, NULL, NULL) < 0) {
            error (_("Failed to wait on socket: %s!"), strerror (errno));
        }
        else if (!thread_fam_run) {
            /* do nothing */
        }
        else if (!FD_ISSET (fd, &readfds)) {
            /* do nothing */
        }
        else if (FAMPending (&fc) == 0) {
            /* do nothing */
        }
        else if (FAMNextEvent (&fc, &fe) < 0) {
            /* do nothing */
        }
        else if (!fam_code_relevant (&fe)) {
            /* do nothing */
        }
        else if (!fam_file_relevant (&fe)) {
            /* do nothing */
        }
        else {
            filelist_t *filelist = (filelist_t *) fe.userdata;
            ho_setref (event.data.user_data, filelist);

#ifdef DEBUG_FAM
            if (mrl_is_directory (filelist->mrl)) {
                debug ("got '%s' for '%s/%s'", fam_code_to_reason (fe.code),
                       filelist->mrl, fe.filename);
            }
            else {
                debug ("got '%s' for '%s'", fam_code_to_reason (fe.code),
                       filelist->mrl);
            }
#endif

            odk_oxine_event_send_synchronized (oxine->odk, &event);
        }
    }

    ho_setref (event.data.user_data, NULL);
    pthread_exit (NULL);
    return NULL;
}
#endif /* HAVE_FAM */


bool
fam_monitor_start (void)
{
    request_list = l_list_new ();
    pthread_mutex_init (&request_list_mutex, NULL);

#ifdef HAVE_FAM
    if (FAMOpen (&fc) != 0) {
        error (_("Could not open connection to file alteration monitor: %s!"),
               FamErrlist[FAMErrno]);
        return false;
    }

    /* I have no idea why this is necessary! */
#if 1
    FAMRequest request;
    if (FAMMonitorDirectory (&fc, OXINE_LOCALEDIR, &request, NULL) != 0) {
        error (_("Could not start monitor for '%s': %s!"),
               OXINE_LOCALEDIR, FamErrlist[FAMErrno]);
    }
    if (FAMCancelMonitor (&fc, &request) != 0) {
        error (_("Could not cancel monitor for '%s': %s!"),
               OXINE_LOCALEDIR, FamErrlist[FAMErrno]);
        return false;
    }
#endif

    thread_fam_run = true;
    if (pthread_create (&thread_fam, NULL, fam_monitor_thread, NULL) != 0) {
        error (_("Could not create file alteration monitor thread: %s!"),
               strerror (errno));
        return false;
    }
#endif /* HAVE_FAM */

    pthread_cond_init (&poll_continue, NULL);
    pthread_mutex_init (&poll_wait_mutex, NULL);

    thread_poll_active = false;
    thread_poll_run = true;
    if (pthread_create (&thread_poll, NULL, filelist_poll_thread, NULL) != 0) {
        error (_("Could not create file alteration monitor thread: %s!"),
               strerror (errno));
        return false;
    }

    return true;
}


bool
fam_monitor_free (void)
{
#ifdef HAVE_FAM
    FAMRequest request;
    thread_fam_run = false;

    /* Start monitoring a new directory to make the FAM thread leave the
     * select call. */
    if (FAMMonitorDirectory (&fc, OXINE_LOCALEDIR, &request, NULL) != 0) {
        error (_("Could not start monitor for '%s': %s!"),
               OXINE_LOCALEDIR, FamErrlist[FAMErrno]);
    }

    pthread_join (thread_fam, NULL);

    if (FAMClose (&fc) != 0) {
        error (_("Could not close connection to "
                 "file alteration monitor: %s!"), strerror (errno));
    }
#endif /* HAVE_FAM */

    thread_poll_run = false;
    thread_poll_active = false;
    pthread_cond_signal (&poll_continue);
    pthread_join (thread_poll, NULL);

    l_list_free (request_list, request_list_free);
    pthread_mutex_destroy (&request_list_mutex);
    pthread_mutex_destroy (&poll_wait_mutex);
    pthread_cond_destroy (&poll_continue);

    info (_("Successfully stopped file alteration monitor thread."));

    return true;
}


void
fam_monitor_poll (bool enable)
{
    if (!thread_poll_active && enable) {
        thread_poll_active = true;
        pthread_cond_signal (&poll_continue);
    }
    thread_poll_active = enable;
}
