
/*
 * Copyright (C) 2004-2005 Maximilian Schwerin
 *
 * This file is part of oxine a free media player.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: environment.c 4004 2009-09-26 12:38:07Z mschwerin $
 *
 */
#include "config.h"

#include <dirent.h>
#include <errno.h>
#include <unistd.h>

#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/types.h>

#include "download.h"
#include "environment.h"
#include "heap.h"
#include "i18n.h"
#include "list.h"
#include "logger.h"
#include "oxine.h"
#include "utils.h"

extern oxine_t *oxine;

static char *dir_home = NULL;
static char *dir_oxine = NULL;
static char *dir_oxine_cache = NULL;
static char *dir_oxine_tmp = NULL;
static char *dir_oxine_playlists = NULL;
static char *dir_oxine_skin = NULL;

static char *file_config = NULL;
static char *file_lirc_config = NULL;
static char *file_jsw_config = NULL;
static char *file_dvb_channels = NULL;
static char *file_favorites = NULL;
static char *file_playlist_rw = NULL;
static char *file_xdg_user_dirs = NULL;

static char *mrl_home = NULL;
static char *mrl_desktop = NULL;
static char *mrl_root = NULL;
static char *mrl_favorites_most = NULL;
static char *mrl_favorites_recent = NULL;
static char *mrl_playlists = NULL;

static char *mrl_xdg_music_dir = NULL;
static char *mrl_xdg_videos_dir = NULL;
static char *mrl_xdg_pictures_dir = NULL;

static char *mrl_mediamarks_music = NULL;
static char *mrl_mediamarks_video = NULL;
static char *mrl_mediamarks_image = NULL;
static char *mrl_mediamarks_media = NULL;
static char *mrl_mediamarks_subtitle = NULL;

static pthread_t cache_thread = 0;


const char *
get_dir_root (void)
{
    return "/";
}


const char *
get_mrl_root (void)
{
    if (!mrl_root) {
        mrl_root = mrl_from_filename (get_dir_root ());
    }

    return (const char *) mrl_root;
}


const char *
get_dir_home (void)
{
    if (!dir_home) {
        char *tmp = getenv ("HOME");
        if (!tmp) {
            /* do nothing */
        }
        else if (strlen (tmp) == 0) {
            /* do nothing */
        }
        else if (str_equals (tmp, "/")) {
            /* do nothing */
        }
        else if (access (tmp, X_OK | W_OK | R_OK) != 0) {
            /* do nothing */
        }
        else {
            dir_home = ho_strdup (tmp);
        }
    }
    if (!dir_home) {
        error (_("Unable to find usable home directory, using '/tmp'."));
        dir_home = ho_strdup ("/tmp");
    }
    if (dir_home[strlen (dir_home) - 1] == '/') {
        dir_home[strlen (dir_home) - 1] = '\0';
    }

    return (const char *) dir_home;
}


const char *
get_mrl_home (void)
{
    if (!mrl_home) {
        mrl_home = mrl_from_filename (get_dir_home ());
    }

    return (const char *) mrl_home;
}


const char *
get_dir_oxine (void)
{
    if (!dir_oxine) {
        dir_oxine = ho_strdup_printf ("%s/.oxine", get_dir_home ());

        mkdir (dir_oxine, 0700);
    }

    return (const char *) dir_oxine;
}


const char *
get_dir_oxine_skin (void)
{
    if (!dir_oxine_skin) {
        const char *skin = config_get_enum ("gui.skin");
        dir_oxine_skin = ho_strdup_printf ("%s/%s", OXINE_SKINDIR, skin);
    }

    return (const char *) dir_oxine_skin;
}


const char *
get_dir_playlists (void)
{
    if (!dir_oxine_playlists) {
        dir_oxine_playlists = ho_strdup_printf ("%s/playlists",
                                                get_dir_oxine ());

        mkdir (dir_oxine_playlists, 0700);
    }

    return (const char *) dir_oxine_playlists;
}


const char *
get_mrl_playlists (void)
{
    if (!mrl_playlists) {
        mrl_playlists = mrl_from_filename (get_dir_playlists ());
    }

    return (const char *) mrl_playlists;
}


const char *
get_dir_oxine_cache (void)
{
    if (!dir_oxine_cache) {
        dir_oxine_cache = ho_strdup_printf ("%s/cache", get_dir_oxine ());

        mkdir (dir_oxine_cache, 0700);
    }

    return (const char *) dir_oxine_cache;
}


const char *
get_dir_oxine_tmp (void)
{
    if (!dir_oxine_tmp) {
        dir_oxine_tmp = ho_strdup_printf ("/tmp/oxine-%d", getpid ());

        mkdir (dir_oxine_tmp, 0700);
    }

    return (const char *) dir_oxine_tmp;
}


void
set_file_config (const char *filename)
{
    ho_free (file_config);
    file_config = ho_strdup (filename);
}


const char *
get_file_config (void)
{
    if (!file_config) {
        file_config = ho_strdup_printf ("%s/config", get_dir_oxine ());
    }

    return (const char *) file_config;
}


const char *
get_mrl_mediamarks_music (void)
{
    if (!mrl_mediamarks_music) {
        char *file = ho_strdup_printf ("%s/mediamarks_music.xml",
                                       get_dir_oxine ());
        mrl_mediamarks_music = mrl_from_filename (file);
        ho_free (file);
    }

    return (const char *) mrl_mediamarks_music;
}


const char *
get_mrl_mediamarks_video (void)
{
    if (!mrl_mediamarks_video) {
        char *file = ho_strdup_printf ("%s/mediamarks_video.xml",
                                       get_dir_oxine ());
        mrl_mediamarks_video = mrl_from_filename (file);
        ho_free (file);
    }

    return (const char *) mrl_mediamarks_video;
}


const char *
get_mrl_mediamarks_image (void)
{
    if (!mrl_mediamarks_image) {
        char *file = ho_strdup_printf ("%s/mediamarks_image.xml",
                                       get_dir_oxine ());
        mrl_mediamarks_image = mrl_from_filename (file);
        ho_free (file);
    }

    return (const char *) mrl_mediamarks_image;
}


const char *
get_mrl_mediamarks_media (void)
{
    if (!mrl_mediamarks_media) {
        char *file = ho_strdup_printf ("%s/mediamarks_media.xml",
                                       get_dir_oxine ());
        mrl_mediamarks_media = mrl_from_filename (file);
        ho_free (file);
    }

    return (const char *) mrl_mediamarks_media;
}


const char *
get_mrl_mediamarks_subtitle (void)
{
    if (!mrl_mediamarks_subtitle) {
        char *file = ho_strdup_printf ("%s/mediamarks_subtitle.xml",
                                       get_dir_oxine ());
        mrl_mediamarks_subtitle = mrl_from_filename (file);
        ho_free (file);
    }

    return (const char *) mrl_mediamarks_subtitle;
}


const char *
get_file_jsw_config (void)
{
    if (!file_jsw_config) {
        file_jsw_config = ho_strdup_printf ("%s/joystick", get_dir_oxine ());
    }

    return (const char *) file_jsw_config;
}


const char *
get_file_lirc_config (void)
{
    if (!file_lirc_config) {
        file_lirc_config = ho_strdup_printf ("%s/lircrc", get_dir_oxine ());
    }

    return (const char *) file_lirc_config;
}


const char *
get_file_dvb_channels (void)
{
    if (!file_dvb_channels) {
        file_dvb_channels = ho_strdup_printf ("%s/.xine/channels.conf",
                                              get_dir_home ());
    }

    return (const char *) file_dvb_channels;
}


const char *
get_file_favorites (void)
{
    if (!file_favorites) {
        file_favorites = ho_strdup_printf ("%s/favorites.xml",
                                           get_dir_oxine ());
    }

    return (const char *) file_favorites;
}


const char *
get_mrl_favorites_most (void)
{
    if (!mrl_favorites_most) {
        char *temp = mrl_from_filename (get_file_favorites ());
        mrl_favorites_most = ho_strdup_printf ("%s?sort=most", temp);
        ho_free (temp);
    }

    return (const char *) mrl_favorites_most;
}


const char *
get_mrl_favorites_recent (void)
{
    if (!mrl_favorites_recent) {
        char *temp = mrl_from_filename (get_file_favorites ());
        mrl_favorites_recent = ho_strdup_printf ("%s?sort=recent", temp);
        ho_free (temp);
    }

    return (const char *) mrl_favorites_recent;
}


const char *
get_file_playlist_rw (void)
{
    if (!file_playlist_rw) {
        file_playlist_rw = ho_strdup_printf ("%s/playlist_rw.oxp",
                                             get_dir_oxine ());
    }

    return (const char *) file_playlist_rw;
}


static const char *
get_xdg_user_dirs (void)
{
    if (!file_xdg_user_dirs) {
        char *XDG_CONFIG_HOME = getenv ("XDG_CONFIG_HOME");
        if (XDG_CONFIG_HOME) {
            file_xdg_user_dirs = ho_strdup_printf ("%s/user-dirs.dirs",
                                                   XDG_CONFIG_HOME);
        }
        else {
            const char *home = get_dir_home ();
            file_xdg_user_dirs = ho_strdup_printf ("%s/.config/"
                                                   "user-dirs.dirs", home);
        }
    }

    return (const char *) file_xdg_user_dirs;
}


/**
 * http://www.freedesktop.org/wiki/Software/xdg-user-dirs
 */
static char *
get_xdg_user_dir (const char *name)
{
    const char *home = get_dir_home ();
    const char *xdg_user_dirs = get_xdg_user_dirs ();
    char *xdg_user_dir = NULL;
    FILE *f = fopen (xdg_user_dirs, "r");
    if (!f) {
        error (_("Could not open '%s': %s!"), xdg_user_dirs,
               strerror (errno));
        return NULL;
    }
    char line[256];
    while (fgets (line, sizeof (line), f)) {
        char *l = trim_whitespace (line);
        /* It's a comment */
        if ((strlen (l) == 0) || (l[0] == '#')) {
            continue;
        }
        if (!str_starts_with (l, name)) {
            continue;
        }
        l += strlen (name);
        if ((l[0] != '=') || (l[1] != '"')) {
            continue;
        }
        l += 2;
        while (l[strlen (l) - 1] == '"') {
            l[strlen (l) - 1] = '\0';
        }
        if (str_starts_with (l, "$HOME")) {
            l += strlen ("$HOME");
            xdg_user_dir = ho_strdup_printf ("%s/%s", home, l);
        }
        else {
            xdg_user_dir = ho_strdup (l);
        }
    }
    return xdg_user_dir;
}


const char *
get_mrl_xdg_music_dir (void)
{
    if (!mrl_xdg_music_dir) {
        char *dir = get_xdg_user_dir ("XDG_MUSIC_DIR");
        if (mrl_is_directory (dir)) {
            mrl_xdg_music_dir = mrl_from_filename (dir);
        }
        else {
            mrl_desktop = ho_strdup_printf ("%s/%s", get_mrl_home (),
                                            _("Music"));
        }
    }

    return (const char *) mrl_xdg_music_dir;
}


const char *
get_mrl_xdg_videos_dir (void)
{
    if (!mrl_xdg_videos_dir) {
        char *dir = get_xdg_user_dir ("XDG_VIDEOS_DIR");
        if (mrl_is_directory (dir)) {
            mrl_xdg_videos_dir = mrl_from_filename (dir);
        }
        else {
            mrl_desktop = ho_strdup_printf ("%s/%s", get_mrl_home (),
                                            _("Videos"));
        }
    }

    return (const char *) mrl_xdg_videos_dir;
}


const char *
get_mrl_xdg_pictures_dir (void)
{
    if (!mrl_xdg_pictures_dir) {
        char *dir = get_xdg_user_dir ("XDG_PICTURES_DIR");
        if (mrl_is_directory (dir)) {
            mrl_xdg_pictures_dir = mrl_from_filename (dir);
        }
        else {
            mrl_desktop = ho_strdup_printf ("%s/%s", get_mrl_home (),
                                            _("Pictures"));
        }
    }

    return (const char *) mrl_xdg_pictures_dir;
}


const char *
get_mrl_desktop (void)
{
    if (!mrl_desktop) {
        char *dir = get_xdg_user_dir ("XDG_DESKTOP_DIR");
        if (mrl_is_directory (dir)) {
            mrl_desktop = mrl_from_filename (dir);
        }
        else {
            mrl_desktop = ho_strdup_printf ("%s/%s", get_mrl_home (),
                                            _("Desktop"));
        }
    }

    return (const char *) mrl_desktop;
}


typedef struct {
    char *filename;
    off_t filesize;
    time_t atime;
} cache_entry_t;


static void
build_cache_list (l_list_t * filelist, off_t * cache_size,
                  const char *dirname)
{
    DIR *dirp = opendir (dirname);
    if (!dirp) {
        error (_("Unable to list contents of directory '%s': %s!"),
               dirname, strerror (errno));
        return;
    }

    struct stat filestat;
    struct dirent *entp;
    while ((entp = readdir (dirp))) {
        if (strcmp (entp->d_name, ".") == 0) {
            continue;
        }
        if (strcmp (entp->d_name, "..") == 0) {
            continue;
        }

        char *filename = ho_strdup_printf ("%s/%s", dirname, entp->d_name);
        if (stat (filename, &filestat) == -1) {
            error (_("Unable to get status for file '%s': %s."),
                   filename, strerror (errno));
        }
        else if (S_ISDIR (filestat.st_mode)) {
            build_cache_list (filelist, cache_size, filename);
        }
        else {
            cache_entry_t *entry = ho_new (cache_entry_t);
            entry->filename = ho_strdup (filename);
            entry->filesize = filestat.st_blocks * 512;
            {
                entry->atime = filestat.st_atime;
            }
            if (entry->atime < filestat.st_mtime) {
                entry->atime = filestat.st_mtime;
            }
            if (entry->atime < filestat.st_ctime) {
                entry->atime = filestat.st_ctime;
            }
            *cache_size += entry->filesize;
            l_list_append (filelist, entry);
        }

        ho_free (filename);
    }
    closedir (dirp);

    /* Try to remove the current directory. If it's not empty this will not
     * succeed which is exactly what we want. */
    rmdir (dirname);
}


static bool
cache_entry_swap (void *data1, void *data2)
{
    cache_entry_t *e1 = (cache_entry_t *) data1;
    cache_entry_t *e2 = (cache_entry_t *) data2;

    return (e2->atime < e1->atime);
}


static void
cache_entry_free (void *p)
{
    cache_entry_t *e = (cache_entry_t *) p;

    ho_free (e->filename);
    ho_free (e);
}


/**
 * We build a list of all files in the cache. The list is sorted by the time
 * of the last access to the file. If the sum of all file sizes exceeds the
 * level set by the user we delete files starting with the oldest until the
 * cache size is small enough.
 */
static void *
cache_clean_thread (void *p)
{
    if (!get_dir_oxine_cache ()) {
        return NULL;
    }

    off_t cache_size_cur = 0LL;
    off_t cache_size_max =
        1024LL * 1024LL * (off_t) config_get_number ("misc.cache.max_size");

    l_list_t *filelist = l_list_new ();
    build_cache_list (filelist, &cache_size_cur, dir_oxine_cache);
    l_list_sort (filelist, cache_entry_swap);

    info (_("The cache directory is %3llu%% (%llu bytes) full."),
          cache_size_cur * 100LL / cache_size_max, cache_size_cur);

    cache_entry_t *e = (cache_entry_t *) l_list_first (filelist);
    while (e && (cache_size_cur > cache_size_max)) {
        if (unlink (e->filename) == -1) {
            error (_("Could not remove '%s': %s!"),
                   e->filename, strerror (errno));
        }
        else {
            cache_size_cur -= e->filesize;
        }
        e = l_list_next (filelist, e);
    }

    l_list_free (filelist, cache_entry_free);

    info (_("The cache directory is %3llu%% (%llu bytes) full."),
          cache_size_cur * 100LL / cache_size_max, cache_size_cur);

    return NULL;
}


void
clean_oxine_cache (void)
{
    if (cache_thread) {
        pthread_join (cache_thread, NULL);
        cache_thread = 0;
    }

    if (pthread_create (&cache_thread, NULL, cache_clean_thread, NULL) != 0) {
        error (_("Could not create cache cleaner thread: %s!"),
               strerror (errno));
        cache_thread = 0;
    }
}


void
environment_free (void)
{
    /* Remove the temporary files. */
    if (dir_oxine_tmp) {
        char *command = ho_strdup_printf ("rm -rf %s", dir_oxine_tmp);
        execute_shell (command, 0);
        ho_free (command);
    }

    if (cache_thread) {
        pthread_join (cache_thread, NULL);
    }

    ho_free (mrl_root);
    ho_free (mrl_home);
    ho_free (mrl_desktop);
    ho_free (mrl_favorites_most);
    ho_free (mrl_favorites_recent);
    ho_free (mrl_playlists);
    ho_free (mrl_mediamarks_music);
    ho_free (mrl_mediamarks_video);
    ho_free (mrl_mediamarks_image);
    ho_free (mrl_mediamarks_media);
    ho_free (mrl_mediamarks_subtitle);

    ho_free (mrl_xdg_music_dir);
    ho_free (mrl_xdg_videos_dir);
    ho_free (mrl_xdg_pictures_dir);

    ho_free (dir_home);
    ho_free (dir_oxine);
    ho_free (dir_oxine_cache);
    ho_free (dir_oxine_playlists);
    ho_free (dir_oxine_skin);
    ho_free (dir_oxine_tmp);

    ho_free (file_config);
    ho_free (file_jsw_config);
    ho_free (file_lirc_config);
    ho_free (file_dvb_channels);
    ho_free (file_favorites);
    ho_free (file_playlist_rw);
    ho_free (file_xdg_user_dirs);
}
