
/*
 * Copyright (C) 2004-2005 Maximilian Schwerin
 *
 * This file is part of oxine a free media player.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: event.c 4023 2010-01-17 10:51:02Z mschwerin $
 *
 */

#include "config.h"

#include <assert.h>
#include <stdlib.h>

#include "disc_fam.h"
#include "environment.h"
#include "event.h"
#include "heap.h"
#include "i18n.h"
#include "logger.h"
#include "oxine.h"
#include "utils.h"

#include "menu_base.h"
#include "menu_dialog.h"
#include "menu_help.h"
#include "menu_filelist.h"
#include "menu_main.h"
#include "menu_playback.h"
#include "menu_playlist.h"
#include "menu_subtitle.h"
#include "menu_settings.h"

extern oxine_t *oxine;


/**
 * This event handler is responsible for handling DVD related keyboard events.
 *
 * @return                      If the keyboard event is consumed by this
 *                              function TRUE is returned.
 */
static bool
oxine_dvd_key_handler (oxine_event_t * event)
{
    if (event->source.key == OXINE_KEY_NULL) {
        return true;
    }
    else if (!odk_current_is_dvd (oxine->odk)) {
        return false;
    }
    else if (odk_get_forward_events_to_xine (oxine->odk)) {
        return false;
    }

    /* we backup the key */
    oxine_key_id_t key_save = event->source.key;
    /* and consume this event */
    event->source.key = OXINE_KEY_NULL;

    switch (key_save) {
    case OXINE_KEY_PREV:
    case OXINE_KEY_NEXT:
        if (odk_current_has_chapters (oxine->odk)) {
            odk_xine_event_send (oxine->odk, true, key_save);
        }
        break;
    case OXINE_KEY_MENU1:
    case OXINE_KEY_MENU2:
    case OXINE_KEY_MENU3:
    case OXINE_KEY_MENU4:
    case OXINE_KEY_MENU5:
    case OXINE_KEY_MENU6:
    case OXINE_KEY_MENU7:
        {
            odk_xine_event_send (oxine->odk, true, key_save);
        }
        break;
    default:
        /* As this was not an event we want to handle, we restore the original
         * key code. */
        event->source.key = key_save;
        break;
    }

    return (event->source.key == OXINE_KEY_NULL);
}


#ifdef HAVE_VDR

/**
 * This event handler is responsible for handling VDR related keyboard events.
 *
 * @return                      If the keyboard event is consumed by this
 *                              function TRUE is returned.
 */
static bool
oxine_vdr_key_handler (oxine_event_t * event)
{
    if (event->source.key == OXINE_KEY_NULL) {
        return true;
    }
    else if (!odk_current_is_vdr (oxine->odk)) {
        return false;
    }

    /* we backup the key */
    oxine_key_id_t key_save = event->source.key;
    /* and consume this event */
    event->source.key = OXINE_KEY_NULL;

    switch (key_save) {
    case OXINE_KEY_ACTIVATE:
    case OXINE_KEY_MENU_HELP:
    case OXINE_KEY_TELEVISION:
    case OXINE_KEY_SKIP:
    case OXINE_KEY_SEEK:
    case OXINE_KEY_SPEED:
    case OXINE_KEY_PLAYMODE:
    case OXINE_KEY_SPU_CHANNEL:
    case OXINE_KEY_SPU_OFFSET:
    case OXINE_KEY_AUDIO_OFFSET:
    case OXINE_KEY_ASPECT_RATIO:
    case OXINE_KEY_EJECT:
    case OXINE_KEY_0:
    case OXINE_KEY_1:
    case OXINE_KEY_2:
    case OXINE_KEY_3:
    case OXINE_KEY_4:
    case OXINE_KEY_5:
    case OXINE_KEY_6:
    case OXINE_KEY_7:
    case OXINE_KEY_8:
    case OXINE_KEY_9:
        /* These keys are ignored when in VDR mode. */
        break;
    case OXINE_KEY_SELECT:
    case OXINE_KEY_PPLAY:
        odk_xine_event_send (oxine->odk, false, XINE_EVENT_VDR_PAUSE);
        break;
    case OXINE_KEY_PLAY:
        odk_xine_event_send (oxine->odk, false, XINE_EVENT_VDR_PLAY);
        break;
    case OXINE_KEY_PAUSE:
        odk_xine_event_send (oxine->odk, false, XINE_EVENT_VDR_PAUSE);
        break;
    case OXINE_KEY_STOP:
        odk_xine_event_send (oxine->odk, false, XINE_EVENT_VDR_STOP);
        break;
    case OXINE_KEY_FASTRWD:
        odk_xine_event_send (oxine->odk, false, XINE_EVENT_VDR_FASTREW);
        break;
    case OXINE_KEY_FASTFWD:
        odk_xine_event_send (oxine->odk, false, XINE_EVENT_VDR_FASTFWD);
        break;
    case OXINE_KEY_PREV:
        odk_xine_event_send (oxine->odk, false, XINE_EVENT_VDR_CHANNELMINUS);
        break;
    case OXINE_KEY_NEXT:
        odk_xine_event_send (oxine->odk, false, XINE_EVENT_VDR_CHANNELPLUS);
        break;
    case OXINE_KEY_BACK:
        odk_xine_event_send (oxine->odk, false, XINE_EVENT_VDR_BACK);
        break;
    case OXINE_KEY_VOLUME:
        if (event->how > 0) {
            odk_xine_event_send (oxine->odk, false, XINE_EVENT_VDR_VOLPLUS);
        }
        else {
            odk_xine_event_send (oxine->odk, false, XINE_EVENT_VDR_VOLMINUS);
        }
        break;
    case OXINE_KEY_VOLUME_MUTE:
        odk_change_stream_param (oxine->odk,
                                 ODK_PARAM_AO_VOLUME_MUTE, 1,
                                 ODK_AO_VOLUME_MUTE_OFF,
                                 ODK_AO_VOLUME_MUTE_ON);
        break;
    case OXINE_KEY_MENU_PLAYBACK:
    case OXINE_KEY_SHOW_OSD:
        odk_xine_event_send (oxine->odk, false, XINE_EVENT_INPUT_MENU1);
        break;
    case OXINE_KEY_AUDIO_CHANNEL:
        odk_xine_event_send (oxine->odk, false, XINE_EVENT_VDR_AUDIO);
        break;
    case OXINE_KEY_DEINTERLACE:
        odk_change_stream_param (oxine->odk, ODK_PARAM_VO_DEINTERLACE,
                                 1, 0, 1);
        break;
    case OXINE_KEY_MENU_SETTINGS:
        odk_xine_event_send (oxine->odk, false, XINE_EVENT_VDR_SETUP);
        break;
    case OXINE_KEY_MENU_MAIN:
    case OXINE_KEY_MENU_MEDIA:
    case OXINE_KEY_MENU_MUSIC:
    case OXINE_KEY_MENU_VIDEO:
    case OXINE_KEY_MENU_IMAGE:
    case OXINE_KEY_MENU_PLAYLIST:
    case OXINE_KEY_QUIT:
    case OXINE_KEY_RESTART:
    case OXINE_KEY_SHUTDOWN:
        /* As we do not want to use the oxine OSD over VDR, we always stop TV
         * playback before showing our menus. */
        odk_xine_event_send (oxine->odk, false, XINE_EVENT_VDR_STOP);
        odk_stop_stream (oxine->odk);
    default:
        /* As this was not an event we want to pass to VDR, we restore the
         * original key code. */
        event->source.key = key_save;
        break;
    }

    return (event->source.key == OXINE_KEY_NULL);
}
#endif


/**
 * This event handler is responsible for handling stream speed related
 * keyboard events.
 *
 * @return                      If the keyboard event is consumed by this
 *                              function TRUE is returned.
 */
static bool
oxine_stream_speed_key_handler (oxine_event_t * event)
{
    if (event->source.key == OXINE_KEY_NULL) {
        return true;
    }
    else if (!odk_current_is_playback_mode (oxine->odk)) {
        return false;
    }

    int cur_speed = odk_get_stream_param (oxine->odk, ODK_PARAM_SPEED);
    int new_speed = cur_speed;

    if (is_current_menu (hide_user_interface)
        || is_current_menu (show_menu_playback)
        || is_current_menu (show_playback_controls)
        || is_current_menu (show_playback_audio_info)
        || is_current_menu (show_playback_video_info)) {
        switch (event->source.key) {
        case OXINE_KEY_SELECT:
            if (cur_speed == ODK_SPEED_PAUSE) {
                new_speed = ODK_SPEED_NORMAL;
            }
            else {
                new_speed = ODK_SPEED_PAUSE;
            }
            break;
        case OXINE_KEY_SPEED:
            if (event->how > 0) {
                if (cur_speed < ODK_SPEED_FAST_4) {
                    new_speed <<= 1;
                }
            }
            else {
                if (cur_speed > ODK_SPEED_SLOW_4) {
                    new_speed >>= 1;
                }
            }
            break;
        default:
            break;
        }
    }

    {
        switch (event->source.key) {
        case OXINE_KEY_PPLAY:
            if (cur_speed == ODK_SPEED_PAUSE) {
                new_speed = ODK_SPEED_NORMAL;
            }
            else {
                new_speed = ODK_SPEED_PAUSE;
            }
            break;
        case OXINE_KEY_PLAY:
            new_speed = ODK_SPEED_NORMAL;
            break;
        case OXINE_KEY_PAUSE:
            new_speed = ODK_SPEED_PAUSE;
            break;
        default:
            break;
        }
    }

    if (new_speed == ODK_SPEED_PAUSE) {
        playlist_set_playstatus (oxine->current_playlist,
                                 PLAYBACK_STATUS_PAUSED);
    }
    else {
        playlist_set_playstatus (oxine->current_playlist,
                                 PLAYBACK_STATUS_PLAYING);
    }

    if (cur_speed != new_speed) {
        odk_set_stream_param (oxine->odk, ODK_PARAM_SPEED, new_speed);
        if (show_stream_parameter (ODK_PARAM_SPEED)) {
            event->hide_gui_timeout = 3000;
        }
    }

    return (event->source.key == OXINE_KEY_NULL);
}


static bool
oxine_stream_seek_key_handler (oxine_event_t * event)
{
    if (event->source.key == OXINE_KEY_NULL) {
        return true;
    }
    else if (!odk_current_is_playback_mode (oxine->odk)) {
        return false;
    }

    int seek = -1;

    /* we backup the key */
    oxine_key_id_t key_save = event->source.key;
    /* and consume this event */
    event->source.key = OXINE_KEY_NULL;

    switch (key_save) {
    case OXINE_KEY_SEEK:
        seek = event->how;
        break;
    case OXINE_KEY_0:
        seek = 0;
        break;
    case OXINE_KEY_1:
        seek = 10;
        break;
    case OXINE_KEY_2:
        seek = 20;
        break;
    case OXINE_KEY_3:
        seek = 30;
        break;
    case OXINE_KEY_4:
        seek = 40;
        break;
    case OXINE_KEY_5:
        seek = 50;
        break;
    case OXINE_KEY_6:
        seek = 60;
        break;
    case OXINE_KEY_7:
        seek = 70;
        break;
    case OXINE_KEY_8:
        seek = 80;
        break;
    case OXINE_KEY_9:
        seek = 90;
        break;
    default:
        /* As this was not an event we want to handle, we restore the original
         * key code. */
        event->source.key = key_save;
        break;
    }

    if (seek != -1) {
        odk_set_stream_param (oxine->odk, ODK_PARAM_SPEED, ODK_SPEED_NORMAL);
        odk_set_stream_param (oxine->odk, ODK_PARAM_POSITION, seek);
        if (show_stream_parameter (ODK_PARAM_POSITION)) {
            event->hide_gui_timeout = 3000;
        }
    }

    return (event->source.key == OXINE_KEY_NULL);
}


#ifdef HAVE_IMAGE_ROTATION

/**
 * This event handler is responsible for handling image playback related
 * keyboard events.
 *
 * @return                      If the keyboard event is consumed by this
 *                              function TRUE is returned.
 */
static bool
oxine_image_key_handler (oxine_event_t * event)
{
    if (event->source.key == OXINE_KEY_NULL) {
        return true;
    }
    else if (!odk_current_is_playback_mode (oxine->odk)) {
        return false;
    }
    else if (!odk_current_is_image (oxine->odk)) {
        return false;
    }

    /* we backup the key */
    oxine_key_id_t key_save = event->source.key;
    /* and consume this event */
    event->source.key = OXINE_KEY_NULL;

    switch (key_save) {
    case OXINE_KEY_MENU1:
    case OXINE_KEY_RED:
        event->how = -90;
    case OXINE_KEY_ROTATE_LEFT:
        if (event->how > 0) {
            event->how *= -1;
        }
        {
            otk_cb_t current_menu = get_current_menu ();
            void *current_menu_data = get_current_menu_data ();
            please_wait (NULL);
            odk_rotate_current_image (oxine->odk, event->how);
            set_current_menu (current_menu, current_menu_data);
            show_menu_current ();
        }
        break;
    case OXINE_KEY_MENU4:
    case OXINE_KEY_BLUE:
        event->how = +90;
    case OXINE_KEY_ROTATE_RIGHT:
        if (event->how < 0) {
            event->how *= -1;
        }
        {
            otk_cb_t current_menu = get_current_menu ();
            void *current_menu_data = get_current_menu_data ();
            please_wait (NULL);
            odk_rotate_current_image (oxine->odk, event->how);
            set_current_menu (current_menu, current_menu_data);
            show_menu_current ();
        }
        break;
    default:
        /* As this was not an event we want to handle, we restore the original
         * key code. */
        event->source.key = key_save;
        break;
    }

    return (event->source.key == OXINE_KEY_NULL);
}
#endif /* HAVE_IMAGE_ROTATION */


/**
 * This event handler is responsible for handling playback related keyboard
 * events.
 *
 * @return                      If the keyboard event is consumed by this
 *                              function TRUE is returned.
 */
static bool
oxine_playback_key_handler (oxine_event_t * event)
{
    if (event->source.key == OXINE_KEY_NULL) {
        return true;
    }
    else if (!odk_current_is_playback_mode (oxine->odk)) {
        return false;
    }
#ifdef HAVE_IMAGE_ROTATION
    else if (oxine_image_key_handler (event)) {
        return true;
    }
#endif
#ifdef HAVE_VDR
    else if (oxine_vdr_key_handler (event)) {
        return true;
    }
#endif
    else if (oxine_dvd_key_handler (event)) {
        return true;
    }
    else if (oxine_stream_speed_key_handler (event)) {
        return true;
    }
    else if (oxine_stream_seek_key_handler (event)) {
        return true;
    }

    switch (event->source.key) {
    case OXINE_KEY_STOP:
        odk_stop_stream (oxine->odk);
        break;
    case OXINE_KEY_PREV:
        if (playlist_get_current_pos (oxine->current_playlist) == 0) {
            odk_set_stream_param (oxine->odk, ODK_PARAM_POSITION, 0);
        }
        else if (odk_get_stream_param (oxine->odk,
                                       ODK_PARAM_POSITION_MSECS) > 2000) {
            odk_set_stream_param (oxine->odk, ODK_PARAM_POSITION, 0);
        }
        else {
            playlist_play_prev (event->how);
        }
        break;
    case OXINE_KEY_NEXT:
        if (!playlist_current_is_last (oxine->current_playlist)) {
            playlist_play_next (event->how);
        }
        break;
    case OXINE_KEY_VOLUME:
        odk_change_stream_param (oxine->odk, ODK_PARAM_AO_VOLUME,
                                 event->how, 0, 100);
        if (show_stream_parameter (ODK_PARAM_AO_VOLUME)) {
            event->hide_gui_timeout = 3000;
        }
        break;
    case OXINE_KEY_VOLUME_MUTE:
        odk_change_stream_param (oxine->odk,
                                 ODK_PARAM_AO_VOLUME_MUTE, 1,
                                 ODK_AO_VOLUME_MUTE_OFF,
                                 ODK_AO_VOLUME_MUTE_ON);
        if (show_stream_parameter (ODK_PARAM_AO_VOLUME_MUTE)) {
            event->hide_gui_timeout = 3000;
        }
        break;
    case OXINE_KEY_AUDIO_CHANNEL:
        {
            int max_channels = odk_get_stream_info (oxine->odk,
                                                    ODK_STREAM_INFO_MAX_AUDIO_CHANNEL);
            if (max_channels >= 1) {
                odk_change_stream_param (oxine->odk,
                                         ODK_PARAM_AUDIO_CHANNEL,
                                         event->how, -2, max_channels - 1);
                if (show_stream_parameter (ODK_PARAM_AUDIO_CHANNEL)) {
                    event->hide_gui_timeout = 3000;
                }
            }
        }
        break;
    case OXINE_KEY_AUDIO_OFFSET:
        odk_change_stream_param (oxine->odk,
                                 ODK_PARAM_AUDIO_OFFSET,
                                 event->how, -90000, 90000);
        if (show_stream_parameter (ODK_PARAM_AUDIO_OFFSET)) {
            event->hide_gui_timeout = 3000;
        }
        break;
    case OXINE_KEY_FASTFWD:
        odk_set_stream_param (oxine->odk, ODK_PARAM_SPEED, ODK_SPEED_FAST_2);
        if (show_stream_parameter (ODK_PARAM_SPEED)) {
            event->hide_gui_timeout = 3000;
        }
        break;
    case OXINE_KEY_FASTRWD:
        warn ("xine-lib does not yet support fast rewind.");
        break;
    case OXINE_KEY_SKIP:
        odk_set_stream_param (oxine->odk, ODK_PARAM_SPEED, ODK_SPEED_NORMAL);
        odk_change_stream_param (oxine->odk, ODK_PARAM_POSITION_MSECS,
                                 event->how * 1000, 0, 0);
        if (show_stream_parameter (ODK_PARAM_POSITION)) {
            event->hide_gui_timeout = 3000;
        }
        break;
    case OXINE_KEY_SATURATION:
        odk_change_stream_param (oxine->odk, ODK_PARAM_VO_SATURATION,
                                 event->how, 0, 65535);
        if (show_stream_parameter (ODK_PARAM_VO_SATURATION)) {
            event->hide_gui_timeout = 3000;
        }
        break;
    case OXINE_KEY_BRIGHTNESS:
        odk_change_stream_param (oxine->odk, ODK_PARAM_VO_BRIGHTNESS,
                                 event->how, 0, 65535);
        if (show_stream_parameter (ODK_PARAM_VO_BRIGHTNESS)) {
            event->hide_gui_timeout = 3000;
        }
        break;
    case OXINE_KEY_CONTRAST:
        odk_change_stream_param (oxine->odk, ODK_PARAM_VO_CONTRAST,
                                 event->how, 0, 65535);
        if (show_stream_parameter (ODK_PARAM_VO_CONTRAST)) {
            event->hide_gui_timeout = 3000;
        }
        break;
    case OXINE_KEY_HUE:
        odk_change_stream_param (oxine->odk, ODK_PARAM_VO_HUE,
                                 event->how, 0, 65535);
        if (show_stream_parameter (ODK_PARAM_VO_HUE)) {
            event->hide_gui_timeout = 3000;
        }
        break;
    case OXINE_KEY_MENU_PLAYBACK:
        if (is_current_menu (show_menu_playback)) {
            if (odk_current_has_video (oxine->odk)) {
                hide_user_interface (oxine);
                set_backto_menu (hide_user_interface, NULL);
            }
            else {
                show_playback_audio_info (oxine);
            }
        }
        else {
            show_menu_playback (oxine);
        }
        break;
    case OXINE_KEY_SHOW_OSD:
        if (is_current_menu (hide_user_interface)
            || is_current_menu (show_playback_controls)
            || is_current_menu (show_playback_audio_info)
            || is_current_menu (show_playback_video_info)) {
            show_menu_playback (oxine);
            event->source.key = OXINE_KEY_NULL;
        }
        else {
            if (odk_current_has_video (oxine->odk)) {
                hide_user_interface (oxine);
                set_backto_menu (hide_user_interface, NULL);
            }
            else {
                show_playback_audio_info (oxine);
            }
            event->source.key = OXINE_KEY_NULL;
        }
        break;
    case OXINE_KEY_BACK:
        if (is_current_menu (hide_user_interface)
            || is_current_menu (show_playback_controls)
            || is_current_menu (show_playback_audio_info)
            || is_current_menu (show_playback_video_info)) {
            show_menu_playback (oxine);
            event->source.key = OXINE_KEY_NULL;
        }
        else if (is_current_menu (show_menu_playback)) {
            if (odk_current_has_video (oxine->odk)) {
                hide_user_interface (oxine);
                set_backto_menu (hide_user_interface, NULL);
            }
            else {
                show_playback_audio_info (oxine);
            }
            event->source.key = OXINE_KEY_NULL;
        }
        break;
    default:
        break;
    }

    switch (event->source.key) {
    case OXINE_KEY_ACTIVATE:
        if (!oxine->stream_parameter_is_visible) {
            int l = odk_get_stream_param (oxine->odk, ODK_PARAM_LENGTH_MSECS);
            if ((l < 1000 * 60 * 2) && odk_current_is_dvd (oxine->odk)) {
                /* If the current stream is a DVD and the length of the stream
                 * is less than 2 minutes we assume that we are in a DVD menu
                 * and do not want the stream parameter to be shown. */
            }
            else if (show_stream_parameter (ODK_PARAM_SPEED)) {
                event->hide_gui_timeout = 3000;
            }
            event->source.key = OXINE_KEY_NULL;
        }
        else if (odk_current_has_video (oxine->odk)) {
            hide_user_interface (oxine);
            set_backto_menu (hide_user_interface, NULL);
            event->source.key = OXINE_KEY_NULL;
        }
        else {
            show_playback_audio_info (oxine);
            event->source.key = OXINE_KEY_NULL;
        }
        break;
    default:
        break;
    }

    /* If the current stream does not have a video component we quit at this
     * point as none of the following events are relevant for streams
     * without video. */
    if (!odk_current_has_video (oxine->odk)) {
        return (event->source.key == OXINE_KEY_NULL);
    }

    /* If we're currently playing an image we quit at this point as none of
     * the following events are relevant for image viewing. */
    if (odk_current_is_image (oxine->odk)) {
        return (event->source.key == OXINE_KEY_NULL);
    }

    switch (event->source.key) {
    case OXINE_KEY_ZOOM:
        odk_change_stream_param (oxine->odk, ODK_PARAM_VO_ZOOM_X,
                                 event->how, 0, 400);
        odk_change_stream_param (oxine->odk, ODK_PARAM_VO_ZOOM_Y,
                                 event->how, 0, 400);
        if (show_stream_parameter (ODK_PARAM_VO_ZOOM_X)) {
            event->hide_gui_timeout = 3000;
        }
        break;
    case OXINE_KEY_ZOOM_X:
        odk_change_stream_param (oxine->odk, ODK_PARAM_VO_ZOOM_X,
                                 event->how, 0, 400);
        if (show_stream_parameter (ODK_PARAM_VO_ZOOM_X)) {
            event->hide_gui_timeout = 3000;
        }
        break;
    case OXINE_KEY_ZOOM_Y:
        odk_change_stream_param (oxine->odk, ODK_PARAM_VO_ZOOM_Y,
                                 event->how, 0, 400);
        if (show_stream_parameter (ODK_PARAM_VO_ZOOM_X)) {
            event->hide_gui_timeout = 3000;
        }
        break;
    case OXINE_KEY_ZOOM_RESET:
        odk_set_stream_param (oxine->odk, ODK_PARAM_VO_ZOOM_X, 100);
        odk_set_stream_param (oxine->odk, ODK_PARAM_VO_ZOOM_Y, 100);
        if (show_stream_parameter (ODK_PARAM_VO_ZOOM_X)) {
            event->hide_gui_timeout = 3000;
        }
        break;
    case OXINE_KEY_SPU_CHANNEL:
        {
            int max_channels = odk_get_stream_info (oxine->odk,
                                                    ODK_STREAM_INFO_MAX_SPU_CHANNEL);
            if (max_channels >= 1) {
                odk_change_stream_param (oxine->odk,
                                         ODK_PARAM_SPU_CHANNEL,
                                         event->how, -2, max_channels - 1);
                if (show_stream_parameter (ODK_PARAM_SPU_CHANNEL)) {
                    event->hide_gui_timeout = 3000;
                }
            }
        }
        break;
    case OXINE_KEY_SPU_OFFSET:
        odk_change_stream_param (oxine->odk, ODK_PARAM_SPU_OFFSET,
                                 event->how, -90000, 90000);
        if (show_stream_parameter (ODK_PARAM_SPU_OFFSET)) {
            event->hide_gui_timeout = 3000;
        }
        break;
    case OXINE_KEY_ASPECT_RATIO:
        odk_change_stream_param (oxine->odk,
                                 ODK_PARAM_VO_ASPECT_RATIO, 1,
                                 ODK_VO_ASPECT_RATIO_AUTO,
                                 ODK_VO_ASPECT_RATIO_DVB);
        if (show_stream_parameter (ODK_PARAM_VO_ASPECT_RATIO)) {
            event->hide_gui_timeout = 3000;
        }
        break;
    case OXINE_KEY_DEINTERLACE:
        odk_change_stream_param (oxine->odk,
                                 ODK_PARAM_VO_DEINTERLACE, 1,
                                 ODK_VO_DEINTERLACE_OFF,
                                 ODK_VO_DEINTERLACE_ON);
        if (show_stream_parameter (ODK_PARAM_VO_DEINTERLACE)) {
            event->hide_gui_timeout = 3000;
        }
        break;
    default:
        break;
    }

    return (event->source.key == OXINE_KEY_NULL);
}


/**
 * This event handler is responsible for handling keyboard events.
 */
static void
oxine_key_handler (oxine_event_t * event)
{
    /* Evaluate all playback related keypress events. */
    if (oxine_playback_key_handler (event)) {
        return;
    }

    switch (event->source.key) {
    case OXINE_KEY_MENU_MAIN:
        if (is_current_menu (show_menu_main)
            && odk_current_is_playback_mode (oxine->odk)) {
            show_menu_playback (oxine);
        }
        else {
            show_menu_main (oxine);
        }
        break;
    case OXINE_KEY_MENU_MEDIA:
        if (is_current_menu (show_menu_media)
            && odk_current_is_playback_mode (oxine->odk)) {
            show_menu_playback (oxine);
        }
        else {
            show_menu_media (oxine);
        }
        break;
    case OXINE_KEY_MENU_MUSIC:
        if (is_current_menu (show_menu_music)
            && odk_current_is_playback_mode (oxine->odk)) {
            show_menu_playback (oxine);
        }
        else {
            show_menu_music (oxine);
        }
        break;
    case OXINE_KEY_MENU_VIDEO:
        if (is_current_menu (show_menu_video)
            && odk_current_is_playback_mode (oxine->odk)) {
            show_menu_playback (oxine);
        }
        else {
            show_menu_video (oxine);
        }
        break;
#ifdef HAVE_IMAGE_PLAYBACK
    case OXINE_KEY_MENU_IMAGE:
        if (is_current_menu (show_menu_image)
            && odk_current_is_playback_mode (oxine->odk)) {
            show_menu_playback (oxine);
        }
        else {
            show_menu_image (oxine);
        }
        break;
#endif
    case OXINE_KEY_MENU_SETTINGS:
        if (is_current_menu (show_menu_settings)
            && odk_current_is_playback_mode (oxine->odk)) {
            show_menu_playback (oxine);
        }
        else {
            show_menu_settings (oxine);
        }
        break;
    case OXINE_KEY_MENU_PLAYLIST:
        if (is_current_menu (show_menu_playlist)
            && odk_current_is_playback_mode (oxine->odk)) {
            show_menu_playback (oxine);
        }
        else {
            show_menu_playlist (oxine);
        }
        break;
    case OXINE_KEY_MENU_HELP:
        if (is_current_menu (show_menu_help)
            && odk_current_is_playback_mode (oxine->odk)) {
            show_menu_playback (oxine);
        }
        else {
            show_menu_help (oxine);
        }
        break;
    case OXINE_KEY_TELEVISION:
        odk_stop_stream (oxine->odk);
        play_tv_cb (oxine);
        break;
    case OXINE_KEY_FULLSCREEN:
        odk_window_set_fullscreen (oxine->odk,
                                   !odk_window_is_fullscreen (oxine->odk));
        break;
    case OXINE_KEY_WINDOW_ICONIFY:
        odk_window_iconify (oxine->odk);
        break;
    case OXINE_KEY_WINDOW_RAISE:
        odk_window_raise (oxine->odk);
        break;
    case OXINE_KEY_WINDOW_LOWER:
        odk_window_lower (oxine->odk);
        break;
    case OXINE_KEY_BACK:
        backto_menu_cb (oxine);
        break;
    case OXINE_KEY_EJECT:
        eject_cb (oxine);
        break;
    case OXINE_KEY_SIDEBAR:
        toggle_sidebar ();
        break;
    case OXINE_KEY_PLAYMODE:
        switch (playlist_get_playmode (oxine->rw_playlist)) {
        case PLAYBACK_MODE_NORMAL:
            playlist_set_playmode (oxine->rw_playlist, PLAYBACK_MODE_REPEAT);
            break;
        case PLAYBACK_MODE_REPEAT:
            playlist_set_playmode (oxine->rw_playlist, PLAYBACK_MODE_RANDOM);
            break;
        case PLAYBACK_MODE_RANDOM:
            playlist_set_playmode (oxine->rw_playlist, PLAYBACK_MODE_NORMAL);
            break;
        }
        show_menu_current ();
        break;
    case OXINE_KEY_RESTART:
        oxine->restart = true;
    case OXINE_KEY_QUIT:
        playlist_rw_change_cb (oxine->rw_playlist);
        odk_exit (oxine->odk);
        break;
    case OXINE_KEY_SHUTDOWN:
        playlist_rw_change_cb (oxine->rw_playlist);
        shutdown_cb (oxine);
        break;
    default:
        break;
    }
}


/**
 * This event handler is responsible for reacting to mouse button events.
 */
static void
oxine_button_handler (oxine_event_t * event)
{
    switch (event->source.button) {
    case OXINE_MOUSE_BUTTON_RIGHT:
    case OXINE_MOUSE_BUTTON_MIDDLE:
        {
            oxine_event_t event;
            oxine_event_init (&event);
            event.type = OXINE_EVENT_KEY;
            event.source.key = OXINE_KEY_SHOW_OSD;
            odk_oxine_event_send (oxine->odk, &event);
        }
        break;
    case OXINE_MOUSE_SCROLLWHEEL_UP:
        {
            oxine_event_t event;
            oxine_event_init (&event);
            event.type = OXINE_EVENT_KEY;
            event.source.key = OXINE_KEY_VOLUME;
            event.how = +5;
            odk_oxine_event_send (oxine->odk, &event);
        }
        break;
    case OXINE_MOUSE_SCROLLWHEEL_DOWN:
        {
            oxine_event_t event;
            oxine_event_init (&event);
            event.type = OXINE_EVENT_KEY;
            event.source.key = OXINE_KEY_VOLUME;
            event.how = -5;
            odk_oxine_event_send (oxine->odk, &event);
        }
        break;
    case OXINE_MOUSE_BUTTON_6:
        {
            oxine_event_t event;
            oxine_event_init (&event);
            event.type = OXINE_EVENT_KEY;
            event.source.key = OXINE_KEY_PREV;
            odk_oxine_event_send (oxine->odk, &event);
        }
        break;
    case OXINE_MOUSE_BUTTON_7:
        {
            oxine_event_t event;
            oxine_event_init (&event);
            event.type = OXINE_EVENT_KEY;
            event.source.key = OXINE_KEY_NEXT;
            odk_oxine_event_send (oxine->odk, &event);
        }
        break;
    default:
        break;
    }
}


/**
 * This event handler is responsible for reacting to mouse movements.
 */
static void
oxine_motion_handler (oxine_event_t * event)
{
#ifdef HAVE_VDR
    if (odk_current_is_vdr (oxine->odk)) {
        return;
    }
#endif

    int y_show = odk_osd_get_height (oxine->odk) - 50;
    int y_hide = odk_osd_get_height (oxine->odk) - 100;
    if (odk_current_is_playback_mode (oxine->odk)) {
        if (is_current_menu (hide_user_interface)
            || is_current_menu (show_playback_controls)
            || is_current_menu (show_playback_video_info)) {
            if (!oxine->playback_controls_are_visible
                && (event->data.mouse.pos.y > y_show)) {
                show_playback_controls (oxine);
                const int timeout = 6 * 1000;
                schedule_hide_gui_job (timeout);
            }
            else if (oxine->playback_controls_are_visible
                     && (event->data.mouse.pos.y > y_hide)) {
                const int timeout = 6 * 1000;
                schedule_hide_gui_job (timeout);
            }
            else if (oxine->playback_controls_are_visible
                     && (event->data.mouse.pos.y < y_hide)) {
                hide_user_interface (oxine);
            }
        }
        else if (is_current_menu (show_playback_audio_info)) {
            if (event->data.mouse.pos.y > y_show) {
                show_menu_playback (oxine);
            }
        }
        else if (is_current_menu (show_menu_playback)) {
            const int y_hide = 20;
            if (event->data.mouse.pos.y < y_hide) {
                if (!odk_current_has_video (oxine->odk)) {
                    show_playback_audio_info (oxine);
                }
                else {
                    hide_user_interface (oxine);
                }
            }
        }
    }

    int x_show = 20;
    int x_hide = odk_osd_get_width (oxine->odk) - 20;
    if (event->data.mouse.pos.x < x_show) {
        show_sidebar ();
    }
    if (event->data.mouse.pos.x > x_hide) {
        hide_sidebar ();
    }
}


/**
 * This is oxine's main event handler. All events generated by X, xine-lib
 * or oxine itself pass through here. 
 *
 * If an underlying level has already reacted to an event (consumed the
 * event) the event source may be OXINE_KEY_NULL or OXINE_MOUSE_BUTTON_NULL.
 * In that case the event is purely informational, no further action can be
 * taken on it.
 */
void
oxine_event_handler (void *p, oxine_event_t * event)
{
    switch (event->type) {
    case OXINE_EVENT_KEY:
        oxine_key_handler (event);
        break;
    case OXINE_EVENT_MOTION:
        oxine_motion_handler (event);
        break;
    case OXINE_EVENT_BUTTON:
        oxine_button_handler (event);
        break;
    case OXINE_EVENT_PLAYBACK_STARTED:
        {
            playlist_set_playstatus (oxine->rw_playlist,
                                     PLAYBACK_STATUS_STOPPED);
            playlist_set_playstatus (oxine->ro_playlist,
                                     PLAYBACK_STATUS_STOPPED);
        }
        if (odk_current_is_playback_mode (oxine->odk)) {
            playlist_set_playstatus (oxine->current_playlist,
                                     PLAYBACK_STATUS_PLAYING);
        }
        /* We ignore this event if we're in logo mode. */
        if (!odk_current_is_playback_mode (oxine->odk)) {
            /* do nothing */
        }
        /* If the stream is a TV, DVD, or VCD stream we hide the GUI. */
        else if (odk_current_is_television (oxine->odk)
                 || odk_current_is_dvd (oxine->odk)
                 || odk_current_is_vcd (oxine->odk)) {
            hide_user_interface (oxine);
        }
        /* If this stream has video we show the video info menu. */
        else if (odk_current_has_video (oxine->odk)) {
            if (odk_current_is_image (oxine->odk)) {
                show_menu_current ();
            }
            else if (is_current_menu (hide_user_interface)) {
                show_playback_video_info (oxine);
            }
            else {
                show_menu_current ();
            }
        }
        /* If this stream does not have video we show the audio info menu. */
        else {
            if (is_current_menu (hide_user_interface)) {
                show_playback_audio_info (oxine);
            }
            else {
                show_menu_current ();
            }
        }
        break;
    case OXINE_EVENT_PLAYBACK_STOPPED:
        playlist_set_playstatus (oxine->current_playlist,
                                 PLAYBACK_STATUS_STOPPED);
        playback_ended_menu_cb (oxine);
        break;
    case OXINE_EVENT_PLAYBACK_FINISHED_WITH_ALTERNATIVES:
        playlist_set_playstatus (oxine->current_playlist,
                                 PLAYBACK_STATUS_STOPPED);
#ifdef XINE_PARAM_GAPLESS_SWITCH
        if (xine_check_version (1, 1, 1)) {
            odk_set_stream_param (oxine->odk, ODK_PARAM_GAPLESS, 1);
        }
#endif
        playlist_play_item (oxine->current_playlist,
                            oxine->current_playlist->current);
        break;
    case OXINE_EVENT_PLAYBACK_FINISHED:
        playlist_set_playstatus (oxine->current_playlist,
                                 PLAYBACK_STATUS_STOPPED);
#ifdef XINE_PARAM_GAPLESS_SWITCH
        if (xine_check_version (1, 1, 1)) {
            odk_set_stream_param (oxine->odk, ODK_PARAM_GAPLESS, 1);
        }
#endif
        playlist_play_next (0);
        break;
    case OXINE_EVENT_PLAYBACK_ERROR:
        show_message_dialog ((otk_cb_t) playlist_play_next, oxine,
                             (otk_cb_t) playback_ended_menu_cb, oxine,
                             DIALOG_OK_CANCEL, NULL,
                             _("Could not play title!"));
        break;
    case OXINE_EVENT_GUI_REPAINT:
        show_menu_current ();
        break;
    case OXINE_EVENT_GUI_HIDE:
        if (odk_current_is_playback_mode (oxine->odk)) {
            if (odk_current_has_video (oxine->odk)) {
                hide_user_interface (oxine);
            }
            else {
                show_playback_audio_info (oxine);
            }
        }
        break;
    case OXINE_EVENT_ADD_ALTERNATIVE:
        if (odk_current_is_playback_mode (oxine->odk)) {
            playlist_t *currentlist = oxine->current_playlist;
            playlist_lock (currentlist);
            playitem_t *currentitem = playlist_get_current (currentlist);
            /* If the current item already has an alternative list, and this
             * is the first alternative we get, we are playing the first
             * alternative from the existing list. As this item gives us
             * alternatives, we remove it. */
            if (currentitem && currentitem->alternative_list
                && (event->data.alternative.pos == 0)) {
                playlist_t *altlist = currentitem->alternative_list;
                playitem_t *altitem = playlist_first (altlist);
                playlist_remove (altlist, altitem);
            }
            if (currentitem) {
                playitem_insert_alternative (currentitem,
                                             event->data.alternative.mrl,
                                             event->data.alternative.title,
                                             event->data.alternative.pos);
            }
            playlist_unlock (currentlist);
        }
        break;
    default:
        break;
    }

    /* When we're in playback mode and are showing a menu we start a timeout
     * of 60 seconds after which we hide the user interface if we're playing a
     * stream with video or show the audio info if it's an audio only stream. */
    switch (event->type) {
    case OXINE_EVENT_OTK_UPDATE:
    case OXINE_EVENT_GUI_HIDE:
    case OXINE_EVENT_PLAYBACK_STOPPED:
    case OXINE_EVENT_PLAYBACK_ERROR:
    case OXINE_EVENT_ADD_ALTERNATIVE:
        break;
    case OXINE_EVENT_KEY:
        if (event->source.key == OXINE_KEY_INPUT) {
            break;
        }
    default:
        if (!odk_current_is_playback_mode (oxine->odk)) {
            /* do nothing */
        }
        else if (is_current_menu (show_playback_video_info)) {
            int timeout = 10 * 1000;
            if (event->hide_gui_timeout) {
                timeout = event->hide_gui_timeout;
            }
            schedule_hide_gui_job (timeout);
        }
        else {
            int timeout = 60 * 1000;
            if (event->hide_gui_timeout) {
                timeout = event->hide_gui_timeout;
            }
            schedule_hide_gui_job (timeout);
        }
        break;
    }

    /* We only want the file alteration monitor to poll folders when we're
     * in one of the filelist menus as we don't care about changes if we're
     * not. This only has an effect on folders we poll ourselves. Folders
     * monitored by FAM are not effected. */
    bool enable_polling = false;
    enable_polling |= filelist_menu_is_current_menu ();
    enable_polling |= is_current_menu (show_menu_subtitle);
    fam_monitor_poll (enable_polling);
}
