
/*
 * Copyright (C) 2002-2003 Stefan Holst
 * Copyright (C) 2004-2005 Maximilian Schwerin
 *
 * This file is part of oxine a free media player.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: filelist.c 4033 2010-09-04 22:51:53Z mschwerin $
 *
 */

#include "config.h"

#include <assert.h>
#include <ctype.h>
#include <crypt.h>
#include <errno.h>
#include <dirent.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>

#include "filelist.h"
#include "disc.h"
#include "disc_image.h"
#include "environment.h"
#include "i18n.h"
#include "heap.h"
#include "list.h"
#include "mutex.h"
#include "logger.h"
#include "mediamarks.h"
#include "playlist_m3u.h"
#include "playlist_xml.h"
#include "utils.h"
#include "vdr.h"

//#define PROFILE_FILELIST_EXPAND

#define INFO_LIST_FORMAT        "<tr>" \
                                "<td width=150 align=right>%s</td>" \
                                "<td width=10></td><td>%s</td>" \
                                "</tr>"

extern oxine_t *oxine;

static filelist_t *toplevel_list = NULL;

static fileitem_len_cb_t fileitem_len_cb = NULL;
static fileitem_pwd_cb_t fileitem_pwd_cb = NULL;

static char *xine_extensions = NULL;

static l_list_t *allowed_xine_extensions;
static l_list_t *allowed_audio_extensions;
static l_list_t *allowed_video_extensions;
static l_list_t *allowed_image_extensions;
static l_list_t *allowed_subtitle_extensions;
static l_list_t *allowed_playlist_extensions;

static char *audio_extensions[] = {
    "16sv", "16sv", "669", "8svx", "8svx", "aac", "abs", "ac3", "aif",
    "aiff", "amf", "anx", "au", "aud", "axa", "axv", "cak", "cpk", "dat",
    "dps", "dts", "film", "flac", "iff", "ik2", "iki", "it", "m4a", "mdl",
    "med", "mka", "mod", "mp+", "mp2", "mp3", "mpa", "mpc", "mpp", "mpega",
    "nsf", "oga", "ogg", "ogx", "ra", "ram", "rpm", "s3m", "shn", "snd",
    "spx", "stm", "str", "svx", "tta", "voc", "vox", "wav", "wma", "wv",
    "wve", "wvp", "xa", "xa1", "xa2", "xap", "xas", "xm", NULL
};

static char *video_extensions[] = {
    "4xm", "anim", "anim", "anim3", "anim5", "anim7", "anim8", "anx", "asf",
    "asp", "avi", "axa", "axv", "bin", "cak", "cin", "cpk", "cue",
    "dat", "dif", "divx", "dv", "film", "flc", "fli", "flv", "iff", "iki",
    "ik2", "img", "iso", "dps", "m2p", "m2t", "m2v", "m4b", "mdf", "mjpg",
    "mkv", "mng", "mov", "mp4", "mpe", "mpeg", "mpg", "mpv", "mv8", "mve",
    "nsv", "nrg", "ogm", "ogv", "ogx", "pes", "pva", "qt", "ram", "rm",
    "rmvb", "roq", "rpm", "rv", "spx", "str", "toc", "trp", "ts", "vmd",
    "vob", "vqa", "wma", "wmv", "wva", "wvx", "xa", "xa1", "xa2",
    "xap", "xas", "y4m", NULL
};

static char *image_extensions[] = {
    "gif", "ham", "ham6", "ham8", "iff", "ilbm", "jpeg", "jpg", "png",
    NULL
};

static char *subtitle_extensions[] = {
    "asc", "smi", "srt", "ssa", "sub", "txt",
    NULL
};

static char *playlist_extensions[] = {
    "asx", "m3u", "pls", "oxp", "wax",
    NULL
};


static char *mediamarks_extensions[] = {
    "oxm",
    NULL
};


static bool
ext_in_list (const char *ext, char *list[])
{
    int i = 0;
    for (; list[i]; i++) {
        if (strcasecmp (list[i], ext) == 0) {
            return true;
        }
    }
    return false;
}


void
filelist_static_init (void)
{
    allowed_xine_extensions = l_list_new ();
    allowed_audio_extensions = l_list_new ();
    allowed_video_extensions = l_list_new ();
    allowed_image_extensions = l_list_new ();
    allowed_subtitle_extensions = l_list_new ();
    allowed_playlist_extensions = l_list_new ();

    char *exts = xine_get_file_extensions (oxine->xine);
    xine_extensions = ho_strdup (exts);

    char *p = xine_extensions;
    while (p) {
        char *q = index (p, ' ');
        if (q) {
            q[0] = '\0';
            q++;
        }
        if (strlen (p) == 0) {
            p = q;
            continue;
        }

        l_list_append (allowed_xine_extensions, p);

        /* It is possible, that some extensions are valid for more than one
         * media type. Thus we have to test all extensions against all media
         * types (audio, video, image, subtitle). */
        bool found = false;
        if (ext_in_list (p, audio_extensions)) {
            l_list_append (allowed_audio_extensions, p);
            found = true;
        }
        if (ext_in_list (p, video_extensions)) {
            l_list_append (allowed_video_extensions, p);
            found = true;
        }
        if (ext_in_list (p, image_extensions)) {
            l_list_append (allowed_image_extensions, p);
            found = true;
        }
        if (ext_in_list (p, subtitle_extensions)) {
            l_list_append (allowed_subtitle_extensions, p);
            found = true;
        }
        if (ext_in_list (p, playlist_extensions)) {
            found = true;
        }
        if (!found) {
            debug ("Unknown file extension: '%s'.", p);
        }

        p = q;
    }

    int i = 0;
    for (i = 0; playlist_extensions[i]; i++) {
        char *ext = playlist_extensions[i];
        l_list_append (allowed_playlist_extensions, ext);
    }

    free (exts);
}


void
filelist_static_free (void)
{
    l_list_free (allowed_xine_extensions, NULL);
    l_list_free (allowed_audio_extensions, NULL);
    l_list_free (allowed_video_extensions, NULL);
    l_list_free (allowed_image_extensions, NULL);
    l_list_free (allowed_subtitle_extensions, NULL);
    l_list_free (allowed_playlist_extensions, NULL);
    ho_free (xine_extensions);

    ho_setref (toplevel_list, NULL);

#ifdef HAVE_VDR
    vdr_static_free ();
#endif
}


static bool
mrl_matches_extensions (const char *mrl, l_list_t * allowed_extensions)
{
    if (!mrl || !allowed_extensions) {
        return false;
    }
    const char *ext = get_extension (mrl);
    if (!ext) {
        return false;
    }

    char *allowed = l_list_first (allowed_extensions);
    while (allowed) {
        if (strcasecmp (allowed, ext) == 0) {
            return true;
        }
        allowed = l_list_next (allowed_extensions, allowed);
    }

    return false;
}


bool
mrl_is_audio (const char *mrl)
{
    if (mrl_matches_extensions (mrl, allowed_audio_extensions))
        return true;
    if (mrl_has_protocol (mrl, "cdda"))
        return true;

    return false;
}


bool
mrl_is_video (const char *mrl)
{
    if (mrl_matches_extensions (mrl, allowed_video_extensions))
        return true;
    if (mrl_has_protocol (mrl, "dvd"))
        return true;
    if (mrl_has_protocol (mrl, "vcd"))
        return true;
#ifdef HAVE_VDR
    if (mrl_has_protocol (mrl, "vdr"))
        return true;
#endif
    if (mrl_is_directory_dvd (mrl))
        return true;
    if (mrl_is_directory_vcd (mrl))
        return true;

    return false;
}


bool
mrl_is_image (const char *mrl)
{
    return mrl_matches_extensions (mrl, allowed_image_extensions);
}


bool
mrl_is_subtitle (const char *mrl)
{
    return mrl_matches_extensions (mrl, allowed_subtitle_extensions);
}


bool
mrl_is_playlist (const char *mrl)
{
    return mrl_matches_extensions (mrl, allowed_playlist_extensions);
}


static bool
mrl_is_multimedia (const char *mrl)
{
    bool allowed = false;

    allowed |= mrl_is_audio (mrl);
    allowed |= mrl_is_video (mrl);
#ifdef HAVE_IMAGE_PLAYBACK
    allowed |= mrl_is_image (mrl);
#endif

    return allowed;
}


bool
fileitem_is_allowed (fileitem_t * fileitem, fileitem_allowed_t allowed)
{
    if (!fileitem) {
        return false;
    }
    if (fileitem->type == FILE_TYPE_FOLDER) {
        return true;
    }
    if (fileitem->type == FILE_TYPE_MEDIAMARKS) {
        return true;
    }
    if (fileitem->type == FILE_TYPE_PLAYLIST) {
        return true;
    }

    char *mrl = fileitem->mrl;
    if (mrl_is_hidden (mrl)) {
        return config_get_bool ("media.files.show_hidden_files");
    }
    if (mrl_is_directory (mrl)
        && !mrl_is_directory_vcd (mrl)
        && !mrl_is_directory_dvd (mrl)) {
        return true;
    }
    if (mrl_is_playlist_oxp (mrl)
        || mrl_is_playlist_m3u (mrl)
        || mrl_is_playlist_pls (mrl)) {
        return true;
    }
    if (mrl_is_mediamarks_oxm (mrl)) {
        return true;
    }

    switch (allowed) {
    case ALLOW_FILES_ALL:
        return true;
    case ALLOW_FILES_MULTIMEDIA:
        return mrl_is_multimedia (mrl);
    case ALLOW_FILES_SUBTITLE:
        return mrl_is_subtitle (mrl);
    case ALLOW_FILES_MUSIC:
        return mrl_is_audio (mrl);
    case ALLOW_FILES_VIDEO:
        return mrl_is_video (mrl);
#ifdef HAVE_IMAGE_PLAYBACK
    case ALLOW_FILES_IMAGE:
        return mrl_is_image (mrl);
#endif
    default:
        break;
    }

    return false;
}


static bool
filelist_is_locked (filelist_t * filelist)
{
#ifdef DEBUG
    return (!filelist || filelist->mutex.__data.__count > 0);
#else
    return true;
#endif
}


static char *
clean_sort_by (char *in)
{
    char *tmp = in;

    if (tmp == NULL) {
        return tmp;
    }

    while (tmp[0] == ' ') {
        tmp = &tmp[1];
    }
    while (tmp[0] == '\t') {
        tmp = &tmp[1];
    }
    while (tmp[0] == '\'') {
        tmp = &tmp[1];
    }
    while (tmp[0] == '\"') {
        tmp = &tmp[1];
    }

    char *out = ho_strdup (tmp);
    ho_free (in);

    return out;
}


bool
fileitem_swap_cb (void *d1, void *d2)
{
    fileitem_t *f1 = (fileitem_t *) d1;
    fileitem_t *f2 = (fileitem_t *) d2;

    assert (f1);
    assert (f2);

#ifdef DEBUG
    assert (mrl_is_valid (f1->mrl));
    assert (mrl_is_valid (f2->mrl));
#endif

    f1->sort_by = clean_sort_by (f1->sort_by);
    f2->sort_by = clean_sort_by (f2->sort_by);

    if (false) {
        /* do nothing */
    }
    else if (!fileitem_is_folder (f1) && fileitem_is_folder (f2)) {
        return true;
    }
    else if (fileitem_is_folder (f1) && fileitem_is_folder (f2)) {
        if (false) {
            /* do nothing */
        }
        else if (!f1->sort_by || !f2->sort_by) {
            /* do nothing */
        }
        else if (str_starts_with (f1->sort_by, "__")
                 && str_starts_with (f2->sort_by, "__")) {
            int result = strcasecmp (f1->sort_by, f2->sort_by);
            return (result > 0);
        }
        else if (!str_starts_with (f1->sort_by, "__")
                 && str_starts_with (f2->sort_by, "__")) {
            return true;
        }
        else if (str_starts_with (f1->sort_by, "__")
                 && !str_starts_with (f2->sort_by, "__")) {
            return false;
        }
        else if (mrl_has_protocol (f2->mrl, "http")
                 && !mrl_has_protocol (f1->mrl, "http")) {
            return true;
        }
        else if (mrl_has_protocol (f1->mrl, "http")
                 && !mrl_has_protocol (f2->mrl, "http")) {
            return false;
        }
        else if (str_equals (f2->mrl, FILELIST_HOME_MRL)) {
            return true;
        }
        else if (str_equals (f1->mrl, FILELIST_HOME_MRL)) {
            return false;
        }
        else if (str_equals (f2->mrl, FILELIST_ROOT_MRL)) {
            return true;
        }
        else if (str_equals (f1->mrl, FILELIST_ROOT_MRL)) {
            return false;
        }
        else if (f1->sort_by && f2->sort_by) {
            int result = strcasecmp (f1->sort_by, f2->sort_by);
            return (result > 0);
        }
    }
    else if (!fileitem_is_folder (f1) && !fileitem_is_folder (f2)) {
        if (false) {
            /* do nothing */
        }
#ifdef HAVE_LIBEXIF
        else if (f1->sort_by_exif && f2->sort_by_exif
                 && config_get_bool ("image.sort.by_date")) {
            int result = strcmp (f1->sort_by_exif, f2->sort_by_exif);
            if (result == 0) {
                result = strcasecmp (f1->sort_by, f2->sort_by);
            }
            return (result > 0);
        }
#endif /* HAVE_LIBEXIF */
        else if (f1->sort_by && f2->sort_by) {
            int result = strcasecmp (f1->sort_by, f2->sort_by);
            return (result > 0);
        }
    }

    return false;
}


/**
 * Callback used by l_list_free and l_list_clear to free a fileitem.
 */
static void
fileitem_free_cb (void *p)
{
    ho_badref (p);
    ho_decref (p);
}


/**
 * Callback used by l_list_free and l_list_clear to free a fileinfo item.
 */
static void
fileinfo_free_cb (void *p)
{
    ho_free (p);
}


/**
 * Never call this function directly. It will be called when the reference
 * count of the fileitem reaches zero.
 */
static void
fileitem_free (void *p)
{
    fileitem_t *fileitem = (fileitem_t *) p;

    assert (fileitem);

    if (fileitem->info_list) {
        l_list_free (fileitem->info_list, fileinfo_free_cb);
        fileitem->info_list = NULL;
    }
    if (fileitem->child_list) {
        fileitem->child_list->parent_item = NULL;
        ho_badref (fileitem->child_list);
        ho_setref (fileitem->child_list, NULL);
    }

    ho_free (fileitem->title_gui);
    ho_free (fileitem->title_search);
    ho_free (fileitem->title);
    ho_free (fileitem->cmd);
    ho_free (fileitem->mrl);
    ho_free (fileitem->thumbnail_mrl);
    ho_free (fileitem->sort_by);
#ifdef HAVE_LIBEXIF
    ho_free (fileitem->sort_by_exif);
#endif
#ifdef HAVE_HAL
    ho_free (fileitem->hal.device);
    ho_free (fileitem->hal.volume_udi);
    ho_free (fileitem->hal.drive_udi);
#endif

    ho_free (fileitem);
}


/**
 * Never call this function directly. It will be called when the reference
 * count of the fileitem reaches zero.
 */
static void
filelist_free (void *p)
{
    filelist_t *filelist = (filelist_t *) p;

    assert (filelist);
    assert (filelist->list);
    assert (!filelist_is_locked (filelist));

    if (filelist->info_list) {
        l_list_free (filelist->info_list, fileinfo_free_cb);
        filelist->info_list = NULL;
    }
    {
        l_list_free (filelist->list, fileitem_free_cb);
        filelist->list = NULL;
    }

    //debug ("%s (%s)", __func__, filelist->mrl);

    pthread_mutex_destroy (&filelist->mutex);
    pthread_mutexattr_destroy (&filelist->mutex_attr);

    ho_free (filelist->title);
    ho_free (filelist->title_complete);
    ho_free (filelist->mrl);
    ho_free (filelist->thumbnail_mrl);
    ho_free (filelist->error);

    ho_free (filelist);
}


static fileitem_t *
_fileitem_get (filelist_t * parent, const char *mrl)
{
    fileitem_t *item = NULL;
    fileitem_t *dest = NULL;

    assert (mrl);
    assert (parent);

    //debug ("%s (%s, %s)", __func__, parent->mrl, mrl);

    filelist_lock (parent);

    item = filelist_first (parent);
    while (item && !dest) {
        if (mrls_are_equal (item->mrl, mrl)) {
            dest = ho_incref (item);
        }
        item = filelist_next (parent, item);
    }

    item = filelist_first (parent);
    while (item && !dest) {
        if (item->child_list) {
            dest = _fileitem_get (item->child_list, mrl);
        }
        item = filelist_next (parent, item);
    }

    filelist_unlock (parent);

    return dest;
}


filelist_t *
filelist_get (const char *mrl)
{
    if (!toplevel_list) {
        fileitem_t *item = NULL;
        filelist_t *list = NULL;

        list = filelist_new (NULL, FILELIST_TOPLEVEL_TITLE,
                             FILELIST_TOPLEVEL_MRL, ALLOW_FILES_MULTIMEDIA);

        filelist_lock (list);

        item = filelist_add (list, FILELIST_MEDIA_TITLE,
                             FILELIST_MEDIA_MRL, FILE_TYPE_FOLDER);
        filelist_expand (item);

        item = filelist_add (list, FILELIST_SUBTITLE_TITLE,
                             FILELIST_SUBTITLE_MRL, FILE_TYPE_FOLDER);
        filelist_expand (item);

        item = filelist_add (list, FILELIST_REMOVABLE_TITLE,
                             FILELIST_REMOVABLE_MRL, FILE_TYPE_FOLDER);
        filelist_expand (item);

        item = filelist_add (list, FILELIST_TEMPORARY_TITLE,
                             FILELIST_TEMPORARY_MRL, FILE_TYPE_FOLDER);
        filelist_expand (item);

        item = filelist_add (list, FILELIST_EXTRACTOR_TITLE,
                             FILELIST_EXTRACTOR_MRL, FILE_TYPE_FOLDER);
        filelist_expand (item);

        filelist_unlock (list);

        toplevel_list = list;
    }
    fileitem_t *item = NULL;
    filelist_t *list = NULL;
    if (mrls_are_equal (mrl, FILELIST_TOPLEVEL_MRL)) {
        list = toplevel_list;
    }
    else if ((item = _fileitem_get (toplevel_list, mrl))) {
        filelist_lock (item->parent_list);
        list = filelist_expand (item);
        filelist_unlock (item->parent_list);
        ho_decref (item);
    }
    return list;
}


filelist_t *
filelist_new (fileitem_t * parent_item,
              const char *title, const char *mrl,
              fileitem_allowed_t allowed_filetypes)
{
    assert (title);
    assert (mrl);

    //debug ("%s (%s)", __func__, mrl);

    filelist_t *filelist = ho_initref (ho_new (filelist_t));
    filelist->r.free = filelist_free;

    filelist->list = l_list_new ();
    filelist->allowed_filetypes = allowed_filetypes;

    filelist->mrl = ho_strdup (mrl);
    filelist->title = ho_strdup (title);
    if (!parent_item || !parent_item->parent_list) {
        filelist->title_complete = ho_strdup (title);
    }
    else if (str_equals (parent_item->parent_list->mrl,
                         FILELIST_TOPLEVEL_MRL)) {
        filelist->title_complete = ho_strdup (title);
    }
    else {
        char *pt = parent_item->parent_list->title_complete;
        filelist->title_complete = ho_strdup_printf ("%s - %s", title, pt);
    }
    filelist->thumbnail_mrl = NULL;

    filelist->parent_item = parent_item;
    pthread_mutexattr_init (&filelist->mutex_attr);
    pthread_mutexattr_settype (&filelist->mutex_attr,
                               PTHREAD_MUTEX_RECURSIVE);
    pthread_mutex_init (&filelist->mutex, &filelist->mutex_attr);

    filelist->top_position = 0;
    filelist->cur_position = 0;
    filelist->wait_for_password = false;
    filelist->playback_length = 0;
    filelist->update_allowed = true;

    return filelist;
}


void
filelist_clear (filelist_t * filelist)
{
    assert (ho_checkref (filelist));

    filelist_lock (filelist);

    l_list_clear (filelist->list, fileitem_free_cb);
    filelist->top_position = 0;
    filelist->cur_position = 0;
    filelist->playback_length = 0;
    ho_free (filelist->error);

    filelist_unlock (filelist);
}


void
filelist_sort (filelist_t * filelist, l_swap_cb_t swap_cb)
{
    assert (ho_checkref (filelist));

    filelist_lock (filelist);

    if (swap_cb) {
        l_list_sort (filelist->list, swap_cb);
    }
    else {
        l_list_sort (filelist->list, fileitem_swap_cb);
    }

    filelist_unlock (filelist);
}


/// Removes any extension we know about from the filename.
static char *
remove_extension (char *title)
{
    int i = 0;
    int l = strlen (title);

    for (i = 0; video_extensions[i]; i++) {
        char *ext = video_extensions[i];
        if (mrl_has_extension (title, ext)) {
            title[l - strlen (ext) - 1] = '\0';
            return title;
        }
    }

    for (i = 0; audio_extensions[i]; i++) {
        char *ext = audio_extensions[i];
        if (mrl_has_extension (title, ext)) {
            title[l - strlen (ext) - 1] = '\0';
            return title;
        }
    }

    for (i = 0; image_extensions[i]; i++) {
        char *ext = image_extensions[i];
        if (mrl_has_extension (title, ext)) {
            title[l - strlen (ext) - 1] = '\0';
            return title;
        }
    }

    for (i = 0; subtitle_extensions[i]; i++) {
        char *ext = subtitle_extensions[i];
        if (mrl_has_extension (title, ext)) {
            title[l - strlen (ext) - 1] = '\0';
            return title;
        }
    }

    for (i = 0; playlist_extensions[i]; i++) {
        char *ext = playlist_extensions[i];
        if (mrl_has_extension (title, ext)) {
            title[l - strlen (ext) - 1] = '\0';
            return title;
        }
    }

    for (i = 0; mediamarks_extensions[i]; i++) {
        char *ext = mediamarks_extensions[i];
        if (mrl_has_extension (title, ext)) {
            title[l - strlen (ext) - 1] = '\0';
            return title;
        }
    }

    return title;
}


static fileitem_t *
fileitem_detect (const char *mrl)
{
    fileitem_type_t type = FILE_TYPE_REGULAR;
    fileitem_t *fileitem = NULL;

    char *add_title = NULL;
    char *gui_title = NULL;
    char *add_mrl = NULL;

    char *filename = mrl_get_filename (mrl);

    char *filename_bn = get_basename (filename);
    replace_chars (filename_bn, '_', ' ');

    char *title = ho_strdup (filename_bn);
    remove_extension (title);

    if (mrls_are_equal (mrl, get_dir_home ())) {
        add_title = ho_strdup (_("Home"));
        type = FILE_TYPE_FOLDER;
    }
    else if (mrls_are_equal (mrl, get_dir_root ())) {
        add_title = ho_strdup (_("Filesystem"));
        type = FILE_TYPE_FOLDER;
    }
    else if (mrls_are_equal (mrl, get_dir_playlists ())) {
        add_title = ho_strdup (_("Playlists"));
        type = FILE_TYPE_FOLDER;
    }
    else if (mrl_is_directory_dvd (mrl)) {
        add_title = ho_strdup (title);
        gui_title = ho_strdup_printf ("[] %s (%s)", title, _("DVD"));
        add_mrl = mrl_build ("dvd", filename, 0);
        type = FILE_TYPE_FOLDER;
    }
    else if (mrl_is_directory_vcd (mrl)) {
        add_title = ho_strdup (title);
        gui_title = ho_strdup_printf ("%s (%s)", title, _("Video CD"));
        add_mrl = mrl_build ("vcd", filename, 0);
        type = FILE_TYPE_REGULAR;
    }
    else if (mrl_is_directory (mrl)) {
        add_title = ho_strdup (title);
        type = FILE_TYPE_FOLDER;
    }
    else if (mrl_is_playlist_m3u (mrl)) {
        if (!strstr (filename, get_dir_playlists ())) {
            add_title = ho_strdup_printf ("%s: %s", _("Playlist"), title);
        }
        type = FILE_TYPE_PLAYLIST;
    }
    else if (mrl_is_playlist_pls (mrl)) {
        if (!strstr (filename, get_dir_playlists ())) {
            add_title = ho_strdup_printf ("%s: %s", _("Playlist"), title);
        }
        type = FILE_TYPE_PLAYLIST;
    }
    else if (mrl_is_playlist_oxp (mrl)) {
        if (!strstr (filename, get_dir_playlists ())) {
            add_title = ho_strdup_printf ("%s: %s", _("Playlist"), title);
        }
        type = FILE_TYPE_PLAYLIST;
    }
    else if (mrl_is_mediamarks_oxm (mrl)) {
        add_title = ho_strdup (title);
        type = FILE_TYPE_MEDIAMARKS;
    }
    else if (mrl_has_protocol (mrl, "dvd")) {
        int number = mrl_get_number (mrl);
        if (number != -1) {
            add_title = ho_strdup_printf ("%s - %s %02d", _("DVD"),
                                          _("Chapter"), number + 1);
            type = FILE_TYPE_REGULAR;
        }
        else {
            add_title = ho_strdup (title);
            gui_title = ho_strdup_printf ("%s (%s)", _("Unknown Disc"),
                                          _("DVD"));
            type = FILE_TYPE_FOLDER;
        }
    }
    else if (mrl_has_protocol (mrl, "vcd")) {
        int number = mrl_get_number (mrl);
        if (number != -1) {
            add_title = ho_strdup_printf ("%s - %s %02d", _("Video CD"),
                                          _("Title"), number + 1);
        }
        else {
            add_title = ho_strdup (title);
            gui_title = ho_strdup_printf ("%s (%s)", _("Unknown Disc"),
                                          _("Video CD"));
        }
        type = FILE_TYPE_REGULAR;
    }
    else if (mrl_has_protocol (mrl, "cdda")) {
        int number = mrl_get_number (mrl);
        if (number != -1) {
            add_title = ho_strdup_printf ("%s - %s %02d", _("Audio CD"),
                                          _("Title"), number + 1);
            type = FILE_TYPE_REGULAR;
        }
        else {
            add_title = ho_strdup (title);
            gui_title = ho_strdup_printf ("%s: %s", _("Audio CD"),
                                          _("Unknown Disc"));
            type = FILE_TYPE_FOLDER;
        }
    }
#ifdef HAVE_LIBCDIO
    else if (mrl_is_iso_image (mrl)) {
        fileitem = disc_image_detect (mrl, title);
        goto out_free;
    }
#endif /* HAVE_LIBCDIO */

    if (!add_title) {
        add_title = ho_strdup_printf ("%s", title);
    }
    if (!add_mrl) {
        add_mrl = ho_strdup (mrl);
    }

    fileitem = ho_initref (ho_new (fileitem_t));
    fileitem->r.free = fileitem_free;
    fileitem->type = type;
    fileitem->mrl = add_mrl;
    fileitem->title = add_title;
    fileitem->title_gui = gui_title;

  out_free:
    ho_free (filename);
    ho_free (filename_bn);
    ho_free (title);

    return fileitem;
}


static void
filelist_add_fileitem (filelist_t * filelist, fileitem_t * fileitem)
{
    fileitem->parent_list = filelist;

    fileitem->sort_by = ho_strdup (fileitem->title);
    replace_special_chars (fileitem->sort_by);
#ifdef HAVE_LIBEXIF
    fileitem->sort_by_exif = exif_info_get (EXIF_INFO_SORT, fileitem->mrl);
#endif /* HAVE_LIBEXIF */

    if (mrl_is_networked (fileitem->mrl)) {
        /* don't query the playback length */
    }
    else if (fileitem->type == FILE_TYPE_REGULAR) {
        fileitem->playback_length = fileitem_len_cb (fileitem->mrl);
        filelist->playback_length += fileitem->playback_length;
    }

    l_list_append (filelist->list, fileitem);
}


fileitem_t *
fileitem_add_alternative (fileitem_t * fileitem, const char *mrl)
{
    fileitem_t *alternative = NULL;

    assert (ho_checkref (fileitem));
    assert (ho_checkref (fileitem->parent_list));
    assert (filelist_is_locked (fileitem->parent_list));
    assert (fileitem->type == FILE_TYPE_REGULAR);
    assert (mrl);

    fileitem_allowed_t allowed = ALLOW_FILES_MULTIMEDIA;
    if (fileitem->parent_list) {
        allowed = fileitem->parent_list->allowed_filetypes;
    }
    if (!fileitem->child_list) {
        fileitem->child_list = filelist_new (fileitem, fileitem->title,
                                             fileitem->mrl, allowed);
    }

    alternative = filelist_add (fileitem->child_list, fileitem->title,
                                mrl, fileitem->type);

    const char *thumbnail_mrl = fileitem_get_cover_image (fileitem);
    if (alternative && thumbnail_mrl) {
        alternative->thumbnail_mrl = ho_strdup (thumbnail_mrl);
    }

    return alternative;
}


fileitem_t *
filelist_add_magic (filelist_t * filelist, const char *title, const char *mrl)
{
    fileitem_t *fileitem = NULL;

    assert (ho_checkref (filelist));
    assert (mrl);

    filelist_lock (filelist);
    if (!mrl_is_valid (mrl)) {
        char *new_mrl = mrl_from_filename (mrl);
        fileitem = filelist_add_magic (filelist, title, new_mrl);
        ho_free (new_mrl);
    }
    else {
        fileitem = fileitem_detect (mrl);
        if (title) {
            ho_free (fileitem->title);
            fileitem->title = ho_strdup (title);
        }
        filelist_add_fileitem (filelist, fileitem);
    }
    filelist_unlock (filelist);

    return fileitem;
}


fileitem_t *
filelist_add (filelist_t * filelist, const char *title, const char *mrl,
              fileitem_type_t type)
{
    fileitem_t *fileitem = NULL;

    assert (ho_checkref (filelist));
    assert (title);
    assert (mrl);
    assert (type != FILE_TYPE_UNKNOWN);

    filelist_lock (filelist);
    if (!mrl_is_valid (mrl)) {
        char *new_mrl = mrl_from_filename (mrl);
        fileitem = filelist_add (filelist, title, new_mrl, type);
        ho_free (new_mrl);
    }
    else {
        fileitem = ho_initref (ho_new (fileitem_t));
        fileitem->r.free = fileitem_free;
        fileitem->title = ho_strdup (title);
        fileitem->mrl = ho_strdup (mrl);
        fileitem->type = type;
        fileitem->is_mountpoint = false;
        filelist_add_fileitem (filelist, fileitem);
    }
    filelist_unlock (filelist);

    return fileitem;
}


/**
 * Reads the titles of DVD. All titles of the DVD are added to the list.
 */
static void
dvd_read (filelist_t * filelist)
{
    char *mrl = filelist->mrl;
    if (!mrl_has_protocol (mrl, "dvd")) {
        return;
    }

    filelist_add (filelist, _("Play DVD"), filelist->mrl,
                  FILE_TYPE_REGULAR);

    xine_stream_t *stream = xine_stream_new (oxine->xine, NULL, NULL);
    if (!stream) {
        return;
    }
    if (xine_open (stream, mrl) != 1) {
        return;
    }

    uint32_t tc = xine_get_stream_info (stream,
                                        XINE_STREAM_INFO_DVD_TITLE_COUNT);

    int i;
    for (i = 0; i < tc; i++) {
        char *track_mrl = ho_strdup_printf ("%s/%d", mrl, i + 1);
        char *track_title = ho_strdup_printf ("%s %02d", _("Title"), i + 1);
        fileitem_t *item = filelist_add (filelist, track_title, track_mrl,
                                         FILE_TYPE_REGULAR);
        fileitem_get_info_list (item);

        ho_free (track_mrl);
        ho_free (track_title);
    }

    xine_close (stream);
    xine_dispose (stream);
}


/**
 * Reads the titles of an audio CD. All titles of the CD are added to the
 * list.
 */
static void
audiocd_read (filelist_t * filelist)
{
    char *mrl = filelist->mrl;
    if (!mrl_has_protocol (mrl, "cdda")) {
        return;
    }

    int num_mrls;
    char **mrls = xine_get_autoplay_mrls (oxine->xine, "CD", &num_mrls);

    if (!mrls || (num_mrls <= 0)) {
        return;
    }

    int i;
    bool use_cddb = config_get_bool ("media.audio_cd.use_cddb");
    for (i = 0; i < num_mrls; i++) {
        char *track_mrl = ho_strdup_printf ("%s/%d", mrl, i + 1);

        char *title = meta_info_get (META_INFO_TITLE, track_mrl);
        if (!title) {
            /* If retrieving information from CDDB fails for one of the track,
             * we do not try to retrieve the information for any of the other
             * tracks. We do this because of the long timeouts that may occur
             * when trying again and again for all tracks. */
            config_set_bool ("media.audio_cd.use_cddb", false);
            title = ho_strdup (_("Unknown Title"));
        }

        char *track_title = ho_strdup_printf ("%02d - %s", i + 1, title);
        fileitem_t *item = filelist_add (filelist, track_title, track_mrl,
                                         FILE_TYPE_REGULAR);
        fileitem_get_info_list (item);

#ifdef HAVE_HAL
        fileitem_t *p = filelist->parent_item;
        if (p->hal.device) {
            item->hal.device = ho_strdup (p->hal.device);
        }
        if (p->hal.volume_udi) {
            item->hal.volume_udi = ho_strdup (p->hal.volume_udi);
        }
        if (p->hal.drive_udi) {
            item->hal.drive_udi = ho_strdup (p->hal.drive_udi);
        }
#endif /* HAVE_HAL */

        ho_free (title);
        ho_free (track_mrl);
        ho_free (track_title);
    }

    config_set_bool ("media.audio_cd.use_cddb", use_cddb);
}


/**
 * Reads a directory. All entries that are currently allowed are added to
 * the list.
 */
static void
directory_read (filelist_t * filelist)
{
    assert (ho_checkref (filelist));
    assert (filelist->mrl);

    bool show_hidden = config_get_bool ("media.files.show_hidden_files");

    char *dirname = mrl_get_filename (filelist->mrl);
    if (dirname[strlen (dirname) - 1] != '/') {
        char *tmp = ho_strdup_printf ("%s/", dirname);
        ho_free (dirname);
        dirname = tmp;
    }

    char *pwdfile = ho_strdup_printf ("%s.password", dirname);
    if (file_exists (pwdfile)) {
        /* Get the password from the file. */
        FILE *f = fopen (pwdfile, "r");
        if (!f) {
            error (_("Could not open '%s': %s!"), pwdfile, strerror (errno));
            info (_("Not granting access to directory '%s'!"), dirname);
            goto out_free;
        }

        char pwd_exp[128];
        if (fgets (pwd_exp, 128, f)) {
            trim_whitespace (pwd_exp);
        }
        else {
            pwd_exp[0] = '\0';
        }
        fclose (f);

        /* Check the password against a user provided password. */
        if (!fileitem_pwd_cb || !fileitem_pwd_cb (dirname, pwd_exp)) {
            filelist->wait_for_password = true;
            error (_("Incorrect password provided for directory '%s'!"),
                   dirname);
            info (_("Not granting access to directory '%s'!"), dirname);
            goto out_free;
        }
    }
    filelist->wait_for_password = false;

    DIR *dirp = opendir (dirname);
    if (!dirp) {
        error (_("Unable to list contents of directory '%s': %s!"),
               dirname, strerror (errno));
        if (errno == EACCES) {
            ho_free (filelist->error);
            filelist->error = ho_strdup (_("Permission denied..."));
        }
        goto out_free;
    }

    struct dirent *entp;
    while ((entp = readdir (dirp))) {
        if (strcmp (entp->d_name, ".") == 0) {
            continue;
        }
        if (strcmp (entp->d_name, "..") == 0) {
            continue;
        }

        fileitem_allowed_t allowed = filelist->allowed_filetypes;
        char *abs = ho_strdup_printf ("%s%s", dirname, entp->d_name);
        char *mrl = mrl_build ("file", abs, 0);
        if (!(mrl_is_hidden (mrl) && !show_hidden)) {
            fileitem_t *fileitem = fileitem_detect (mrl);
            if (fileitem) {
                assert (fileitem->type != FILE_TYPE_UNKNOWN);
                assert (fileitem->title);
                assert (fileitem->mrl);
                assert (mrl_is_valid (fileitem->mrl));
                if (fileitem_is_allowed (fileitem, allowed)) {
                    filelist_add_fileitem (filelist, fileitem);
                }
                else {
                    ho_decref (fileitem);
                }
            }
        }
        ho_free (abs);
        ho_free (mrl);
    }
    closedir (dirp);

    filelist_sort (filelist, NULL);

  out_free:
    ho_free (dirname);
    ho_free (pwdfile);
}


bool
fileitem_is_folder (fileitem_t * fileitem)
{
    bool has_sublist = false;

    switch (fileitem->type) {
    case FILE_TYPE_FOLDER:
    case FILE_TYPE_MEDIAMARKS:
        has_sublist = true;
        break;
    default:
        break;
    }

    return has_sublist;
}


static filelist_t *
_filelist_expand (fileitem_t * fileitem, bool update)
{
    int top_position = 0;
    int cur_position = 0;

#ifdef PROFILE_FILELIST_EXPAND
    struct timeval ts;
    struct timeval te;
    struct timezone tz;
    gettimeofday (&ts, &tz);
#endif

    assert (ho_checkref (fileitem));
    assert (filelist_is_locked (fileitem->parent_list));
    assert (fileitem->mrl);
    assert (fileitem->type != FILE_TYPE_UNKNOWN);

    /* Check if this fileitem will have a child list.
     *
     * - Regular files and playlists only get a child list, if they get
     *   content from an external command.
     * - Folders and mediamarks will always get a child list.
     */
    switch (fileitem->type) {
    case FILE_TYPE_REGULAR:
    case FILE_TYPE_PLAYLIST:
        if (!fileitem->cmd) {
            goto out_quit;
        }
        break;
    case FILE_TYPE_FOLDER:
    case FILE_TYPE_MEDIAMARKS:
        break;
    default:
        {
            goto out_quit;
        }
        break;
    }

    /* What files are allowed in the child list of this fileitem. */
    fileitem_allowed_t allowed = ALLOW_FILES_MULTIMEDIA;
    if (fileitem->parent_list) {
        allowed = fileitem->parent_list->allowed_filetypes;
    }

    /* If the fileitem does not yet have a child item, we create it. */
    if (!fileitem->child_list) {
        fileitem->child_list = filelist_new (fileitem, fileitem->title,
                                             fileitem->mrl, allowed);
    }
    /* Audio CDs don't usually change their content. Once we've expanded them
     * we leave them alone. */
    else if (mrl_has_protocol (fileitem->mrl, "cdda")) {
        goto out_quit;
    }
    /* If the fileitem already has a child list (meaning we have expanded this
     * fileitem before) and update is FALSE, we quit. */
    else if (!update) {
        goto out_quit;
    }

    /* Lock the child list to make sure nobody can access it until we're done. */
    filelist_lock (fileitem->child_list);

    /* Save the GUI positions of this filelist to restore them when we're done. */
    top_position = fileitem->child_list->top_position;
    cur_position = fileitem->child_list->cur_position;

    /* If an error message is set, we clear the child list. */
    if (fileitem->child_list->error) {
        filelist_clear (fileitem->child_list);
    }
    /* VDR lists must be completely updated, as the IDs change every time an
     * entry is added or removed. */
    else if (mrl_has_protocol (fileitem->mrl, "vdr")) {
        filelist_clear (fileitem->child_list);
    }
    /* We update the favorites filelist completely to make sure the timestamp
     * and listened count are updated. */
    else if (mrls_are_equal (fileitem->mrl, get_file_favorites ())) {
        filelist_clear (fileitem->child_list);
    }

    /* If the fileitem has a thumbnail image, we set this image for the child
     * list also. */
    if (fileitem->thumbnail_mrl) {
        char *thumbnail_mrl = fileitem->thumbnail_mrl;
        ho_free (fileitem->child_list->thumbnail_mrl);
        fileitem->child_list->thumbnail_mrl = ho_strdup (thumbnail_mrl);
    }

    /* Free the info list of the parent item, as it may not be up to date
     * anymore when finished. */
    if (fileitem->info_list) {
        filelist_lock (fileitem->parent_list);
        l_list_t *info_list = fileitem->info_list;
        l_list_free (info_list, fileinfo_free_cb);
        fileitem->info_list = NULL;
        filelist_unlock (fileitem->parent_list);
    }
    /* Free the info list of the child list, as it may not be up to date
     * anymore when finished. */
    if (fileitem->child_list->info_list) {
        l_list_t *info_list = fileitem->child_list->info_list;
        l_list_free (info_list, fileinfo_free_cb);
        fileitem->child_list->info_list = NULL;
    }

    /* We don't want to work directly on an existing child list so as not to
     * loose any previously expanded sublists. So we create a temporary list,
     * work on that and merge it with the existing child list at the end. */
    filelist_t *temp_list = filelist_new (fileitem, fileitem->title,
                                          fileitem->mrl, allowed);
    filelist_lock (temp_list);

    /* If the fileitem has an associated command, we execute that command to
     * build the child list. The content of this child list is written to a
     * temporary mediamarks file by the script and read from there. */
    fileitem_type_t type = fileitem->type;
    if (fileitem->cmd) {
        char *org_mrl = fileitem->mrl;
        char *new_mrl = ho_strdup_printf ("%s/mediamarks-%d.xml",
                                          get_dir_oxine_tmp (), time (NULL));
        char *command = ho_strdup_printf ("%s %s %s", fileitem->cmd,
                                          org_mrl, new_mrl);
        if (execute_shell (command, 0)) {
            ho_free (temp_list->mrl);
            temp_list->mrl = new_mrl;
            type = FILE_TYPE_MEDIAMARKS;
        }
        else {
            ho_free (new_mrl);
        }
        ho_free (command);
    }

    char *query = mrl_get_query (fileitem->mrl);
    char *mrl = mrl_without_query (fileitem->mrl);
    switch (type) {
    case FILE_TYPE_MEDIAMARKS:
        mediamarks_read (temp_list);
        break;
    case FILE_TYPE_FOLDER:
        if (mrl_has_protocol (mrl, "cdda")) {
            audiocd_read (temp_list);
        }
        else if (mrl_has_protocol (mrl, "dvd")) {
            dvd_read (temp_list);
        }
        else if (mrls_are_equal (mrl, FILELIST_REMOVABLE_MRL)) {
            /* Ignore it, this gets filled in the HAL thread. */
        }
        else if (mrls_are_equal (mrl, FILELIST_TEMPORARY_MRL)) {
            /* Ignore it, this is only a container for temporary fileitems. */
        }
        else if (mrls_are_equal (mrl, FILELIST_EXTRACTOR_MRL)) {
            /* Ignore it, this is only a container for extractor fileitems. */
        }
        else if (mrls_are_equal (mrl, FILELIST_MEDIA_MRL)) {
            temp_list->allowed_filetypes = ALLOW_FILES_MULTIMEDIA;
            if (!mediamarks_read (temp_list)) {
                filelist_add (temp_list, FILELIST_HOME_TITLE,
                              FILELIST_HOME_MRL, FILE_TYPE_FOLDER);
                filelist_add (temp_list, FILELIST_ROOT_TITLE,
                              FILELIST_ROOT_MRL, FILE_TYPE_FOLDER);
                mediamarks_write (temp_list);
            }
            {
                filelist_add (temp_list, FILELIST_MUSIC_TITLE,
                              FILELIST_MUSIC_MRL, FILE_TYPE_FOLDER);
                filelist_add (temp_list, FILELIST_VIDEO_TITLE,
                              FILELIST_VIDEO_MRL, FILE_TYPE_FOLDER);
                filelist_add (temp_list, FILELIST_IMAGE_TITLE,
                              FILELIST_IMAGE_MRL, FILE_TYPE_FOLDER);
            }
            filelist_sort (temp_list, NULL);
        }
        else if (mrls_are_equal (mrl, FILELIST_SUBTITLE_MRL)) {
            temp_list->allowed_filetypes = ALLOW_FILES_SUBTITLE;
            if (!mediamarks_read (temp_list)) {
                filelist_add (temp_list, FILELIST_HOME_TITLE,
                              FILELIST_HOME_MRL, FILE_TYPE_FOLDER);
                filelist_add (temp_list, FILELIST_ROOT_TITLE,
                              FILELIST_ROOT_MRL, FILE_TYPE_FOLDER);
                mediamarks_write (temp_list);
            }
            filelist_sort (temp_list, NULL);
        }
        else if (mrls_are_equal (mrl, FILELIST_MUSIC_MRL)) {
            temp_list->allowed_filetypes = ALLOW_FILES_MUSIC;
            if (!mediamarks_read (temp_list)) {
                filelist_add (temp_list, FILELIST_HOME_TITLE,
                              FILELIST_HOME_MRL, FILE_TYPE_FOLDER);
                filelist_add (temp_list, FILELIST_ROOT_TITLE,
                              FILELIST_ROOT_MRL, FILE_TYPE_FOLDER);
                filelist_add (temp_list, FILELIST_XDG_MUSIC_TITLE,
                              FILELIST_XDG_MUSIC_MRL, FILE_TYPE_FOLDER);
                mediamarks_write (temp_list);
            }
            if (config_get_bool ("misc.favorites.enable")) {
                const char *mrl = FILELIST_FAVORITES_MOST_MRL;
                const char *title = FILELIST_FAVORITES_MOST_TITLE;
                fileitem_t *item = filelist_add (temp_list, title,
                                                 mrl, FILE_TYPE_FOLDER);
                filelist_expand (item);
                ho_free (item->sort_by);
                item->sort_by = ho_strdup ("__00001");
            }
            if (config_get_bool ("misc.favorites.enable")) {
                const char *mrl = FILELIST_FAVORITES_RECENT_MRL;
                const char *title = FILELIST_FAVORITES_RECENT_TITLE;
                fileitem_t *item = filelist_add (temp_list, title,
                                                 mrl, FILE_TYPE_FOLDER);
                filelist_expand (item);
                ho_free (item->sort_by);
                item->sort_by = ho_strdup ("__00002");
            }
            if (config_get_bool ("playlist.metafolder.show")) {
                const char *mrl = FILELIST_PLAYLISTS_MRL;
                const char *title = FILELIST_PLAYLISTS_TITLE;
                fileitem_t *item = filelist_add (temp_list, title,
                                                 mrl, FILE_TYPE_FOLDER);
                ho_free (item->sort_by);
                item->sort_by = ho_strdup ("__00003");
            }
            if (file_exists (OXINE_AUDIODIR)) {
                const char *mrl = OXINE_AUDIODIR;
                const char *title = _("Mediamarks");
                fileitem_t *item = filelist_add (temp_list, title,
                                                 mrl, FILE_TYPE_FOLDER);
                filelist_expand (item);
                ho_free (item->sort_by);
                item->sort_by = ho_strdup ("__00004");
            }
            filelist_sort (temp_list, NULL);
        }
        else if (mrls_are_equal (mrl, FILELIST_VIDEO_MRL)) {
            temp_list->allowed_filetypes = ALLOW_FILES_VIDEO;
            if (!mediamarks_read (temp_list)) {
                filelist_add (temp_list, FILELIST_HOME_TITLE,
                              FILELIST_HOME_MRL, FILE_TYPE_FOLDER);
                filelist_add (temp_list, FILELIST_ROOT_TITLE,
                              FILELIST_ROOT_MRL, FILE_TYPE_FOLDER);
                filelist_add (temp_list, FILELIST_XDG_VIDEOS_TITLE,
                              FILELIST_XDG_VIDEOS_MRL, FILE_TYPE_FOLDER);
                mediamarks_write (temp_list);
            }
#ifdef HAVE_VDR
            if ((config_get_number ("television.type") == 2)
                && config_get_bool ("television.enable")
                && (config_get_bool ("television.show.vfolder.recordings")
                    || config_get_bool ("television.show.vfolder.timers")
                    || config_get_bool ("television.show.vfolder.channels"))
                ) {
                const char *mrl = FILELIST_VDR_MRL;
                const char *title = _("Television");
                fileitem_t *item = filelist_add (temp_list, title,
                                                 mrl, FILE_TYPE_FOLDER);
                ho_free (item->sort_by);
                item->sort_by = ho_strdup ("__00001");
            }
#endif
            if (config_get_bool ("playlist.metafolder.show")) {
                const char *mrl = FILELIST_PLAYLISTS_MRL;
                const char *title = _("Playlists");
                fileitem_t *item = filelist_add (temp_list, title,
                                                 mrl, FILE_TYPE_FOLDER);
                ho_free (item->sort_by);
                item->sort_by = ho_strdup ("__00002");
            }
            if (file_exists (OXINE_VIDEODIR)) {
                const char *mrl = OXINE_VIDEODIR;
                const char *title = _("Mediamarks");
                fileitem_t *item = filelist_add (temp_list, title,
                                                 mrl, FILE_TYPE_FOLDER);
                filelist_expand (item);
                ho_free (item->sort_by);
                item->sort_by = ho_strdup ("__00003");
            }
            filelist_sort (temp_list, NULL);
        }
        else if (mrls_are_equal (mrl, FILELIST_IMAGE_MRL)) {
            temp_list->allowed_filetypes = ALLOW_FILES_IMAGE;
            if (!mediamarks_read (temp_list)) {
                filelist_add (temp_list, FILELIST_HOME_TITLE,
                              FILELIST_HOME_MRL, FILE_TYPE_FOLDER);
                filelist_add (temp_list, FILELIST_ROOT_TITLE,
                              FILELIST_ROOT_MRL, FILE_TYPE_FOLDER);
                filelist_add (temp_list, FILELIST_XDG_PICTURES_TITLE,
                              FILELIST_XDG_PICTURES_MRL, FILE_TYPE_FOLDER);
                mediamarks_write (temp_list);
            }
            if (file_exists (OXINE_IMAGEDIR)) {
                const char *mrl = OXINE_IMAGEDIR;
                const char *title = _("Mediamarks");
                fileitem_t *item = filelist_add (temp_list, title,
                                                 mrl, FILE_TYPE_FOLDER);
                filelist_expand (item);
                ho_free (item->sort_by);
                item->sort_by = ho_strdup ("__00001");
            }
            filelist_sort (temp_list, NULL);
        }
        else if (mrls_are_equal (mrl, FILELIST_FAVORITES_MOST_MRL)) {
            if (favorites_read (temp_list)) {
                char *sort = query_get_parameter (query, "sort");
                if (strc_equals (sort, "recent")) {
                    filelist_sort (temp_list, favorites_swap_time_cb);
                }
                else {
                    filelist_sort (temp_list, favorites_swap_count_cb);
                }
                ho_free (sort);
            }
        }
#ifdef HAVE_VDR
        else if (mrls_are_equal (mrl, FILELIST_VDR_MRL)) {
            if (config_get_bool ("television.show.vfolder.channels")) {
                const char *mrl = FILELIST_VDR_CHA_MRL;
                const char *title = _("Channels");
                fileitem_t *item = filelist_add (temp_list, title,
                                                 mrl, FILE_TYPE_FOLDER);
                ho_free (item->sort_by);
                item->sort_by = ho_strdup ("__00001");
            }
            if (config_get_bool ("television.show.vfolder.recordings")) {
                const char *mrl = FILELIST_VDR_REC_MRL;
                const char *title = _("Recordings");
                fileitem_t *item = filelist_add (temp_list, title,
                                                 mrl, FILE_TYPE_FOLDER);
                ho_free (item->sort_by);
                item->sort_by = ho_strdup ("__00002");
            }
            if (config_get_bool ("television.show.vfolder.timers")) {
                const char *mrl = FILELIST_VDR_TIM_MRL;
                const char *title = _("Timers");
                fileitem_t *item = filelist_add (temp_list, title,
                                                 mrl, FILE_TYPE_FOLDER);
                ho_free (item->sort_by);
                item->sort_by = ho_strdup ("__00003");
            }
        }
        else if (strc_starts_with (mrl, FILELIST_VDR_CHA_MRL)) {
            vdr_channels_read (temp_list);
        }
        else if (strc_starts_with (mrl, FILELIST_VDR_REC_MRL)) {
            vdr_recordings_read (temp_list);
        }
        else if (strc_starts_with (mrl, FILELIST_VDR_TIM_MRL)) {
            vdr_timers_read (temp_list);
        }
#endif
#ifdef HAVE_HAL
        else if (fileitem->hal.device && volume_mount (fileitem->hal.device)) {
            /* After mounting the volume we have to retrieve the mountpoint.
             * This is necessary because HAL does not give us a mountpoint
             * before the disc is mounted. */
            ho_free (fileitem->mrl);
            fileitem->mrl = volume_get_mountpoint (fileitem->hal.device);
            ho_free (mrl);
            mrl = ho_strdup (fileitem->mrl);
            /* After fixing the MRLs we can read this volume just like any
             * other directory. */
            directory_read (temp_list);
        }
#endif
        else if (fileitem->is_mountpoint && volume_mount (mrl)) {
            directory_read (temp_list);
        }
        else if (mrl_is_directory (mrl)) {
            directory_read (temp_list);
        }
        else {
            temp_list->error = ho_strdup (_("Unknown folder type"));
        }
        break;
    default:
        break;
    }
    ho_free (mrl);
    ho_free (query);

    /* We reset the MRL of the temporary list to the MRL of the parent item.
     * This is necessary in case we loaded the child list via an external
     * command as the child lists MRL would then be that of the locally
     * stored mediamarks file. */
    ho_free (temp_list->mrl);
    temp_list->mrl = ho_strdup (fileitem->mrl);

    /* If the child list already contains items we move any child list from
     * those items to the newly created items in the temporary filelist and
     * clear the child list. */
    if (l_list_length (fileitem->child_list->list) > 0) {
        fileitem_t *new_item = filelist_first (temp_list);
        while (new_item) {
            fileitem_t *old_item = filelist_first (fileitem->child_list);
            while (old_item) {
                if (str_equals (old_item->mrl, new_item->mrl)) {
                    if (old_item->child_list) {
                        ho_setref (new_item->child_list,
                                   old_item->child_list);
                        new_item->child_list->parent_item = new_item;
                        ho_setref (old_item->child_list, NULL);
                    }
                    l_list_remove (fileitem->child_list->list, old_item);
                    fileitem_free_cb (old_item);
                    break;
                }
                old_item = filelist_next (fileitem->child_list, old_item);
            }
            new_item = filelist_next (temp_list, new_item);
        }
        l_list_clear (fileitem->child_list->list, fileitem_free_cb);
    }

    /* Now we move all items from the temporary list to the child list. */
    {
        fileitem_t *new_item = filelist_first (temp_list);
        while (new_item) {
            l_list_remove (temp_list->list, new_item);
            l_list_append (fileitem->child_list->list, new_item);
            new_item->parent_list = fileitem->child_list;
            new_item = filelist_first (temp_list);
        }
    }

    /* Last we copy all the flags from the temporary list to the child list. */
    ho_free (fileitem->child_list->error);
    fileitem->child_list->error = ho_strdup (temp_list->error);
    fileitem->child_list->wait_for_password = temp_list->wait_for_password;
    fileitem->child_list->playback_length = temp_list->playback_length;
    fileitem->child_list->allowed_filetypes = temp_list->allowed_filetypes;

    /* Unlock and free the temporary filelist. */
    filelist_unlock (temp_list);
    ho_setref (temp_list, NULL);

    /* If the URL contains a keyword in the query string we filter the
     * results by that keyword. We do not do this for lists built by an
     * external script. They are supposed to do that themselves. */
    char *keyword = mrl_get_query_parameter (fileitem->mrl, "oxine_keyword");
    replace_special_chars (keyword);
    if (!fileitem->cmd && keyword && (strlen (keyword) > 0)) {
        fileitem_t *item = filelist_first (fileitem->child_list);
        while (item) {
            fileitem_t *next = filelist_next (fileitem->child_list, item);
            if (!item->title_search) {
                item->title_search = ho_strdup (item->title);
                replace_special_chars (item->title_search);
            }
            if (!strc_contains (item->title_search, keyword)) {
                filelist_remove (fileitem->child_list, item);
            }
            item = next;
        }
    }
    ho_free (keyword);

    FILE *f = NULL;
    char *cache_dir = mrl_get_cache_dir (fileitem->mrl);
    char *cache_file = ho_strdup_printf ("%s/count", cache_dir);
    if (mrl_has_protocol (fileitem->mrl, "cdda")) {
        /* do nothing */
    }
#ifdef HAVE_HAL
    else if (fileitem->hal.device) {
        /* do nothing */
    }
#endif
    else if (!mkdir_recursive (cache_dir, 0700)) {
        /* do nothing */
    }
    else if ((f = fopen (cache_file, "w"))) {
        fprintf (f, "%d", l_list_length (fileitem->child_list->list));
        fclose (f);
    }
    ho_free (cache_file);
    ho_free (cache_dir);

    filelist_unlock (fileitem->child_list);

#ifdef PROFILE_FILELIST_EXPAND
    gettimeofday (&te, &tz);
    if (te.tv_usec < ts.tv_usec) {
        te.tv_sec -= 1;
        te.tv_usec += 1000 * 1000;
    }
    time_t msec = 0;
    msec += (te.tv_usec - ts.tv_usec) / 1000;
    msec += (te.tv_sec - ts.tv_sec) * 1000;
    debug ("filelist_expand: %4d ms, update=%s", msec,
           update ? "true" : "false");
#endif

  out_quit:
    return fileitem->child_list;
}


filelist_t *
filelist_expand (fileitem_t * fileitem)
{
    assert (ho_checkref (fileitem));
    assert (filelist_is_locked (fileitem->parent_list));
    return _filelist_expand (fileitem, false);
}


filelist_t *
filelist_update (filelist_t * filelist)
{
    assert (ho_checkref (filelist));
    assert (ho_checkref (filelist->parent_item));
    assert (filelist_is_locked (filelist->parent_item->parent_list));
    return _filelist_expand (filelist->parent_item, true);
}


void
filelist_remove (filelist_t * filelist, fileitem_t * fileitem)
{
    filelist_lock (filelist);

    assert (ho_checkref (filelist));
    assert (ho_checkref (fileitem));
    assert (filelist == fileitem->parent_list);

    filelist->playback_length -= fileitem->playback_length;
    l_list_remove (filelist->list, fileitem);
    ho_badref (fileitem);
    ho_decref (fileitem);

    filelist_unlock (filelist);
}


fileitem_t *
filelist_first (filelist_t * filelist)
{
    assert (ho_checkref (filelist));
    assert (filelist_is_locked (filelist));

    return (fileitem_t *) l_list_first (filelist->list);
}


fileitem_t *
filelist_prev (filelist_t * filelist, fileitem_t * fileitem)
{
    assert (ho_checkref (filelist));
    assert (ho_checkref (fileitem));
    assert (filelist == fileitem->parent_list);
    assert (filelist_is_locked (fileitem->parent_list));

    return (fileitem_t *) l_list_prev (filelist->list, fileitem);
}


fileitem_t *
filelist_next (filelist_t * filelist, fileitem_t * fileitem)
{
    assert (ho_checkref (filelist));
    assert (ho_checkref (fileitem));
    assert (filelist == fileitem->parent_list);
    assert (filelist_is_locked (fileitem->parent_list));

    return (fileitem_t *) l_list_next (filelist->list, fileitem);
}


fileitem_t *
filelist_get_by_mrl (filelist_t * filelist, const char *mrl)
{
    assert (ho_checkref (filelist));

    if (!mrl) {
        return NULL;
    }

    filelist_lock (filelist);

    fileitem_t *founditem = NULL;
    fileitem_t *fileitem = filelist_first (filelist);
    while (fileitem) {
        if (mrls_are_equal (fileitem->mrl, mrl)) {
            founditem = fileitem;
            break;
        }
        fileitem = filelist_next (filelist, fileitem);
    }

    filelist_unlock (filelist);

    return founditem;
}


fileitem_t *
filelist_get_by_pos (filelist_t * filelist, int pos)
{
    assert (ho_checkref (filelist));

    filelist_lock (filelist);

    if (pos >= filelist_length (filelist)) {
        return NULL;
    }

    int i = 0;
    fileitem_t *founditem = NULL;
    fileitem_t *fileitem = filelist_first (filelist);
    while (fileitem) {
        if (i == pos) {
            founditem = fileitem;
            break;
        }
        fileitem = filelist_next (filelist, fileitem);
        i += 1;
    }

    filelist_unlock (filelist);

    return founditem;
}


bool
filelist_contains (filelist_t * filelist, const char *mrl)
{
    assert (ho_checkref (filelist));

    filelist_lock (filelist);

    bool contains = false;
    if (mrl) {
        fileitem_t *fileitem = filelist_first (filelist);
        while (fileitem) {
            if (strcasecmp (fileitem->mrl, mrl) == 0) {
                contains = true;
                break;
            }
            fileitem = filelist_next (filelist, fileitem);
        }
    }

    filelist_unlock (filelist);

    return contains;
}


int
filelist_length (filelist_t * filelist)
{
    int result = -1;

    if (filelist) {
        filelist_lock (filelist);
        result = l_list_length (filelist->list);
        filelist_unlock (filelist);
    }

    return result;
}


void
filelist_set_pwd_callback (fileitem_pwd_cb_t cb)
{
    fileitem_pwd_cb = cb;
}


void
filelist_set_length_cb (fileitem_len_cb_t cb)
{
    fileitem_len_cb = cb;
}


void
filelist_set_allowed (filelist_t * filelist, fileitem_allowed_t allowed)
{
    assert (ho_checkref (filelist));

    filelist_lock (filelist);

    filelist->allowed_filetypes = allowed;

    fileitem_t *item = filelist_first (filelist);
    while (item) {
        if (item->child_list) {
            filelist_set_allowed (item->child_list, allowed);
        }
        item = filelist_next (filelist, item);
    }

    filelist_unlock (filelist);
}


const char *
filelist_get_cover_image (filelist_t * filelist)
{
    const char *mrl = NULL;

    assert (ho_checkref (filelist));

    filelist_lock (filelist);

    if (is_cover_image_valid (filelist->thumbnail_mrl)) {
        mrl = filelist->thumbnail_mrl;
    }
    else {
        ho_free (filelist->thumbnail_mrl);
        filelist->thumbnail_mrl = get_cover_image_from_mrl (filelist->mrl);
        mrl = filelist->thumbnail_mrl;
    }

    filelist_unlock (filelist);

    return mrl;
}


const char *
fileitem_get_cover_image (fileitem_t * fileitem)
{
    const char *mrl = NULL;

    assert (ho_checkref (fileitem));
    assert (ho_checkref (fileitem->parent_list));
    assert (filelist_is_locked (fileitem->parent_list));

    if (is_cover_image_valid (fileitem->thumbnail_mrl)) {
        mrl = fileitem->thumbnail_mrl;
    }
    else if (mrl_is_image (fileitem->mrl)) {
        ho_free (fileitem->thumbnail_mrl);
        fileitem->thumbnail_mrl = ho_strdup (fileitem->mrl);
        mrl = fileitem->thumbnail_mrl;
    }
    else {
        ho_free (fileitem->thumbnail_mrl);
        fileitem->thumbnail_mrl = get_cover_image_from_mrl (fileitem->mrl);
        mrl = fileitem->thumbnail_mrl;
    }

    return mrl;
}


int
fileitem_get_child_count (fileitem_t * fileitem)
{
    assert (ho_checkref (fileitem));
    assert (ho_checkref (fileitem->parent_list));
    assert (filelist_is_locked (fileitem->parent_list));

    char *cache_dir = mrl_get_cache_dir (fileitem->mrl);
    char *cache_file = ho_strdup_printf ("%s/count", cache_dir);

    int child_count = -1;
    if (fileitem->child_list) {
        child_count = filelist_length (fileitem->child_list);
    }
    else if (file_exists (cache_file)) {
        int size = 0;
        char *buffer = read_entire_file (cache_file, &size);
        if (buffer) {
            child_count = atoi (buffer);
        }
        ho_free (buffer);
    }
    else {
        switch (fileitem->type) {
        case FILE_TYPE_FOLDER:
            if (mrl_is_directory (fileitem->mrl)) {
                filelist_expand (fileitem);
            }
            break;
        default:
            if (!mrl_is_networked (fileitem->mrl)) {
                filelist_expand (fileitem);
            }
            break;
        }

        child_count = filelist_length (fileitem->child_list);
    }
    ho_free (cache_dir);
    ho_free (cache_file);

    return child_count;
}


void
fileitem_free_info_list (fileitem_t * fileitem)
{
    assert (ho_checkref (fileitem));
    assert (ho_checkref (fileitem->parent_list));
    assert (filelist_is_locked (fileitem->parent_list));

    if (fileitem->info_list) {
        l_list_free (fileitem->info_list, fileinfo_free_cb);
        fileitem->info_list = NULL;
    }
}


l_list_t *
filelist_get_info_list (filelist_t * filelist)
{
    assert (ho_checkref (filelist));
    assert (filelist_is_locked (filelist));

    if (filelist->info_list) {
        return filelist->info_list;
    }

    filelist->info_list = l_list_new ();

    {
        int count = filelist_length (filelist);
        int length = filelist->playback_length;
        char *time = meta_info_get_playback_length_str (length);

        if (count == 0) {
            char *t = ho_strdup (_("The current folder contains "
                                   "no entries."));
            l_list_append (filelist->info_list, t);
        }
        else if (length && time) {
            char *f = _("The current folder contains %d %s with a total "
                        "playback time of %s.");
            char *e = (count == 1) ? _("entry") : _("entries");
            char *t = ho_strdup_printf (f, count, e, time);
            l_list_append (filelist->info_list, t);
        }
        else {
            char *f = _("The current folder contains %d %s.");
            char *e = (count == 1) ? _("entry") : _("entries");
            char *t = ho_strdup_printf (f, count, e);
            l_list_append (filelist->info_list, t);
        }
        ho_free (time);
    }

    return filelist->info_list;
}


static void
fileitem_info_list_append (l_list_t * info_list, const char *title,
                           const char *value)
{
    char *t = xml_escape_string (value, XML_ESCAPE_NO_QUOTE);
    char *i = ho_strdup_printf (INFO_LIST_FORMAT, title, t);
    l_list_append (info_list, i);
    ho_free (t);
}


l_list_t *
fileitem_get_info_list (fileitem_t * fileitem)
{
    assert (ho_checkref (fileitem));
    assert (ho_checkref (fileitem->parent_list));
    assert (filelist_is_locked (fileitem->parent_list));

    if (fileitem->info_list) {
        return fileitem->info_list;
    }

    fileitem->info_list = l_list_new ();
    switch (fileitem->type) {
    case FILE_TYPE_REGULAR:
        if (false) {
            /* do nothing */
        }
        else if (mrl_is_networked (fileitem->mrl) && fileitem->title) {
            fileitem_info_list_append (fileitem->info_list, _("Title:"),
                                       fileitem->title);
        }
#ifdef HAVE_VDR
        else if (strc_starts_with (fileitem->mrl, FILELIST_VDR_TIM_MRL)) {
            vdr_timer_read_info (fileitem);
        }
#endif /* HAVE_VDR */
#ifdef HAVE_LIBEXIF
        else if (mrl_is_image (fileitem->mrl)) {
            char *info = NULL;
            char *date = exif_info_get (EXIF_INFO_DATE, fileitem->mrl);
            char *time = exif_info_get (EXIF_INFO_TIME, fileitem->mrl);
            if (date && time) {
                info = ho_strdup_printf (_("The selected image was "
                                           "taken on %s at %s."), date, time);
            }
            else if (date) {
                info = ho_strdup_printf (_("The selected image was "
                                           "taken on %s."), date);
            }
            else if (time) {
                info = ho_strdup_printf (_("The selected image was "
                                           "taken at %s."), time);
            }
            if (info) {
                l_list_append (fileitem->info_list, info);
            }

            ho_free (date);
            ho_free (time);
        }
#endif /* HAVE_LIBEXIF */
        else {
            char *mrl = fileitem->mrl;
            time_t now = time (NULL);
            struct tm *tm_now = localtime (&now);
            int current_year = tm_now->tm_year + 1900;

            char *title = meta_info_get (META_INFO_TITLE, mrl);
            if (title && (strlen (title) > 0)) {
                fileitem_info_list_append (fileitem->info_list, _("Title:"),
                                           title);
            }
            else if (fileitem->title) {
                fileitem_info_list_append (fileitem->info_list, _("Title:"),
                                           fileitem->title);
            }
            ho_free (title);

            char *album = meta_info_get (META_INFO_ALBUM, mrl);
            if (album && (strlen (album) > 0)) {
                fileitem_info_list_append (fileitem->info_list, _("Album:"),
                                           album);
            }
            ho_free (album);

            char *artist = meta_info_get (META_INFO_ARTIST, mrl);
            if (artist && (strlen (artist) > 0)) {
                fileitem_info_list_append (fileitem->info_list, _("Artist:"),
                                           artist);
            }
            ho_free (artist);

            char *genre = meta_info_get (META_INFO_GENRE, mrl);
            if (genre && (strlen (genre) > 0)) {
                fileitem_info_list_append (fileitem->info_list, _("Genre:"),
                                           genre);
            }
            ho_free (genre);

            char *year = meta_info_get (META_INFO_YEAR, mrl);
            if (year && (atoi (year) > 0) && (atoi (year) <= current_year)) {
                fileitem_info_list_append (fileitem->info_list, _("Year:"),
                                           year);
            }
            ho_free (year);

            char *track = meta_info_get (META_INFO_TRACK_NUMBER, mrl);
            if (track && (strlen (track) > 0)) {
                fileitem_info_list_append (fileitem->info_list, _("Track:"),
                                           track);
            }
            ho_free (track);

            int length = fileitem->playback_length;
            char *time = meta_info_get_playback_length_str (length);
            if (length && time) {
                fileitem_info_list_append (fileitem->info_list, _("Length:"),
                                           time);
            }
            ho_free (time);

            char *quality = mrl_get_quality (mrl);
            if (quality && (strlen (quality) > 0)) {
                fileitem_info_list_append (fileitem->info_list, _("Quality:"),
                                           _(quality));
            }
            ho_free (quality);
        }
        break;
    case FILE_TYPE_PLAYLIST:
        {
            fileitem_info_list_append (fileitem->info_list, _("Title:"),
                                       fileitem->title);
        }
        break;
    case FILE_TYPE_MEDIAMARKS:
    case FILE_TYPE_FOLDER:
        if (false) {
            /* do nothing */
        }
#ifdef HAVE_VDR
        else if (mrls_are_equal (fileitem->mrl, FILELIST_VDR_MRL)) {
            char *t = _("The selected folder contains television folders.");
            l_list_append (fileitem->info_list, ho_strdup (t));
        }
        else if (mrls_are_equal (fileitem->mrl, FILELIST_VDR_REC_MRL)) {
            char *t = _("The selected folder contains "
                        "television recordings.");
            l_list_append (fileitem->info_list, ho_strdup (t));
        }
        else if (mrls_are_equal (fileitem->mrl, FILELIST_VDR_TIM_MRL)) {
            char *t = _("The selected folder contains television timer.");
            l_list_append (fileitem->info_list, ho_strdup (t));
        }
        else if (mrls_are_equal (fileitem->mrl, FILELIST_VDR_CHA_MRL)) {
            char *t = _("The selected folder contains television channels.");
            l_list_append (fileitem->info_list, ho_strdup (t));
        }
#endif
#ifdef HAVE_HAL
        else if (fileitem->hal.device) {
            int count = fileitem_get_child_count (fileitem);
            /* As we may expand the fileitem in fileitem_get_child_count and
             * the info list is freed in filelist_expand, we have to
             * recreate it here. */
            if (!fileitem->info_list) {
                fileitem->info_list = l_list_new ();
            }

            /* A fileitem of the type FOLDER that has a non null device
             * can only be an audio CD or a mountable volume. */
            char *device = fileitem->hal.device;
            if (volume_get_type (device) == VOLUME_TYPE_AUDIO_CD) {
                char *f = _("The selected Audio CD contains %d %s.");
                char *e = (count == 1) ? _("entry") : _("entries");
                char *t = ho_strdup_printf (f, count, e);
                l_list_append (fileitem->info_list, t);
            }
            else if (!volume_is_mounted (device)) {
                char *t = _("The selected volume is not mounted.");
                l_list_append (fileitem->info_list, ho_strdup (t));
            }
            else if (count == 0) {
                char *t = _("The selected volume is mounted "
                            "and contains no entries.");
                l_list_append (fileitem->info_list, ho_strdup (t));
            }
            else {
                char *f = _("The selected volume is mounted "
                            "and contains %d %s.");
                char *e = (count == 1) ? _("entry") : _("entries");
                char *t = ho_strdup_printf (f, count, e);
                l_list_append (fileitem->info_list, t);
            }
        }
#endif
        else {
            int count = fileitem_get_child_count (fileitem);
            /* As we may expand the fileitem in fileitem_get_child_count and
             * the info list is freed in filelist_expand, we have to
             * recreate it here. */
            if (!fileitem->info_list) {
                fileitem->info_list = l_list_new ();
            }

            if (count == -1) {
                /* do nothing */
            }
            else if (count == 0) {
                char *t = _("The selected folder contains no entries.");
                l_list_append (fileitem->info_list, ho_strdup (t));
            }
            else if (mrls_are_equal (fileitem->mrl, get_dir_playlists ())) {
                char *f = _("The selected folder contains %d saved %s.");
                char *e = (count == 1) ? _("playlist") : _("playlists");
                char *t = ho_strdup_printf (f, count, e);
                l_list_append (fileitem->info_list, t);
            }
            else {
                char *f = _("The selected folder contains %d %s.");
                char *e = (count == 1) ? _("entry") : _("entries");
                char *t = ho_strdup_printf (f, count, e);
                l_list_append (fileitem->info_list, t);
            }

            if (fileitem->listened_count) {
                char *f = _("The selected folder has been "
                            "listened to %d %s.");
                char *e = (count == 1) ? _("time") : _("times");
                char *t = ho_strdup_printf (f, fileitem->listened_count, e);
                l_list_append (fileitem->info_list, t);
            }
            if (fileitem->listened_time > 0) {
                const char *df = config_get_string ("gui.date_format");
                struct tm ts;
                localtime_r (&fileitem->listened_time, &ts);
                char *d = ho_strdup_strftime (df, &ts);

                char *f = _("The selected folder was listened "
                            "to on %s for the last time.");
                char *t = ho_strdup_printf (f, d);
                l_list_append (fileitem->info_list, t);
            }
        }
        break;
    default:
        break;
    }

    return fileitem->info_list;
}


#ifdef DEBUG
void
filelist_print (filelist_t * filelist)
{
    assert (ho_checkref (filelist));

    debug ("=============================================");
    debug ("Filelist: %s", filelist->title);
    debug ("          %s", filelist->mrl);

    fileitem_t *item = filelist_first (filelist);
    while (item) {
        debug ("Fileitem: %s", item->title);
        debug ("          %s", item->mrl);
        item = filelist_next (filelist, item);
    }
    debug ("=============================================");
}
#endif


bool
filelist_check_for_changes (filelist_t * filelist)
{
    bool list_has_changed = false;

    filelist_t *cur_list = NULL;
    filelist_t *new_list = NULL;
    filelist_t *sub_list = NULL;
    fileitem_t *cur_item = NULL;
    fileitem_t *new_item = NULL;

    if (!filelist->update_allowed) {
        return false;
    }
    if (!filelist->parent_item) {
        return false;
    }

    filelist_lock (filelist);

    ho_setref (cur_list, filelist);
    ho_setref (cur_item, filelist->parent_item);

    /* As the filelist we're looking at may be visible at the moment we do
     * not want to change it. */
    fileitem_allowed_t allowed = cur_list->allowed_filetypes;
    new_list = filelist_new (NULL, FILELIST_TOPLEVEL_TITLE,
                             FILELIST_TOPLEVEL_MRL, allowed);
    filelist_lock (new_list);

    /* We add and expand a copy of the parent item of the list we want to
     * check to our new filelist. */
    new_item = filelist_add (new_list, cur_item->title,
                             cur_item->mrl, cur_item->type);
    new_item->cmd = cur_item->cmd ? ho_strdup (cur_item->cmd) : NULL;
    filelist_unlock (filelist);
    filelist_expand (new_item);
    filelist_lock (filelist);

    ho_setref (sub_list, new_item->child_list);
    filelist_lock (sub_list);

    {
        int cur_length = filelist_length (cur_list);
        int sub_length = filelist_length (sub_list);
        list_has_changed = (cur_length != sub_length);
    }

    if (!list_has_changed) {
        fileitem_t *cur_item = filelist_first (cur_list);
        fileitem_t *sub_item = filelist_first (sub_list);
        while (cur_item && sub_item && !list_has_changed) {
#ifdef HAVE_VDR
            if (cur_item->vdr.is_active != sub_item->vdr.is_active) {
                list_has_changed = true;
            }
            if (cur_item->vdr.is_recording != sub_item->vdr.is_recording) {
                list_has_changed = true;
            }
            if (cur_item->vdr.is_scheduled != sub_item->vdr.is_scheduled) {
                list_has_changed = true;
            }
#endif /* HAVE_VDR */
            if (!str_equals (cur_item->mrl, sub_item->mrl)) {
                list_has_changed = true;
            }
            if (!str_equals (cur_item->title, sub_item->title)) {
                list_has_changed = true;
            }
            cur_item = filelist_next (cur_list, cur_item);
            sub_item = filelist_next (sub_list, sub_item);
        }
    }

    filelist_unlock (sub_list);
    filelist_unlock (new_list);

    ho_setref (sub_list, NULL);
    ho_setref (new_list, NULL);

    filelist_unlock (filelist);

    return list_has_changed;
}
