
/*
 * Copyright (C) 2002-2003 Stefan Holst
 * Copyright (C) 2004-2005 Maximilian Schwerin
 *
 * This file is part of oxine a free media player.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: filelist.h 4004 2009-09-26 12:38:07Z mschwerin $
 *
 */

#ifndef HAVE_FILELIST_H
#define HAVE_FILELIST_H

#include <assert.h>

#include "heap.h"
#include "list.h"
#include "mutex.h"
#include "types.h"

#define FILELIST_TOPLEVEL_MRL           "computer://"
#define FILELIST_TOPLEVEL_TITLE         _("Computer")

#define FILELIST_REMOVABLE_MRL          "computer://removable"
#define FILELIST_REMOVABLE_TITLE        _("Removable Drives")

#define FILELIST_TEMPORARY_MRL          "computer://temporary"
#define FILELIST_TEMPORARY_TITLE         _("Temporary")

#define FILELIST_EXTRACTOR_MRL          "computer://extractor"
#define FILELIST_EXTRACTOR_TITLE         _("CD / DVD Extractor")

#define FILELIST_MEDIA_MRL              get_mrl_mediamarks_media ()
#define FILELIST_MEDIA_TITLE            _("Media")
#define FILELIST_MUSIC_MRL              get_mrl_mediamarks_music ()
#define FILELIST_MUSIC_TITLE            _("Music")
#define FILELIST_VIDEO_MRL              get_mrl_mediamarks_video ()
#define FILELIST_VIDEO_TITLE            _("Videos")
#define FILELIST_IMAGE_MRL              get_mrl_mediamarks_image ()
#define FILELIST_IMAGE_TITLE            _("Images")
#define FILELIST_SUBTITLE_MRL           get_mrl_mediamarks_subtitle ()
#define FILELIST_SUBTITLE_TITLE         _("Subtitle")

#define FILELIST_HOME_MRL               get_mrl_home ()
#define FILELIST_HOME_TITLE             _("Home")
#define FILELIST_ROOT_MRL               get_mrl_root ()
#define FILELIST_ROOT_TITLE             _("Filesystem")
#define FILELIST_DESKTOP_MRL            get_mrl_desktop ()
#define FILELIST_DESKTOP_TITLE          _("Desktop")

#define FILELIST_XDG_MUSIC_MRL          get_mrl_xdg_music_dir ()
#define FILELIST_XDG_MUSIC_TITLE        _("Music")
#define FILELIST_XDG_VIDEOS_MRL         get_mrl_xdg_videos_dir ()
#define FILELIST_XDG_VIDEOS_TITLE       _("Videos")
#define FILELIST_XDG_PICTURES_MRL       get_mrl_xdg_pictures_dir ()
#define FILELIST_XDG_PICTURES_TITLE     _("Images")

#define FILELIST_FAVORITES_MOST_MRL     get_mrl_favorites_most ()
#define FILELIST_FAVORITES_MOST_TITLE   _("Favorites - Most listened to")

#define FILELIST_FAVORITES_RECENT_MRL   get_mrl_favorites_recent ()
#define FILELIST_FAVORITES_RECENT_TITLE _("Favorites - Recently listened to")

#define FILELIST_PLAYLISTS_MRL          get_mrl_playlists ()
#define FILELIST_PLAYLISTS_TITLE        _("Playlists")

#ifdef HAVE_VDR
#define FILELIST_VDR_MRL                "vdr://"
#define FILELIST_VDR_EPG_MRL            "vdr://epg"
#define FILELIST_VDR_CHA_MRL            "vdr://channels"
#define FILELIST_VDR_REC_MRL            "vdr://recordings"
#define FILELIST_VDR_TIM_MRL            "vdr://timers"
#endif

typedef enum {
    FILE_TYPE_UNKNOWN,
    /// A virtual or real folder.
    FILE_TYPE_FOLDER,
    /// A virtual folder containing mediamarks.
    FILE_TYPE_MEDIAMARKS,
    /// A playlist file
    FILE_TYPE_PLAYLIST,
    /// A normal file or MRL.
    FILE_TYPE_REGULAR,
} fileitem_type_t;

typedef enum {
    /// Allow only music files to be displayed.
    ALLOW_FILES_MUSIC,
    /// Allow only video files to be displayed.
    ALLOW_FILES_VIDEO,
    /// Allow only image files to be displayed.
    ALLOW_FILES_IMAGE,
    /// Allow only subtitle files to be displayed.
    ALLOW_FILES_SUBTITLE,
    /// Allow all multimedia files (music, video, subtitles) to be displayed.
    ALLOW_FILES_MULTIMEDIA,
    /// Allow any file to be displayed.
    ALLOW_FILES_ALL
} fileitem_allowed_t;

typedef struct filelist_s filelist_t;
typedef struct fileitem_s fileitem_t;

#ifdef HAVE_VDR
typedef struct {

    /**
     * The EPG-ID of this fileitem.
     */
    int epg_id;

    /**
     * The channel ID of this fileitem.
     */
    int channel_id;

    /**
     * The start time of this fileitem.
     */
    time_t time_beg;

    /**
     * The end time of this fileitem.
     */
    time_t time_end;

    /**
     * This is used by timers.
     */
    bool is_active;

    /**
     * This is used by timers.
     */
    bool is_recording;

    /**
     * This is used by timers.
     */
    bool is_scheduled;
} vdr_info_t;
#endif /* HAVE_VDR */

#ifdef HAVE_HAL
typedef struct {

    /**
     * The device file corresponding to this fileitem.
     *
     * This contains the full pathname of the device. For a CDROM this would
     * be something like /dev/cdrom.
     *
     * This is used for removable media.
     */
    char *device;

    /**
     * The unique device identifier (UDI) of a volume.
     *
     * This is an identifier, the Unique Device Identifer, that is unique for
     * a device object - that is, no other device object can have the same UDI
     * at the same time. The UDI is computed using bus-specific information
     * and is meant to be unique across device insertions and independent of
     * the physical port or slot the device may be plugged into.
     *
     * This is used for removable media.
     */
    char *volume_udi;

    /**
    * The unique device identifier (UDI) of a drive.
    *
     * This is an identifier, the Unique Device Identifer, that is unique for
     * a device object - that is, no other device object can have the same UDI
     * at the same time. The UDI is computed using bus-specific information
     * and is meant to be unique across device insertions and independent of
     * the physical port or slot the device may be plugged into.
     *
     * This is used for removable media.
     */
    char *drive_udi;
} hal_info_t;
#endif /* HAVE_HAL */

/**
 * This datastructure represents one entry of a filelist.
 */
struct fileitem_s {
    referable_t r;

    /**
     * The original title of this item.
     */
    char *title;

    /**
     * The title of this item that is visible in the GUI. This string may
     * contain more information (e.g. start time of a recording) than the
     * title alone.
     */
    char *title_gui;

    /**
     * The title of this item that is used for searching. This is
     * essentially the original title without special characters.
     */
    char *title_search;

    /**
     * The command to create the content of this fileitem.
     */
    char *cmd;

    /**
     * The MRL of this item.
     */
    char *mrl;

    /**
     * The MRL of the thumbnail belonging to this fileitem.
     */
    char *thumbnail_mrl;

    /**
     * Sort by this string.
     */
    char *sort_by;

#ifdef HAVE_LIBEXIF

    /**
     * Sort by the EXIF date.
     */
    char *sort_by_exif;
#endif

    /**
     * Is this location a mountpoint.
     */
    bool is_mountpoint;

    /**
     * The type of this fileitem.
     */
    fileitem_type_t type;

    /**
     * The list of items below this item. If this fileitem is of the type
     * FOLDER, MEDIAMARKS or PLAYLIST the entries in this list are child
     * items. If this fileitem is of the type REGULAR the entries are
     * alternative MRLs of the main MRL.
     */
    filelist_t *child_list;

    /**
     * The parent of this fileitem.
     */
    filelist_t *parent_list;

    /**
     * The playback length of this item in seconds.
     */
    int playback_length;

    /**
     * A list of strings with info about this fileitem.
     */
    l_list_t *info_list;

    /**
     * The number of times this item has already been listened to. This is
     * used in the favorites list.
     */
    int listened_count;

    /**
     * The last time (unix time) this item was listened to. This is used in
     * the favorites list.
     */
    time_t listened_time;

#ifdef HAVE_VDR
    vdr_info_t vdr;
#endif                                                          /* HAVE_VDR */
#ifdef HAVE_HAL
    hal_info_t hal;
#endif                                                          /* HAVE_HAL */
};


/**
 * This datastructure represents a filelist.
 */
struct filelist_s {
    referable_t r;

    /**
     * The linked list containing the fileitems.
     */
    l_list_t *list;

    /**
     * The title of this list.
     */
    char *title;

    /**
     * The complete title of a list. This is created by concatenating the
     * title of the parent list with the title of this list.
     */
    char *title_complete;

    /**
     * The MRL of this list.
     */
    char *mrl;

    /**
     * A thumbnail of this list.
     */
    char *thumbnail_mrl;

    /**
     * This is TRUE if a password has to be provided.
     */
    bool wait_for_password;

    /**
     * This is non-NULL if an error occured while reading this list.
     */
    char *error;

    /**
     * The parent item of this list.
     */
    fileitem_t *parent_item;

    /**
     * The type of files allowed in this filelist.
     */
    fileitem_allowed_t allowed_filetypes;

    /**
     * A mutex to protect the list.
     */
    pthread_mutex_t mutex;
    pthread_mutexattr_t mutex_attr;

    /**
     * This is used to save the top position of list when displaying it.
     */
    int top_position;

    /**
     * This is used to save the current position in the list when displaying it.
     */
    int cur_position;

    /**
     * The playback length of this list in seconds.
     */
    int playback_length;

    /**
     * If TRUE this filelist may be monitored for changes.
     */
    bool update_allowed;

    /**
     * A list of strings with info about this filelist.
     */
    l_list_t *info_list;
};


/**
 * A callback that is used to retrieve the playback length of a fileitem.
 *
 * @param mrl                   The MRL of the item to retrieve the playback
 *                              length of.
 */
typedef int (*fileitem_len_cb_t) (const char *mrl);


/**
 * A callback that is used to retrieve the password of a fileitem. The
 * callback is used to check the encrypted password <code>pwd</code> for an
 * MRL against a user provided password. If the callback returns TRUE access
 * to the MRL is allowed.
 *
 * @param mrl                   The MRL of the item we want to retrieve the
 *                              password for.
 */
typedef bool (*fileitem_pwd_cb_t) (const char *mrl, const char *pwd);


/**
 * Initializes the static data needed by the filelist module.
 */
void filelist_static_init (void);


/**
 * Frees the static data needed by the filelist module.
 */
void filelist_static_free (void);


/**
 * Sets a callback function for providing a password for a MRL. The callback
 * is used to check the encrypted password <code>pwd</code> for an MRL
 * against a user provided password. If the callback returns TRUE access to
 * the MRL is allowed.
 */
void filelist_set_pwd_callback (fileitem_pwd_cb_t cb);


/**
 * Set the callback used to retrieve the playback length in seconds of a
 * fileitem.
 */
void filelist_set_length_cb (fileitem_len_cb_t cb);


/**
 * Sets the filetypes that are allowed in a filelist and all sublists.
 *
 * @param filelist              A filelist object.
 * @param allowed_filetypes     What filetypes may be added to this filelist.
 */
void filelist_set_allowed (filelist_t * filelist, fileitem_allowed_t allowed);


/**
 * Returns the filelist for the given MRL. If no such filelist exists, NULL
 * is returned.
 *
 * @param mrl                   The MRL of the filelist to return.
 */
filelist_t *filelist_get (const char *mrl);


/**
 * Creates a new filelist. Only files of the allowed type will be added when
 * expanding a directory.
 *
 * The returned filelist has a reference count of one. It will be free when
 * the reference count reaches 0.
 *
 * @param parent                The parent of this filelist. This can be NULL.
 * @param title                 The title of this filelist.
 * @param mrl                   The MRL of this filelist.
 * @param allowed_filetypes     This defines what filetypes may be added to
 *                              this filelist.
 */
filelist_t *filelist_new (fileitem_t * parent,
                          const char *title, const char *mrl,
                          fileitem_allowed_t allowed_filetypes);


/**
 * Removes all fileitems from the list.
 *
 * @param filelist              A pointer to a filelist data structure.
 */
void filelist_clear (filelist_t * filelist);


/**
 * Locks the filelist mutex. Make sure to always lock filelists from top to
 * bottom. NEVER lock the parent of a filelist that you've already locked!
 *
 * @param filelist              A pointer to a filelist data structure.
 */
#define filelist_lock(filelist)     mutex_lock (&(filelist)->mutex)


/**
 * Tries to lock the filelist mutex.
 *
 * @param filelist              A pointer to a filelist data structure.
 */
#define filelist_trylock(filelist)  mutex_trylock (&(filelist)->mutex)


/**
 * Unlocks the filelist mutex.
 *
 * @param filelist              A pointer to a filelist data structure.
 */
#define filelist_unlock(filelist)   mutex_unlock (&(filelist)->mutex)


/**
 * Sorts a filelist.
 *
 * @param filelist              A pointer to a filelist data structure.
 * @param swap_cb               A callback, that can be used to determin if
 *                              two entries must be swapped. This may be NULL
 *                              in which case the standard behaviour is used.
 */
void filelist_sort (filelist_t * filelist, l_swap_cb_t swap_cb);


/**
 * Adds a new entry to the list. Use this method if you know all the
 * parameters.
 *
 * The returned fileitem has a reference count of one. It will be free when
 * the reference count reaches 0.
 *
 * @param filelist              A pointer to a filelist data structure.
 * @param title                 The title of the new entry.
 * @param mrl                   The MRL of the new entry.
 * @param type                  The type of the new entry.
 */
fileitem_t *filelist_add (filelist_t * filelist, const char *title,
                          const char *mrl, fileitem_type_t type);


/**
 * Adds a new entry to the list. Use this method if do not know some of the
 * parameters and want them to be detected automaticaly.
 *
 * The returned fileitem has a reference count of one. It will be free when
 * the reference count reaches 0.
 *
 * @param filelist              A pointer to a filelist data structure.
 * @param title                 The title of the new entry. If this is NULL
 *                              the title will be generated automatically.
 * @param mrl                   The MRL of the new entry.
 */
fileitem_t *filelist_add_magic (filelist_t * filelist, const char *title,
                                const char *mrl);


/**
 * Removes the fileitem from the list.
 *
 * @param filelist              A pointer to a filelist data structure.
 * @param fileitem              The fileitem to remove.
 */
void filelist_remove (filelist_t * filelist, fileitem_t * fileitem);


/**
 * Re-reads the items of this filelist.
 *
 * @param filelist              The filelist to update.
 * @return                      The filelist
 */
filelist_t *filelist_update (filelist_t * filelist);


/**
 * Depending on the type of the fileitem it is expanded. Expanding means
 * that e.g. a directory is read and any files in the directory are added to
 * a sublist of this fileitem.
 *
 * @param fileitem              The fileitem to expand.
 * @return                      The sublist (if any).
 */
filelist_t *filelist_expand (fileitem_t * fileitem);


/**
 * Returns the first fileitem in the list.
 *
 * @param filelist              A pointer to a filelist data structure.
 */
fileitem_t *filelist_first (filelist_t * filelist);


/**
 * Returns the previous fileitem in the list.
 *
 * @param filelist              A pointer to a filelist data structure.
 * @param fileitem              The next fileitem.
 */
fileitem_t *filelist_prev (filelist_t * filelist, fileitem_t * fileitem);


/**
 * Returns the next fileitem in the list.
 *
 * @param filelist              A pointer to a filelist data structure.
 * @param fileitem              The previous fileitem.
 */
fileitem_t *filelist_next (filelist_t * filelist, fileitem_t * fileitem);


/**
 * Returns (if its there) the fileitem with the passed MRL.
 *
 * @param filelist              A pointer to a filelist data structure.
 * @param mrl                   This is the MRL to search for.
 */
fileitem_t *filelist_get_by_mrl (filelist_t * filelist, const char *mrl);


/**
 * Returns the fileitem at the passed position.
 *
 * @param filelist              A pointer to a filelist data structure.
 * @param pos                   This position of the item to return.
 */
fileitem_t *filelist_get_by_pos (filelist_t * filelist, int pos);


/**
 * Determines, if the list contains an entry with the given MRL.
 *
 * @param filelist              A pointer to a filelist data structure.
 * @param mrl                   This is the MRL to search for.
 */
bool filelist_contains (filelist_t * filelist, const char *mrl);


/**
 * Returns the number of fileitems in this list. If the passed filelist is
 * NULL the function returns -1.
 *
 * @param filelist              A pointer to a filelist data structure.
 */
int filelist_length (filelist_t * filelist);


/**
 * Determines if the file is allowed depending on the allowed filetypes. The
 * suffix of the file is compared to all suffixes of the allowed type.
 *
 * @param fileitem              A pointer to a fileitem data structure.
 * @param allowed               The filetypes that are allowed.
 */
bool fileitem_is_allowed (fileitem_t *fileitem, fileitem_allowed_t allowed);


/**
 * Returns the MRL of a thumbnail image of a filelist.
 *
 * @param filelist              A pointer to a filelist data structure.
 */
const char *filelist_get_cover_image (filelist_t * filelist);


/**
 * Returns the MRL of a thumbnail image of a fileitem.
 *
 * @param fileitem              A pointer to a fileitem data structure.
 */
const char *fileitem_get_cover_image (fileitem_t * fileitem);


#ifdef DEBUG

/**
 * Prints the filelist to the console.
 *
 * @param filelist              A pointer to a filelist data structure.
 */
void filelist_print (filelist_t * filelist);
#endif


/**
 * Returns TRUE if the passed MRL points to an audio file.
 */
bool mrl_is_audio (const char *mrl);


/**
 * Returns TRUE if the passed MRL points to a video file.
 */
bool mrl_is_video (const char *mrl);


/**
 * Returns TRUE if the passed MRL points to an image file.
 */
bool mrl_is_image (const char *mrl);


/**
 * Returns TRUE if the passed MRL points to a subtitle file.
 */
bool mrl_is_subtitle (const char *mrl);


/**
 * Returns TRUE if the passed MRL points to a playlist file.
 */
bool mrl_is_playlist (const char *mrl);


/**
 * Returns TRUE if the fileitem is a folder.
 *
 * @param fileitem              A pointer to a fileitem data structure.
 */
bool fileitem_is_folder (fileitem_t * fileitem);


/**
 * Returns the number of items in this items child list. If the item has not
 * yet been expanded this is done (only for local directories and mediamarks).
 *
 * @param fileitem              A pointer to a fileitem data structure.
 * @return                      The number of items in the child list or -1
 *                              if this item does not have a child list.
 */
int fileitem_get_child_count (fileitem_t * fileitem);


/**
 * Returns a list containing lines of text with information about the
 * specified fileitem.
 *
 * @param fileitem              A pointer to a fileitem data structure.
 */
l_list_t *fileitem_get_info_list (fileitem_t * fileitem);


/**
 * Returns a list containing lines of text with information about the
 * specified filelist.
 *
 * @param filelist              A pointer to a filelist data structure.
 */
l_list_t *filelist_get_info_list (filelist_t * filelist);


/**
 * Frees the info list of a fileitem.
 *
 * @param fileitem              A pointer to a fileitem data structure.
 */
void fileitem_free_info_list (fileitem_t * fileitem);


/**
 * Checks a filelist for changes. Only local folders, VDR timers and VDR
 * recordings are checked. For all other types of filelists the function
 * will always return FALSE.
 *
 * @param filelist              A pointer to a filelist data structure.
 * @return                      Returns TRUE if the content of the filelist
 *                              has changed.
 */
bool filelist_check_for_changes (filelist_t * filelist);


/**
 * Adds an alternative MRL to the given fileitem. The fileitem must be of
 * the type REGULAR.
 *
 * @param fileitem              The fileitem to add the alternative MRL to.
 * @param mrl                   The alternative MRL.
 *
 * @return                      The fileitem of the alternative MRL.
 */
fileitem_t *fileitem_add_alternative (fileitem_t * fileitem, const char *mrl);


/**
 * Callback to swap fileitems. This is the callback passed to
 * <code>l_list_sort</code>, to determin, if two fileitems have
 * to be reordered.
 */
bool fileitem_swap_cb (void *d1, void *d2);


#endif /* HAVE_FILELIST_H */
