
/*
 * Copyright (C) 2002-2003 Stefan Holst
 * Copyright (C) 2004-2005 Maximilian Schwerin
 *
 * This file is part of oxine a free media player.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: heap.h 3944 2009-07-09 20:21:13Z mschwerin $
 */

#ifndef HAVE_HEAP_H
#define HAVE_HEAP_H

#include "config.h"

#include <assert.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>

#include <sys/types.h>

#include "logger.h"
#include "types.h"

/**
 * A cookie which is used to try and detect memory errors.
 */
#define REFERABLE_COOKIE        0x7329ABCD


/**
 * This data structure can be embedded at the start of higher level structs
 * to keep a reference count.
 */
typedef struct {
#ifdef DEBUG
    int cookie;
#endif
    /**
     * The reference counter.
     */
    int reference_count;
    /**
     * If this is not zero new references to the object may not be made.
     */
    bool invalid;
    /**
     * A user defined function to free the object.
     */
    void (*free) (void *);
} referable_t;


static inline void *
_gen_malloc (const char *file, int line, size_t size)
{
    void *ptr = malloc (size);

    if (!ptr) {
        _log (LEVEL_FATAL, NULL, file, line, "Could not allocate memory!");
        abort ();
    }

    memset (ptr, 0, size);

    return ptr;
}


static inline void *
_gen_realloc (const char *file, int line, void *ptr, size_t size)
{
    void *p = realloc (ptr, size);

    if (!p) {
        _log (LEVEL_FATAL, NULL, file, line, "Could not allocate memory!");
        abort ();
    }

    return p;
}


static inline void *
_gen_free (const char *file, int line, void *ptr)
{
    free (ptr);
    return NULL;
}


static inline void *
_gen_memdup (const char *file, int line, const void *ptr, size_t size)
{
    void *p = _gen_malloc (file, line, size);
    memcpy (p, ptr, size);
    return p;
}


static inline char *
_gen_strdup (const char *file, int line, const char *ptr)
{
    if (!ptr) {
        return NULL;
    }
    else {
        return strdup (ptr);
    }
}

char *_gen_strdup_printf (char *, int, const char *, ...);

char *_gen_strdup_strftime (char *, int, const char *, const struct tm *);

/**
 * Allocates memory for the given type. The memory newly allocated memory is
 * set to 0. If the allocation fails the programm is aborted with an error
 * message.
 *
 * @param type                  The type of the variable for which the
 *                              memory is to be allocated.
 */
#define ho_new(type)            (_gen_malloc(__FILE__, __LINE__, sizeof(type)))

/**
 * Allocates memory of a given size. The memory newly allocated memory is
 * set to 0. If the allocation fails the programm is aborted with an error
 * message.
 *
 * @param size                  The size in bytes of the memory that is to
 *                              be allocated.
 */
#define ho_malloc(size)         (_gen_malloc(__FILE__, __LINE__, (size_t)size))

/**
 * Reallocates memory.
 */
#define ho_realloc(ptr, size)   (_gen_realloc(__FILE__, __LINE__, ptr, (size_t)size))

/**
 * Frees the memory pointed to by the parameter and sets the parameter to NULL.
 *
 * @param ptr                   A pointer to a memory previously allocated on the heap.
 */
#define ho_free(ptr)            (ptr = _gen_free(__FILE__, __LINE__, ptr))

/**
 * Duplicates an object of the given type on the heap. If the passed object
 * is NULL the function returns NULL.
 *
 * @param ptr                   A pointer to the memory that is to be
 *                              duplicated.
 * @param type                  The type of the memory that is to be duplicated.
 */
#define ho_dup(ptr, type)       (_gen_memdup(__FILE__, __LINE__, ptr, sizeof(type)))

/**
 * Duplicates an object on the heap. If the passed object is NULL the
 * function returns NULL.
 *
 * @param ptr                   A pointer to the memory that is to be
 *                              duplicated.
 * @param size                  The size in bytes of the memory that is to
 *                              be duplicated.
 */
#define ho_memdup(ptr, size)    (_gen_memdup(__FILE__, __LINE__, ptr, (size_t)size))

/**
 * Duplicates a string on the heap. If the passed string is NULL the
 * function returns NULL.
 *
 * @param ptr                   A pointer to a string.
 */
#define ho_strdup(ptr)          (_gen_strdup(__FILE__, __LINE__, ptr))

/**
 * Runs sprintf on the argument list and returns a duplicate of the result
 * allocated on the heap.
 */
#define ho_strdup_printf(...)   (_gen_strdup_printf(__FILE__, __LINE__, __VA_ARGS__))

/**
 * Runs strftime on the argument list and returns a duplicate of the result
 * allocated on the heap.
 */
#define ho_strdup_strftime(...) (_gen_strdup_strftime(__FILE__, __LINE__, __VA_ARGS__))


/**
 * Initializes the referable object.
 */
static inline void *
ho_initref (void *p)
{
    referable_t *r = (referable_t *) p;
#ifdef DEBUG
    r->cookie = REFERABLE_COOKIE;
#endif
    r->reference_count = 1;
    r->invalid = false;
    return p;
}


/**
 * Check if a reference is still valid.
 */
static inline bool
ho_checkref (void *p)
{
    referable_t *r = (referable_t *) p;
#ifdef DEBUG
    assert (r);
    assert (r->reference_count > 0);
    assert (r->cookie == REFERABLE_COOKIE);
#endif
    return !r->invalid;
}


/**
 * Invalidate a referable object.
 */
static inline void *
ho_badref (void *p)
{
    referable_t *r = (referable_t *) p;
#ifdef DEBUG
    assert (r);
    assert (r->reference_count > 0);
    assert (r->cookie == REFERABLE_COOKIE);
#endif
    r->invalid = true;
    return p;
}


/**
 * Increment the reference count of a referable object.
 */
static inline void *
ho_incref (void *p)
{
    referable_t *r = (referable_t *) p;
#ifdef DEBUG
    assert (r);
    assert (!r->invalid);
    assert (r->cookie == REFERABLE_COOKIE);
#endif
    r->reference_count++;
    return p;
}


/**
 * Decrement the reference count of a referable object. If the reference
 * counter reaches zero it will be freed automatically.
 */
static inline void *
ho_decref (void *p)
{
    referable_t *r = (referable_t *) p;
#ifdef DEBUG
    assert (r);
    assert (r->reference_count > 0);
    assert (r->cookie == REFERABLE_COOKIE);
    r->reference_count--;
#endif
    if (r->reference_count > 0) {
        return p;
    }
    else if (r->free) {
#ifdef DEBUG
        r->cookie = 0;
#endif
        r->free (p);
        return NULL;
    }
    else {
#ifdef DEBUG
        r->cookie = 0;
#endif
        ho_free (p);
        return NULL;
    }
}


/**
 * Set the variable, which is pointed to by V to P. If the variable is set
 * the reference of the object pointed to is decremented. After setting the
 * variable the reference of the object pointed to is incremented.
 */
#define ho_setref(v,p)                      \
{                                           \
    if (p) {                                \
        ho_incref (p);                      \
    }                                       \
    if (v) {                                \
        ho_decref (v);                      \
    }                                       \
    v = p;                                  \
}


#endif /* HAVE_HEAP_H */
