
/*
 * Copyright (C) 2004-2005 Maximilian Schwerin
 *
 * This file is part of oxine a free media player.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: list.c 3782 2009-04-24 10:43:17Z mschwerin $
 *
 */

#include "config.h"

#include <assert.h>
#include <stdlib.h>
#include <stdio.h>

#include "heap.h"
#include "list.h"

#ifdef L_LIST_THREAD_SAFE
#include "mutex.h"
#endif

typedef struct l_item_s l_item_t;

struct l_item_s {
    void *data;
    l_item_t *next;
    l_item_t *prev;
};

struct l_list_s {
    l_item_t *first;
    l_item_t *last;

    l_item_t *current;

    int length;

#ifdef L_LIST_THREAD_SAFE
    /**
     * A mutex to protect the list.
     */
    pthread_mutex_t mutex;
    pthread_mutexattr_t mutex_attr;
#endif
};


#ifdef L_LIST_THREAD_SAFE

#define l_lock(list)            mutex_lock (&list->mutex);

#define l_unlock(list)          mutex_unlock (&list->mutex);


void
_l_list_lock (l_list_t * list, const char *file, int line)
{
    assert (list);
#ifdef DEBUG
    _mutex_lock (&list->mutex, file, line);
#else
    mutex_lock (&list->mutex);
#endif
}


void
_l_list_unlock (l_list_t * list, const char *file, int line)
{
    assert (list);
#ifdef DEBUG
    _mutex_unlock (&list->mutex, file, line);
#else
    mutex_unlock (&list->mutex);
#endif
}

#else

#define l_lock(list)

#define l_unlock(list)

#endif /* L_LIST_THREAD_SAFE */


#ifdef L_LIST_TEST

/**
 * Tests the list for errors.
 *
 * @param list                  The list to test.
 */
static void
l_list_test (l_list_t * list)
{
    assert (list);

    l_lock (list)

    /* Check if we find the correct number of entries when walking forwards. */
    {
        int i = 0;
        l_item_t *item = list->first;
        while (item) {
            i++;
            item = item->next;
        }
        assert (i == list->length);
    }
    /* Check if we find the correct number of entries when walking backwards. */
    {
        int i = 0;
        l_item_t *item = list->last;
        while (item) {
            i++;
            item = item->prev;
        }
        assert (i == list->length);
    }
    /* Check that we get the same items at the same position when walking
     * forwards and backwards. */
    {
        l_item_t *items[list->length];
        int i = 0;
        l_item_t *item = list->first;
        while (item) {
            items[i] = item;
            i++;
            item = item->next;
        }

        i = list->length - 1;
        item = list->last;
        while (item) {
            assert (items[i] == item);
            i--;
            item = item->prev;
        }
    }

    l_unlock (list)
}

#else

#define l_list_test(list)

#endif /* L_LIST_TEST */


l_list_t *
l_list_new (void)
{
    l_list_t *list = ho_new (l_list_t);

    list->length = 0;

    list->first = NULL;
    list->last = NULL;
    list->current = NULL;

#ifdef L_LIST_THREAD_SAFE
    pthread_mutexattr_init (&list->mutex_attr);
    pthread_mutexattr_settype (&list->mutex_attr, PTHREAD_MUTEX_RECURSIVE);
    pthread_mutex_init (&list->mutex, &list->mutex_attr);
#endif

    return list;
}


void
l_list_free (l_list_t * list, l_free_cb_t free_data_cb)
{
    assert (list);

    l_list_clear (list, free_data_cb);

#ifdef L_LIST_THREAD_SAFE
    pthread_mutex_destroy (&list->mutex);
    pthread_mutexattr_destroy (&list->mutex_attr);
#endif

    ho_free (list);
}


void
l_list_clear (l_list_t * list, l_free_cb_t free_data_cb)
{
    assert (list);

    l_lock (list)
    l_list_test (list);

    while (list->first) {
        l_item_t *i = list->first;
        list->first = i->next;
        if (free_data_cb && i->data) {
            free_data_cb (i->data);
        }
        ho_free (i);
    }

    list->length = 0;
    list->first = NULL;
    list->last = NULL;
    list->current = NULL;

    l_list_test (list);
    l_unlock (list)
}


#if 0
static void
l_list_bubble_sort (l_list_t * list, l_swap_cb_t swap_cb)
{
    l_item_t *p1 = list->first;

    while (p1) {
        l_item_t *p2 = p1;
        while (p2) {
            if (swap_cb (p1->data, p2->data)) {
                void *tmp = p1->data;
                p1->data = p2->data;
                p2->data = tmp;
            }
            p2 = p2->next;
        }
        p1 = p1->next;
    }
}
#endif


/**
 * Helper function for mergesort.
 *
 * @param list                  The main list.
 * @param l                     The left partial list.
 * @param r                     The right partial list.
 * @param swap_cb               This is a pointer to a callback function
 *                              that is used to determin if two entries
 *                              must be swapped. This may not be NULL.
 */
static inline void
l_list_merge_merge (l_list_t * list, l_list_t * l, l_list_t * r,
                    l_swap_cb_t swap_cb)
{
    l_list_clear (list, NULL);

    l_item_t *li = l->first;
    l_item_t *ri = r->first;

    while (li && ri) {
        if (swap_cb (ri->data, li->data)) {
            l_list_append (list, li->data);
            li = li->next;
        }
        else {
            l_list_append (list, ri->data);
            ri = ri->next;
        }
    }
    while (li) {
        l_list_append (list, li->data);
        li = li->next;
    }
    while (ri) {
        l_list_append (list, ri->data);
        ri = ri->next;
    }
}


/**
 * A mergesort implementation.
 *
 * @param list                  The list to sort.
 * @param swap_cb               This is a pointer to a callback function
 *                              that is used to determin if two entries
 *                              must be swapped. This may not be NULL.
 */
static inline void
l_list_merge_sort (l_list_t * list, l_swap_cb_t swap_cb)
{
    int j = 0;

    l_list_t *l = l_list_new ();
    l_list_t *r = l_list_new ();

    l_item_t *i = list->first;
    while (j < (list->length / 2)) {
        l_list_append (l, i->data);
        i = i->next;
        j = j + 1;
    }
    while (i) {
        l_list_append (r, i->data);
        i = i->next;
    }

    l_list_sort (l, swap_cb);
    l_list_sort (r, swap_cb);

    l_list_merge_merge (list, l, r, swap_cb);

    l_list_free (l, NULL);
    l_list_free (r, NULL);
}


void
l_list_sort (l_list_t * list, l_swap_cb_t swap_cb)
{
    assert (list);
    assert (swap_cb);

    l_lock (list)
    l_list_test (list);

    if (list->length > 1) {
        l_list_merge_sort (list, swap_cb);
    }

    l_list_test (list);
    l_unlock (list)
}


void
l_list_move_down (l_list_t * list, void *data)
{
    assert (list);
    assert (data);

    l_lock (list)
    l_list_test (list);

    l_item_t *p1 = list->first;

    while (p1 && p1->next) {
        if (p1->data == data) {
            p1->data = p1->next->data;
            p1->next->data = data;
            break;
        }
        p1 = p1->next;
    }

    l_list_test (list);
    l_unlock (list)
}


void
l_list_move_up (l_list_t * list, void *data)
{
    assert (list);
    assert (data);

    l_lock (list)
    l_list_test (list);

    l_item_t *p1 = list->first;

    while (p1 && p1->next) {
        if (p1->next->data == data) {
            p1->next->data = p1->data;
            p1->data = data;
            break;
        }
        p1 = p1->next;
    }

    l_list_test (list);
    l_unlock (list)
}


void
l_list_concat (l_list_t * l1, l_list_t * l2)
{
    assert (l1);
    assert (l2);

    l_lock (l1);
    l_list_test (l1);
    l_lock (l2);
    l_list_test (l2);

    if (l1->first) {
        assert (l1->last);
        if (l2->first) {
            assert (l2->last);
            l1->last->next = l2->first;
            l2->first->prev = l1->last;
            l1->last = l2->last;
        }
    }
    else {
        l1->first = l2->first;
        l1->last = l2->last;
    }

    l1->current = NULL;
    l1->length += l2->length;

    l_list_test (l1);
    l_unlock (l1);
    l_list_test (l2);
    l_unlock (l2);
}


void *
l_list_insert (l_list_t * list, int position, void *data)
{
    assert (list);
    assert (data);
    assert (position >= 0);

    l_lock (list)
    l_list_test (list);

    l_item_t *item = ho_new (l_item_t);
    item->data = data;

    if (!list->first) {
        item->prev = NULL;
        item->next = NULL;
        list->first = item;
        list->last = item;
    }
    else if (position == 0) {
        item->prev = NULL;
        item->next = list->first;
        list->first->prev = item;
        list->first = item;
    }
    else if (position >= list->length) {
        item->prev = list->last;
        item->next = NULL;
        list->last->next = item;
        list->last = item;
    }
    else {
        int p = 0;
        l_item_t *cur = list->first;
        while (cur && p != position) {
            p += 1;
            cur = cur->next;
        }
        item->prev = cur->prev;
        item->next = cur;

        cur->prev->next = item;
        cur->prev = item;
    }

    list->length++;
    list->current = NULL;

#ifdef L_LIST_TEST
    l_list_test (list);
#endif

    l_list_test (list);
    l_unlock (list)

    return data;
}


void *
l_list_append (l_list_t * list, void *data)
{
    assert (list);
    assert (data);

    l_lock (list)
    l_list_insert (list, list->length, data);
    l_unlock (list)

    return data;
}


void *
l_list_prepend (l_list_t * list, void *data)
{
    assert (list);
    assert (data);

    l_lock (list)
    l_list_insert (list, 0, data);
    l_unlock (list)

    return data;
}


void *
l_list_remove (l_list_t * list, void *data)
{
    assert (list);
    assert (data);

    l_lock (list)
    l_list_test (list);

    if (list->first) {
        l_item_t *current = list->first;
        while (current) {
            if (current->data == data) {
                if (current->next)
                    current->next->prev = current->prev;
                if (current->prev)
                    current->prev->next = current->next;
                if (list->first == current)
                    list->first = current->next;
                if (list->last == current)
                    list->last = current->prev;

                ho_free (current);

                list->length--;
                list->current = NULL;
                break;
            }
            current = current->next;
        }
    }

    l_list_test (list);
    l_unlock (list)

    return data;
}


void *
l_list_first (l_list_t * list)
{
    assert (list);
    void *data = NULL;

    l_lock (list)
    l_list_test (list);

    if (list->first) {
        list->current = list->first;
        data = list->first->data;
    }

    l_unlock (list)

    return data;
}


void *
l_list_last (l_list_t * list)
{
    assert (list);
    void *data = NULL;

    l_lock (list)
    l_list_test (list);

    if (list->last) {
        list->current = list->last;
        data = list->last->data;
    }

    l_unlock (list)

    return data;
}


void *
l_list_next (l_list_t * list, void *data)
{
    assert (list);
    assert (data);
    void *next = NULL;

    l_lock (list)
    l_list_test (list);

    if (list->current && list->current->data == data) {
        list->current = list->current->next;
    }
    else {
        list->current = list->first;
        while (list->current) {
            if (list->current && (list->current->data == data)) {
                list->current = list->current->next;
                break;
            }
            list->current = list->current->next;
        }
    }
    if (list->current) {
        next = list->current->data;
    }

    l_unlock (list)

    return next;
}


void *
l_list_prev (l_list_t * list, void *data)
{
    assert (list);
    assert (data);
    void *prev = NULL;

    l_lock (list)
    l_list_test (list);

    if (list->current && list->current->data == data) {
        list->current = list->current->prev;
    }
    else {
        list->current = list->last;
        while (list->current) {
            if (list->current && (list->current->data == data)) {
                list->current = list->current->prev;
                break;
            }
            list->current = list->current->prev;
        }
    }
    if (list->current) {
        prev = list->current->data;
    }

    l_unlock (list)

    return prev;
}


int
l_list_length (l_list_t * list)
{
    assert (list);
    return list->length;
}
