
/*
 * Copyright (C) 2002-2003 Stefan Holst
 * Copyright (C) 2004-2005 Maximilian Schwerin
 *
 * This file is part of oxine a free media player.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: mediamarks.c 3828 2009-04-29 07:31:57Z mschwerin $
 *
 */
#include "config.h"

#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>

#include "codeset.h"
#include "environment.h"
#include "filelist.h"
#include "heap.h"
#include "i18n.h"
#include "list.h"
#include "logger.h"
#include "mediamarks.h"
#include "utils.h"
#include "xmlparser.h"

#define OXINE_MEDIAMARKS_XML    "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
#define OXINE_MEDIAMARKS_DTD    "<!DOCTYPE oxine_mediamarks SYSTEM \"http://oxine.sf.net/DTD/mediamarks.dtd\">\n\n"


static char *
href_to_mrl (char * href, const char *xml_mrl)
{
    if (!href) {
        return NULL;
    }
    else if (mrl_is_valid (href)) {
        return href;
    }
    else if (href[0] != '/') {
        char *filename = mrl_get_filename (xml_mrl);
        char *pathname = get_dirname (filename);
        char *absolute = ho_strdup_printf ("%s/%s", pathname, href);
        char *mrl = mrl_from_filename (absolute);
        ho_free (filename);
        ho_free (pathname);
        ho_free (absolute);
        ho_free (href);
        return mrl;
    }
    else {
        char *mrl = mrl_from_filename (href);
        ho_free (href);
        return mrl;
    }
}


static void
info_free (void * p)
{
    ho_free (p);
}


static void
read_info (const char *xml_mrl, l_list_t * info_list, xml_node_t * node)
{
    recode_t *xr = recode_init ("UTF-8", USE_SYSTEM_ENCODING);

    while (node) {
        if (strcasecmp (node->name, "line") == 0) {
            char *line = recode (xr, node->data);
            l_list_append (info_list, line);
        }

        node = node->next;
    }

    recode_done (xr);
}


static void
read_entry (const char *xml_mrl, filelist_t * filelist, xml_node_t * node)
{
    fileitem_type_t type = FILE_TYPE_UNKNOWN;
    char *title = NULL;
    char *mrl = NULL;
    char *cmd = NULL;
    char *thumbnail_mrl = NULL;
    char *sort_by = NULL;
    int listened_count = 0;
    time_t listened_time = 0;
    bool is_mountpoint = false;
    l_list_t *info_list = NULL;

    recode_t *xr = recode_init ("UTF-8", USE_SYSTEM_ENCODING);

    while (node) {
        if (strcasecmp (node->name, "title") == 0) {
            title = recode (xr, node->data);
        }
        else if (strcasecmp (node->name, "thumbnail_mrl") == 0) {
            char *href = recode (xr, xml_parser_get_property (node, "href"));
            thumbnail_mrl = href_to_mrl (href, xml_mrl);
        }
        else if (strcasecmp (node->name, "mrl") == 0) {
            char *href = recode (xr, xml_parser_get_property (node, "href"));
            mrl = href_to_mrl (href, xml_mrl);

            const char *entry_type = xml_parser_get_property (node, "type");
            if (!entry_type) {
                /* do nothing */
            }
            else if (strcasecmp (entry_type, "regular") == 0) {
                type = FILE_TYPE_REGULAR;
            }
            else if (strcasecmp (entry_type, "directory") == 0) {
                type = FILE_TYPE_FOLDER;
            }
            else if (strcasecmp (entry_type, "mountpoint") == 0) {
                type = FILE_TYPE_FOLDER;
                is_mountpoint = true;
            }
            else if (strcasecmp (entry_type, "mediamarks") == 0) {
                type = FILE_TYPE_MEDIAMARKS;
            }
            else if (strcasecmp (entry_type, "playlist") == 0) {
                type = FILE_TYPE_PLAYLIST;
            }
        }
        else if (strcasecmp (node->name, "sort_by") == 0) {
            sort_by = recode (xr, node->data);
        }
        else if (strcasecmp (node->name, "cmd") == 0) {
            cmd = recode (xr, xml_parser_get_property (node, "href"));
        }
        else if (strcasecmp (node->name, "listened_count") == 0) {
            listened_count = atoi (node->data);
        }
        else if (strcasecmp (node->name, "listened_time") == 0) {
            listened_time = atoi (node->data);
        }
        else if (strcasecmp (node->name, "info") == 0) {
            info_list = l_list_new ();
            read_info (xml_mrl, info_list, node->child);
        }
        else {
            error (_("Unknown name '%s' for node!"), node->name);
        }

        node = node->next;
    }

    recode_done (xr);

    fileitem_t *fileitem = NULL;
    if (mrl && title && (type != FILE_TYPE_UNKNOWN)) {
        fileitem = filelist_add (filelist, title, mrl, type);
    }
    else if (mrl) {
        fileitem = filelist_add_magic (filelist, title, mrl);
    }
    else {
        error (_("Found an entry that did not contain a MRL!"));
    }

    if (fileitem) {
        fileitem->cmd = cmd;
        fileitem->listened_count = listened_count;
        fileitem->listened_time = listened_time;
        fileitem->thumbnail_mrl = thumbnail_mrl;
        fileitem->is_mountpoint = is_mountpoint;
        fileitem->info_list = info_list;

        if (sort_by) {
            ho_free (fileitem->sort_by);
            fileitem->sort_by = sort_by;
        }
        if (type != FILE_TYPE_UNKNOWN) {
            fileitem->type = type;
        }
    }
    else {
        ho_free (cmd);
        ho_free (thumbnail_mrl);
        ho_free (sort_by);

        if (info_list) {
            l_list_free (info_list, info_free);
        }
    }

    ho_free (mrl);
    ho_free (title);
}


static void
read_nodes (const char *xml_mrl, filelist_t * filelist, xml_node_t * node)
{
    while (node) {
        if (!strcasecmp (node->name, "entry")) {
            read_entry (xml_mrl, filelist, node->child);
        }

        else {
            error (_("Unknown name '%s' for node!"), node->name);
        }
        node = node->next;
    }
}


static bool
_mediamarks_read (filelist_t * filelist, const char *rootnode_name,
                  l_swap_cb_t swap_cb)
{
    bool result = false;
    bool sort = true;

    int size = 0;
    char *buffer = read_entire_file (filelist->mrl, &size);
    if (!buffer) {
        goto buffer_free;
    }

    xml_node_t *rootnode;
    xml_parser_init (buffer, size, XML_PARSER_CASE_INSENSITIVE);
    if (xml_parser_build_tree (&rootnode) != XML_PARSER_OK) {
        error (_("Parsing '%s' failed!"), filelist->mrl);
        goto buffer_free;
    }

    if (strcasecmp (rootnode->name, rootnode_name) != 0) {
        error (_("Root node of '%s' must be '%s'!"),
               filelist->mrl, rootnode_name);
        goto parser_free;
    }

    const char *str_sort = xml_parser_get_property (rootnode, "sort");
    if (str_sort && (strcasecmp (str_sort, "false") == 0)) {
        sort = false;
    }

    const char *str_update = xml_parser_get_property (rootnode, "update");
    if (str_update && (strcasecmp (str_update, "false") == 0)) {
        filelist->update_allowed = false;
    }
    else {
        filelist->update_allowed = true;
    }

    read_nodes (filelist->mrl, filelist, rootnode->child);

    if (sort) {
        filelist_sort (filelist, swap_cb);
    }

    result = true;
  parser_free:
    xml_parser_free_tree (rootnode);
  buffer_free:
    ho_free (buffer);

    return result;
}


bool
mediamarks_read (filelist_t * filelist)
{
    return _mediamarks_read (filelist, "oxine_mediamarks", NULL);
}


bool
mediamarks_write (filelist_t * filelist)
{
    char *filename = mrl_get_filename (filelist->mrl);

    FILE *f = fopen (filename, "w");
    if (f == NULL) {
        error (_("Could not open '%s': %s!"), filename, strerror (errno));
        ho_free (filename);
        return false;
    }

    recode_t *xr = recode_init (USE_SYSTEM_ENCODING, "UTF-8");

    fprintf (f, OXINE_MEDIAMARKS_XML);
    fprintf (f, OXINE_MEDIAMARKS_DTD);

    fprintf (f, "<!--\n\n");
    fprintf (f, "This file was automatically created by oxine.\n");
    fprintf (f, "If you want to add other entries feel free to do so.\n\n");
    fprintf (f, "An example mediamarks file can be found at\n");
    fprintf (f, "%s/mediamarks_example.xml\n\n", OXINE_DATADIR);
    fprintf (f, "-->\n\n");

    fprintf (f, "<oxine_mediamarks>\n");
    fileitem_t *item = filelist_first (filelist);
    while (item) {
        if (item->type == FILE_TYPE_FOLDER) {
            char *esc_str0 = xml_escape (item->title,
                                         xr, XML_ESCAPE_NO_QUOTE);
            char *esc_str1 = xml_escape (item->mrl,
                                         xr, XML_ESCAPE_DOUBLE_QUOTE);
            fprintf (f, "  <entry>\n");
            fprintf (f, "    <title>%s</title>\n", esc_str0);
            fprintf (f, "    <mrl href=\"%s\" />\n", esc_str1);
            fprintf (f, "  </entry>\n");
            ho_free (esc_str0);
            ho_free (esc_str1);
        }

        item = filelist_next (filelist, item);
    }
    fprintf (f, "</oxine_mediamarks>\n");
    fclose (f);

    recode_done (xr);
    ho_free (filename);

    return true;
}


bool
favorites_swap_time_cb (void *d1, void *d2)
{
    fileitem_t *f1 = (fileitem_t *) d1;
    fileitem_t *f2 = (fileitem_t *) d2;

    if (f2->listened_time == f1->listened_time) {
        return fileitem_swap_cb (d1, d2);
    }
    if (f2->listened_time > f1->listened_time) {
        return true;
    }

    return false;
}


bool
favorites_swap_count_cb (void *d1, void *d2)
{
    fileitem_t *f1 = (fileitem_t *) d1;
    fileitem_t *f2 = (fileitem_t *) d2;

    if (f2->listened_count == f1->listened_count) {
        return fileitem_swap_cb (d1, d2);
    }
    if (f2->listened_count > f1->listened_count) {
        return true;
    }

    return false;
}


bool
favorites_read (filelist_t * filelist)
{
    if (!file_exists (filelist->mrl)) {
        favorites_write (filelist);
    }

    return _mediamarks_read (filelist, "oxine_favorites",
                             favorites_swap_count_cb);
}
