
/*
 * Copyright (C) 2004-2005 Maximilian Schwerin
 *
 * This file is part of oxine a free media player.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: mediamarks_favorites.c 3796 2009-04-27 10:31:37Z mschwerin $
 *
 */
#include "config.h"

#include <assert.h>
#include <errno.h>
#include <limits.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "codeset.h"
#include "environment.h"
#include "filelist.h"
#include "heap.h"
#include "i18n.h"
#include "logger.h"
#include "mediamarks.h"
#include "mediamarks_favorites.h"
#include "utils.h"

#define OXINE_FAVORITES_XML     "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
#define OXINE_FAVORITES_DTD     "<!DOCTYPE oxine_favorites SYSTEM \"http://oxine.sf.net/DTD/favorites.dtd\">\n\n"


bool
favorites_write (filelist_t * favorites_filelist)
{
    const char *fav_mrl = get_file_favorites ();
    FILE *f = fopen (fav_mrl, "w");
    if (f == NULL) {
        error (_("Could not open '%s': %s!"), fav_mrl, strerror (errno));
        return false;
    }

    recode_t *xr = recode_init (USE_SYSTEM_ENCODING, "UTF-8");

    fprintf (f, OXINE_FAVORITES_XML);
    fprintf (f, OXINE_FAVORITES_DTD);

    fprintf (f, "<!--\n\n");
    fprintf (f, "WARNING:\n\n");
    fprintf (f, "This file was automatically created by oxine.\n");
    fprintf (f, "Any changes made to this file will be lost!\n\n");
    fprintf (f, "-->\n\n");

    fprintf (f, "<oxine_favorites>\n");
    fileitem_t *item = filelist_first (favorites_filelist);
    while (item) {
        if (mrl_is_directory (item->mrl)) {
            char *esc_str0 = xml_escape (item->title,
                                         xr, XML_ESCAPE_NO_QUOTE);
            char *esc_str1 = xml_escape (item->mrl,
                                         xr, XML_ESCAPE_DOUBLE_QUOTE);
            char *esc_str2 = xml_escape (item->thumbnail_mrl,
                                         xr, XML_ESCAPE_DOUBLE_QUOTE);
            fprintf (f, "  <entry>\n");
            if (esc_str0) {
                fprintf (f, "    <title>%s</title>\n", esc_str0);
            }
            if (esc_str1) {
                fprintf (f, "    <mrl           href=\"%s\" />\n", esc_str1);
            }
            if (esc_str2) {
                fprintf (f, "    <thumbnail_mrl href=\"%s\" />\n", esc_str2);
            }
            {
                fprintf (f, "    <listened_count>%d</listened_count>\n",
                         item->listened_count);
                fprintf (f, "    <listened_time>%lu</listened_time>\n",
                         item->listened_time);
            }
            fprintf (f, "  </entry>\n");
            ho_free (esc_str0);
            ho_free (esc_str1);
            ho_free (esc_str2);
        }

        item = filelist_next (favorites_filelist, item);
    }
    fprintf (f, "</oxine_favorites>\n");
    fclose (f);

    recode_done (xr);

    return true;
}


void
favorites_add_directory (filelist_t * favorites_filelist,
                         const char *title, const char *mrl)
{
    if (!config_get_bool ("misc.favorites.enable")) {
        return;
    }
    if (!mrl_is_directory (mrl)) {
        debug ("This is not a directory (ignoring): '%s'!", mrl);
        return;
    }

    /* Make sure we don't work with a softlink. */
    char *absolute = mrl_resolve_softlink (mrl);

    /* Lock access to the favorites filelist. */
    filelist_lock (favorites_filelist);

    /* Try to find an entry with the same MRL. */
    fileitem_t *item = filelist_first (favorites_filelist);
    while (item && !mrls_are_equal (item->mrl, absolute)) {
        item = filelist_next (favorites_filelist, item);
    }

    /* If no entry was found, we add a new one. */
    if (!item) {
        item = filelist_add (favorites_filelist, title,
                             absolute, FILE_TYPE_FOLDER);
    }

    /* Increment listened counter. */
    if (item) {
        item->listened_count++;
        item->listened_time = time (NULL);
    }

    /* Write the list back to the favorites file. */
    favorites_write (favorites_filelist);

    /* Grant access to the favorites filelist. */
    filelist_unlock (favorites_filelist);

    ho_free (absolute);
}


void
favorites_remove_directory (filelist_t * favorites_filelist,
                            fileitem_t * fileitem)
{
    /* Remove the specified item. */
    filelist_remove (favorites_filelist, fileitem);

    /* Write the list back to the favorites file. */
    filelist_lock (favorites_filelist);
    favorites_write (favorites_filelist);
    filelist_unlock (favorites_filelist);
}
