
/*
 * Copyright (C) 2004-2005 Maximilian Schwerin
 *
 * This file is part of oxine a free media player.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: menu_dialog.c 3648 2008-12-16 21:23:51Z mschwerin $
 *
 */
#include "config.h"

#include <assert.h>
#include <stdarg.h>
#include <stdio.h>

#include "environment.h"
#include "download.h"
#include "i18n.h"
#include "heap.h"
#include "logger.h"
#include "otk.h"
#include "odk.h"
#include "oxine.h"
#include "utils.h"

#include "menu_base.h"
#include "menu_dialog.h"

#define LINE_HEIGHT             35
#define WIDGET_HEIGHT           40

#define SPACER_HEIGHT           20
#define SPACER_WIDTH            20

#define BUTTON_HEIGHT           40
#define BUTTON_WIDTH            160


extern oxine_t *oxine;

typedef struct {
    otk_cb_t lcb;
    void *lcb_data;
    char *lcb_label;

    otk_cb_t rcb;
    void *rcb_data;
    char *rcb_label;

    int align_message;
    char *message;

    char *background_mrl;

    bool with_external_widget;
    int widget_x;
    int widget_y;
    int widget_w;
    int widget_h;
} dialog_data_t;

static dialog_data_t dialog_data;

void
repaint_dialog (void *data_p)
{
    dialog_data_t * temp = ho_dup (&dialog_data, dialog_data_t);
    temp->lcb_label = ho_strdup (dialog_data.lcb_label);
    temp->rcb_label = ho_strdup (dialog_data.rcb_label);
    temp->message = ho_strdup (dialog_data.message);
    temp->background_mrl = ho_strdup (dialog_data.background_mrl);

    show_dialog (temp->lcb,
                 temp->lcb_data,
                 temp->lcb_label,
                 temp->rcb,
                 temp->rcb_data,
                 temp->rcb_label,
                 temp->with_external_widget,
                 temp->background_mrl,
                 temp->align_message, temp->message);

    ho_free (temp->lcb_label);
    ho_free (temp->rcb_label);
    ho_free (temp->message);
    ho_free (temp->background_mrl);
    ho_free (temp);
}


void
show_dialog (otk_cb_t lcb, void *lcb_data, const char *lcb_label,
             otk_cb_t rcb, void *rcb_data, const char *rcb_label,
             bool with_external_widget,
             const char *background_mrl, int align_message,
             const char *format, ...)
{
    char msg[1024];
    va_list args;
    va_start (args, format);
    vsnprintf (msg, 1024, format, args);
    va_end (args);

    ho_free (dialog_data.lcb_label);
    ho_free (dialog_data.rcb_label);
    ho_free (dialog_data.message);
    ho_free (dialog_data.background_mrl);

    dialog_data.lcb = lcb;
    dialog_data.lcb_data = lcb_data;
    dialog_data.lcb_label = ho_strdup (lcb_label);
    dialog_data.rcb = rcb;
    dialog_data.rcb_data = rcb_data;
    dialog_data.rcb_label = ho_strdup (rcb_label);
    dialog_data.message = ho_strdup (msg);
    dialog_data.align_message = align_message;
    dialog_data.background_mrl = NULL;
    dialog_data.background_mrl = ho_strdup (background_mrl);

    int w = 2 * (odk_osd_get_width (oxine->odk) / 3);
    int h = SPACER_HEIGHT * 2 + SPACER_HEIGHT / 2;

    char *cline = msg;
    while (cline) {
        char *nline = index (cline, '\n');
        if (nline) {
            nline[0] = 0;
        }
        if (strlen (cline) > 0) {
            h += LINE_HEIGHT;
        }
        else {
            h += SPACER_HEIGHT / 2;
        }
        if (nline) {
            nline[0] = '\n';
            nline++;
        }
        cline = nline;
    }
    h += (lcb != NULL) ? (BUTTON_HEIGHT + SPACER_HEIGHT) : 0;
    h += with_external_widget ? (WIDGET_HEIGHT + SPACER_HEIGHT) : 0;

    int x = (odk_osd_get_width (oxine->odk) - w) / 2;
    int y = (odk_osd_get_height (oxine->odk) - h) / 2;
    int a = align_message;

    create_new_window (false, true);

    /* Display the border. */
    otk_border_new (oxine->otk, x, y, w, h);
    x += SPACER_WIDTH;
    w -= SPACER_WIDTH * 2;
    y += SPACER_HEIGHT;

    /* Display the broken up message. */
    cline = msg;
    while (cline) {
        char *nline = index (cline, '\n');
        if (nline) {
            nline[0] = 0;
            nline++;
        }
        if (strlen (cline) > 0) {
            otk_label_new (oxine->otk, x, y, w, LINE_HEIGHT, a, cline);
            y += LINE_HEIGHT;
        }
        else {
            y += SPACER_HEIGHT / 2;
        }
        cline = nline;
    }
    y += SPACER_HEIGHT;

    if (with_external_widget) {
        dialog_data.widget_x = x;
        dialog_data.widget_y = y;
        dialog_data.widget_w = w;
        dialog_data.widget_h = WIDGET_HEIGHT;
        y += WIDGET_HEIGHT + SPACER_HEIGHT;
    }

    int x0 = x + (w - 1 * BUTTON_WIDTH) / 2;
    int x1 = x + (w - 2 * BUTTON_WIDTH) / 2;
    int x2 = x1 + BUTTON_WIDTH + SPACER_WIDTH;

    otk_widget_t *b;
    if (lcb && rcb) {
        b = otk_text_button_new (oxine->otk, x1, y,
                                 BUTTON_WIDTH, BUTTON_HEIGHT,
                                 lcb_label, lcb, lcb_data);
        otk_widget_set_alignment (b, OTK_ALIGN_CENTER);
        otk_widget_set_focused (b, true);

        b = otk_text_button_new (oxine->otk, x2, y,
                                 BUTTON_WIDTH, BUTTON_HEIGHT,
                                 rcb_label, rcb, rcb_data);
        otk_widget_set_alignment (b, OTK_ALIGN_CENTER);
    }
    else if (lcb) {
        b = otk_text_button_new (oxine->otk, x0, y,
                                 BUTTON_WIDTH, BUTTON_HEIGHT,
                                 lcb_label, lcb, lcb_data);
        otk_widget_set_alignment (b, OTK_ALIGN_CENTER);
        otk_widget_set_focused (b, true);
    }

    set_current_menu (repaint_dialog, NULL);

    show_user_interface (NULL);
    show_menu_background (background_mrl);
}


void
get_dialog_widget_size (int *x, int *y, int *w, int *h)
{
    *x = dialog_data.widget_x;
    *y = dialog_data.widget_y;
    *w = dialog_data.widget_w;
    *h = dialog_data.widget_h;
}


void
free_menu_dialog (void)
{
    ho_free (dialog_data.lcb_label);
    ho_free (dialog_data.rcb_label);
    ho_free (dialog_data.message);
    ho_free (dialog_data.background_mrl);
}


void
show_message_dialog (otk_cb_t lcb, void *lcb_data,
                     otk_cb_t rcb, void *rcb_data,
                     message_dialog_type_t type,
                     const char *background_mrl, const char *format, ...)
{
    int alignment = OTK_ALIGN_CENTER | OTK_ALIGN_VCENTER;

    char message[1024];
    va_list args;
    va_start (args, format);
    vsnprintf (message, 1024, format, args);
    va_end (args);

    const char *lcb_label = NULL;
    const char *rcb_label = NULL;

    if (type == DIALOG_OK) {
        lcb_label = _("OK");
        rcb = NULL;
    }
    else if (type == DIALOG_BACK) {
        lcb_label = _("Back");
        rcb = NULL;
    }
    else if (type == DIALOG_CANCEL) {
        lcb_label = _("Cancel");
        rcb = NULL;
    }
    else if (type == DIALOG_OK_CANCEL) {
        lcb_label = _("OK");
        rcb_label = _("Cancel");
    }
    else if (type == DIALOG_YES_NO) {
        lcb_label = _("Yes");
        rcb_label = _("No");
    }
    else if (type == DIALOG_RETRY_CANCEL) {
        lcb_label = _("Retry");
        rcb_label = _("Cancel");
    }

    show_dialog (lcb, lcb_data, lcb_label,
                 rcb, rcb_data, rcb_label,
                 false, background_mrl, alignment, message);
}
