
/*
 * Copyright (C) 2004-2005 Maximilian Schwerin
 *
 * This file is part of oxine a free media player.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: menu_filelist.c 4027 2010-01-17 12:39:40Z mschwerin $
 *
 */
#include "config.h"

#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include "disc.h"
#include "download.h"
#include "environment.h"
#include "extractor.h"
#include "filelist.h"
#include "heap.h"
#include "i18n.h"
#include "logger.h"
#include "mediamarks.h"
#include "mediamarks_favorites.h"
#include "meta_info.h"
#include "oxine.h"
#include "playlist.h"
#include "utils.h"
#include "vdr.h"

#include "menu_base.h"
#include "menu_dialog.h"
#include "menu_extractor.h"
#include "menu_fileinfo.h"
#include "menu_filelist.h"
#include "menu_main.h"
#include "menu_playback.h"
#include "menu_playlist.h"

extern oxine_t *oxine;

typedef enum {
    MENU_FILELIST_MEDIA,
    MENU_FILELIST_MUSIC,
    MENU_FILELIST_VIDEO,
    MENU_FILELIST_IMAGE,
} menu_filelist_menu_t;

typedef struct {
    l_list_t *prev_list;
    l_list_t *next_list;

    fileitem_t *focused;

    filelist_t *cur_filelist;
    filelist_t *top_filelist;
    filelist_t *fav_filelist;

    otk_cb_t menu_cb;
    const char *background;
} menu_filelist_data_t;

#define FILELIST_MENU_DATA_COUNT    4
static menu_filelist_data_t mdata[FILELIST_MENU_DATA_COUNT] = {
    {NULL, NULL, NULL, NULL, NULL, NULL, NULL, false},
    {NULL, NULL, NULL, NULL, NULL, NULL, NULL, false},
    {NULL, NULL, NULL, NULL, NULL, NULL, NULL, false},
    {NULL, NULL, NULL, NULL, NULL, NULL, NULL, false},
};

static pthread_t preload_thread;

static menu_filelist_data_t *cur_menu = &mdata[0];

static otk_widget_t *button_back = NULL;
static otk_widget_t *button_prev = NULL;
static otk_widget_t *button_next = NULL;
static otk_widget_t *button_home = NULL;
static otk_widget_t *button_hide = NULL;
static otk_widget_t *edit_search = NULL;

static int thumbnail_job = -1;
static int fileinfo_job = -1;

#define FILELIST_MENU_BUTTONS_COUNT 8
static otk_widget_t *filelist_menu_buttons[FILELIST_MENU_BUTTONS_COUNT];
static otk_widget_t *filelist_menu_list = NULL;
static otk_widget_t *filelist_menu_short = NULL;
static otk_widget_t *filelist_menu_window = NULL;
static otk_widget_t *filelist_menu_title = NULL;
static otk_widget_t *filelist_menu_info = NULL;
static int filelist_menu_info_width = 0;
#ifdef HAVE_OSD_IMAGE
static otk_widget_t *filelist_menu_image = NULL;
#endif

static otk_cb_t show_last_filelist_menu_cb = show_menu_media;

/*
 * ***************************************************************************
 * Some prototypes of methods declared in this file
 * ***************************************************************************
 */
static void filelist_menu_update_buttons (void);
static void filelist_menu_update_list (void);
static void filelist_menu_update_title (void);

static void filelist_menu_set_filelist (filelist_t * newlist,
                                        bool force_update, bool allow_update);
static bool filelist_menu_get_selected (fileitem_t *** items,
                                        int *num_selected);

static void listentry_focus_enter_cb (void *entry_cb_data);
static void listentry_focus_leave_cb (void *entry_cb_data);

static void listentry_activate_cb (void *entry_cb_data);
static void listentry_remove_cb (void *entry_cb_data);
static void listentry_select_cb (void *entry_cb_data);
static void listentry_copy_cb (void *entry_cb_data);

static void filelist_addall_now (void *p);
static void filelist_playall_now (void *p);
static void filelist_playall_cb (void *entry_cb_data);

static void focused_add_cb (void *p);
static void selected_add_cb (void *p);
static void selected_play_cb (void *p);
static void selected_remove_cb (void *p);
static void selected_copy_cb (void *p);

static void filelist_menu_clear_lists (void);
static void filelist_menu_please_wait (void);


static bool
_filelist_menu_is_current_menu ()
{
    return filelist_menu_is_current_menu ();
}


static void
filelist_tracker_ifree (void *p)
{
    ho_decref (p);
}


static void
filelist_tracker_reset (l_list_t * tracker)
{
    l_list_clear (tracker, filelist_tracker_ifree);
}


static void
filelist_tracker_append (l_list_t * tracker, filelist_t * filelist)
{
    ho_incref (filelist);
    l_list_append (tracker, filelist);
}


static void
filelist_tracker_remove (l_list_t * tracker, filelist_t * filelist)
{
    l_list_remove (tracker, filelist);
    ho_decref (filelist);
}


static void
filelist_tracker_clean (l_list_t * tracker)
{
    filelist_t *filelist = l_list_first (tracker);
    while (filelist) {
        /* If we're the only one holding a reference to a filelist it's
         * probably not valid any more. */
        if (filelist->r.reference_count == 1) {
            filelist_t *next = l_list_next (tracker, filelist);
            filelist_tracker_remove (tracker, filelist);
            filelist = next;
        }
        else {
            filelist = l_list_next (tracker, filelist);
        }
    }
}


static bool
current_is_toplevel (void)
{
    return (cur_menu->cur_filelist == cur_menu->top_filelist);
}


static bool
current_is_recent (void)
{
    return mrls_are_equal (cur_menu->cur_filelist->mrl,
                           FILELIST_FAVORITES_RECENT_MRL);
}


static bool
current_is_favorites (void)
{
    return mrls_are_equal (cur_menu->cur_filelist->mrl,
                           FILELIST_FAVORITES_MOST_MRL);
}


static bool
current_is_playlists (void)
{
    return mrls_are_equal (cur_menu->cur_filelist->mrl, get_mrl_playlists ());
}


static bool
current_is_vdr_recordings (void)
{
#ifdef HAVE_VDR
    fileitem_t *parent = cur_menu->cur_filelist->parent_item;
    if (!parent) {
        return false;
    }
    return strc_starts_with (parent->mrl, FILELIST_VDR_REC_MRL);
#else
    return false;
#endif
}


static bool
current_is_vdr_timers (void)
{
#ifdef HAVE_VDR
    fileitem_t *parent = cur_menu->cur_filelist->parent_item;
    if (!parent) {
        return false;
    }
    return strc_starts_with (parent->mrl, FILELIST_VDR_TIM_MRL);
#else
    return false;
#endif
}


static bool
current_is_vdr_channels (void)
{
#ifdef HAVE_VDR
    fileitem_t *parent = cur_menu->cur_filelist->parent_item;
    if (!parent) {
        return false;
    }
    return strc_starts_with (parent->mrl, FILELIST_VDR_CHA_MRL);
#else
    return false;
#endif
}


/**
 * Returns TRUE if the fileitem may be played.
 */
static bool
fileitem_is_playable (fileitem_t * fileitem)
{
    bool playable = true;
    playable &= !mrls_are_equal (fileitem->mrl, FILELIST_MEDIA_MRL);
    playable &= !mrls_are_equal (fileitem->mrl, FILELIST_MUSIC_MRL);
    playable &= !mrls_are_equal (fileitem->mrl, FILELIST_VIDEO_MRL);
    playable &= !mrls_are_equal (fileitem->mrl, FILELIST_IMAGE_MRL);
    playable &= !mrls_are_equal (fileitem->mrl, FILELIST_HOME_MRL);
    playable &= !mrls_are_equal (fileitem->mrl, FILELIST_ROOT_MRL);
    playable &= !mrls_are_equal (fileitem->mrl, FILELIST_FAVORITES_MOST_MRL);
    playable &=
        !mrls_are_equal (fileitem->mrl, FILELIST_FAVORITES_RECENT_MRL);
    playable &= !mrls_are_equal (fileitem->mrl, FILELIST_PLAYLISTS_MRL);
    playable &= !mrls_are_equal (fileitem->mrl, OXINE_AUDIODIR);
    playable &= !mrls_are_equal (fileitem->mrl, OXINE_VIDEODIR);
    playable &= !mrls_are_equal (fileitem->mrl, OXINE_IMAGEDIR);
#ifdef HAVE_VDR
    playable &= !mrls_are_equal (fileitem->mrl, FILELIST_VDR_REC_MRL);
    playable &= !strc_starts_with (fileitem->mrl, FILELIST_VDR_EPG_MRL);
    playable &= !strc_starts_with (fileitem->mrl, FILELIST_VDR_CHA_MRL);
    playable &= !strc_starts_with (fileitem->mrl, FILELIST_VDR_TIM_MRL);
#endif
    return playable;
}


/**
 * Returns TRUE if the fileitem may be added to the playlist.
 */
static bool
fileitem_is_addable (fileitem_t * fileitem)
{
    bool addable = true;
    addable &= fileitem_is_playable (fileitem);
#ifdef HAVE_VDR
    addable &= !strc_starts_with (fileitem->mrl, FILELIST_VDR_REC_MRL);
#endif
    return addable;
}


static void
filelist_menu_update_buttons (void)
{
    char *label = NULL;
    bool enabled = false;

    if (!_filelist_menu_is_current_menu ()) {
        return;
    }

    bool has_prev = (l_list_length (cur_menu->prev_list) > 0);
    bool has_next = (l_list_length (cur_menu->next_list) > 0);

    otk_widget_set_enabled (button_back, !current_is_toplevel ());
    otk_widget_set_enabled (button_home, !current_is_toplevel ());
    otk_widget_set_enabled (button_prev, has_prev);
    otk_widget_set_enabled (button_next, has_next);

    if (!oxine->is_sidebar_visible) {
        return;
    }

#ifdef DEBUG
    int j = 0;
    for (j = 0; j < FILELIST_MENU_BUTTONS_COUNT; j++) {
        assert (filelist_menu_buttons[j]);
    }
#endif

    int i;
    int num_selected;
    fileitem_t **items;
    filelist_menu_get_selected (&items, &num_selected);

    /* Play the selected items */
    if (false) {
        /* do nothing */
    }
#ifdef HAVE_IMAGE_PLAYBACK
    else if (is_current_menu (show_menu_image)) {
        label = _("View");
    }
#endif
#ifdef HAVE_VDR
    else if (current_is_vdr_channels ()) {
        label = _("Record");
    }
#endif
    else {
        label = _("Play");
    }
    {
        enabled = false;
        for (i = 0; i < num_selected; i++) {
            enabled |= fileitem_is_playable (items[i]);
        }
    }
    otk_widget_set_enabled (filelist_menu_buttons[0], enabled);
    otk_button_set_text (filelist_menu_buttons[0], label);

    /* Load saved playlists */
    if (current_is_playlists ()) {
        enabled = (num_selected == 1);
        label = _("Load");
    }
    /* Add to playlist */
    else {
        label = _("Add");
        enabled = false;
        for (i = 0; i < num_selected; i++) {
            enabled |= fileitem_is_addable (items[i]);
        }
    }
    otk_widget_set_enabled (filelist_menu_buttons[1], enabled);
    otk_button_set_text (filelist_menu_buttons[1], label);

    /* Reset the favorites list */
    if (current_is_favorites () || current_is_recent ()) {
        enabled = (filelist_length (cur_menu->cur_filelist) > 0);
        label = _("Reset");
    }
    /* Remove saved playlists */
    else if (current_is_playlists ()) {
        enabled = (num_selected > 0);
        label = _("Remove");
    }
    /* Remove VDR recordings */
    else if (current_is_vdr_recordings ()) {
        enabled = (num_selected > 0);
        label = _("Remove");
    }
    /* Remove VDR timers */
    else if (current_is_vdr_timers ()) {
        enabled = (num_selected > 0);
        label = _("Remove");
    }
    /* Remove regular files */
    else {
        enabled = (num_selected > 0);
        enabled &= !current_is_toplevel ();

        /**
         * @todo Implement removing for regular files.
         */
        enabled = false;
        label = _("Remove");
    }
    otk_widget_set_enabled (filelist_menu_buttons[2], enabled);
    otk_button_set_text (filelist_menu_buttons[2], label);

    /* Extract DVD's, video and audio CD's */
#ifdef HAVE_HAL
    bool contains_disc = false;
    for (i = 0; i < num_selected; i++) {
        contains_disc |= strc_starts_with (items[i]->mrl, "dvd:");
        contains_disc |= strc_starts_with (items[i]->mrl, "vcd:");
        contains_disc |= strc_starts_with (items[i]->mrl, "cdda:");
    }
#endif /* HAVE_HAL */

    if (false) {
        /* do nothing */
    }
#ifdef HAVE_HAL
    else if (contains_disc && (num_selected == 1)) {
        enabled = true;
        label = _("Extract");
    }
#endif /* HAVE_HAL */
#ifdef HAVE_VDR
    else if (current_is_vdr_timers ()) {
        enabled = (num_selected == 1);
        if ((num_selected == 1) && !items[0]->vdr.is_active) {
            label = _("Activate");
        }
        else {
            label = _("Deactivate");
        }
    }
#endif /* HAVE_VDR */
    else {
        enabled = (num_selected > 0);
        enabled &= !current_is_toplevel ();
        enabled &= !current_is_favorites ();
        enabled &= !current_is_recent ();
        enabled &= !current_is_playlists ();
        enabled &= !current_is_vdr_recordings ();
        enabled &= !current_is_vdr_channels ();
        enabled &= !current_is_vdr_timers ();

        /**
         * @todo Implement copying for regular files.
         */
        enabled = false;
        label = _("Copy");
    }
    otk_widget_set_enabled (filelist_menu_buttons[3], enabled);
    otk_button_set_text (filelist_menu_buttons[3], label);

    /* Current title */
    {
        enabled = odk_current_is_playback_mode (oxine->odk);
    }
    otk_widget_set_visible (filelist_menu_buttons[7], enabled);

    ho_free (items);
}


static void
filelist_menu_list_save_position (void)
{
    cur_menu->cur_filelist->top_position =
        otk_list_get_pos (filelist_menu_list);
    cur_menu->cur_filelist->cur_position =
        otk_list_get_focus (filelist_menu_list);
}


static bool
filelist_menu_add_fileitem (fileitem_t * fileitem)
{
    if (!fileitem) {
        return false;
    }
    else if (!ho_checkref (fileitem)) {
        return false;
    }
    else if (!fileitem_is_addable (fileitem)) {
        return false;
    }

    filelist_expand (fileitem);

    if (!playlist_add_fileitem (oxine->rw_playlist, fileitem)) {
        return false;
    }
    else if (!cur_menu->fav_filelist) {
        return true;
    }
    else if (fileitem->type != FILE_TYPE_FOLDER) {
        return true;
    }

    favorites_add_directory (cur_menu->fav_filelist,
                             fileitem->title, fileitem->mrl);

    return true;
}


/**
 * This function is called whenever a listentry referencing a fileitem is
 * called. We can then decrement the reference counter of the fileitem.
 */
static void
fileitem_decref_cb (void *p)
{
    fileitem_t *item = (fileitem_t *) p;
    ho_decref (item);
}


static void
filelist_menu_addto_list (fileitem_t * item)
{
    otk_widget_t *w;

    char *device = NULL;
#ifdef HAVE_HAL
    if (item->hal.device) {
        device = item->hal.device;
    }
#endif

    const char *title = NULL;
    switch (item->type) {
    case FILE_TYPE_FOLDER:
    case FILE_TYPE_MEDIAMARKS:
        if (!item->title_gui) {
            item->title_gui = ho_strdup_printf ("[] %s", item->title);
        }
    default:
        if (item->title_gui) {
            title = item->title_gui;
        }
        else {
            title = item->title;
        }
        break;
    }
    assert (title);

    /* Increment the reference counter to the fileitem so it is not freed
     * while still referenced by the OTK list. */
    ho_incref (item);

    /* If this entry refers to a DVD and is not an image on disc, we use the
     * DVD disc callback. */
    if (strc_starts_with (item->mrl, "dvd:")
        && mrl_is_device (item->mrl)) {
        w = otk_listentry_new (filelist_menu_list, title,
                               play_dvd_cb, device,
                               listentry_select_cb, item,
                               listentry_remove_cb, item,
                               listentry_copy_cb, item,
                               NULL, NULL, NULL, NULL);
    }
    /* If this entry refers to a video CD and is not an image on disc, we use
     * the video CD disc callback. */
    else if (strc_starts_with (item->mrl, "vcd:")
             && mrl_is_device (item->mrl)) {
        w = otk_listentry_new (filelist_menu_list, title,
                               play_vcd_cb, device,
                               listentry_select_cb, item,
                               listentry_remove_cb, item,
                               listentry_copy_cb, item,
                               NULL, NULL, NULL, NULL);
    }
#ifdef HAVE_VDR
    else if (strc_starts_with (item->mrl, FILELIST_VDR_TIM_MRL)
             && (item->type == FILE_TYPE_REGULAR)) {
        w = otk_listentry_new (filelist_menu_list, title,
                               NULL, NULL,
                               listentry_select_cb, item,
                               listentry_remove_cb, item,
                               NULL, NULL, NULL, NULL, NULL, NULL);
    }
    else if (strc_starts_with (item->mrl, FILELIST_VDR_EPG_MRL)
             && (item->type == FILE_TYPE_REGULAR)) {
        w = otk_listentry_new (filelist_menu_list, title,
                               listentry_activate_cb, item,
                               listentry_select_cb, item,
                               NULL, NULL, NULL, NULL, NULL, NULL,
                               NULL, NULL);
    }
#endif
    /* Else we use the normal activate callback. */
    else {
        w = otk_listentry_new (filelist_menu_list, title,
                               listentry_activate_cb, item,
                               listentry_select_cb, item,
                               listentry_remove_cb, item,
                               listentry_copy_cb, item,
                               NULL, NULL, NULL, NULL);
    }

    otk_widget_set_destruct_callback (w, fileitem_decref_cb, item);
    otk_widget_set_focus_callbacks (w,
                                    listentry_focus_enter_cb, item,
                                    listentry_focus_leave_cb, item);
}


/**
 * Returns TRUE if this fileitem is to be considered for the playall metatrack.
 */
static bool
fileitem_good_for_playall_metatrack (fileitem_t * item)
{
    bool good = false;
    good |= mrl_has_protocol (item->mrl, "file");
    good |= mrl_has_protocol (item->mrl, "cdda");
    return good;
}


static void
filelist_menu_update_list (void)
{
    if (!_filelist_menu_is_current_menu ()) {
        return;
    }

    /* If we're currently not in the toplevel filelist and the filelist does
     * NOT have a parent we're probably in a removable disc, that has been
     * removed (see disc.c). To rescue ourself of segmentation faults we jump
     * to the toplevel filelist and clear the prev and next lists. */
    if (!current_is_toplevel () && !cur_menu->cur_filelist->parent_item) {
        filelist_menu_set_filelist (cur_menu->top_filelist, false, true);
        filelist_tracker_reset (cur_menu->prev_list);
        filelist_tracker_reset (cur_menu->next_list);
    }

    fileitem_t *item = cur_menu->cur_filelist->parent_item;
    char *keyword = mrl_get_query_parameter (item->mrl, "oxine_keyword");
    if (keyword && (strlen (keyword) > 0)) {
        otk_editbox_set_text (edit_search, keyword);
    }
    else {
        otk_editbox_set_text (edit_search, _("Search for..."));
    }
    ho_free (keyword);

    ho_setref (cur_menu->focused, NULL);
    otk_list_clear (filelist_menu_list);
    otk_list_clear (filelist_menu_info);

    filelist_lock (cur_menu->cur_filelist);

    /* Depending on the location and content of the current folder we might
     * want to display a play-all metaitem at the top of the list. */
    char *mrl = cur_menu->cur_filelist->mrl;
    if (config_get_bool ("misc.play_all_metatrack.use")
        && !current_is_toplevel ()
        && !current_is_favorites ()
        && !current_is_recent ()
        && !current_is_playlists ()
        && !mrl_has_protocol (mrl, "dvb")
        && !mrl_has_protocol (mrl, "vdr")
        && !mrl_has_protocol (mrl, "ftp")
        && !mrl_has_protocol (mrl, "http")
        && !mrls_are_equal (mrl, get_mrl_mediamarks_music ())
        && !mrls_are_equal (mrl, get_mrl_mediamarks_video ())
        && !mrls_are_equal (mrl, get_mrl_mediamarks_image ())
        && !mrls_are_equal (mrl, get_mrl_root ())
        && !mrls_are_equal (mrl, get_mrl_home ())) {
        bool has_regulars = false;
        bool has_subfolders = false;

        fileitem_t *item = filelist_first (cur_menu->cur_filelist);
        while (item) {
            switch (item->type) {
            case FILE_TYPE_REGULAR:
                has_regulars |= fileitem_good_for_playall_metatrack (item);
                break;
            case FILE_TYPE_FOLDER:
            case FILE_TYPE_MEDIAMARKS:
                has_subfolders |= fileitem_good_for_playall_metatrack (item);
                break;
            default:
                break;
            }
            item = filelist_next (cur_menu->cur_filelist, item);
        }

        /* If the current folder only contains on regular entry we don't
         * show the play-all metaitem. */
        has_regulars &= (filelist_length (cur_menu->cur_filelist) > 1);

        if (has_regulars || has_subfolders) {
            otk_cb_t cb = NULL;
            const char *title = NULL;

            const char *title_add = NULL;
            const char *title_play = NULL;

            if (has_subfolders) {
                title_add = _("Add all titles from this folder"
                              " and all subfolders to the playlist...");
                title_play = _("Play all titles in this folder"
                               " and all subfolders...");
            }
            else {
                title_add = _("Add all titles in this folder"
                              " to the playlist...");
                title_play = _("Play all titles in this folder...");
            }

            int policy = config_get_number ("playlist.addto.policy");
            if ((policy == 0) || !odk_current_is_playback_mode (oxine->odk)) {
                cb = filelist_playall_now;
                title = title_play;
            }
            else {
                cb = filelist_addall_now;
                title = title_add;
            }

            otk_widget_t *w = otk_listentry_new (filelist_menu_list, title,
                                                 filelist_playall_cb, cb,
                                                 listentry_select_cb,
                                                 NULL, NULL, NULL, NULL, NULL,
                                                 NULL, NULL, NULL, NULL);
            otk_widget_set_focus_callbacks (w, listentry_focus_enter_cb, NULL,
                                            listentry_focus_leave_cb, NULL);
        }
    }

    /* next we add all the entries from the current filelist */
    if (current_is_favorites () || current_is_recent ()) {
        int i = 0;
        fileitem_t *item = filelist_first (cur_menu->cur_filelist);
        for (i = 0; (i < 20) && item; i++) {
            if (item->listened_count > 1) {
                filelist_menu_addto_list (item);
            }
            else if (current_is_recent ()) {
                filelist_menu_addto_list (item);
            }
            item = filelist_next (cur_menu->cur_filelist, item);
        }
    }
    else {
        fileitem_t *item = filelist_first (cur_menu->cur_filelist);
        while (item) {
            filelist_menu_addto_list (item);
            item = filelist_next (cur_menu->cur_filelist, item);
        }
    }

    /* last we add all removable drives currently available */
#ifdef HAVE_HAL
    if (!current_is_toplevel ()) {
        /* do nothing */
    }
    else if (filelist_trylock (oxine->hal_volume_list) != EBUSY) {
        fileitem_allowed_t a = cur_menu->top_filelist->allowed_filetypes;
        fileitem_t *item = filelist_first (oxine->hal_volume_list);
        while (item) {
            bool add = false;
            /* We only allow audio CDs in the audio menu and DVDs and VCDs
             * in the video menu. */
            add |= fileitem_is_allowed (item, a);
            /* Mountable discs (e.g. external HD, CDROM) are allowed in all
             * menus. */
            add |= item->is_mountpoint;
            /* We only add volumes. Empty drives are not shown in this list
             * as they are not really of interest to a user. */
            add &= (item->hal.volume_udi != NULL);

            if (add) {
                filelist_menu_addto_list (item);
            }
            item = filelist_next (oxine->hal_volume_list, item);
        }
        filelist_unlock (oxine->hal_volume_list);
    }
#endif

    int top_position = cur_menu->cur_filelist->top_position;
    int cur_position = cur_menu->cur_filelist->cur_position;
    otk_list_set_pos (filelist_menu_list, top_position);
    otk_list_set_focus (filelist_menu_list, cur_position);

    filelist_unlock (cur_menu->cur_filelist);
}


static void
filelist_menu_update_title (void)
{
    if (!_filelist_menu_is_current_menu ()) {
        /* do nothing */
    }
    else if (cur_menu->cur_filelist == cur_menu->top_filelist) {
        otk_label_set_text (filelist_menu_title,
                            _("Please choose an entry..."));
    }
    else {
        char *title = cur_menu->cur_filelist->title_complete;
        otk_label_set_text (filelist_menu_title, title);
    }
}


static void
filelist_menu_set_filelist (filelist_t * newlist, bool force_update,
                            bool allow_update)
{
    filelist_t *prev = (filelist_t *) l_list_last (cur_menu->prev_list);
    filelist_t *next = (filelist_t *) l_list_last (cur_menu->next_list);

    /* Cancel any running thumbnail job. */
    cancel_event_job (&thumbnail_job);
    /* Cancel any running file info job. */
    cancel_event_job (&fileinfo_job);

    /* Stop monitoring the old location. */
    fam_monitor_del_list (cur_menu->cur_filelist);

    if (!newlist) {
        newlist = cur_menu->top_filelist;
    }
#ifdef HAVE_HAL
    /* If we're trying to set the HAL volume list as the new filelist, we
     * set the toplevel filelist instead. */
    else if (newlist == oxine->hal_volume_list) {
        filelist_t *filelist = cur_menu->top_filelist;
        filelist_menu_set_filelist (filelist, false, allow_update);
        return;
    }
#endif
    /* If the new filelist we're trying to set points to a local file or
     * directory, that does not exist any more, we walk up one step. */
    else if (mrl_is_local (newlist->mrl) && !file_exists (newlist->mrl)) {
        filelist_t *filelist = newlist->parent_item->parent_list;
        filelist_menu_set_filelist (filelist, force_update, allow_update);
        return;
    }

    /* We update the new list if its on a local filesystem to make sure we're
     * in sync with the filesystem. */
    force_update |= mrl_is_local (newlist->mrl);
    if (force_update && allow_update) {
        filelist_menu_please_wait ();
        filelist_menu_clear_lists ();
        filelist_lock (newlist->parent_item->parent_list);
        filelist_update (newlist);
        filelist_unlock (newlist->parent_item->parent_list);
    }

    /* If we're entering the directory we're already in, we don't want to
     * update the trackers in any way. */
    if (newlist == cur_menu->cur_filelist) {
        /* do nothing */
    }
    /* We are returning to the directory on top of the prev stack. We 
     * remove it from the prev stack and add the current directory to the 
     * next stack. */
    else if (newlist == prev) {
        filelist_tracker_remove (cur_menu->prev_list, newlist);
        filelist_tracker_append (cur_menu->next_list, cur_menu->cur_filelist);
    }
    /* We are returning to the directory on top of the next stack. We 
     * remove it from the next stack and add the current directory to the 
     * prev stack. */
    else if (newlist == next) {
        filelist_tracker_remove (cur_menu->next_list, newlist);
        filelist_tracker_append (cur_menu->prev_list, cur_menu->cur_filelist);
    }
    /* We are entering a directory that is neither on top of the prev stack
     * nor on top of the next stack. We clear the next stack and add the
     * current directory to the prev stack. */
    else {
        filelist_tracker_reset (cur_menu->next_list);
        filelist_tracker_append (cur_menu->prev_list, cur_menu->cur_filelist);
    }

    /* Make sure any filelists living only in the trackers are removed and
     * thus freed. */
    filelist_tracker_clean (cur_menu->prev_list);
    filelist_tracker_clean (cur_menu->next_list);

    /* We update the position in the list. */
    if (otk_list_get_length (filelist_menu_list) > 0) {
        filelist_menu_list_save_position ();
    }
    ho_setref (cur_menu->cur_filelist, newlist);

#ifdef HAVE_OSD_IMAGE
    /* We schedule a new thumbnail job. */
    oxine_event_t *thumbnail_event = oxine_event_new (oxine);
    thumbnail_event->type = OXINE_EVENT_THUMBNAIL_UPDATE;
    thumbnail_job = schedule_event_job (100, thumbnail_event);
#endif

    /* We schedule a new fileinfo job. */
    oxine_event_t *fileinfo_event = oxine_event_new (oxine);
    fileinfo_event->type = OXINE_EVENT_FILEINFO_UPDATE;
    fileinfo_job = schedule_event_job (100, fileinfo_event);

    /* Start monitoring the new location. */
    fam_monitor_add_list (newlist);

    /* Depending on where we are we allow single or multiple entries of the
     * list to be selected. */
    bool selection_single = false;
    selection_single |= current_is_vdr_timers ();
    selection_single |= current_is_vdr_recordings ();
    selection_single |= current_is_vdr_channels ();

    if (selection_single) {
        otk_list_set_allow_select (filelist_menu_list,
                                   OTK_LIST_SELECTION_SINGLE);
    }
    else {
        otk_list_set_allow_select (filelist_menu_list,
                                   OTK_LIST_SELECTION_MULTIPLE);
    }

    filelist_menu_update_list ();
    filelist_menu_update_title ();
    filelist_menu_update_buttons ();
}


static void
fileitem_remove_yes_cb (void *entry_cb_data)
{
    fileitem_t *fileitem = (fileitem_t *) entry_cb_data;
    filelist_t *filelist = fileitem->parent_list;

    assert (fileitem);
    assert (filelist);

    please_wait (cur_menu->background);

    filelist_lock (filelist);

    if (current_is_playlists ()) {
        info (_("Removing '%s'."), fileitem->mrl);
        char *filename = mrl_get_filename (fileitem->mrl);
        if (unlink (filename) == 0) {
            filelist_remove (cur_menu->cur_filelist, fileitem);
        }
        else {
            error (_("Could not remove '%s': %s!"),
                   filename, strerror (errno));
        }
        ho_free (filename);
    }
    else if (current_is_favorites () || current_is_recent ()) {
        info (_("Removing '%s' from favorites."), fileitem->mrl);
        favorites_remove_directory (cur_menu->cur_filelist, fileitem);
    }
#ifdef HAVE_VDR
    else if ((fileitem->type == FILE_TYPE_REGULAR)
             && strc_starts_with (fileitem->mrl, FILELIST_VDR_REC_MRL)) {
        info (_("Removing '%s'."), fileitem->mrl);
        if (vdr_recording_remove (fileitem)) {
            /* After removing an entry we have to reread the entries, as the
             * numbers of the recordings may have changed. */
            filelist_clear (filelist);
        }
        else {
            error (_("Could not remove '%s': %s!"),
                   fileitem->mrl, _("SVDRP error"));
        }
    }
    else if ((fileitem->type == FILE_TYPE_REGULAR)
             && strc_starts_with (fileitem->mrl, FILELIST_VDR_TIM_MRL)) {
        info (_("Removing '%s'."), fileitem->mrl);
        if (vdr_timer_remove (fileitem)) {
            /* After removing an entry we have to reread the entries, as the
             * numbers of the timers may have changed. */
            filelist_clear (filelist);
        }
        else {
            error (_("Could not remove '%s': %s!"),
                   fileitem->mrl, _("SVDRP error"));
        }
    }
#endif
#ifdef HAVE_HAL
    else if (fileitem->parent_list == oxine->hal_volume_list) {
        /* Only optical drives are ejected. All other external volumes are
         * unmounted. */
        LibHalDriveType type = hal_get_drive_type (fileitem->hal.device);
        if (type != LIBHAL_DRIVE_TYPE_CDROM) {
            volume_umount (fileitem->hal.device);
        }
        else {
            drive_eject (fileitem->hal.device);
        }
    }
#endif
    else if (mrl_is_device (fileitem->mrl)) {
        char *device = mrl_get_filename (fileitem->mrl);
        drive_eject (device);
        ho_free (device);
    }
    else {
        todo ("Implement removing of 'normal' files.");
    }

    filelist_unlock (filelist);

    show_menu_filelist (NULL);
}


static bool
fileitem_remove_ok (fileitem_t * fileitem)
{
    if (false) {
        /* do nothing */
    }
    else if (current_is_playlists ()) {
        return true;
    }
    else if (current_is_favorites () || current_is_recent ()) {
        return true;
    }
#ifdef HAVE_VDR
    else if ((fileitem->type == FILE_TYPE_REGULAR)
             && strc_starts_with (fileitem->mrl, FILELIST_VDR_REC_MRL)) {
        return true;
    }
    else if ((fileitem->type == FILE_TYPE_REGULAR)
             && strc_starts_with (fileitem->mrl, FILELIST_VDR_TIM_MRL)) {
        return true;
    }
#endif
#ifdef HAVE_HAL
    else if (fileitem->parent_list == oxine->hal_volume_list) {
        return true;
    }
#endif
    else if (mrl_is_device (fileitem->mrl)) {
        return true;
    }
    else {
        return false;
    }
}


static void
listentry_remove_cb (void *entry_cb_data)
{
    fileitem_t *fileitem = (fileitem_t *) entry_cb_data;

    if (!ho_checkref (fileitem)) {
        /* do nothing */
    }
    else if (!fileitem_remove_ok (fileitem)) {
        char *t = _("The selected entry may not be removed.");
        show_message_dialog (fileitem_remove_yes_cb, fileitem,
                             cur_menu->menu_cb, NULL,
                             DIALOG_BACK, cur_menu->background, t);
    }
#ifdef HAVE_HAL
    else if (fileitem->parent_list == oxine->hal_volume_list) {
        fileitem_remove_yes_cb (fileitem);
    }
#endif
    else if (mrl_is_device (fileitem->mrl)) {
        fileitem_remove_yes_cb (fileitem);
    }
    else {
        char *f = _("Are you sure you want to delete this entry?");
        char *t = ho_strdup_printf ("%s\n\n- %s -", f, fileitem->title);
        show_message_dialog (fileitem_remove_yes_cb, fileitem,
                             cur_menu->menu_cb, NULL,
                             DIALOG_YES_NO, cur_menu->background, t);
        ho_free (t);
    }
}


static void
listentry_copy_cb (void *entry_cb_data)
{
#ifdef HAVE_EXTRACTOR
    fileitem_t *fileitem = (fileitem_t *) entry_cb_data;
    filelist_t *filelist = fileitem->parent_list;

    filelist_lock (filelist);
    if (!ho_checkref (fileitem)) {
        /* do nothing */
    }
    else if (extractor_start (1, &fileitem)) {
        show_menu_extractor (NULL);
    }
    else {
        show_message_dialog (backto_menu_cb, oxine, NULL, NULL,
                             DIALOG_OK, cur_menu->background,
                             _("The extractor is currently busy!"));
    }
    filelist_unlock (filelist);
#endif
}


/*
 * This is the callback for a click on an item in the list.
 */
static void
listentry_select_cb (void *entry_cb_data)
{
    /* Only the 'Play all' metaentries should have entry_cb_data == NULL. As
     * they are always at the top of the list, we can deselect them here. */
    if (!entry_cb_data) {
        otk_list_set_selected (filelist_menu_list, 0, false);
    }
    else {
        filelist_menu_update_buttons ();
    }
}


static void
filelist_menu_please_wait (void)
{
    otk_label_set_text (filelist_menu_title, _("Please wait..."));
    otk_widget_set_enabled (button_back, false);
    otk_widget_set_enabled (button_home, false);
    otk_widget_set_enabled (button_prev, false);
    otk_widget_set_enabled (button_next, false);
    otk_draw (oxine->otk);
}


static void
filelist_menu_clear_lists (void)
{
    ho_setref (cur_menu->focused, NULL);
    otk_list_clear (filelist_menu_list);
    otk_list_clear (filelist_menu_info);
    otk_draw_needy (oxine->otk);
}


#ifdef HAVE_VDR
static void
create_timer_yes_cb (void *p)
{
    fileitem_t *fileitem = (fileitem_t *) p;

    if (vdr_timer_create (fileitem)) {
        show_message_dialog (cur_menu->menu_cb, NULL,
                             cur_menu->menu_cb, NULL,
                             DIALOG_BACK, cur_menu->background,
                             _("Successfully created new timer!"));
    }
    else {
        show_message_dialog (cur_menu->menu_cb, NULL,
                             cur_menu->menu_cb, NULL,
                             DIALOG_BACK, cur_menu->background,
                             _("Failed to create new timer!"));
    }
}
#endif /* HAVE_VDR */


/*
 * This is the callback for a doubleclick on an item in the list.
 */
static void
listentry_activate_cb (void *entry_cb_data)
{
    fileitem_t *fileitem = (fileitem_t *) entry_cb_data;
    filelist_t *filelist = fileitem->parent_list;
    if (!ho_checkref (fileitem)) {
        return;
    }

    filelist_menu_please_wait ();
    filelist_menu_list_save_position ();

    /* When changing folders we clear the lists, so that when this
     * takes some time the user gets a visual feedback that something
     * is happening. */
    switch (fileitem->type) {
    case FILE_TYPE_FOLDER:
    case FILE_TYPE_MEDIAMARKS:
        filelist_menu_clear_lists ();
        break;
    default:
        break;
    }

    filelist_lock (filelist);

    /* Make sure the fileitem is expanded. */
    bool allow_update = fileitem->child_list ? true : false;
    filelist_expand (fileitem);

    switch (fileitem->type) {
    case FILE_TYPE_FOLDER:
    case FILE_TYPE_MEDIAMARKS:
        if (fileitem->child_list->wait_for_password) {
            /* do nothing */
        }
        else if (fileitem->child_list->error) {
            if (otk_list_get_length (filelist_menu_list) == 0) {
                filelist_menu_update_list ();
                filelist_menu_update_buttons ();
            }
            char *error = ho_strdup_printf ("%s: %s", _("Error"),
                                            fileitem->child_list->error);
            otk_label_set_text (filelist_menu_title, error);
            otk_draw_needy (oxine->otk);
            ho_free (error);
        }
        else {
            filelist_menu_set_filelist (fileitem->child_list,
                                        false, allow_update);
        }
        break;
    case FILE_TYPE_PLAYLIST:
        if (mrl_is_playlist_oxp (fileitem->mrl)) {
            playlist_clear (oxine->rw_playlist);

            filelist_lock (fileitem->parent_list);
            playlist_add_fileitem (oxine->rw_playlist, fileitem);
            filelist_unlock (fileitem->parent_list);

            /*
             * FIXME: Starting the new item leads to the change callback being
             * called. If we're already in playback mode, the current title
             * and starting time of the loaded playlist is overwritten.
             */
            playitem_t *playitem = playlist_get_current (oxine->rw_playlist);
            playlist_play_item (oxine->rw_playlist, playitem);
        }
        else {
            int policy = config_get_number ("playlist.addto.policy");
            if ((policy == 0) || !odk_current_is_playback_mode (oxine->odk)) {
                playlist_clear (oxine->rw_playlist);
            }

            filelist_lock (fileitem->parent_list);
            playlist_add_fileitem (oxine->rw_playlist, fileitem);
            filelist_unlock (fileitem->parent_list);

            if ((policy == 0) || !odk_current_is_playback_mode (oxine->odk)) {
                set_playback_ended_menu (cur_menu->menu_cb, NULL);
                playlist_play_first (oxine->rw_playlist);
            }
            else {
                char *msg = _("Adding new items to end of current playlist.");
                show_message_dialog (show_menu_filelist, NULL, NULL, NULL,
                                     DIALOG_OK, cur_menu->background, msg);
            }
        }
        break;
    case FILE_TYPE_REGULAR:
        if (false) {
            /* do nothing */
        }
#ifdef HAVE_VDR
        else if (strc_starts_with (fileitem->mrl, FILELIST_VDR_EPG_MRL)) {
            char *f = _("Are you sure you want to record this show?");
            char *t = ho_strdup_printf ("%s\n\n%s", f, fileitem->title);
            show_message_dialog (create_timer_yes_cb, fileitem,
                                 cur_menu->menu_cb, NULL,
                                 DIALOG_YES_NO, cur_menu->background, t);
            ho_free (t);
        }
        else if (strc_starts_with (fileitem->mrl, FILELIST_VDR_CHA_MRL)) {
            bool result = vdr_channel_play (fileitem);
            if (result) {
                hide_user_interface (NULL);
                result = vdr_play ();
            }
            if (!result) {
                const char *m = _("Failed to start television playback!");
                show_message_dialog (cur_menu->menu_cb, NULL,
                                     cur_menu->menu_cb, NULL,
                                     DIALOG_BACK, cur_menu->background, m);
            }
            else {
                set_playback_ended_menu (cur_menu->menu_cb, NULL);
            }
        }
        else if (strc_starts_with (fileitem->mrl, FILELIST_VDR_REC_MRL)) {
            bool result = vdr_recording_play (fileitem);
            if (result) {
                hide_user_interface (NULL);
                result = vdr_play ();
            }
            if (!result) {
                const char *m = _("Failed to play television recording!");
                show_message_dialog (cur_menu->menu_cb, NULL,
                                     cur_menu->menu_cb, NULL,
                                     DIALOG_BACK, cur_menu->background, m);
            }
            else {
                set_playback_ended_menu (cur_menu->menu_cb, NULL);
            }
        }
#endif
        else {
            int policy = config_get_number ("playlist.addto.policy");
            if ((policy == 0) || !odk_current_is_playback_mode (oxine->odk)) {
                playlist_clear (oxine->rw_playlist);
            }

            playitem_t *play = NULL;
            if (is_current_menu (show_menu_video)) {
                filelist_lock (fileitem->parent_list);
                play = playlist_add_fileitem (oxine->rw_playlist, fileitem);
                filelist_unlock (fileitem->parent_list);
            }
            else {
                fileitem_t *item = filelist_first (filelist);
                while (item) {
                    if (item == fileitem) {
                        play = playlist_add_fileitem (oxine->rw_playlist,
                                                      item);
                    }
                    else if (item->type == FILE_TYPE_REGULAR) {
                        playlist_add_fileitem (oxine->rw_playlist, item);
                    }
                    item = filelist_next (filelist, item);
                }
            }

            if ((policy == 0) || !odk_current_is_playback_mode (oxine->odk)) {
                set_playback_ended_menu (cur_menu->menu_cb, NULL);
                playlist_play_item (oxine->rw_playlist, play);
            }
            else {
                char *msg = _("Adding new items to end of current playlist.");
                show_message_dialog (show_menu_filelist, NULL, NULL, NULL,
                                     DIALOG_OK, cur_menu->background, msg);
            }

            if (cur_menu->fav_filelist) {
                filelist_t *list = fileitem->parent_list;
                favorites_add_directory (cur_menu->fav_filelist,
                                         list->title, list->mrl);
            }
        }
        break;
    default:
        break;
    }

    filelist_unlock (filelist);
}


#ifdef HAVE_OSD_IMAGE
static void
filelist_menu_thumbnail_preload (fileitem_t * fileitem)
{
    if (!fileitem) {
        return;
    }

    char *mrl = NULL;
    const char *thumbnail_mrl = fileitem_get_cover_image (fileitem);

    if (thumbnail_mrl) {
        if (is_downloadable (thumbnail_mrl)) {
            mrl = download_to_cache (thumbnail_mrl, NULL, false);
        }
        else {
            mrl = ho_strdup (thumbnail_mrl);
        }

        if (!file_exists (mrl)) {
            ho_free (mrl);
        }
    }

    if (mrl) {
        int a = ODK_ALIGN_CENTER | ODK_ALIGN_VCENTER;
        int x = filelist_menu_image->x;
        int y = filelist_menu_image->y;
        int w = filelist_menu_image->w;
        int h = filelist_menu_image->h;

        odk_osd_preload_image (oxine->odk, mrl, x, y, w, h, false, 0, a);
    }

    ho_free (mrl);
}


static char *
filelist_menu_thumbnail_get (char *mrl)
{
    if (mrl && is_downloadable (mrl)) {
        char *new_mrl = download_to_cache (mrl, NULL, false);
        ho_free (mrl);
        mrl = new_mrl;
    }
    if (mrl && !file_exists (mrl)) {
        ho_free (mrl);
    }

    return mrl;
}


static void
filelist_menu_thumbnail_set ()
{
    assert (_filelist_menu_is_current_menu ());
    filelist_t *filelist = cur_menu->cur_filelist;
    fileitem_t *fileitem = cur_menu->focused;
    if (fileitem && ho_checkref (fileitem)) {
        filelist = fileitem->parent_list;
    }

    if (filelist_trylock (filelist) == EBUSY) {
        oxine_event_t *thumbnail_event = oxine_event_new (oxine);
        thumbnail_event->type = OXINE_EVENT_THUMBNAIL_UPDATE;
        thumbnail_job = schedule_event_job (300, thumbnail_event);
        return;
    }

    char *mrl = NULL;
    if (fileitem && ho_checkref (fileitem)) {
        mrl = (char *) fileitem_get_cover_image (fileitem);
        mrl = mrl ? ho_strdup (mrl) : NULL;
        mrl = filelist_menu_thumbnail_get (mrl);
    }
    if (!mrl && ho_checkref (filelist)) {
        mrl = (char *) filelist_get_cover_image (filelist);
        mrl = mrl ? ho_strdup (mrl) : NULL;
        mrl = filelist_menu_thumbnail_get (mrl);
    }
    if (!mrl && (filelist->allowed_filetypes == ALLOW_FILES_VIDEO)) {
        mrl = ho_strdup_printf ("%s/icons/logo_folder_video.png",
                                get_dir_oxine_skin ());
        mrl = filelist_menu_thumbnail_get (mrl);
    }
    if (!mrl && (filelist->allowed_filetypes == ALLOW_FILES_MUSIC)) {
        mrl = ho_strdup_printf ("%s/icons/logo_folder_music.png",
                                get_dir_oxine_skin ());
        mrl = filelist_menu_thumbnail_get (mrl);
    }
    if (!mrl && (filelist->allowed_filetypes == ALLOW_FILES_IMAGE)) {
        mrl = ho_strdup_printf ("%s/icons/logo_folder_image.png",
                                get_dir_oxine_skin ());
        mrl = filelist_menu_thumbnail_get (mrl);
    }

    otk_image_set_mrl (filelist_menu_image, mrl);

    /* We try and preload the thumbnail images of the next and of the
     * previous item in the list. */
    if (fileitem && ho_checkref (fileitem)) {
        filelist_menu_thumbnail_preload (filelist_next (filelist, fileitem));
        filelist_menu_thumbnail_preload (filelist_prev (filelist, fileitem));
    }

    filelist_unlock (filelist);

    ho_free (mrl);
}
#endif /* HAVE_OSD_IMAGE */


static void
filelist_menu_fileinfo_update ()
{
    assert (_filelist_menu_is_current_menu ());
    filelist_t *filelist = cur_menu->cur_filelist;
    fileitem_t *fileitem = cur_menu->focused;
    if (fileitem && ho_checkref (fileitem)) {
        filelist = fileitem->parent_list;
    }

    otk_list_clear (filelist_menu_info);

    if (filelist_trylock (filelist) == EBUSY) {
        oxine_event_t *thumbnail_event = oxine_event_new (oxine);
        thumbnail_event->type = OXINE_EVENT_FILEINFO_UPDATE;
        thumbnail_job = schedule_event_job (300, thumbnail_event);
        return;
    }

    if (ho_checkref (filelist)) {
        l_list_t *infolist = filelist_get_info_list (filelist);
        if (infolist) {
            char *infoitem = (char *) l_list_first (infolist);
            while (infoitem) {
                otk_infoitem_new (filelist_menu_info, infoitem);
                infoitem = (char *) l_list_next (infolist, infoitem);
            }
        }
    }
    if (fileitem && ho_checkref (fileitem)) {
        if (otk_list_get_length (filelist_menu_info)) {
            otk_list_separator_new (filelist_menu_info);
        }
        l_list_t *infolist = fileitem_get_info_list (fileitem);
        if (infolist) {
            char *infoitem = (char *) l_list_first (infolist);
            while (infoitem) {
                otk_infoitem_new (filelist_menu_info, infoitem);
                infoitem = (char *) l_list_next (infolist, infoitem);
            }
        }
    }

    filelist_unlock (filelist);
}


static void
filelist_menu_filelist_update (filelist_t * filelist)
{
    assert (filelist);

    if (!_filelist_menu_is_current_menu ()) {
        /* do nothing */
    }
    else if (!ho_checkref (filelist)) {
        /* do nothing */
    }
    else if (filelist != cur_menu->cur_filelist) {
        /* do nothing */
    }
    /* Check to see if the current folder still exists. If it does not
     * exist, we change to the parent directory. And check the new current
     * filelist for changes. */
    else if (mrl_is_local (filelist->mrl) && !file_exists (filelist->mrl)) {
        filelist->top_position = otk_list_get_pos (filelist_menu_list);
        filelist->cur_position = otk_list_get_focus (filelist_menu_list);
        filelist = filelist->parent_item->parent_list;
        filelist_menu_set_filelist (filelist, true, true);
    }
    else {
        filelist->top_position = otk_list_get_pos (filelist_menu_list);
        filelist->cur_position = otk_list_get_focus (filelist_menu_list);
        filelist_menu_set_filelist (filelist, true, true);
    }
}


static void
listentry_focus_enter_cb (void *entry_cb_data)
{
    fileitem_t *fileitem = (fileitem_t *) entry_cb_data;
    if (fileitem && !ho_checkref (fileitem)) {
        return;
    }
    if (fileitem && fileitem->title) {
        int c = fileitem->title[0];
        otk_list_focus_by_index_char (filelist_menu_short, c);
    }
    else {
        otk_list_set_pos (filelist_menu_short, 0);
    }

    /* We cancel any fileinfo job that might be running. */
    cancel_event_job (&fileinfo_job);
    /* We cancel any thumbnail job that might be running. */
    cancel_event_job (&thumbnail_job);

#ifdef HAVE_OSD_IMAGE
    /* We want to wait for 300ms before showing the thumbnail. This is a
     * good idea, because if the user scrolls down a list quickly, showing
     * the thumbnail at once would slow down scrolling. */
    oxine_event_t *thumbnail_event = oxine_event_new (oxine);
    thumbnail_event->type = OXINE_EVENT_THUMBNAIL_UPDATE;
    thumbnail_job = schedule_event_job (300, thumbnail_event);
#endif /* HAVE_OSD_IMAGE */

    /* We want to wait for 300ms before showing the fileinfo. This is a
     * good idea, because if the user scrolls down a list quickly, showing
     * the fileinfo at once would slow down scrolling. */
    oxine_event_t *fileinfo_event = oxine_event_new (oxine);
    fileinfo_event->type = OXINE_EVENT_FILEINFO_UPDATE;
    fileinfo_job = schedule_event_job (300, fileinfo_event);

    /* Set the focused fileitem. */
    ho_setref (cur_menu->focused, fileitem);
}


static void
listentry_focus_leave_cb (void *entry_cb_data)
{
    fileitem_t *fileitem = (fileitem_t *) entry_cb_data;
    if (fileitem && !ho_checkref (fileitem)) {
        return;
    }

    /* We do the same thing as before. */
    listentry_focus_enter_cb (entry_cb_data);

    /* Make sure the title is updated when leaving an item that has an error. */
    filelist_menu_update_title ();
}


static void
filelist_up_cb (void *p)
{
    if (cur_menu->cur_filelist != cur_menu->top_filelist) {
        fileitem_t *new_item = cur_menu->cur_filelist->parent_item;
        filelist_t *new_list = new_item->parent_list;
        filelist_menu_set_filelist (new_list, false, true);
    }
}


static void
filelist_home_cb (void *p)
{
    if (cur_menu->cur_filelist != cur_menu->top_filelist) {
        filelist_t *new_list = cur_menu->top_filelist;
        filelist_menu_set_filelist (new_list, false, true);
    }
}


static void
filelist_prev_cb (void *p)
{
    filelist_t *last = (filelist_t *) l_list_last (cur_menu->prev_list);
    if (last) {
        filelist_menu_set_filelist (last, false, true);
    }
}


static void
filelist_next_cb (void *p)
{
    filelist_t *last = (filelist_t *) l_list_last (cur_menu->next_list);
    if (last) {
        filelist_menu_set_filelist (last, false, true);
    }
}


/**
 * Toggles the sidebar making sure that previously selected items in the
 * list are selected after the change also.
 */
static void
filelist_toggle_sidebar_cb (void *p)
{
    int i = 0;
    int n = 0;
    int *selected = otk_list_get_selected_positions (filelist_menu_list, &n);
    bool focused = otk_widget_is_focused (button_hide);

    toggle_sidebar ();

    for (i = 0; i < n; i++) {
        otk_list_set_selected (filelist_menu_list, selected[i], true);
    }
    if (focused) {
        otk_widget_set_focused (button_hide, true);
    }

    filelist_menu_update_buttons ();
    show_user_interface (NULL);

    ho_free (selected);
}


static void
filelist_menu_event_handler (void *p, oxine_event_t * event)
{
    switch (event->type) {
    case OXINE_EVENT_FILELIST_UPDATE:
        filelist_menu_filelist_update ((filelist_t *) event->data.user_data);
        break;
    default:
        break;
    }

    if (!_filelist_menu_is_current_menu ()) {
        /* Cancel any running thumbnail job. */
        cancel_event_job (&thumbnail_job);
        /* Cancel any running file info job. */
        cancel_event_job (&fileinfo_job);
        return;
    }

    switch (event->type) {
#ifdef HAVE_OSD_IMAGE
    case OXINE_EVENT_THUMBNAIL_UPDATE:
        filelist_menu_thumbnail_set ();
        break;
#endif
    case OXINE_EVENT_FILEINFO_UPDATE:
        filelist_menu_fileinfo_update ();
        break;
    case OXINE_EVENT_FILELIST_REENTER:
        ((otk_cb_t) event->data.user_data) ((void *) 0x02);
        break;
    case OXINE_EVENT_FILELIST_CLEAR:
        filelist_menu_clear_lists ();
        break;
    case OXINE_EVENT_KEY:
        switch (event->source.key) {
        case OXINE_KEY_HOME:
            {
                filelist_home_cb (NULL);
            }
            event->source.key = OXINE_KEY_NULL;
            break;
        case OXINE_KEY_BACK:
            if (current_is_toplevel ()) {
                show_menu_main (NULL);
            }
            else {
                filelist_up_cb (NULL);
            }
            event->source.key = OXINE_KEY_NULL;
            break;
        case OXINE_KEY_PREV:
            {
                filelist_prev_cb (NULL);
            }
            event->source.key = OXINE_KEY_NULL;
            break;
        case OXINE_KEY_NEXT:
            {
                filelist_next_cb (NULL);
            }
            event->source.key = OXINE_KEY_NULL;
            break;
        case OXINE_KEY_INSERT:
            if (!otk_list_has_focus (filelist_menu_list)) {
                /* do nothing */
            }
            else if (otk_list_get_selected_count (filelist_menu_list) > 0) {
                selected_add_cb (NULL);
            }
            else {
                focused_add_cb (NULL);
            }
            event->source.key = OXINE_KEY_NULL;
            break;
        case OXINE_KEY_SIDEBAR:
            {
                filelist_toggle_sidebar_cb (NULL);
            }
            event->source.key = OXINE_KEY_NULL;
            break;
        case OXINE_KEY_RED:
            event->source.key = OXINE_KEY_ACTIVATE;
            odk_oxine_event_send (oxine->odk, event);
            event->source.key = OXINE_KEY_NULL;
            break;
        case OXINE_KEY_GREEN:
            event->source.key = OXINE_KEY_INSERT;
            odk_oxine_event_send (oxine->odk, event);
            event->source.key = OXINE_KEY_NULL;
            break;
        case OXINE_KEY_YELLOW:
            event->source.key = OXINE_KEY_REMOVE;
            odk_oxine_event_send (oxine->odk, event);
            event->source.key = OXINE_KEY_NULL;
            break;
        case OXINE_KEY_BLUE:
            if (!otk_list_has_focus (filelist_menu_list)) {
                /* do nothing */
            }
            else if (!cur_menu->focused) {
                /* do nothing */
            }
            else if (ho_checkref (cur_menu->focused)) {
                filelist_lock (cur_menu->focused->parent_list);
                show_menu_fileitem_info (cur_menu->focused);
                filelist_unlock (cur_menu->focused->parent_list);
            }
            event->source.key = OXINE_KEY_NULL;
            break;
        default:
            break;
        }
        break;
    default:
        break;
    }
}


static bool
filelist_menu_get_selected (fileitem_t *** items, int *num_selected)
{
    *items = (fileitem_t **) otk_list_get_selected_values (filelist_menu_list,
                                                           num_selected);

    return (*items && num_selected);
}


static void
focused_add_cb (void *p)
{
    fileitem_t *item = cur_menu->focused;
    filelist_lock (cur_menu->cur_filelist);
    if (filelist_menu_add_fileitem (item)) {
        filelist_menu_list_save_position ();
        show_menu_playlist (oxine);
    }
    filelist_unlock (cur_menu->cur_filelist);
}


static void
selected_add_cb (void *p)
{
    int i;
    int num_selected;
    bool added = false;
    fileitem_t **fileitems;

    if (!filelist_menu_get_selected (&fileitems, &num_selected)) {
        return;
    }

    filelist_lock (cur_menu->cur_filelist);
    for (i = 0; i < num_selected; i++) {
        added |= filelist_menu_add_fileitem (fileitems[i]);
    }
    if (added) {
        filelist_menu_list_save_position ();
        show_menu_playlist (oxine);
    }
    filelist_unlock (cur_menu->cur_filelist);

    ho_free (fileitems);
}


static void
selected_play_cb (void *p)
{
    int i;
    int num_selected;
    bool added = false;
    fileitem_t **fileitems;

    if (!filelist_menu_get_selected (&fileitems, &num_selected)) {
        return;
    }

    filelist_lock (cur_menu->cur_filelist);
    playlist_clear (oxine->rw_playlist);
    for (i = 0; i < num_selected; i++) {
        added |= filelist_menu_add_fileitem (fileitems[i]);
    }
    if (added) {
        filelist_menu_list_save_position ();
        set_playback_ended_menu (cur_menu->menu_cb, NULL);
        playlist_play_first (oxine->rw_playlist);
    }
    filelist_unlock (cur_menu->cur_filelist);

    ho_free (fileitems);
}


static void
selected_copy_cb (void *p)
{
    if (false) {
        /* do nothing */
    }
#ifdef HAVE_VDR
    else if (current_is_vdr_timers ()) {
        int num_selected;
        fileitem_t **fileitems;

        if (!filelist_menu_get_selected (&fileitems, &num_selected)) {
            /* do nothing */
        }
        else if (!vdr_timer_set_active (fileitems[0],
                                        !fileitems[0]->vdr.is_active)) {
            /* do nothing */
        }
        else {
            filelist_menu_set_filelist (cur_menu->cur_filelist, true, true);
        }

        ho_free (fileitems);
    }
#endif
#ifdef HAVE_EXTRACTOR
    else {
        int num_selected;
        fileitem_t **fileitems;

        filelist_lock (cur_menu->cur_filelist);
        if (!filelist_menu_get_selected (&fileitems, &num_selected)) {
            /* do nothing */
        }
        else if (extractor_start (num_selected, fileitems)) {
            show_menu_extractor (oxine);
        }
        else {
            show_message_dialog (backto_menu_cb, oxine, NULL, NULL,
                                 DIALOG_OK, cur_menu->background,
                                 _("The extractor is currently busy!"));
        }
        filelist_unlock (cur_menu->cur_filelist);

        ho_free (fileitems);
    }
#endif
}


static void
selected_remove_cb (void *p)
{
    /* If we're in the favorites filelist, the remove button is a reset
     * button that removes all items from the favorites filelist. */
    if (current_is_favorites () || current_is_recent ()) {
        otk_list_clear (filelist_menu_list);
        otk_list_clear (filelist_menu_info);

        filelist_lock (cur_menu->cur_filelist);
        filelist_clear (cur_menu->cur_filelist);
        favorites_write (cur_menu->cur_filelist);
        filelist_unlock (cur_menu->cur_filelist);

        filelist_menu_update_list ();
        filelist_menu_update_buttons ();
        filelist_menu_update_title ();
    }
    /* In all other lists we get all selected items and call the remove
     * callback for all of them. */
    else {
        int num_selected;
        fileitem_t **fileitems;

        if (filelist_menu_get_selected (&fileitems, &num_selected)) {
            int i = 0;
            for (; i < num_selected; i++) {
                listentry_remove_cb (fileitems[i]);
            }
            ho_free (fileitems);
        }
    }
}


static void
filelist_addall_now (void *p)
{
    filelist_t *filelist = (filelist_t *) p;

    please_wait (cur_menu->background);

    filelist_lock (filelist);
    fileitem_t *fileitem = filelist_first (filelist);
    while (fileitem) {
        playlist_add_fileitem (oxine->rw_playlist, fileitem);
        fileitem = filelist_next (filelist, fileitem);
    }
    if (cur_menu->fav_filelist) {
        favorites_add_directory (cur_menu->fav_filelist,
                                 filelist->title, filelist->mrl);
    }
    filelist_unlock (filelist);

    show_menu_filelist (NULL);
}


static void
filelist_playall_now (void *p)
{
    filelist_t *filelist = (filelist_t *) p;

    please_wait (cur_menu->background);
    playlist_clear (oxine->rw_playlist);

    filelist_lock (filelist);
    fileitem_t *fileitem = filelist_first (filelist);
    while (fileitem) {
        playlist_add_fileitem (oxine->rw_playlist, fileitem);
        fileitem = filelist_next (filelist, fileitem);
    }
    filelist_unlock (filelist);

    if (cur_menu->fav_filelist) {
        favorites_add_directory (cur_menu->fav_filelist,
                                 filelist->title, filelist->mrl);
    }

    set_playback_ended_menu (cur_menu->menu_cb, NULL);
    playlist_play_first (oxine->rw_playlist);
}


static void
filelist_playall_cb (void *entry_cb_data)
{
    otk_cb_t cb = (otk_cb_t) entry_cb_data;
    filelist_t *filelist = cur_menu->cur_filelist;
    int subfolder_count = 0;

    filelist_lock (filelist);
    fileitem_t *fileitem = filelist_first (filelist);
    while (fileitem) {
        switch (fileitem->type) {
        case FILE_TYPE_FOLDER:
        case FILE_TYPE_MEDIAMARKS:
            subfolder_count++;
            break;
        default:
            break;
        }
        fileitem = filelist_next (filelist, fileitem);
    }
    filelist_unlock (filelist);

    if (subfolder_count > 5) {
        char *msg = _("This folder contains %d subfolders.\n\n"
                      "Adding all subfolders to the playlist\n"
                      "recursively may take a very long time!\n\n"
                      "Are you sure you want to continue?");
        show_message_dialog (cb, filelist, backto_menu_cb, NULL,
                             DIALOG_YES_NO, cur_menu->background,
                             msg, subfolder_count);
    }
    else {
        cb (filelist);
    }
}


static void
searchbox_activate_cb (void *p, char *keyword)
{
    filelist_t *cur_filelist = NULL;

    ho_setref (cur_filelist, cur_menu->cur_filelist);
    filelist_lock (cur_filelist);

    char *search_mrl = mrl_attach_query (cur_menu->cur_filelist->mrl,
                                         "oxine_keyword", keyword);
    ho_free (cur_menu->cur_filelist->parent_item->mrl);
    cur_menu->cur_filelist->parent_item->mrl = search_mrl;

    filelist_menu_set_filelist (cur_menu->cur_filelist, true, true);
    otk_widget_set_focused (edit_search, true);

    filelist_unlock (cur_filelist);
    ho_setref (cur_filelist, NULL);
}


static void
searchbox_focus_enter_cb (void *p)
{
    char *keyword = otk_editbox_get_text (edit_search);
    if (str_equals (keyword, _("Search for..."))) {
        otk_editbox_set_text (edit_search, "");
    }
    ho_free (keyword);
}


static void
searchbox_focus_leave_cb (void *p)
{
    char *keyword = otk_editbox_get_text (edit_search);
    if (strlen (keyword) == 0) {
        otk_editbox_set_text (edit_search, _("Search for..."));
    }
    ho_free (keyword);
}


static void
shortcut_activate_cb (void *c)
{
    otk_widget_t *w = otk_list_focus_by_index_char (filelist_menu_list,
                                                    (int) c);
    if (w) {
        otk_widget_set_focused (w, true);
    }
}


static void
filelist_menu_show_search (int x, int y, int w, int h)
{
    edit_search = otk_editbox_new (oxine->otk, x, y, w, h, 100, NULL,
                                   NULL, searchbox_activate_cb, NULL);
    otk_widget_set_focus_callbacks (edit_search,
                                    searchbox_focus_enter_cb, NULL,
                                    searchbox_focus_leave_cb, NULL);
    otk_editbox_set_text (edit_search, _("Search for..."));
}


static void
filelist_menu_show_list (int x, int y, int w, int h)
{
    int a = (OTK_ALIGN_LEFT | OTK_ALIGN_VCENTER);

    int x0 = x;
    int y0 = y;
    int w0 = w - 5 * 46 + 6 - 10;
    filelist_menu_title = otk_label_new (oxine->otk, x0, y0, w0, 35, a,
                                         _("Choose an entry..."));
    otk_widget_set_fontsize (filelist_menu_title, 32);

    int x1 = x0 + w0 + 10;
    int y1 = y;
    button_prev = otk_vector_button_new (oxine->otk, x1, y1, 40, 35,
                                         OSD_VECTOR_ARROW_LEFT, 20, 20,
                                         filelist_prev_cb, oxine);
    x1 += 46;
    button_next = otk_vector_button_new (oxine->otk, x1, y1, 40, 35,
                                         OSD_VECTOR_ARROW_RIGHT, 20, 20,
                                         filelist_next_cb, oxine);
    x1 += 46;
    button_back = otk_vector_button_new (oxine->otk, x1, y1, 40, 35,
                                         OSD_VECTOR_ARROW_UP, 20, 20,
                                         filelist_up_cb, oxine);
    x1 += 46;
    button_home = otk_vector_button_new (oxine->otk, x1, y1, 40, 35,
                                         OSD_VECTOR_HOME, 20, 20,
                                         filelist_home_cb, oxine);
    x1 += 46;
    button_hide = otk_vector_button_new (oxine->otk, x1, y1, 40, 35,
                                         OSD_VECTOR_SIDEBAR, 20, 20,
                                         filelist_toggle_sidebar_cb, oxine);

    int x2 = x;
    int y2 = y + 40;
    int w2 = w - 40;
    int h2 = h - 85;
    filelist_menu_list = otk_list_new (oxine->otk,
                                       x2, y2, w2, h2, 30, 33, true, true,
                                       OTK_LIST_SELECTION_MULTIPLE, oxine);
    int x3 = x2 + w2 + 5;
    int y3 = y + 40;
    int w3 = 35;
    int h3 = h - 85;
    filelist_menu_short = otk_list_new (oxine->otk,
                                        x3, y3, w3, h3, 30, 33, true, false,
                                        OTK_LIST_SELECTION_NONE, oxine);

    int c;
    char b[2] = "A";
    for (c = '0'; c <= '9'; c++) {
        b[0] = c;
        otk_listentry_new (filelist_menu_short, b,
                           shortcut_activate_cb, (void *) c,
                           NULL, NULL, NULL, NULL, NULL, NULL,
                           NULL, NULL, NULL, NULL);
    }
    for (c = 'A'; c <= 'Z'; c++) {
        b[0] = c;
        otk_listentry_new (filelist_menu_short, b,
                           shortcut_activate_cb, (void *) c,
                           NULL, NULL, NULL, NULL, NULL, NULL,
                           NULL, NULL, NULL, NULL);
    }

    int x4 = x2;
    int y4 = y2 + h2 + 5;
    int w4 = w;
    int h4 = 40;
    filelist_menu_show_search (x4, y4, w4, h4);

    otk_widget_set_enabled (button_back, false);
    otk_widget_set_enabled (button_home, false);
    otk_widget_set_enabled (button_prev, false);
    otk_widget_set_enabled (button_next, false);
}


static void
filelist_menu_show_info (int x, int y, int w, int h)
{
    int x1 = x;
    int y1 = y;
    int w1 = w;
    int h1 = h;

#ifdef HAVE_OSD_IMAGE
    int x3 = x;
    int y3 = y;
    int w3 = h;
    int h3 = h;
    otk_border_new (oxine->otk, x3, y3, w3, h3);

    int a2 = (ODK_ALIGN_CENTER | ODK_ALIGN_VCENTER);
    int w2 = h - 10;
    int h2 = h - 10;
    int x2 = x + 5 + (w2 / 2);
    int y2 = y + 5 + (h2 / 2);

    filelist_menu_image = otk_image_new (oxine->otk, x2, y2,
                                         w2, h2, NULL, a2, false);

    x1 += w3 + 5;
    w1 -= w3 + 5;
#endif /* HAVE_OSD_IMAGE */

    filelist_menu_info = otk_infolist_new (oxine->otk, x1, y1,
                                           w1, h1, 26, 26);
    filelist_menu_info_width = w1 - 40;
}


static void
filelist_menu_show_sidebar (int x, int y, int w, int h)
{
    otk_widget_t *b = NULL;

    int x1 = x;
    int y1 = y;

    b = otk_text_button_new (oxine->otk, x1, y1, w, 35, _("Play"),
                             selected_play_cb, oxine);
    filelist_menu_buttons[0] = b;
    y1 += 40;

    b = otk_text_button_new (oxine->otk, x1, y1, w, 35, _("Add"),
                             selected_add_cb, oxine);
    filelist_menu_buttons[1] = b;
    y1 += 40;

    b = otk_text_button_new (oxine->otk, x1, y1, w, 35, _("Remove"),
                             selected_remove_cb, oxine);
    filelist_menu_buttons[2] = b;
    y1 += 40;

    b = otk_text_button_new (oxine->otk, x1, y1, w, 35, _("Copy"),
                             selected_copy_cb, oxine);
    filelist_menu_buttons[3] = b;
    y1 += 40;

    b = otk_text_button_new (oxine->otk, x1, y1, w, 35, _("Eject"),
                             eject_cb, oxine);
    filelist_menu_buttons[4] = b;
    y1 += 50;

    b = otk_text_button_new (oxine->otk, x1, y1, w, 35, _("Playlist"),
                             show_menu_playlist, oxine);
    filelist_menu_buttons[5] = b;
    y1 += 40;

    b = otk_text_button_new (oxine->otk, x1, y1, w, 35, _("Mainmenu"),
                             show_menu_main, oxine);
    filelist_menu_buttons[6] = b;
    y1 += 40;

    b = otk_text_button_new (oxine->otk, x1, y1, w, 35, _("Current title"),
                             show_menu_playback, oxine);
    filelist_menu_buttons[7] = b;
    y1 += 50;

    otk_widget_set_enabled (filelist_menu_buttons[0], false);
    otk_widget_set_enabled (filelist_menu_buttons[1], false);
    otk_widget_set_enabled (filelist_menu_buttons[2], false);
    otk_widget_set_enabled (filelist_menu_buttons[3], false);
    otk_widget_set_visible (filelist_menu_buttons[7], false);
}


/**
 * Trys to lock a filelist. If this is not possible at the moment we show
 * 'Please wait' and try again 50 ms later.
 */
static bool
filelist_menu_testlock (filelist_t * filelist, const char *background,
                        otk_cb_t menu_cb, void *menu_cb_data)
{
    /* If the current filelist is not locked, we continue normally. */
    if (filelist_trylock (filelist) != EBUSY) {
        filelist_unlock (filelist);
        return true;
    }
    /* If the current filelist is locked (or still locked) we wait another
     * 500 ms for it to become available. */
    else {
        please_wait (background);
        set_current_menu (menu_cb, menu_cb_data);
        oxine_event_t *reenter_event = oxine_event_new (oxine);
        reenter_event->type = OXINE_EVENT_FILELIST_REENTER;
        reenter_event->data.user_data = menu_cb;
        schedule_event_job (50, reenter_event);
        return false;
    }
}


static void
filelist_menu_enter (menu_filelist_menu_t menu_index, const char *mrl,
                     fileitem_allowed_t allowed_filetypes,
                     const char *background, otk_cb_t menu_cb, void *p)
{
    odk_add_event_handler (oxine->odk, filelist_menu_event_handler,
                           oxine, EVENT_HANDLER_PRIORITY_NORMAL);

    filelist_t *filelist = filelist_get (FILELIST_MEDIA_MRL);
    if (!filelist_menu_testlock (filelist, background, menu_cb, p)) {
        return;
    }
    filelist = filelist_get (mrl);
    if (!filelist_menu_testlock (filelist, background, menu_cb, p)) {
        return;
    }

    /* Cancel any running thumbnail job. */
    cancel_event_job (&thumbnail_job);
    /* Cancel any running file info job. */
    cancel_event_job (&fileinfo_job);

    /* Make sure the position in the list of the last used filelist menu is
     * saved before we (maybe) change to one of the other filelist menus. */
    if (filelist_menu_list && cur_menu->cur_filelist) {
        otk_set_current_window (oxine->otk, filelist_menu_window);
        filelist_menu_list_save_position ();
    }

    /* Remove all of the current filelist from the file alteration monitor.
     * This is necessary as we don't want filelists from say the video
     * filelist menu to be polled while we're in the music filelist menu. */
    int i = 0;
    for (i = 0; i < FILELIST_MENU_DATA_COUNT; i++) {
        fam_monitor_del_list (mdata[i].cur_filelist);
    }

    /* Set the current menu to the index of the new fielist menu. */
    assert (menu_index < FILELIST_MENU_DATA_COUNT);
    cur_menu = &mdata[menu_index];

    /* If this filelist menu does not have a current filelist yet (meaning
     * we have not been here before) we create and fill it. */
    if (!cur_menu->cur_filelist) {
        please_wait (cur_menu->background);

        ho_setref (cur_menu->top_filelist, filelist_get (mrl));
        ho_setref (cur_menu->cur_filelist, cur_menu->top_filelist);
        ho_setref (cur_menu->fav_filelist, NULL);

        if (menu_index == MENU_FILELIST_MUSIC) {
            ho_setref (cur_menu->fav_filelist,
                       filelist_get (FILELIST_FAVORITES_MOST_MRL));
        }

        cur_menu->prev_list = l_list_new ();
        cur_menu->next_list = l_list_new ();

        cur_menu->menu_cb = menu_cb;
        cur_menu->background = background;
    }

    if (filelist_menu_window) {
        otk_window_destroy (filelist_menu_window);
        filelist_menu_window = NULL;
    }

    /* Make sure all widgets are set to NULL so we detect bad access' to
     * them more easily. */
#ifdef DEBUG
    {
        button_back = NULL;
        button_prev = NULL;
        button_next = NULL;
        button_home = NULL;
        button_hide = NULL;
        edit_search = NULL;

        filelist_menu_list = NULL;
        filelist_menu_short = NULL;
        filelist_menu_window = NULL;
        filelist_menu_title = NULL;
        filelist_menu_info = NULL;
#ifdef HAVE_OSD_IMAGE
        filelist_menu_image = NULL;
#endif

        int i = 0;
        for (i = 0; i < FILELIST_MENU_BUTTONS_COUNT; i++) {
            filelist_menu_buttons[i] = NULL;
        }
    }
#endif

    filelist_menu_window = create_new_window (true, true);

    int x = 20;
    int y = 100;
    int w = odk_osd_get_width (oxine->odk) - 40;
    int h = odk_osd_get_height (oxine->odk) - 120;

    int x1 = x;
    int y1 = y;
    int w1 = w;
    int h1 = h - 190;

    int x2 = x;
    int y2 = y1 + h1 + 5;
    int w2 = w;
    int h2 = h - h1 - 5;

    int x3 = x;
    int y3 = y;
    int w3 = h2;
    int h3 = h1;

    if (oxine->is_sidebar_visible) {
        filelist_menu_show_sidebar (x3, y3, w3, h3);
        x1 += w3 + 5;
        w1 -= w3 + 5;
    }
    {
        filelist_menu_show_list (x1, y1, w1, h1);
        filelist_menu_show_info (x2, y2, w2, h2);
    }

    show_last_filelist_menu_cb = menu_cb;
    set_backto_menu (menu_cb, NULL);
    set_current_menu (menu_cb, NULL);

    show_user_interface (NULL);
    show_menu_background (background);

    filelist_menu_set_filelist (cur_menu->cur_filelist, false, true);
}


void
show_menu_music (void *p)
{
    filelist_menu_enter (MENU_FILELIST_MUSIC, FILELIST_MUSIC_MRL,
                         ALLOW_FILES_MUSIC, "menu_music.png",
                         show_menu_music, p);
}


void
show_menu_video (void *p)
{
    filelist_menu_enter (MENU_FILELIST_VIDEO, FILELIST_VIDEO_MRL,
                         ALLOW_FILES_VIDEO, "menu_video.png",
                         show_menu_video, p);
}


#ifdef HAVE_IMAGE_PLAYBACK
void
show_menu_image (void *p)
{
    filelist_menu_enter (MENU_FILELIST_IMAGE, FILELIST_IMAGE_MRL,
                         ALLOW_FILES_IMAGE, "menu_image.png",
                         show_menu_image, p);
}
#endif


void
show_menu_media (void *p)
{
    filelist_menu_enter (MENU_FILELIST_MEDIA, FILELIST_MEDIA_MRL,
                         ALLOW_FILES_MULTIMEDIA, "menu_media.png",
                         show_menu_media, p);
}


void
show_menu_filelist (void *p)
{
    show_last_filelist_menu_cb (oxine);
}


static void *
filelist_menu_preload_thread (void *data)
{
    filelist_get (FILELIST_MEDIA_MRL);
    usleep (1000 * 200);
    filelist_get (FILELIST_MUSIC_MRL);
    usleep (1000 * 200);
    filelist_get (FILELIST_IMAGE_MRL);
    usleep (1000 * 200);
    filelist_get (FILELIST_VIDEO_MRL);
#ifdef HAVE_VDR
    filelist_get (FILELIST_VDR_MRL);
    filelist_get (FILELIST_VDR_TIM_MRL);
    filelist_get (FILELIST_VDR_REC_MRL);
    filelist_get (FILELIST_VDR_CHA_MRL);
#endif

    return NULL;
}


void
filelist_menu_preload (void)
{
    if (pthread_create (&preload_thread, NULL,
                        filelist_menu_preload_thread, NULL) != 0) {
        error (_("Could not create filelist menu preload thread: %s!"),
               strerror (errno));
    }
}


bool
filelist_menu_is_current_menu (void)
{
    bool result = false;

    result |= is_current_menu (show_menu_music);
    result |= is_current_menu (show_menu_video);
#ifdef HAVE_IMAGE_PLAYBACK
    result |= is_current_menu (show_menu_image);
#endif
    result |= is_current_menu (show_menu_media);

    return result;
}


void
free_menu_filelist (void)
{
    cancel_event_job (&thumbnail_job);
    cancel_event_job (&fileinfo_job);

    int i = 0;
    for (i = 0; i < FILELIST_MENU_DATA_COUNT; i++) {
        cur_menu = &mdata[i];

        if (cur_menu->next_list) {
            l_list_free (cur_menu->next_list, filelist_tracker_ifree);
            cur_menu->next_list = NULL;
        }
        if (cur_menu->prev_list) {
            l_list_free (cur_menu->prev_list, filelist_tracker_ifree);
            cur_menu->prev_list = NULL;
        }

        ho_setref (cur_menu->focused, NULL);
        ho_setref (cur_menu->fav_filelist, NULL);
        ho_setref (cur_menu->cur_filelist, NULL);
        ho_setref (cur_menu->top_filelist, NULL);
    }

    void *preload_res = NULL;
    pthread_join (preload_thread, &preload_res);
}
