
/*
 * Copyright (C) 2004-2005 Maximilian Schwerin
 *
 * This file is part of oxine a free media player.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: menu_main.c 4020 2010-01-12 14:01:52Z mschwerin $
 *
 */
#include "config.h"

#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "disc.h"
#include "download.h"
#include "environment.h"
#include "extractor.h"
#include "heap.h"
#include "i18n.h"
#include "logger.h"
#include "utils.h"
#include "oxine.h"
#include "vdr.h"

#include "menu_base.h"
#include "menu_dialog.h"
#include "menu_extractor.h"
#include "menu_help.h"
#include "menu_main.h"
#include "menu_filelist.h"
#include "menu_playback.h"
#include "menu_playlist.h"
#include "menu_settings.h"

extern oxine_t *oxine;

#ifdef HAVE_HAL
static otk_widget_t *eject_list;
#endif
static otk_widget_t *shutdown_list;
static char *shutdown_commands[3] = { NULL, NULL, NULL };


void
play_dvb_cb (void *cb_p)
{
    please_wait (NULL);

    if (play_autoplay (_("Television"), "DVB")) {
        set_backto_menu (show_menu_main, NULL);
        set_playback_ended_menu (show_menu_main, NULL);
    }
    else {
        show_message_dialog (play_dvb_cb, NULL, show_menu_main, NULL,
                             DIALOG_RETRY_CANCEL, NULL,
                             _("Failed to start television playback!"));
    }
}


void
play_v4l_cb (void *cb_data)
{
    please_wait (NULL);

    if (play_autoplay (_("Television"), "V4L")) {
        set_backto_menu (show_menu_main, NULL);
        set_playback_ended_menu (show_menu_main, NULL);
    }
    else {
        show_message_dialog (play_v4l_cb, NULL, show_menu_main, NULL,
                             DIALOG_RETRY_CANCEL, NULL,
                             _("Failed to start television playback!"));
    }
}


#ifdef HAVE_VDR
void
play_vdr_cb (void *cb_data)
{
    please_wait (NULL);

    if (vdr_play ()) {
        set_backto_menu (show_menu_main, NULL);
        set_playback_ended_menu (show_menu_main, NULL);
    }
    else {
        show_message_dialog (play_vdr_cb, NULL, show_menu_main, NULL,
                             DIALOG_RETRY_CANCEL, NULL,
                             _("Failed to start television playback!"));
    }
}
#endif


void
play_tv_cb (void *cb_data)
{
    int type = config_get_number ("television.type");

    switch (type) {
    case 0:
        play_v4l_cb (cb_data);
        break;
    case 1:
        play_dvb_cb (cb_data);
        break;
#ifdef HAVE_VDR
    case 2:
        play_vdr_cb (cb_data);
        break;
#endif
    default:
        error (_("Unknown type '%d' for television."), type);
        break;
    }
}


void
play_mrl_cb (void *mrl_p)
{
    char *mrl = (char *) mrl_p;

    set_backto_menu (show_menu_main, NULL);
    set_playback_ended_menu (show_menu_main, NULL);

    please_wait (NULL);

    playlist_clear (oxine->ro_playlist);
    playlist_add (oxine->ro_playlist, NULL, mrl);
    playlist_play_first (oxine->ro_playlist);
}


void
play_cdda_cb (void *device_p)
{
    set_backto_menu (show_menu_main, NULL);
    set_playback_ended_menu (show_menu_main, NULL);

    please_wait (NULL);

    char *device = NULL;
    if (device_p) {
        device = ho_strdup ((char *) device_p);
    }
    else {
        device = ho_strdup (config_get_string ("media.audio_cd.device"));
    }

    /* If the drive is ejected we inject the drive. */
    if (drive_is_ejected (device)) {
        drive_inject (device);
    }

    volume_type_t type = volume_get_type (device);
    /* If we are certain that the disc in the specified audio CD drive is
     * indeed an audio CD we play this disc. */
    if (type == VOLUME_TYPE_AUDIO_CD) {
        char *mrl = mrl_build ("cdda", device, 0);
        filelist_t *filelist = filelist_get (FILELIST_TEMPORARY_MRL);
        filelist_lock (filelist);
        fileitem_t *fileitem = filelist_add_magic (filelist, NULL, mrl);
        if (fileitem) {
            playlist_clear (oxine->ro_playlist);
            playlist_add_fileitem (oxine->ro_playlist, fileitem);
            playlist_play_first (oxine->ro_playlist);
            filelist_remove (filelist, fileitem);
        }
        filelist_unlock (filelist);
        ho_free (mrl);
    }
    /* If there is no disc in the drive or if there is a disc in the drive but
     * we know what kind of disc this is we tell the user to insert a audio CD. */
    else if (type != VOLUME_TYPE_UNKNOWN) {
        show_message_dialog (play_cdda_cb, device,
                             backto_menu_cb, oxine,
                             DIALOG_OK_CANCEL, NULL,
                             _("Please insert an Audio CD!"));
    }
    /* If none of the above apply we just try our luck and hope xine-lib can
     * solve this problem for us. */
    else {
        play_autoplay (_("Audio CD"), "CD");
    }

    ho_free (device);
}


void
play_vcd_cb (void *device_p)
{
    set_backto_menu (show_menu_main, NULL);
    set_playback_ended_menu (show_menu_main, NULL);

    please_wait (NULL);

    char *device = NULL;
    if (device_p) {
        device = ho_strdup ((char *) device_p);
    }
    else {
        device = ho_strdup (config_get_string ("media.vcd.device"));
    }

    /* If the drive is ejected we inject the drive. */
    if (drive_is_ejected (device)) {
        drive_inject (device);
    }

    volume_type_t type = volume_get_type (device);
    /* If we are certain that the disc in the specified VCD drive is indeed a
     * VCD we play this disc. */
    if (type == VOLUME_TYPE_VIDEO_CD) {
        char *mrl = mrl_build ("vcd", device, 0);
        playlist_clear (oxine->ro_playlist);
        playlist_add (oxine->ro_playlist, _("Video CD"), mrl);
        playlist_play_first (oxine->ro_playlist);
        ho_free (mrl);
    }
    /* If there is no disc in the drive or if there is a disc in the drive but
     * we know what kind of disc this is we tell the user to insert a VCD. */
    else if (type != VOLUME_TYPE_UNKNOWN) {
        show_message_dialog (play_vcd_cb, device,
                             backto_menu_cb, oxine,
                             DIALOG_OK_CANCEL, NULL,
                             _("Please insert a Video CD!"));
    }
    /* If none of the above apply we just try our luck and hope xine-lib can
     * solve this problem for us. */
    else {
        play_autoplay (_("Video CD"), "VCD");
    }

    ho_free (device);
}


void
play_dvd_cb (void *device_p)
{
    set_backto_menu (show_menu_main, NULL);
    set_playback_ended_menu (show_menu_main, NULL);

    please_wait (NULL);

    char *device = NULL;
    if (device_p) {
        device = ho_strdup ((char *) device_p);
    }
    else {
        device = ho_strdup (config_get_string ("media.dvd.device"));
    }

    /* If the drive is ejected we inject the drive. */
    if (drive_is_ejected (device)) {
        drive_inject (device);
    }

    volume_type_t type = volume_get_type (device);
    /* If we are certain that the disc in the specified DVD drive is indeed a
     * DVD we play this disc. */
    if (type == VOLUME_TYPE_VIDEO_DVD) {
        char *mrl = mrl_build ("dvd", device, 0);
        playlist_clear (oxine->ro_playlist);
        playlist_add (oxine->ro_playlist, _("DVD"), mrl);
        playlist_play_first (oxine->ro_playlist);
        ho_free (mrl);
    }
    /* If there is no disc in the drive or if there is a disc in the drive but
     * we know what kind of disc this is we tell the user to insert a DVD. */
    else if (type != VOLUME_TYPE_UNKNOWN) {
        show_message_dialog (play_dvd_cb, device,
                             backto_menu_cb, oxine,
                             DIALOG_OK_CANCEL, NULL,
                             _("Please insert a DVD!"));
    }
    /* If none of the above apply we just try our luck and hope xine-lib can
     * solve this problem for us. */
    else {
        play_autoplay (_("DVD"), "DVD");
    }

    ho_free (device);
}


static void
shutdown_execute_cb (void *cb_data)
{
    int p = otk_list_get_pos (shutdown_list);

    config_set_number ("shutdown.last", p);

    oxine_event_t event;
    oxine_event_init (&event);
    event.type = OXINE_EVENT_KEY;
    if (p == 0) {
        event.source.key = OXINE_KEY_QUIT;
    }
    else if (p == 1) {
        event.source.key = OXINE_KEY_RESTART;
    }
    else {
        const char *cmd = shutdown_commands[p - 2];
        if (cmd && (strlen (cmd) > 0)) {
            oxine->shutdown_command = ho_strdup (cmd);
        }
        event.source.key = OXINE_KEY_QUIT;
    }
    odk_oxine_event_send (oxine->odk, &event);
}


void
shutdown_cb (void *cb_data)
{
    if (!config_get_bool ("shutdown.ask")) {
        odk_exit (oxine->odk);
        return;
    }

    int x = 0;
    int y = 0;
    int w = 0;
    int h = 0;

    int i = 0;
    const char *command;

    ho_free (shutdown_commands[0]);
    shutdown_commands[0] = NULL;
    ho_free (shutdown_commands[1]);
    shutdown_commands[1] = NULL;
    ho_free (shutdown_commands[2]);
    shutdown_commands[2] = NULL;

    char *message = NULL;
    if (false) {
        /* do nothing */
    }
#ifdef HAVE_VDR
    else if (vdr_is_recording ()) {
        message = ho_strdup_printf ("%s\n%s\n\n%s", _("Please note:"),
                                    _("The television is recording "
                                      "at the moment!"),
                                    _("How do you want to proceed?"));
    }
#endif
    else {
        message = ho_strdup_printf (_("How do you want to proceed?"));
    }
    show_dialog (shutdown_execute_cb, NULL, _("OK"),
                 backto_menu_cb, NULL, _("Cancel"),
                 true, "menu_quit.png",
                 OTK_ALIGN_LEFT | OTK_ALIGN_VCENTER, message);
    ho_free (message);

    get_dialog_widget_size (&x, &y, &w, &h);
    shutdown_list = otk_selector_new (oxine->otk, x, y, w, NULL, NULL);

    {
        otk_selectoritem_new (shutdown_list, _("Quit oxine"),
                              shutdown_execute_cb, oxine);
    }
    {
        otk_selectoritem_new (shutdown_list, _("Restart oxine"),
                              shutdown_execute_cb, oxine);
    }

    command = config_get_string ("shutdown.shutdown_command");
    if (command && (strlen (command) > 0)) {
        shutdown_commands[i++] = ho_strdup (command);
        otk_selectoritem_new (shutdown_list,
                              _("Shutdown the computer"),
                              shutdown_execute_cb, oxine);
    }

    command = config_get_string ("shutdown.standby_command");
    if (command && (strlen (command) > 0)) {
        shutdown_commands[i++] = ho_strdup (command);
        otk_selectoritem_new (shutdown_list,
                              _("Switch computer to standby"),
                              shutdown_execute_cb, oxine);
    }

    command = config_get_string ("shutdown.reboot_command");
    if (command && (strlen (command) > 0)) {
        shutdown_commands[i++] = ho_strdup (command);
        otk_selectoritem_new (shutdown_list,
                              _("Reboot the computer"),
                              shutdown_execute_cb, oxine);
    }

    int last_command = config_get_number ("shutdown.last");
    otk_list_set_pos (shutdown_list, last_command);
    otk_list_set_focus (shutdown_list, last_command);

    set_current_menu (shutdown_cb, NULL);
    show_user_interface (NULL);
}


static void
eject_execute_cb (void *device_p)
{
#ifdef HAVE_HAL
    filelist_lock (oxine->hal_volume_list);
#endif

    char *device = NULL;
    if (device_p) {
        device = (char *) device_p;
    }
    else {
#ifdef HAVE_HAL
        int p = otk_list_get_pos (eject_list);
        int i = 0;
        fileitem_t *item = filelist_first (oxine->hal_volume_list);
        while (item) {
            if (i == p) {
                device = item->hal.device;
                break;
            }
            item = filelist_next (oxine->hal_volume_list, item);
            i++;
        }
#endif
    }

    assert (device);

    /* If the tray of the drive is currently open (ejected) we try to close
     * the tray. */
    if (drive_is_ejected (device)) {
        otk_cb_t cb = backto_menu_cb;
        please_wait (NULL);

        if (!drive_inject (device)) {
            show_message_dialog (eject_execute_cb, device, cb, oxine,
                                 DIALOG_RETRY_CANCEL, NULL,
                                 _("Could not inject device!"));
        }
        else {
            cb (oxine);
        }
    }

    /* If the tray of the drive is currently closed we try to open the tray. */
    else {
        otk_cb_t cb = backto_menu_cb;
        please_wait (NULL);

        /* If we're playing from the disc we're about to eject, we stop the
         * stream first. */
        if (odk_current_is_playback_mode (oxine->odk)) {
            const char *current_mrl = odk_current_get_mrl (oxine->odk);
            char *mountpoint = volume_get_mountpoint (device);
            if ((mountpoint && (strstr (current_mrl, mountpoint) != NULL))
                || (odk_current_is_dvd (oxine->odk)
                    && (strstr (current_mrl, device) != NULL))
                || (odk_current_is_vcd (oxine->odk)
                    && (strstr (current_mrl, device) != NULL))
                || (odk_current_is_cdda (oxine->odk)
                    && (strstr (current_mrl, device) != NULL))) {
                cb = playback_ended_menu_cb;
                odk_stop_stream (oxine->odk);
            }
            ho_free (mountpoint);
        }

#ifdef HAVE_EXTRACTOR
        /* If we're extracting from the disc we're about to eject, we stop
         * the extractor. */
        if (extractor_is_running ()) {
            int pos = extractor_get_progress_all () - 1;
            filelist_t *list = extractor_get_list ();
            fileitem_t *item = filelist_get_by_pos (list, pos);

            if (strstr (item->mrl, device) != NULL) {
                extractor_stop ();
            }
        }
#endif

        /* Now we first try to unmount and then try to eject the device. */
        if (volume_is_mounted (device) && !volume_umount (device)) {
            show_message_dialog (eject_execute_cb, device, cb, oxine,
                                 DIALOG_RETRY_CANCEL, NULL,
                                 _("Could not unmount device!"));
        }
        else if (!drive_eject (device)) {
            show_message_dialog (eject_execute_cb, device, cb, oxine,
                                 DIALOG_RETRY_CANCEL, NULL,
                                 _("Could not eject device!"));
        }
        else {
            cb (oxine);
        }
    }

#ifdef HAVE_HAL
    filelist_unlock (oxine->hal_volume_list);
#endif
}


void
eject_cb (void *cb_data)
{
#ifdef HAVE_HAL
    if (filelist_length (oxine->hal_volume_list) == 0) {
        return;
    }
    else if (filelist_length (oxine->hal_volume_list) == 1) {
        if (!config_get_bool ("misc.eject.one_device.confirm")) {
            fileitem_t *item = filelist_first (oxine->hal_volume_list);
            eject_execute_cb (item->hal.device);
            return;
        }
    }

    int x = 0;
    int y = 0;
    int w = 0;
    int h = 0;

    show_dialog (eject_execute_cb, NULL, _("Eject"),
                 backto_menu_cb, NULL, _("Cancel"),
                 true, "menu_eject.png",
                 OTK_ALIGN_LEFT | OTK_ALIGN_VCENTER,
                 _("Select device to eject:"));

    get_dialog_widget_size (&x, &y, &w, &h);
    eject_list = otk_selector_new (oxine->otk, x, y, w, NULL, NULL);

    filelist_lock (oxine->hal_volume_list);
    fileitem_t *item = filelist_first (oxine->hal_volume_list);
    while (item) {
        otk_selectoritem_new (eject_list, item->title,
                              eject_execute_cb, item->hal.device);
        item = filelist_next (oxine->hal_volume_list, item);
    }
    filelist_unlock (oxine->hal_volume_list);

    otk_list_set_pos (eject_list, 0);
    otk_list_set_focus (eject_list, 0);

    set_current_menu (eject_cb, NULL);
    show_user_interface (NULL);
#else
    const char *device = config_get_string ("media.audio_cd.device");
    eject_execute_cb ((void *) device);
#endif
}


static void
menulist_focus_enter (void *p)
{
    menuitem_t *item = (menuitem_t *) p;

    if (item->background_mrl) {
        show_menu_background (item->background_mrl);
    }
    else if (item->cb == play_tv_cb) {
        show_menu_background ("menu_main_tv.png");
    }
    else if (item->cb == play_dvd_cb) {
        show_menu_background ("menu_main_dvd.png");
    }
    else if (item->cb == play_vcd_cb) {
        show_menu_background ("menu_main_vcd.png");
    }
    else if (item->cb == play_cdda_cb) {
        show_menu_background ("menu_main_cdda.png");
    }
    else if (item->cb == show_menu_music) {
        show_menu_background ("menu_main_music.png");
    }
    else if (item->cb == show_menu_video) {
        show_menu_background ("menu_main_video.png");
    }
#ifdef HAVE_IMAGE_PLAYBACK
    else if (item->cb == show_menu_image) {
        show_menu_background ("menu_main_image.png");
    }
#endif
    else if (item->cb == show_menu_playlist) {
        show_menu_background ("menu_main_playlist.png");
    }
#ifdef HAVE_EXTRACTOR
    else if (item->cb == show_menu_extractor) {
        show_menu_background ("menu_main_extractor.png");
    }
#endif
    else if (item->cb == show_menu_settings) {
        show_menu_background ("menu_main_settings.png");
    }
    else if (item->cb == show_menu_help) {
        show_menu_background ("menu_main_help.png");
    }
    else if (item->cb == shutdown_cb) {
        show_menu_background ("menu_main_quit.png");
    }
    else if (item->cb == show_menu_playback) {
        show_menu_background ("menu_main_playback.png");
    }
    else {
        show_menu_background ("menu_main.png");
    }
}


static void
menulist_fill (otk_widget_t * menu, menulist_t * menulist)
{
    menuitem_t *item = l_list_first (menulist->list);
    while (item) {
        if ((item->cb == show_menu_playback)
            && !odk_current_is_playback_mode (oxine->odk)) {
            /* If we're not in playback mode we hide any menu item pointing
             * to the playback menu. */
        }
        else if ((item->cb == show_menu_playback_settings)
                 && !odk_current_is_playback_mode (oxine->odk)) {
            /* If we're not in playback mode we hide any menu item pointing
             * to the playback settings menu. */
        }
#ifdef HAVE_EXTRACTOR
        else if ((item->cb == show_menu_extractor)
                 && !extractor_is_running ()) {
            /* If the extractor is currently not running we hide any menu
             * item pointing to the extractor menu. */
        }
#endif
        else {
            otk_widget_t *w = otk_menuitem_new (menu, _(item->title),
                                                item->cb, item->cb_data);
            otk_widget_set_focus_callbacks (w, menulist_focus_enter, item,
                                            NULL, NULL);
        }

        item = l_list_next (menulist->list, item);
    }
}


static otk_widget_t *
menulist_show (menulist_t * menulist)
{
    otk_widget_t *l;

    int max_w = odk_osd_get_width (oxine->odk);
    int max_h = odk_osd_get_height (oxine->odk);

    int w = (3 * max_w) / 2;
    if ((menulist->w != -1) && (menulist->w < max_w)) {
        w = menulist->w;
    }
    int x = (max_w - w) / 2;
    if ((menulist->x != -1) && (menulist->x + w < max_w)) {
        x = menulist->x;
    }
    int y = 100;
    if ((menulist->y != -1) && (menulist->y < max_h - 20)) {
        y = menulist->y;
    }
    int h = (max_h - y - 20);
    if ((menulist->h != -1) && (menulist->h + y < max_h - 20)) {
        h = menulist->h;
    }

    l = otk_menulist_new (oxine->otk, x, y, w, h, 45, 55, NULL);
    otk_widget_set_fontsize (l, 30);
    otk_widget_set_alignment (l, OTK_ALIGN_CENTER | OTK_ALIGN_VCENTER);
    menulist_fill (l, menulist);

    return l;
}


static void
main_menu_event_handler (void *cb_data, oxine_event_t * event)
{
    if (!is_current_menu (show_menu_main)
        && !is_current_menu (show_menu_sub)) {
        return;
    }
    if (event->type != OXINE_EVENT_KEY) {
        return;
    }

    switch (event->source.key) {
    case OXINE_KEY_BACK:
        if (is_current_menu (show_menu_sub)) {
            show_menu_main (oxine);
        }
        else if (!odk_current_is_logo_mode (oxine->odk)) {
            show_menu_playback (oxine);
        }
        event->source.key = OXINE_KEY_NULL;
        break;
    default:
        break;
    }
}


void
show_menu_sub (void *menulist_p)
{
    menulist_t *menulist = (menulist_t *) menulist_p;

    static menulist_t *sub_menu_menulist = NULL;
    static otk_widget_t *sub_menu_list = NULL;
    static otk_widget_t *sub_menu_window = NULL;

    int top_pos = 0;
    int cur_pos = 0;

    if (sub_menu_menulist != menulist) {
        otk_window_destroy (sub_menu_window);
        sub_menu_window = NULL;
    }

    if (sub_menu_window) {
        otk_set_current_window (oxine->otk, sub_menu_window);
        top_pos = otk_list_get_pos (sub_menu_list);
        cur_pos = otk_list_get_focus (sub_menu_list);
        int length = otk_list_get_length (sub_menu_list);
        otk_list_clear (sub_menu_list);
        menulist_fill (sub_menu_list, menulist);
        top_pos += otk_list_get_length (sub_menu_list) - length;
        cur_pos += otk_list_get_length (sub_menu_list) - length;
    }
    else {
        sub_menu_window = create_new_window (true, true);
        sub_menu_list = menulist_show (menulist);
    }

    if (odk_current_is_logo_mode (oxine->odk)) {
        set_playback_ended_menu (show_menu_sub, menulist);
    }

    set_backto_menu (show_menu_sub, menulist);
    set_current_menu (show_menu_sub, menulist);

    sub_menu_menulist = menulist;
    otk_list_set_focus (sub_menu_list, cur_pos);
    otk_list_set_pos (sub_menu_list, top_pos);

    show_user_interface (NULL);
}


void
show_menu_main (void *cb_data)
{
    static otk_widget_t *main_menu_list = NULL;
    static otk_widget_t *main_menu_window = NULL;

    int top_pos = 0;
    int cur_pos = 0;

    if (main_menu_window) {
        otk_set_current_window (oxine->otk, main_menu_window);
        int length = otk_list_get_length (main_menu_list);
        top_pos = otk_list_get_pos (main_menu_list);
        cur_pos = otk_list_get_focus (main_menu_list);
        otk_list_clear (main_menu_list);
        menulist_fill (main_menu_list, oxine->main_menu);
        top_pos += otk_list_get_length (main_menu_list) - length;
        cur_pos += otk_list_get_length (main_menu_list) - length;
    }
    else {
        odk_add_event_handler (oxine->odk, main_menu_event_handler, oxine,
                               EVENT_HANDLER_PRIORITY_NORMAL);
        main_menu_window = create_new_window (true, true);
        main_menu_list = menulist_show (oxine->main_menu);
    }

    if (odk_current_is_logo_mode (oxine->odk)) {
        set_playback_ended_menu (show_menu_main, NULL);
    }
    {
        set_backto_menu (show_menu_main, NULL);
        set_current_menu (show_menu_main, NULL);
    }

    /* We have to set the focus at this point to make sure the focus
     * callback is called which is necessary to load the background
     * image without which we cannot live!! */
    otk_list_set_focus (main_menu_list, cur_pos);
    otk_list_set_pos (main_menu_list, top_pos);

    show_user_interface (NULL);
}


void
init_menu_main ()
{
    char *menu_xml = NULL;

    menu_xml = ho_strdup_printf ("%s/mainmenu.xml", get_dir_oxine ());
    if (file_exists (menu_xml)) {
        if ((oxine->main_menu = menulist_new (menu_xml, NULL))) {
            goto out_free;
        }
    }
    ho_free (menu_xml);

    menu_xml = ho_strdup_printf ("%s/mainmenu.xml", get_dir_oxine_skin ());
    if (file_exists (menu_xml)) {
        if ((oxine->main_menu = menulist_new (menu_xml, NULL))) {
            goto out_free;
        }
    }
    ho_free (menu_xml);

    menu_xml = ho_strdup_printf ("%s/mainmenu.xml", OXINE_DATADIR);
    if (file_exists (menu_xml)) {
        if ((oxine->main_menu = menulist_new (menu_xml, NULL))) {
            goto out_free;
        }
    }
    ho_free (menu_xml);

    fatal (_("Could not load main menu!"));
    abort ();

  out_free:
    ho_free (menu_xml);
}


void
free_menu_main ()
{
    if (oxine->main_menu) {
        menulist_free (oxine->main_menu);
        oxine->main_menu = NULL;
    }

    ho_free (shutdown_commands[0]);
    ho_free (shutdown_commands[1]);
    ho_free (shutdown_commands[2]);
}
