
/*
 * Copyright (C) 2004-2005 Maximilian Schwerin
 *
 * This file is part of oxine a free media player.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: menu_playback.c 4032 2010-09-04 22:51:24Z mschwerin $
 *
 */
#include "config.h"

#include <assert.h>
#include <ctype.h>
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#ifdef HAVE_AEX
#include "aex.h"
#endif
#include "codeset.h"
#include "download.h"
#include "environment.h"
#include "heap.h"
#include "i18n.h"
#include "lang.h"
#include "logger.h"
#include "meta_info.h"
#include "oxine.h"
#include "menu_base.h"
#include "menu_filelist.h"
#include "menu_main.h"
#include "menu_playback.h"
#include "menu_playlist.h"
#include "utils.h"

extern oxine_t *oxine;

//#define SHOW_BORDERS    1

static enum {
    BUTTON_FOCUS_PLAY,
    BUTTON_FOCUS_VOLUME,
    BUTTON_FOCUS_NEXT,
    BUTTON_FOCUS_PREV,
    BUTTON_FOCUS_EJECT,
    BUTTON_FOCUS_ROTATE_LEFT,
    BUTTON_FOCUS_ROTATE_RIGHT,
    BUTTON_FOCUS_PLAYMODE,
} playback_menu_focus_buttons = BUTTON_FOCUS_PLAY;

static bool playback_menu_show_volume = false;

static bool playback_menu_show_elapsed = true;

/*
 * **************************************************************************
 * Callbacks of the stream parameters menu
 * **************************************************************************
 */
static int
volume_set (void *p, int value)
{
    return odk_set_stream_param (oxine->odk, ODK_PARAM_AO_VOLUME, value);
}


static int
volume_get (void *p)
{
    return odk_get_stream_param (oxine->odk, ODK_PARAM_AO_VOLUME);
}


static int
audio_channel_change (void *p, int pos)
{
    odk_set_stream_param (oxine->odk, ODK_PARAM_AUDIO_CHANNEL, pos - 1);

    return pos;
}


static int
audio_offset_set (void *p, int value)
{
    return odk_set_stream_param (oxine->odk, ODK_PARAM_AUDIO_OFFSET, value);
}


static int
audio_offset_get (void *p)
{
    return odk_get_stream_param (oxine->odk, ODK_PARAM_AUDIO_OFFSET);
}


static int
spu_channel_change (void *p, int pos)
{
    odk_set_stream_param (oxine->odk, ODK_PARAM_SPU_CHANNEL, pos - 1);

    return pos;
}


static int
spu_offset_set (void *p, int value)
{
    return odk_set_stream_param (oxine->odk, ODK_PARAM_SPU_OFFSET, value);
}


static int
spu_offset_get (void *p)
{
    return odk_get_stream_param (oxine->odk, ODK_PARAM_SPU_OFFSET);
}


static int
deinterlace_change (void *p, int pos)
{
    return odk_set_stream_param (oxine->odk, ODK_PARAM_VO_DEINTERLACE, pos);
}


static int
aspect_ratio_change (void *p, int pos)
{
    int value = ODK_VO_ASPECT_RATIO_AUTO;
    switch (pos) {
    case 1:
        value = ODK_VO_ASPECT_RATIO_SQUARE;
        break;
    case 2:
        value = ODK_VO_ASPECT_RATIO_4_3;
        break;
    case 3:
        value = ODK_VO_ASPECT_RATIO_ANAMORPHIC;
        break;
    case 4:
        value = ODK_VO_ASPECT_RATIO_DVB;
        break;
    default:
        break;
    }
    odk_set_stream_param (oxine->odk, ODK_PARAM_VO_ASPECT_RATIO, value);

    return pos;
}


static int
saturation_get (void *p)
{
    int value = odk_get_stream_param (oxine->odk, ODK_PARAM_VO_SATURATION);
    return round ((double) value / (double) ODK_VO_SATURATION_STEP);
}


static int
saturation_set (void *p, int value)
{
    odk_set_stream_param (oxine->odk, ODK_PARAM_VO_SATURATION,
                          value * ODK_VO_SATURATION_STEP);
    return saturation_get (p);
}


static int
hue_get (void *p)
{
    int value = odk_get_stream_param (oxine->odk, ODK_PARAM_VO_HUE);
    return value / ODK_VO_HUE_STEP;
}


static int
hue_set (void *p, int value)
{
    return odk_set_stream_param (oxine->odk, ODK_PARAM_VO_HUE,
                                 value * ODK_VO_HUE_STEP);
    return hue_get (p);
}


static int
contrast_get (void *p)
{
    int value = odk_get_stream_param (oxine->odk, ODK_PARAM_VO_CONTRAST);
    return round ((double) value / (double) ODK_VO_CONTRAST_STEP);
}


static int
contrast_set (void *p, int value)
{
    return odk_set_stream_param (oxine->odk, ODK_PARAM_VO_CONTRAST,
                                 value * ODK_VO_CONTRAST_STEP);
    return contrast_get (p);
}


static int
brightness_get (void *p)
{
    int value = odk_get_stream_param (oxine->odk, ODK_PARAM_VO_BRIGHTNESS);
    return round ((double) value / (double) ODK_VO_BRIGHTNESS_STEP);
}


static int
brightness_set (void *p, int value)
{
    return odk_set_stream_param (oxine->odk, ODK_PARAM_VO_BRIGHTNESS,
                                 value * ODK_VO_BRIGHTNESS_STEP);
    return brightness_get (p);
}


static int
zoom_get (void *p)
{
    return odk_get_stream_param (oxine->odk, ODK_PARAM_VO_ZOOM_X);
}


static int
zoom_set (void *p, int value)
{
    odk_set_stream_param (oxine->odk, ODK_PARAM_VO_ZOOM_Y, value);
    odk_set_stream_param (oxine->odk, ODK_PARAM_VO_ZOOM_X, value);

    return zoom_get (p);
}


static int
playmode_change (void *p, int pos)
{
    playlist_set_playmode (oxine->current_playlist, pos);

    return pos;
}


static void
set_volume_button (void *p, otk_widget_t * button)
{
    bool mute = odk_get_stream_param (oxine->odk, ODK_PARAM_AO_VOLUME_MUTE);

    if (mute) {
        otk_button_set_vector (button, OSD_VECTOR_VOLUME_MUTE);
    }
    else {
        otk_button_set_vector (button, OSD_VECTOR_VOLUME);
    }
}


static void
playmode_toggle (void *p)
{
    playback_menu_focus_buttons = BUTTON_FOCUS_PLAYMODE;

    oxine_event_t event;
    oxine_event_init (&event);
    event.type = OXINE_EVENT_KEY;
    event.source.key = OXINE_KEY_PLAYMODE;
    odk_oxine_event_send (oxine->odk, &event);
}


static void
time_show_toggle (void *p)
{
    playback_menu_show_elapsed = !playback_menu_show_elapsed;

    show_menu_current ();
}


static void
volume_mute_toggle (void *p)
{
    oxine_event_t event;
    oxine_event_init (&event);
    event.type = OXINE_EVENT_KEY;
    event.source.key = OXINE_KEY_VOLUME_MUTE;
    odk_oxine_event_send (oxine->odk, &event);
}


static bool
is_normal_title (void)
{
    return (!odk_current_is_dvd (oxine->odk)
            && !odk_current_is_vcd (oxine->odk)
            && !odk_current_is_v4l (oxine->odk)
            && !odk_current_is_dvb (oxine->odk));
}


#if 0
#ifdef HAVE_AEX
static int
aex_speakers_change (void *p, int num_current)
{
    if (num_current == 0) {
        xine_cfg_entry_t config;
        config_lookup_entry ("audio.driver.use_aex", &config);
        config.num_value = false;
        config_update_entry (&config);
    }
    else {
        aex_t *aex = aex_get_by_pos (num_current - 1);

        xine_cfg_entry_t config;
        config_lookup_entry ("audio.device.aex_host", &config);
        config.str_value = aex->address;
        config_update_entry (&config);
        config_lookup_entry ("audio.device.aex_port", &config);
        config.num_value = aex->port;
        config_update_entry (&config);
        config_lookup_entry ("audio.driver.use_aex", &config);
        config.num_value = true;
        config_update_entry (&config);
    }

    return num_current;
}
#endif
#endif


/*
 * **************************************************************************
 * GUI of the stream parameter menu
 * **************************************************************************
 */
#define BUTTON_Y                (odk_osd_get_height (oxine->odk) - 55)

#define LABEL_X                 (20)
#define LABEL_W                 ((odk_osd_get_width (oxine->odk) - 40) * 60 / 100)

#define EDIT_X                  (LABEL_X + LABEL_W + 10)
#define EDIT_W                  (odk_osd_get_width (oxine->odk) - 20 - EDIT_X)
#define EDIT_H                  (40)


static void
show_stream_param_slider (oxine_t * oxine, odk_stream_param_t param,
                          otk_int_get_cb_t get_cb,
                          otk_int_set_cb_t set_cb,
                          bool with_reset_button, int y,
                          char *value_format, double value_scale,
                          int def_value, int min_step,
                          int min_value, int max_value)
{
    const int al = OTK_ALIGN_LEFT | OTK_ALIGN_VCENTER;
    otk_widget_t *w;

    char *param_name = odk_get_stream_param_name (oxine->odk, param);
    w = otk_label_new (oxine->otk, LABEL_X, y, LABEL_W, EDIT_H,
                       al, param_name);
    ho_free (param_name);

    w = otk_slider_new (oxine->otk, EDIT_X, y, EDIT_W, EDIT_H, 8, 36, true,
                        EDIT_H, EDIT_H, OTK_SLIDER_HORIZONTAL, value_format,
                        value_scale, true, with_reset_button,
                        def_value, min_step, min_value, max_value,
                        get_cb, oxine, set_cb, oxine);
    otk_widget_set_updated (w, true);
}


static void
show_stream_param_selector (oxine_t * oxine, odk_stream_param_t param,
                            otk_int_set_cb_t cb,
                            int min_value, int max_value, int y)
{
    const int al = OTK_ALIGN_LEFT | OTK_ALIGN_VCENTER;
    otk_widget_t *w;

    char *param_name = odk_get_stream_param_name (oxine->odk, param);
    w = otk_label_new (oxine->otk, LABEL_X, y, LABEL_W, EDIT_H,
                       al, param_name);
    ho_free (param_name);

    w = otk_selector_new (oxine->otk, EDIT_X, y, EDIT_W, cb, oxine);
    int i = min_value;
    for (; i <= max_value; i++) {
        char *t = odk_get_stream_param_value_as_string (oxine->odk, param, i);
        otk_listentry_new (w, t, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
                           NULL, NULL, NULL, NULL, NULL);
        ho_free (t);
    }

    int position = odk_get_stream_param (oxine->odk, param) - min_value;
    otk_list_set_pos (w, position);
}


static void
stream_parameter_menu_event_handler (void *p, oxine_event_t * event)
{
    if (!is_current_menu (show_menu_playback_settings)
        && !is_current_menu (show_menu_playback_settings_two)) {
        return;
    }
    if (event->type != OXINE_EVENT_KEY) {
        return;
    }

    switch (event->source.key) {
    case OXINE_KEY_BACK:
        show_menu_main (NULL);
        event->source.key = OXINE_KEY_NULL;
        break;
    default:
        break;
    }
}


void
show_menu_playback_settings_two (void *p)
{
    /* Just to make sure that we never enter the playback menu when not in
     * playback mode just because the developer fucked up. */
    if (odk_current_is_logo_mode (oxine->odk)) {
        debug ("HIT THE PROGRAMMER ON THE HEAD!");
        show_menu_main (NULL);
        return;
    }
    else if (!odk_current_is_playback_mode (oxine->odk)) {
        return;
    }
    else if (is_current_menu (show_menu_playback_settings_two)) {
        return;
    }

    odk_add_event_handler (oxine->odk, stream_parameter_menu_event_handler,
                           oxine, EVENT_HANDLER_PRIORITY_NORMAL);

    create_new_window (false, true);
    int y = 100;

    int sc = odk_get_stream_info (oxine->odk,
                                  ODK_STREAM_INFO_MAX_SPU_CHANNEL);

    /* Subtitle Channel */
    if (sc > 1) {
        show_stream_param_selector (oxine, ODK_PARAM_SPU_CHANNEL,
                                    spu_channel_change,
                                    SPU_CHANNEL_AUTO, sc - 1, y);
        y += 50;
    }
    /* Subtitle Offset */
    if (sc > 0) {
        show_stream_param_slider (oxine, ODK_PARAM_SPU_OFFSET,
                                  spu_offset_get, spu_offset_set,
                                  true, y, "%4.2f ms", 1.0 / 90.0,
                                  ODK_SPU_OFFSET_DEF, ODK_SPU_OFFSET_STEP,
                                  ODK_SPU_OFFSET_MIN, ODK_SPU_OFFSET_MAX);
        y += 50;
    }
    /* Zoom */
    {
        show_stream_param_slider (oxine, ODK_PARAM_VO_ZOOM_X,
                                  zoom_get, zoom_set,
                                  true, y, "%2.0f %%", 1,
                                  ODK_VO_ZOOM_DEF, ODK_VO_ZOOM_STEP,
                                  ODK_VO_ZOOM_MIN, ODK_VO_ZOOM_MAX);
        y += 50;
    }
    /* Aspect Ratio */
    {
        show_stream_param_selector (oxine, ODK_PARAM_VO_ASPECT_RATIO,
                                    aspect_ratio_change,
                                    ODK_VO_ASPECT_RATIO_AUTO,
                                    ODK_VO_ASPECT_RATIO_DVB, y);
        y += 50;
    }
    /* Deinterlace */
    {
        show_stream_param_selector (oxine, ODK_PARAM_VO_DEINTERLACE,
                                    deinterlace_change,
                                    ODK_VO_DEINTERLACE_OFF,
                                    ODK_VO_DEINTERLACE_ON, y);
        y += 50;
    }

    otk_widget_t *b;
    b = otk_text_button_new (oxine->otk, 20, BUTTON_Y, 180, 35,
                             _("Next Page"),
                             show_menu_playback_settings, oxine);
    otk_widget_set_alignment (b, OTK_ALIGN_CENTER);
    otk_widget_set_focused (b, true);

    b = otk_text_button_new (oxine->otk, 210, BUTTON_Y, 180, 35,
                             _("Back"), show_menu_main, oxine);
    otk_widget_set_alignment (b, OTK_ALIGN_CENTER);

    set_backto_menu (show_menu_playback_settings_two, NULL);
    set_current_menu (show_menu_playback_settings_two, NULL);

    show_user_interface (NULL);
    show_menu_background (NULL);
}


void
show_menu_playback_settings (void *p)
{
    /* Just to make sure that we never enter the playback menu when not in
     * playback mode just because the developer fucked up. */
    if (odk_current_is_logo_mode (oxine->odk)) {
        debug ("HIT THE PROGRAMMER ON THE HEAD!");
        show_menu_main (NULL);
        return;
    }
    else if (!odk_current_is_playback_mode (oxine->odk)) {
        return;
    }
    else if (is_current_menu (show_menu_playback_settings)) {
        return;
    }

    odk_add_event_handler (oxine->odk, stream_parameter_menu_event_handler,
                           oxine, EVENT_HANDLER_PRIORITY_NORMAL);

    create_new_window (false, true);
    int y = 100;

    int ac = odk_get_stream_info (oxine->odk,
                                  ODK_STREAM_INFO_MAX_AUDIO_CHANNEL);

    /* Volume */
    otk_widget_t *w;
#if 0
#ifdef HAVE_AEX
    {
        xine_audio_port_t *audio_port = xine_open_audio_driver (oxine->xine,
                                                                "aex", NULL);

        bool use_aex = config_get_bool ("audio.driver.use_aex");
        const char *aex_host = config_get_string ("audio.device.aex_host");

        if (!audio_port) {
            show_message_dialog (show_menu_playback_settings, oxine,
                                 NULL, NULL, DIALOG_OK, NULL,
                                 _("Could not open AEX audio driver."));
            return;
        }

        otk_label_new (oxine->otk, LABEL_X, y + EDIT_H / 2, LABEL_W,
                       OTK_ALIGN_LEFT | OTK_ALIGN_VCENTER, _("Speakers"));
        w = otk_selector_new (oxine->otk, EDIT_X, y, EDIT_W,
                              aex_speakers_change, NULL);

        otk_selectoritem_new (w, _("local speakers"), NULL, NULL);
        int i = 1;
        int position = 0;
        aex_monitor_lock ();
        aex_t *aex = aex_first ();
        while (aex) {
            char *title = ho_strdup_printf ("%s (%s)", aex->title,
                                            _("Airport Express"));
            otk_selectoritem_new (w, title, NULL, NULL);
            ho_free (title);
            if (aex_host && (strcasecmp (aex->address, aex_host) == 0)) {
                position = i;
            }
            aex = aex_next (aex);
            i++;
        }
        aex_monitor_unlock ();

        otk_list_set_pos (w, use_aex ? position : 0);

        y += 60;
    }
#endif
#endif
    if (is_normal_title ()) {
        const int al = OTK_ALIGN_LEFT | OTK_ALIGN_VCENTER;

        w = otk_label_new (oxine->otk, LABEL_X, y, LABEL_W, EDIT_H,
                           al, _("Playback Mode"));
        w = otk_selector_new (oxine->otk, EDIT_X, y, EDIT_W,
                              playmode_change, oxine);

        otk_selectoritem_new (w, _("Normal Mode"), NULL, NULL);
        otk_selectoritem_new (w, _("Loop Mode"), NULL, NULL);
        otk_selectoritem_new (w, _("Shuffle Mode"), NULL, NULL);

        playback_mode_t mode =
            playlist_get_playmode (oxine->current_playlist);
        otk_list_set_pos (w, mode);
        y += 60;
    }
    if (odk_current_has_audio (oxine->odk)) {
        odk_stream_param_t param = ODK_PARAM_AO_VOLUME;
        const int al = OTK_ALIGN_LEFT | OTK_ALIGN_VCENTER;

        char *param_name = odk_get_stream_param_name (oxine->odk, param);
        w = otk_label_new (oxine->otk, LABEL_X, y, LABEL_W, EDIT_H,
                           al, param_name);
        ho_free (param_name);

        w = otk_slider_new (oxine->otk, EDIT_X, y, EDIT_W - EDIT_H - 2,
                            EDIT_H, 8, 36, true, EDIT_H, EDIT_H,
                            OTK_SLIDER_HORIZONTAL, "%2.0f %%", 1, true,
                            false, 50, 5, 0, 100, volume_get, oxine,
                            volume_set, oxine);
        otk_widget_set_updated (w, true);

        w = otk_vector_button_new (oxine->otk, EDIT_X + EDIT_W - EDIT_H, y,
                                   EDIT_H, EDIT_H, OSD_VECTOR_VOLUME,
                                   30, 30, volume_mute_toggle, oxine);
        set_volume_button (oxine, w);
        otk_button_set_update_cb (w, set_volume_button, oxine);
        otk_widget_set_updated (w, true);

        y += 50;
    }
    /* Audio Channel */
    if (odk_current_has_audio (oxine->odk) && (ac > 1)) {
        show_stream_param_selector (oxine, ODK_PARAM_AUDIO_CHANNEL,
                                    audio_channel_change,
                                    AUDIO_CHANNEL_AUTO, ac - 1, y);
        y += 50;
    }
    /* Audio-Video Offset */
    if (odk_current_has_audio (oxine->odk) && (ac > 0)
        && odk_current_has_video (oxine->odk)) {
        show_stream_param_slider (oxine, ODK_PARAM_AUDIO_OFFSET,
                                  audio_offset_get, audio_offset_set,
                                  true, y, "%4.2f ms", 1.0 / 90.0,
                                  ODK_AUDIO_OFFSET_DEF, ODK_AUDIO_OFFSET_STEP,
                                  ODK_AUDIO_OFFSET_MIN, ODK_AUDIO_OFFSET_MAX);
        y += 50;
    }
    if ((odk_current_has_audio (oxine->odk)) && (ac > 0)) {
        y += 10;
    }
    if (odk_current_has_video (oxine->odk)) {
        show_stream_param_slider (oxine, ODK_PARAM_VO_BRIGHTNESS,
                                  brightness_get, brightness_set,
                                  true, y, "%2.0f %%", 1, 50, 1, 0, 100);
        y += 50;
        show_stream_param_slider (oxine, ODK_PARAM_VO_CONTRAST,
                                  contrast_get, contrast_set,
                                  true, y, "%2.0f %%", 1, 50, 1, 0, 100);
        y += 50;
        show_stream_param_slider (oxine, ODK_PARAM_VO_SATURATION,
                                  saturation_get, saturation_set,
                                  true, y, "%2.0f %%", 1, 50, 1, 0, 100);
        y += 50;
        show_stream_param_slider (oxine, ODK_PARAM_VO_HUE,
                                  hue_get, hue_set,
                                  true, y, "%2.0f %%", 1, 50, 1, 0, 100);
        y += 50;
    }
    if (odk_current_has_video (oxine->odk)) {
        w = otk_text_button_new (oxine->otk, 20, BUTTON_Y, 180, 35,
                                 _("Next Page"),
                                 show_menu_playback_settings_two, oxine);
        otk_widget_set_alignment (w, OTK_ALIGN_CENTER);
        otk_widget_set_focused (w, true);
        w = otk_text_button_new (oxine->otk, 210, BUTTON_Y, 180, 35,
                                 _("Back"), show_menu_main, oxine);
        otk_widget_set_alignment (w, OTK_ALIGN_CENTER);
    }
    else {
        w = otk_text_button_new (oxine->otk, 20, BUTTON_Y, 180, 35,
                                 _("Back"), show_menu_main, oxine);
        otk_widget_set_alignment (w, OTK_ALIGN_CENTER);
        otk_widget_set_focused (w, true);
    }

    set_backto_menu (show_menu_playback_settings, NULL);
    set_current_menu (show_menu_playback_settings, NULL);

    show_user_interface (NULL);
    show_menu_background (NULL);
}

/*
 * **************************************************************************
 * Callbacks of the playing menu
 * **************************************************************************
 */

static void
hide_volume_slider_cb (void *p)
{
    playback_menu_focus_buttons = BUTTON_FOCUS_VOLUME;
    playback_menu_show_volume = false;
    show_menu_playback (p);
}


static void
show_volume_slider_cb (void *p)
{
    playback_menu_focus_buttons = BUTTON_FOCUS_VOLUME;
    playback_menu_show_volume = true;
    show_menu_playback (p);
}


static void
track_select_cb (void *entry_cb_data)
{
    playitem_t *item = (playitem_t *) entry_cb_data;

    playlist_play_item (oxine->current_playlist, item);

    /* Save current channel to config file. */
    if (odk_current_is_dvb (oxine->odk)) {
        xine_config_save (oxine->xine, get_file_config ());
    }
}


static void
playlist_prev_cb (void *p)
{
    playback_menu_focus_buttons = BUTTON_FOCUS_PREV;

    oxine_event_t event;
    oxine_event_init (&event);
    event.type = OXINE_EVENT_KEY;
    event.source.key = OXINE_KEY_PREV;

    odk_oxine_event_send (oxine->odk, &event);
}


static void
playlist_next_cb (void *p)
{
    playback_menu_focus_buttons = BUTTON_FOCUS_NEXT;

    oxine_event_t event;
    oxine_event_init (&event);
    event.type = OXINE_EVENT_KEY;
    event.source.key = OXINE_KEY_NEXT;

    odk_oxine_event_send (oxine->odk, &event);
}


static void
media_pplay_cb (void *p)
{
    playback_menu_focus_buttons = BUTTON_FOCUS_PLAY;

    oxine_event_t event;
    oxine_event_init (&event);
    event.type = OXINE_EVENT_KEY;
    event.source.key = OXINE_KEY_PPLAY;

    odk_oxine_event_send (oxine->odk, &event);
}


static void
media_stop_cb (void *p)
{
    oxine_event_t event;
    oxine_event_init (&event);
    event.type = OXINE_EVENT_KEY;
    event.source.key = OXINE_KEY_STOP;

    odk_oxine_event_send (oxine->odk, &event);
}


static void
media_eject_cb (void *p)
{
    playback_menu_focus_buttons = BUTTON_FOCUS_EJECT;

    oxine_event_t event;
    oxine_event_init (&event);
    event.type = OXINE_EVENT_KEY;
    event.source.key = OXINE_KEY_EJECT;

    odk_oxine_event_send (oxine->odk, &event);
}


static int
set_stream_progress (void *p, int position)
{
    odk_set_stream_param (oxine->odk, ODK_PARAM_POSITION, position);

    return position;
}


static int
get_stream_progress (void *p)
{
    return odk_get_stream_param (oxine->odk, ODK_PARAM_POSITION);
}


#ifdef HAVE_IMAGE_ROTATION
static void
rotate_right_cb (void *p)
{
    playback_menu_focus_buttons = BUTTON_FOCUS_ROTATE_RIGHT;

    oxine_event_t event;
    oxine_event_init (&event);
    event.type = OXINE_EVENT_KEY;
    event.source.key = OXINE_KEY_ROTATE_RIGHT;
    event.how = +90;

    odk_oxine_event_send (oxine->odk, &event);
}


static void
rotate_left_cb (void *p)
{
    playback_menu_focus_buttons = BUTTON_FOCUS_ROTATE_LEFT;

    oxine_event_t event;
    oxine_event_init (&event);
    event.type = OXINE_EVENT_KEY;
    event.source.key = OXINE_KEY_ROTATE_LEFT;
    event.how = -90;

    odk_oxine_event_send (oxine->odk, &event);
}
#endif /* HAVE_IMAGE_ROTATION */


/* 
 * **************************************************************************
 * GUI-Layout of the playing menu
 * **************************************************************************
 */
otk_widget_t *label_elapsed = NULL;
otk_widget_t *label_length = NULL;

static void
set_time_string (void *p, otk_widget_t * widget)
{
    int elapsed;
    int length;

    if (!odk_get_pos_length (oxine->odk, NULL, &elapsed, &length)) {
        return;
    }

    int el_d;
    int el_h;
    int el_m;
    int el_s;

    int le_d;
    int le_h;
    int le_m;
    int le_s;

    if ((length == 0) || (length < elapsed)) {
        length = elapsed;
    }
    if (!playback_menu_show_elapsed) {
        elapsed = length - elapsed;
    }

    elapsed /= 1000;
    length /= 1000;

    parse_time (elapsed, &el_d, &el_h, &el_m, &el_s);
    parse_time (length, &le_d, &le_h, &le_m, &le_s);

    char elapsed_str[64];
    if ((el_d > 0) || (le_d > 0)) {
        snprintf (elapsed_str, 64, "%d:%02d:%02d:%02d",
                  el_d, el_h, el_m, el_s);
    }
    else if ((el_h > 0) || (le_h > 0)) {
        snprintf (elapsed_str, 64, "%d:%02d:%02d", el_h, el_m, el_s);
    }
    else {
        snprintf (elapsed_str, 64, "%02d:%02d", el_m, el_s);
    }
    otk_label_set_text (label_elapsed, elapsed_str);

    char length_str[64];
    if (le_d > 0) {
        snprintf (length_str, 64, "%d:%02d:%02d:%02d",
                  le_d, le_h, le_m, le_s);
    }
    else if (le_h > 0) {
        snprintf (length_str, 64, "%d:%02d:%02d", le_h, le_m, le_s);
    }
    else {
        snprintf (length_str, 64, "%02d:%02d", le_m, le_s);
    }
    otk_label_set_text (label_length, length_str);
}


static void
set_pause_button (void *p, otk_widget_t * button)
{
    int speed = odk_get_stream_param (oxine->odk, ODK_PARAM_SPEED);

    if (speed == ODK_SPEED_PAUSE) {
        otk_button_set_text (button, ">");
    }
    else {
        otk_button_set_text (button, "<");
    }
}


static void
set_title_string (void *p, otk_widget_t * widget)
{
    if (!odk_current_is_dvd (oxine->odk)) {
        return;
    }

    uint32_t cc = odk_get_stream_info (oxine->odk,
                                       ODK_STREAM_INFO_DVD_TITLE_COUNT);
    uint32_t cn = odk_get_stream_info (oxine->odk,
                                       ODK_STREAM_INFO_DVD_TITLE_NUMBER);

    char *title = ho_strdup_printf (_("%d of %d"), cn, cc);
    otk_label_set_text (widget, title);
    ho_free (title);
}


static void
set_chapter_string (void *p, otk_widget_t * widget)
{
    if (!odk_current_is_dvd (oxine->odk)
        || !odk_current_has_chapters (oxine->odk)) {
        return;
    }

    uint32_t cc = odk_get_stream_info (oxine->odk,
                                       ODK_STREAM_INFO_DVD_CHAPTER_COUNT);
    uint32_t cn = odk_get_stream_info (oxine->odk,
                                       ODK_STREAM_INFO_DVD_CHAPTER_NUMBER);

    char *chapter = ho_strdup_printf (_("%d of %d"), cn, cc);
    otk_label_set_text (widget, chapter);
    ho_free (chapter);
}


static bool
show_playback_menu_thumbnail (int x, int y, int w, int h)
{
    bool has_thumbnail = false;

#ifdef HAVE_OSD_IMAGE
    if (odk_current_is_image (oxine->odk)) {
        return false;
    }
    if (odk_current_has_video (oxine->odk)) {
        return false;
    }

    char *mrl = NULL;
    const char *thumbnail_mrl = NULL;

    playlist_t *playlist = oxine->current_playlist;
    playlist_lock (playlist);
    playitem_t *playitem = playlist_get_current (playlist);
    thumbnail_mrl = playitem_get_cover_image (playitem);
    playlist_unlock (playlist);

    if (thumbnail_mrl) {
        if (is_downloadable (thumbnail_mrl)) {
            mrl = download_to_cache (thumbnail_mrl, NULL, false);
        }
        else {
            mrl = ho_strdup (thumbnail_mrl);
        }
        if (!file_exists (mrl)) {
            ho_free (mrl);
        }
    }

    if (mrl) {
        int x1 = x + (w / 2);
        int y1 = y + (h / 2);
        int a1 = (ODK_ALIGN_CENTER | ODK_ALIGN_VCENTER);

        otk_image_new (oxine->otk, x1, y1, w, h, mrl, a1, true);
        has_thumbnail = true;
    }

    ho_free (mrl);
#endif
#if SHOW_BORDERS
    if (has_thumbnail) {
        otk_border_new (oxine->otk, x, y, w, h);
    }
#endif

    return has_thumbnail;
}


static void
show_playback_menu_meta_info (int x, int y, int w, int h, int align)
{
    otk_widget_t *widget = NULL;

#if SHOW_BORDERS
    otk_border_new (oxine->otk, x, y, w, h);
#endif

    static int current_year = 0;
    if (!current_year) {
        time_t current = time (NULL);
        struct tm *brokentime = localtime (&current);
        current_year = brokentime->tm_year + 1900;
    }

    char *title = odk_get_meta_info (oxine->odk, META_INFO_TITLE);
    if (!title) {
        if (odk_current_get_title (oxine->odk)) {
            title = ho_strdup (odk_current_get_title (oxine->odk));
        }
        else {
            title = ho_strdup (_("Unknown Title"));
        }
    }
    char *artist = odk_get_meta_info (oxine->odk, META_INFO_ARTIST);
    char *album = odk_get_meta_info (oxine->odk, META_INFO_ALBUM);
    char *year = odk_get_meta_info (oxine->odk, META_INFO_YEAR);
    char *genre = odk_get_meta_info (oxine->odk, META_INFO_GENRE);
    char *track = odk_get_meta_info (oxine->odk, META_INFO_TRACK_NUMBER);
#ifdef XINE_META_INFO_COMPOSER
    char *composer = odk_get_meta_info (oxine->odk, META_INFO_COMPOSER);
#endif /* XINE_META_INFO_COMPOSER */
#ifdef HAVE_LIBEXIF
    const char *mrl = odk_current_get_mrl (oxine->odk);
    char *date = exif_info_get (EXIF_INFO_DATE, mrl);
    char *time = exif_info_get (EXIF_INFO_TIME, mrl);
#endif /* HAVE_LIBEXIF */

    {
        h -= 130;
    }
    if (album && (strlen (album) > 0)) {
        h -= 40;
    }
    if (artist && (strlen (artist) > 0)) {
        h -= 40;
    }
#ifdef XINE_META_INFO_COMPOSER
    if (composer && (strlen (composer) > 0)) {
        h -= 40;
    }
#endif /* XINE_META_INFO_COMPOSER */
    if (year && (atoi (year) > 0) && (atoi (year) <= current_year)) {
        h -= 40;
    }
    if (track && (strlen (track) > 0)) {
        h -= 40;
    }
    if (genre && (strlen (genre) > 0)) {
        h -= 40;
    }
#ifdef HAVE_LIBEXIF
    if (date && strlen (date) > 0) {
        h -= 40;
    }
    if (time && strlen (time) > 0) {
        h -= 40;
    }
#endif /* HAVE_LIBEXIF */
    if (odk_current_is_image (oxine->odk)) {
        h -= 40;                                                /* Image number */
        h += 50;                                                /* Length       */
        h += 40;                                                /* Elapsed      */
    }
    if (odk_current_is_dvd (oxine->odk)) {
        h -= 80;
    }

    int a1 = (OTK_ALIGN_RIGHT | OTK_ALIGN_VCENTER);
    int w1 = (align & OTK_ALIGN_CENTER) ? (w / 2) : (w / 3);
    int x1 = x;
    int y1 = y + (h / 2);

    int a2 = (OTK_ALIGN_LEFT | OTK_ALIGN_VCENTER);
    int w2 = w - w1 - 10;
    int x2 = x1 + w1 + 10;

    {
        otk_label_new (oxine->otk, x1, y1, w1, 40, a1, _("Title:"));
        widget = otk_label_new (oxine->otk, x2, y1, w2, 40, a2, title);
        otk_label_set_scrolling (widget, true);
        y1 += 40;
    }
    if (album && (strlen (album) > 0)) {
        otk_label_new (oxine->otk, x1, y1, w1, 40, a1, _("Album:"));
        otk_label_new (oxine->otk, x2, y1, w2, 40, a2, album);
        y1 += 40;
    }
    if (artist && (strlen (artist) > 0)) {
        otk_label_new (oxine->otk, x1, y1, w1, 40, a1, _("Artist:"));
        otk_label_new (oxine->otk, x2, y1, w2, 40, a2, artist);
        y1 += 40;
    }
#ifdef XINE_META_INFO_COMPOSER
    if (composer && (strlen (composer) > 0)) {
        otk_label_new (oxine->otk, x1, y1, w1, 40, a1, _("Composer:"));
        otk_label_new (oxine->otk, x2, y1, w2, 40, a2, composer);
        y1 += 40;
    }
#endif /* XINE_META_INFO_COMPOSER */
    if (genre && (strlen (genre) > 0)) {
        genre[0] = toupper (genre[0]);
        otk_label_new (oxine->otk, x1, y1, w1, 40, a1, _("Genre:"));
        otk_label_new (oxine->otk, x2, y1, w2, 40, a2, genre);
        y1 += 40;
    }
    if (year && (atoi (year) > 0) && (atoi (year) <= current_year)) {
        otk_label_new (oxine->otk, x1, y1, w1, 40, a1, _("Year:"));
        otk_label_new (oxine->otk, x2, y1, w2, 40, a2, year);
        y1 += 40;
    }
    if (track && (strlen (track) > 0)) {
        otk_label_new (oxine->otk, x1, y1, w1, 40, a1, _("Track:"));
        otk_label_new (oxine->otk, x2, y1, w2, 40, a2, track);
        y1 += 40;
    }
#ifdef HAVE_LIBEXIF
    if (date && strlen (date) > 0) {
        date[0] = toupper (date[0]);
        otk_label_new (oxine->otk, x1, y1, w1, 40, a1, _("Date:"));
        otk_label_new (oxine->otk, x2, y1, w2, 40, a2, date);
        y1 += 40;
    }
    if (time && strlen (time) > 0) {
        time[0] = toupper (time[0]);
        otk_label_new (oxine->otk, x1, y1, w1, 40, a1, _("Time:"));
        otk_label_new (oxine->otk, x2, y1, w2, 40, a2, time);
        y1 += 40;
    }
#endif /* HAVE_LIBEXIF */
    if (odk_current_is_image (oxine->odk)) {
        int len = playlist_length (oxine->current_playlist);
        int cur = playlist_get_current_pos (oxine->current_playlist);
        char *num = ho_strdup_printf (_("%d of %d"), cur + 1, len);
        otk_label_new (oxine->otk, x1, y1, w1, 40, a1, _("Image:"));
        otk_label_new (oxine->otk, x2, y1, w2, 40, a2, num);
        ho_free (num);
        y1 += 40;
    }
    if (odk_current_is_dvd (oxine->odk)) {
        otk_label_new (oxine->otk, x1, y1, w1, 40, a1, _("Title:"));
        widget = otk_label_new (oxine->otk, x2, y1, w2, 40, a2, "00/00");
        set_title_string (oxine, widget);
        otk_label_set_update_cb (widget, set_title_string, oxine);
        otk_widget_set_updated (widget, true);
        y1 += 40;
    }
    if (odk_current_is_dvd (oxine->odk)
        && odk_current_has_chapters (oxine->odk)) {
        otk_label_new (oxine->otk, x1, y1, w1, 40, a1, _("Chapter:"));
        widget = otk_label_new (oxine->otk, x2, y1, w2, 40, a2, "00/00");
        set_chapter_string (oxine, widget);
        otk_label_set_update_cb (widget, set_chapter_string, oxine);
        otk_widget_set_updated (widget, true);
        y1 += 40;
    }
    if (!odk_current_is_image (oxine->odk)) {
        y1 += 10;
        otk_label_new (oxine->otk, x1, y1, w1, 40, a1, _("Length:"));
        widget = otk_label_new (oxine->otk, x2, y1, w2, 40, a2, "00:00");
        label_length = widget;
        y1 += 40;
    }
    if (!odk_current_is_image (oxine->odk)) {
        const char *s = playback_menu_show_elapsed ? _("Elapsed:") : _("Remaining:");
        otk_label_new (oxine->otk, x1, y1, w1, 40, a1, s);
        widget = otk_label_new (oxine->otk, x2, y1, w2, 40, a2, "00:00");
        label_elapsed = widget;
        set_time_string (oxine, widget);
        otk_label_set_update_cb (widget, set_time_string, oxine);
        otk_widget_set_updated (widget, true);

        int x3 = x1 - w1;
        int y3 = y1;
        int w3 = w1 + w2 + (x2 - x1);
        int h3 = 30;
        otk_invisible_button_new (oxine->otk, x3, y3, w3, h3,
                                  time_show_toggle, NULL);

        y1 += 40;
    }

    ho_free (title);
    ho_free (artist);
    ho_free (album);
    ho_free (year);
    ho_free (genre);
    ho_free (track);
#ifdef HAVE_LIBEXIF
    ho_free (date);
    ho_free (time);
#endif /* HAVE_LIBEXIF */
}


static void
playback_menu_show_channel_list (int x, int y, int w, int h)
{
    if (!odk_current_is_dvb (oxine->odk)) {
        return;
    }

    otk_widget_t *l = otk_list_new (oxine->otk, x, y, w, h, 30, 33,
                                    true, true, OTK_LIST_SELECTION_NONE,
                                    oxine);

    playitem_t *channel = playlist_first (oxine->current_playlist);
    while (channel) {
        otk_listentry_new (l, channel->title, track_select_cb, channel,
                           NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
                           NULL, NULL);
        channel = playlist_next (oxine->current_playlist, channel);
    }

    int pos = playlist_get_current_pos (oxine->current_playlist);
    otk_list_set_selected (l, pos, 1);
    otk_list_set_pos (l, pos);
    otk_list_set_focus (l, pos);
}


static void
playback_menu_button_focus (otk_widget_t * b, int focused_button)
{
    if (playback_menu_focus_buttons == focused_button) {
        otk_widget_set_focused (b, true);
    }
}


static void
playback_menu_show_controls (int x, int y, int w, int h, bool with_volume)
{
    otk_widget_t *b;

#if SHOW_BORDERS
    otk_border_new (oxine->otk, x, y, w, h);
#endif

    assert (!odk_current_is_television (oxine->odk));

    int x0 = x;

    if (odk_current_is_image (oxine->odk)) {
        with_volume = false;
    }

    /* If we're playing a playlist and the playlist has more than one entry, a
     * lot of buttons have to be shown. */
    int length = playlist_length (oxine->current_playlist);
    bool playlist_with_more_than_one_entry = (length > 1);

    /* Show the play/ pause, stop, next/ prev buttons. */
    {
        b = otk_text_button_new (oxine->otk, x, y, h, h,
                                 ">", media_pplay_cb, oxine);
        otk_widget_set_fonttype (b, "cetus");
        otk_widget_set_fontsize (b, 20);
        otk_widget_set_alignment (b, OTK_ALIGN_CENTER);
        set_pause_button (oxine, b);
        otk_button_set_update_cb (b, set_pause_button, oxine);
        otk_widget_set_updated (b, true);
        playback_menu_button_focus (b, BUTTON_FOCUS_PLAY);
        x += 60;
    }

    y += 5;
    {
        b = otk_text_button_new (oxine->otk, x, y, 40, 40,
                                 "}", media_stop_cb, oxine);
        otk_widget_set_fonttype (b, "cetus");
        otk_widget_set_fontsize (b, 20);
        otk_widget_set_alignment (b, OTK_ALIGN_CENTER);
        x += 50;
    }

    if ((playlist_with_more_than_one_entry)
        || (odk_current_is_dvd (oxine->odk)
            && odk_current_has_chapters (oxine->odk))) {
        b = otk_text_button_new (oxine->otk, x, y, 40, 40, "[",
                                 playlist_prev_cb, oxine);
        otk_widget_set_fonttype (b, "cetus");
        otk_widget_set_fontsize (b, 20);
        otk_widget_set_alignment (b, OTK_ALIGN_CENTER);
        playback_menu_button_focus (b, BUTTON_FOCUS_PREV);
        x += 50;

        b = otk_text_button_new (oxine->otk, x, y, 40, 40,
                                 "]", playlist_next_cb, oxine);
        otk_widget_set_fonttype (b, "cetus");
        otk_widget_set_fontsize (b, 20);
        otk_widget_set_alignment (b, OTK_ALIGN_CENTER);
        playback_menu_button_focus (b, BUTTON_FOCUS_NEXT);
        x += 50;
    }

    if (is_normal_title ()) {
        const char *label = _("N");
        switch (oxine->current_playlist->playmode) {
        case PLAYBACK_MODE_NORMAL:
            break;
        case PLAYBACK_MODE_REPEAT:
            label = _("L");
            break;
        case PLAYBACK_MODE_RANDOM:
            label = _("S");
            break;
        }
        b = otk_text_button_new (oxine->otk, x, y, 40, 40,
                                 label, playmode_toggle, oxine);
        otk_widget_set_alignment (b, OTK_ALIGN_CENTER);
        playback_menu_button_focus (b, BUTTON_FOCUS_PLAYMODE);
        x += 50;
    }

    {
        b = otk_text_button_new (oxine->otk, x, y, 40, 40,
                                 "{", media_eject_cb, oxine);
        otk_widget_set_fonttype (b, "cetus");
        otk_widget_set_fontsize (b, 20);
        otk_widget_set_alignment (b, OTK_ALIGN_CENTER);
        playback_menu_button_focus (b, BUTTON_FOCUS_EJECT);
        x += 50;
    }

#ifdef HAVE_IMAGE_ROTATION
    if (odk_current_is_image (oxine->odk)) {
        b = otk_vector_button_new (oxine->otk, x, y, 40, 40,
                                   OSD_VECTOR_ARROW_LEFT, 20, 20,
                                   rotate_left_cb, oxine);
        playback_menu_button_focus (b, BUTTON_FOCUS_ROTATE_LEFT);
        x += 50;
        b = otk_vector_button_new (oxine->otk, x, y, 40, 40,
                                   OSD_VECTOR_ARROW_RIGHT, 20, 20,
                                   rotate_right_cb, oxine);
        playback_menu_button_focus (b, BUTTON_FOCUS_ROTATE_RIGHT);
        x += 50;
    }
#endif /* HAVE_IMAGE_ROTATION */

    /* Show the progress bar. */
    int w1 = w - x + x0;
    if (with_volume) {
        w1 -= 50;
    }
    {
        b = otk_slider_new (oxine->otk, x, y, w1, 40,
                            0, 20, false, 0, 0,
                            OTK_SLIDER_HORIZONTAL, NULL, 0,
                            false, false, 0, 1, 0, 100,
                            get_stream_progress, oxine,
                            set_stream_progress, oxine);
        otk_widget_set_updated (b, true);
    }

    /* Show the volume controls. */
    if (!with_volume) {
        return;
    }

    otk_cb_t volume_button_cb = NULL;
    if (playback_menu_show_volume) {
        volume_button_cb = hide_volume_slider_cb;
    }
    else {
        volume_button_cb = show_volume_slider_cb;
    }

    x += w1 + 10;
    {
        b = otk_vector_button_new (oxine->otk, x, y, 40, 40,
                                   OSD_VECTOR_VOLUME, 30, 30,
                                   volume_button_cb, oxine);
        set_volume_button (oxine, b);
        otk_button_set_update_cb (b, set_volume_button, oxine);
        otk_widget_set_updated (b, true);
        playback_menu_button_focus (b, BUTTON_FOCUS_VOLUME);
    }

    /* Show the volume slider. */
    if (playback_menu_show_volume) {
        b = otk_slider_new (oxine->otk, x, y - 280, 40, 270, 36, 8, true, 40,
                            30, OTK_SLIDER_VERTICAL, NULL, 0, true, false, 0,
                            5, 0, 100, volume_get, oxine, volume_set, oxine);
        otk_widget_set_updated (b, true);
    }
}


void
show_menu_playback (void *p)
{
    /* Just to make sure that we never enter the playback menu when not in
     * playback mode just because the developer fucked up. */
    if (odk_current_is_logo_mode (oxine->odk)) {
        debug ("HIT THE PROGRAMMER ON THE HEAD!");
        show_menu_main (NULL);
        return;
    }
    else if (!odk_current_is_playback_mode (oxine->odk)) {
        return;
    }

    create_new_window (false, true);

    if (odk_current_is_dvb (oxine->odk)) {
        int x = 20;
        int y = 100;
        int w = odk_osd_get_width (oxine->odk) - 40;
        int h = odk_osd_get_height (oxine->odk) - 120;

        if (oxine->is_sidebar_visible) {
            x += 200;
            w -= 200;
        }

        playback_menu_show_channel_list (x, y, w, h);
    }
    else {
        int x = 20;
        int y = 100;
        int w = odk_osd_get_width (oxine->odk) - 40;
        int h = odk_osd_get_height (oxine->odk) - 120;

        int x1 = x;
        int w1 = w / 3;
        bool has_thumbnail = show_playback_menu_thumbnail (x1, y, w1, h);

        int a2 = has_thumbnail ? OTK_ALIGN_LEFT : OTK_ALIGN_CENTER;
        int x2 = has_thumbnail ? x + w / 3 : x;
        int w2 = has_thumbnail ? w - w / 3 : w;
        show_playback_menu_meta_info (x2, y, w2, h, a2);

        int x3 = 20;
        int y3 = odk_osd_get_height (oxine->odk) - 65;
        int w3 = odk_osd_get_width (oxine->odk) - 40;
        int h3 = 50;
        playback_menu_show_controls (x3, y3, w3, h3, true);
    }

    set_backto_menu (show_menu_playback, NULL);
    set_current_menu (show_menu_playback, NULL);

    show_user_interface (NULL);
    show_menu_background (NULL);
}


void
show_playback_controls (void *p)
{
    /* Just to make sure that we never enter the playback menu when not in
     * playback mode just because the developer fucked up. */
    if (odk_current_is_logo_mode (oxine->odk)) {
        debug ("HIT THE PROGRAMMER ON THE HEAD!");
        show_menu_main (NULL);
        return;
    }
    else if (!odk_current_is_playback_mode (oxine->odk)) {
        return;
    }

    create_new_window (false, false);
    playback_menu_focus_buttons = BUTTON_FOCUS_PLAY;

    int x = 20;
    int y = odk_osd_get_height (oxine->odk) - 70;
    int w = odk_osd_get_width (oxine->odk) - 40;
    int h = 50;
    playback_menu_show_controls (x, y, w, h, false);

    oxine->playback_controls_are_visible = true;

    set_current_menu (show_playback_controls, NULL);

    show_user_interface (NULL);
    show_menu_background (NULL);
}


/*
 * **************************************************************************
 * GUI-Layout of the info menu
 * **************************************************************************
 */
void
show_playback_video_info (void *p)
{
    /* Just to make sure that we never enter the playback menu when not in
     * playback mode just because the developer fucked up. */
    if (odk_current_is_logo_mode (oxine->odk)) {
        debug ("HIT THE PROGRAMMER ON THE HEAD!");
        show_menu_main (NULL);
        return;
    }
    else if (!odk_current_is_playback_mode (oxine->odk)) {
        return;
    }

    create_new_window (false, false);

    int x = 20;
    int y = 100;
    int w = odk_osd_get_width (oxine->odk) - 40;
    int h = odk_osd_get_height (oxine->odk) - 120;
    int a = OTK_ALIGN_CENTER;

    show_playback_menu_meta_info (x, y, w, h, a);

    set_backto_menu (show_playback_video_info, NULL);
    set_current_menu (show_playback_video_info, NULL);
    otk_draw (oxine->otk);
}


void
show_playback_audio_info (void *p)
{
    /* Just to make sure that we never enter the playback menu when not in
     * playback mode just because the developer fucked up. */
    if (odk_current_is_logo_mode (oxine->odk)) {
        debug ("HIT THE PROGRAMMER ON THE HEAD!");
        show_menu_main (NULL);
        return;
    }
    else if (!odk_current_is_playback_mode (oxine->odk)) {
        return;
    }

    oxine->stream_parameter_is_visible = false;
    create_new_window (false, true);

    int x = 20;
    int y = 100;
    int w = odk_osd_get_width (oxine->odk) - 40;
    int h = odk_osd_get_height (oxine->odk) - 120;

    int x1 = x;
    int w1 = w / 3;
    bool has_thumbnail = show_playback_menu_thumbnail (x1, y, w1, h);

    int x2 = has_thumbnail ? x + w1 : x;
    int w2 = has_thumbnail ? w - w1 : w;
    int a2 = has_thumbnail ? OTK_ALIGN_LEFT : OTK_ALIGN_CENTER;
    show_playback_menu_meta_info (x2, y, w2, h, a2);

    set_backto_menu (show_playback_audio_info, NULL);
    set_current_menu (show_playback_audio_info, NULL);

    show_menu_background (NULL);
    otk_draw (oxine->otk);
}
