
/*
 * Copyright (C) 2004-2005 Maximilian Schwerin
 *
 * This file is part of oxine a free media player.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: menu_playlist.c 4026 2010-01-17 12:31:40Z mschwerin $
 *
 */
#include "config.h"

#include <assert.h>
#include <errno.h>
#include <math.h>
#include <stdio.h>
#include <unistd.h>

#include "disc.h"
#include "download.h"
#include "environment.h"
#include "filelist.h"
#include "heap.h"
#include "i18n.h"
#include "logger.h"
#include "oxine.h"
#include "playlist.h"
#include "utils.h"

#include "menu_base.h"
#include "menu_fileinfo.h"
#include "menu_filelist.h"
#include "menu_main.h"
#include "menu_playback.h"
#include "menu_playlist.h"
#include "menu_subtitle.h"

extern oxine_t *oxine;

static int thumbnail_job = -1;
static int fileinfo_job = -1;

static playitem_t *focused = NULL;

static otk_widget_t *button_hide = NULL;
static otk_widget_t *button_mode = NULL;
static otk_widget_t *button_move_up = NULL;
static otk_widget_t *button_move_down = NULL;

#define PLAYLIST_BUTTONS_NUM 8
static otk_widget_t *playlist_menu_buttons[PLAYLIST_BUTTONS_NUM];
static otk_widget_t *playlist_menu_window = NULL;
static otk_widget_t *playlist_menu_title = NULL;
static otk_widget_t *playlist_menu_list = NULL;
static otk_widget_t *playlist_menu_info = NULL;
static int playlist_menu_info_width = 0;
#ifdef HAVE_OSD_IMAGE
static otk_widget_t *playlist_menu_image = NULL;
#endif

static void playlist_menu_update_list (void);
static void playlist_menu_update_buttons (void);

static void playitem_focus_enter_cb (void *entry_cb_data);
static void playitem_focus_leave_cb (void *entry_cb_data);

static void playlist_activate_cb (void *entry_cb_data);
static void playlist_remove_cb (void *entry_cb_data);
static void playlist_select_cb (void *entry_cb_data);

static void playlist_save_cb (void *p);

static playitem_t *playlist_get_selected (void);


static void
playlist_menu_update_buttons (void)
{
    assert (is_current_menu (show_menu_playlist));

    bool enabled;
    playlist_t *playlist = oxine->rw_playlist;
    playitem_t *selected = playlist_get_selected ();

    playlist_lock (playlist);

    /* Move item up */
    enabled = (selected != NULL)
        && (selected != playlist_first (playlist));
    otk_widget_set_enabled (button_move_up, enabled);

    /* Move item down */
    enabled = (selected != NULL)
        && (playlist_next (playlist, selected) != NULL);
    otk_widget_set_enabled (button_move_down, enabled);

    if (!oxine->is_sidebar_visible) {
        goto unlock_out;
    }

    /* Save, clear and play playlist */
    enabled = (playlist_length (playlist) > 0);
    otk_widget_set_enabled (playlist_menu_buttons[0], enabled);
    otk_widget_set_enabled (playlist_menu_buttons[3], enabled);
    otk_widget_set_enabled (playlist_menu_buttons[4], enabled);

    /* Remove item, move item up or down */
    enabled = (otk_list_get_selected_count (playlist_menu_list) > 0);
    otk_widget_set_enabled (playlist_menu_buttons[2], enabled);

    /* Add/ remove subtitle */
    enabled = selected && mrl_is_video (selected->mrl);
    otk_widget_set_enabled (playlist_menu_buttons[5], enabled);
    if (selected && (selected->subtitle_mrl != NULL)) {
        otk_button_set_text (playlist_menu_buttons[5], _("Delete subtitle"));
    }
    else {
        otk_button_set_text (playlist_menu_buttons[5], _("Add subtitle"));
    }

    /* Current title */
    enabled = odk_current_is_playback_mode (oxine->odk);
    otk_widget_set_visible (playlist_menu_buttons[7], enabled);

    /* Focus the add button on empty playlist */
    if (playlist_length (playlist) == 0) {
        otk_widget_set_focused (playlist_menu_buttons[1], true);
    }

  unlock_out:
    playlist_unlock (playlist);
}


/**
 * This function is called whenever a listentry referencing a playitem is
 * called. We can then decrement the reference counter of the playitem.
 */
static void
playitem_decref_cb (void *p)
{
    playitem_t *item = (playitem_t *) p;
    ho_decref (item);
}


static void
playlist_menu_update_list (void)
{
    playlist_t *playlist = oxine->rw_playlist;

    assert (is_current_menu (show_menu_playlist));

    playlist_lock (playlist);

    ho_setref (focused, NULL);
    otk_list_clear (playlist_menu_list);
    otk_list_clear (playlist_menu_info);

    if (playlist_length (playlist) == 0) {
#ifdef HAVE_OSD_IMAGE
        /* We want to wait for 300ms before showing the thumbnail. This is a
         * good idea, because if the user scrolls down a list quickly, showing
         * the thumbnail at once would slow down scrolling. */
        oxine_event_t *thumbnail_event = oxine_event_new (oxine);
        thumbnail_event->type = OXINE_EVENT_THUMBNAIL_UPDATE;
        thumbnail_job = schedule_event_job (300, thumbnail_event);
#endif /* HAVE_OSD_IMAGE */

        /* We want to wait for 300ms before showing the fileinfo. This is a
         * good idea, because if the user scrolls down a list quickly, showing
         * the fileinfo at once would slow down scrolling. */
        oxine_event_t *fileinfo_event = oxine_event_new (oxine);
        fileinfo_event->type = OXINE_EVENT_FILEINFO_UPDATE;
        fileinfo_job = schedule_event_job (300, fileinfo_event);

        goto unlock_out;
    }

    playitem_t *item = playlist_first (playlist);
    while (item) {
        char *f = "<tr><td>%s%s</td><td width=30 "
            "font=cetus align=center>%s</td></tr>";
        char *title;
        char *subtitle;
        char *status = "";

        if (item->subtitle_mrl) {
            subtitle = ho_strdup_printf ("(+%s)", _("Subtitle"));
        }
        else {
            subtitle = ho_strdup ("");
        }

        if (item != playlist_get_current (playlist)) {
            status = "";
        }
        else {
            switch (playlist_get_playstatus (playlist)) {
            case PLAYBACK_STATUS_STOPPED:
                status = "}";
                break;
            case PLAYBACK_STATUS_PAUSED:
                status = "&lt;";
                break;
            case PLAYBACK_STATUS_PLAYING:
                status = "&gt;";
                break;
            }
        }

        {
            title = ho_strdup_printf (f, item->title, subtitle, status);
        }

        otk_widget_t *w = otk_listentry_new (playlist_menu_list, title,
                                             playlist_activate_cb, item,
                                             playlist_select_cb, item,
                                             playlist_remove_cb, item,
                                             NULL, NULL, NULL, NULL,
                                             NULL, NULL);
        ho_free (subtitle);
        ho_free (title);
        ho_incref (item);

        otk_widget_set_destruct_callback (w, playitem_decref_cb, item);
        otk_widget_set_focus_callbacks (w, playitem_focus_enter_cb, item,
                                        playitem_focus_leave_cb, item);

        item = playlist_next (playlist, item);
    }

    if (playlist->cur_position == -1) {
        playlist->cur_position = 0;
        playitem_t *c = playlist->current;
        playitem_t *i = playlist_first (playlist);
        while (i && c && (i != c)) {
            playlist->cur_position += 1;
            i = playlist_next (playlist, i);
        }
        int v = otk_list_get_visible_count (playlist_menu_list) / 2;
        playlist->top_position = playlist->cur_position - v;
    }

    otk_list_set_pos (playlist_menu_list, playlist->top_position);
    otk_list_set_focus (playlist_menu_list, playlist->cur_position);

  unlock_out:
    playlist_unlock (playlist);
}


#ifdef HAVE_OSD_IMAGE
static void
playlist_menu_thumbnail_preload (playitem_t * playitem)
{
    if (!playitem) {
        return;
    }

    char *mrl = NULL;
    const char *thumbnail_mrl = playitem_get_cover_image (playitem);

    if (thumbnail_mrl) {
        if (is_downloadable (thumbnail_mrl)) {
            mrl = download_to_cache (thumbnail_mrl, NULL, false);
        }
        else {
            mrl = ho_strdup (thumbnail_mrl);
        }

        if (!file_exists (mrl)) {
            ho_free (mrl);
        }
    }

    if (mrl) {
        int a = ODK_ALIGN_CENTER | ODK_ALIGN_VCENTER;
        int x = playlist_menu_image->x;
        int y = playlist_menu_image->y;
        int w = playlist_menu_image->w;
        int h = playlist_menu_image->h;

        odk_osd_preload_image (oxine->odk, mrl, x, y, w, h, false, 0, a);
    }

    ho_free (mrl);
}


static char *
playlist_menu_thumbnail_get (char *mrl)
{
    if (mrl && is_downloadable (mrl)) {
        char *new_mrl = download_to_cache (mrl, NULL, false);
        ho_free (mrl);
        mrl = new_mrl;
    }
    if (mrl && !file_exists (mrl)) {
        ho_free (mrl);
    }

    return mrl;
}


static void
playlist_menu_thumbnail_set ()
{
    playlist_t *playlist = oxine->rw_playlist;
    playitem_t *playitem = focused;

    assert (is_current_menu (show_menu_playlist));

    playlist_lock (playlist);

    char *mrl = NULL;
    if (playitem && ho_checkref (playitem)) {
        mrl = (char *) playitem_get_cover_image (playitem);
        mrl = mrl ? ho_strdup (mrl) : NULL;
        mrl = playlist_menu_thumbnail_get (mrl);
    }
    if (!mrl) {
        mrl = ho_strdup_printf ("%s/icons/logo_playlist.png",
                                get_dir_oxine_skin ());
        mrl = playlist_menu_thumbnail_get (mrl);
    }
    if (!mrl) {
        mrl = ho_strdup_printf ("%s/icons/logo_file_playlist.png",
                                get_dir_oxine_skin ());
        mrl = playlist_menu_thumbnail_get (mrl);
    }

    otk_image_set_mrl (playlist_menu_image, mrl);

    /* We try and preload the thumbnail images of the next and of the
     * previous item in the list. */
    if (playitem && ho_checkref (playitem)) {
        playlist_menu_thumbnail_preload (playlist_next (playlist, playitem));
        playlist_menu_thumbnail_preload (playlist_prev (playlist, playitem));
    }

    playlist_unlock (playlist);

    ho_free (mrl);
}
#endif /* HAVE_OSD_IMAGE */


static void
playlist_menu_update_fileinfo ()
{
    playlist_t *playlist = oxine->rw_playlist;
    playitem_t *playitem = focused;

    assert (is_current_menu (show_menu_playlist));
    otk_list_clear (playlist_menu_info);

    playlist_lock (playlist);
    if (ho_checkref (playlist)) {
        l_list_t *infolist = playlist_get_info_list (playlist);
        if (infolist) {
            char *infoitem = (char *) l_list_first (infolist);
            while (infoitem) {
                otk_infoitem_new (playlist_menu_info, infoitem);
                infoitem = (char *) l_list_next (infolist, infoitem);
            }
        }
    }
    if (playitem && ho_checkref (playitem)) {
        if (otk_list_get_length (playlist_menu_info)) {
            otk_list_separator_new (playlist_menu_info);
        }
        l_list_t *infolist = playitem_get_info_list (playitem);
        if (infolist) {
            char *infoitem = (char *) l_list_first (infolist);
            while (infoitem) {
                otk_infoitem_new (playlist_menu_info, infoitem);
                infoitem = (char *) l_list_next (infolist, infoitem);
            }
        }
    }
    playlist_unlock (playlist);
}


static void
playitem_focus_enter_cb (void *entry_cb_data)
{
    playitem_t *playitem = (playitem_t *) entry_cb_data;
    if (playitem && !ho_checkref (playitem)) {
        return;
    }

    /* We cancel any thumbnail job that might be running. */
    cancel_event_job (&thumbnail_job);
    /* We cancel any fileinfo job that might be running. */
    cancel_event_job (&fileinfo_job);

#ifdef HAVE_OSD_IMAGE
    /* We want to wait for 300ms before showing the thumbnail. This is a
     * good idea, because if the user scrolls down a list quickly, showing
     * the thumbnail at once would slow down scrolling. */
    oxine_event_t *thumbnail_event = oxine_event_new (oxine);
    thumbnail_event->type = OXINE_EVENT_THUMBNAIL_UPDATE;
    thumbnail_job = schedule_event_job (300, thumbnail_event);
#endif /* HAVE_OSD_IMAGE */

    /* We want to wait for 300ms before showing the fileinfo. This is a
     * good idea, because if the user scrolls down a list quickly, showing
     * the fileinfo at once would slow down scrolling. */
    oxine_event_t *fileinfo_event = oxine_event_new (oxine);
    fileinfo_event->type = OXINE_EVENT_FILEINFO_UPDATE;
    fileinfo_job = schedule_event_job (300, fileinfo_event);

    /* Set the focused playitem. */
    ho_setref (focused, playitem);
}


static void
playitem_focus_leave_cb (void *entry_cb_data)
{
}


static void
playlist_activate_cb (void *entry_cb_data)
{
    playitem_t *playitem = (playitem_t *) entry_cb_data;
    playlist_t *playlist = oxine->rw_playlist;
    if (playitem && !ho_checkref (playitem)) {
        return;
    }

    set_playback_ended_menu (show_menu_playlist, NULL);
    playlist_play_item (playlist, playitem);
}


static void
playlist_select_cb (void *entry_cb_data)
{
    playlist_menu_update_buttons ();
    show_user_interface (oxine);
}


static void
playlist_remove_cb (void *entry_cb_data)
{
    playlist_t *playlist = oxine->rw_playlist;
    playitem_t *playitem = (playitem_t *) entry_cb_data;
    if (playitem && !ho_checkref (playitem)) {
        return;
    }

    playlist_lock (playlist);

    /* If we're playing the title that we're about to remove we stop
     * playback before removing it. */
    bool playing = true;
    playing &= odk_current_is_playback_mode (oxine->odk);
    playing &= (oxine->current_playlist == playlist);
    playing &= (playitem == playlist_get_current (playlist));
    if (playing) {
        oxine_event_t event;
        oxine_event_init (&event);
        event.type = OXINE_EVENT_KEY;
        event.source.key = OXINE_KEY_STOP;
        odk_oxine_event_send (oxine->odk, &event);
    }

    playlist_remove (playlist, playitem);

    playlist->top_position = otk_list_get_pos (playlist_menu_list);
    playlist->cur_position = otk_list_get_focus (playlist_menu_list);

    playlist_unlock (playlist);

    playlist_menu_update_list ();
    playlist_menu_update_buttons ();
    show_user_interface (oxine);
}


static void
playlist_clear_cb (void *p)
{
    playlist_t *playlist = oxine->rw_playlist;

    playlist_lock (playlist);

    /* If we're playing a title from this playlist we stop playback before
     * clearing the list. */
    bool playing = true;
    playing &= odk_current_is_playback_mode (oxine->odk);
    playing &= (oxine->current_playlist == playlist);
    if (playing) {
        oxine_event_t event;
        oxine_event_init (&event);
        event.type = OXINE_EVENT_KEY;
        event.source.key = OXINE_KEY_STOP;
        odk_oxine_event_send (oxine->odk, &event);
    }

    playlist_clear (playlist);

    playlist_unlock (playlist);

    playlist_menu_update_list ();
    playlist_menu_update_buttons ();
    show_user_interface (oxine);
}


static playitem_t *
playlist_get_selected (void)
{
    return (playitem_t *) otk_list_get_selected_value (playlist_menu_list);
}


static int
playlist_get_selected_pos (void)
{
    return otk_list_get_selected_position (playlist_menu_list);
}


static void
playlist_remove_subtitle_cb (void *p)
{
    playlist_t *playlist = oxine->rw_playlist;
    playitem_t *selected = playlist_get_selected ();
    playlist_lock (playlist);
    if (selected && ho_checkref (selected) && selected->subtitle_mrl) {
        int pos = playlist_get_selected_pos ();
        int top = otk_list_get_pos (playlist_menu_list);

        playlist->top_position = top;

        ho_free (selected->subtitle_mrl);
        selected->subtitle_mrl = NULL;

        playlist_menu_update_list ();
        otk_list_set_selected (playlist_menu_list, pos, true);

        playlist_menu_update_buttons ();
        otk_widget_set_focused (playlist_menu_buttons[5], true);

        show_user_interface (oxine);
    }
    playlist_unlock (playlist);
}


static void
playlist_subtitle_cb (void *p)
{
    playlist_t *playlist = oxine->rw_playlist;
    playitem_t *selected = playlist_get_selected ();
    playlist_lock (playlist);
    if (selected && ho_checkref (selected) && selected->subtitle_mrl) {
        playlist_remove_subtitle_cb (oxine);
    }
    else {
        show_menu_subtitle (selected);
    }
    playlist_unlock (playlist);
}


static void
playlist_selected_remove_cb (void *p)
{
    playitem_t *selected = playlist_get_selected ();
    playlist_t *playlist = oxine->rw_playlist;

    if (!selected) {
        return;
    }
    else if (!odk_current_is_playback_mode (oxine->odk)) {
        /* ok to continue */
    }
    else if (selected != playlist_get_current (playlist)) {
        /* ok to continue */
    }
    else {
        return;
    }

    int position = playlist_get_selected_pos ();
    playlist->top_position = otk_list_get_pos (playlist_menu_list);

    playlist_remove (playlist, selected);

    playlist_menu_update_list ();
    if (position >= playlist_length (playlist)) {
        otk_list_set_selected (playlist_menu_list, position - 1, true);
    }
    else {
        otk_list_set_selected (playlist_menu_list, position, true);
    }

    playlist_menu_update_buttons ();
    otk_widget_set_focused (playlist_menu_buttons[2], true);

    show_user_interface (oxine);
}


static void
playlist_move_up_selected_cb (void *p)
{
    playitem_t *selected = playlist_get_selected ();
    playlist_t *playlist = oxine->rw_playlist;

    playlist_lock (playlist);
    if (selected) {
        int vis = otk_list_get_visible_count (playlist_menu_list);
        int top = otk_list_get_pos (playlist_menu_list);
        int pos = playlist_get_selected_pos () - 1;

        if (pos < top + (vis / 2)) {
            top -= 1;
        }

        playlist->top_position = top;
        playlist_move_up (playlist, selected);

        playlist_menu_update_list ();
        otk_list_set_selected (playlist_menu_list, pos, true);

        playlist_menu_update_buttons ();
        if (otk_widget_is_enabled (button_move_up)) {
            otk_widget_set_focused (button_move_up, true);
        }
        else if (otk_widget_is_enabled (button_move_down)) {
            otk_widget_set_focused (button_move_down, true);
        }

        show_user_interface (oxine);
    }
    playlist_unlock (playlist);
}


static void
playlist_move_down_selected_cb (void *p)
{
    playitem_t *selected = playlist_get_selected ();
    playlist_t *playlist = oxine->rw_playlist;

    playlist_lock (playlist);
    if (selected) {
        int vis = otk_list_get_visible_count (playlist_menu_list);
        int top = otk_list_get_pos (playlist_menu_list);
        int pos = playlist_get_selected_pos () + 1;

        if (pos > top + (vis / 2)) {
            top += 1;
        }

        playlist->top_position = top;
        playlist_move_down (playlist, selected);

        playlist_menu_update_list ();
        otk_list_set_selected (playlist_menu_list, pos, true);

        playlist_menu_update_buttons ();
        if (otk_widget_is_enabled (button_move_down)) {
            otk_widget_set_focused (button_move_down, true);
        }
        else if (otk_widget_is_enabled (button_move_up)) {
            otk_widget_set_focused (button_move_up, true);
        }

        show_user_interface (oxine);
    }
    playlist_unlock (playlist);
}


static void
playlist_play_cb (void *p)
{
    playlist_t *playlist = oxine->rw_playlist;
    playitem_t *playitem = playlist_get_current (playlist);
    if (!playitem) {
        set_playback_ended_menu (show_menu_playlist, NULL);
        playlist_play_first (playlist);
    }
    else {
        set_playback_ended_menu (show_menu_playlist, NULL);
        playlist_play_item (playlist, playitem);
    }
}


static void
playlist_save_now_cb (void *p, char *title)
{
    playlist_t *playlist = oxine->rw_playlist;

    if (!title || (strlen (title) == 0)) {
        ho_free (title);
        playlist_save_cb (NULL);
    }
    else {
        struct tm tc;
        time_t time_now = time (NULL);
        localtime_r (&time_now, &tc);
        char *time = ho_strdup_strftime ("%d.%m.%y, %H:%M", &tc);
        const char *dir = get_dir_playlists ();
        char *mrl = ho_strdup_printf ("%s/%s %s.oxp", dir, time, title);
        playlist_xml_save (playlist, mrl);
        ho_free (time);
        ho_free (mrl);
        ho_free (title);
        show_menu_playlist (NULL);
    }
}


static void
playlist_save_ok_cb (void *p)
{
    otk_widget_t *widget = (otk_widget_t *) p;
    char *title = otk_editbox_get_text (widget);

    playlist_save_now_cb (oxine, title);
}


static void
playlist_save_cb (void *p)
{
    playlist_t *playlist = oxine->rw_playlist;

    playlist_lock (playlist);
    if (playlist_length (playlist) > 0) {
        char *title = NULL;

        if (playlist_length (playlist) == 1) {
            playitem_t *item = playlist_first (playlist);
            title = ho_strdup (item->title);
        }
        if (title == NULL) {
            playitem_t *item = playlist_first (playlist);
            title = meta_info_get (META_INFO_ALBUM, item->mrl);
            while (item && title) {
                char *t = meta_info_get (META_INFO_ALBUM, item->mrl);
                if (!str_equals (title, t)) {
                    ho_free (title);
                }
                item = playlist_next (playlist, item);
            }
        }
        if (title == NULL) {
            playitem_t *item = playlist_first (playlist);
            title = meta_info_get (META_INFO_ARTIST, item->mrl);
            while (item && title) {
                char *t = meta_info_get (META_INFO_ARTIST, item->mrl);
                if (!str_equals (title, t)) {
                    ho_free (title);
                }
                item = playlist_next (playlist, item);
            }
        }

        otk_widget_t *widget = NULL;
        create_new_window (false, true);

        int a = (OTK_ALIGN_LEFT | OTK_ALIGN_BOTTOM);

        int w = 600;
        int h = 200;
        int x = (odk_osd_get_width (oxine->odk) - w) / 2;
        int y = (odk_osd_get_height (oxine->odk) - h) / 2;

        widget = otk_border_new (oxine->otk, x, y, w, h);

        x += 20;
        y += 20;
        w -= 40;
        widget = otk_label_new (oxine->otk, x, y, w, 30, a,
                                _("Please enter a name for the playlist:"));

        y += 50;
        widget = otk_editbox_new (oxine->otk, x, y, w, 40, 100, NULL, NULL,
                                  playlist_save_now_cb, widget);
        otk_widget_set_focused (widget, true);
        if (title) {
            otk_editbox_set_text (widget, title);
            ho_free (title);
        }

        y += 60;
        widget = otk_text_button_new (oxine->otk, x + 130, y, 160, 40,
                                      _("OK"), playlist_save_ok_cb, widget);
        otk_widget_set_alignment (widget, OTK_ALIGN_CENTER);

        widget = otk_text_button_new (oxine->otk, x + 310, y, 160, 40,
                                      _("Cancel"), show_menu_playlist, NULL);
        otk_widget_set_alignment (widget, OTK_ALIGN_CENTER);

        set_current_menu (playlist_save_cb, NULL);
        show_user_interface (NULL);
    }
    playlist_unlock (playlist);
}


/**
 * Toggles the sidebar making sure that previously selected items in the
 * list are selected after the change also.
 */
static void
playlist_toggle_sidebar_cb (void *p)
{
    int i = 0;
    int n = 0;
    int *selected = otk_list_get_selected_positions (playlist_menu_list, &n);
    bool focused = otk_widget_is_focused (button_hide);

    toggle_sidebar ();

    for (i = 0; i < n; i++) {
        otk_list_set_selected (playlist_menu_list, selected[i], true);
    }
    if (focused) {
        otk_widget_set_focused (button_hide, true);
    }

    playlist_menu_update_buttons ();
    show_user_interface (NULL);

    ho_free (selected);
}


static void
playlist_toggle_playmode_cb (void *p)
{
    int i = 0;
    int n = 0;
    int *selected = otk_list_get_selected_positions (playlist_menu_list, &n);
    bool focused = otk_widget_is_focused (button_mode);

    oxine_event_t event;
    oxine_event_init (&event);
    event.type = OXINE_EVENT_KEY;
    event.source.key = OXINE_KEY_PLAYMODE;
    odk_oxine_event_send (oxine->odk, &event);

    for (i = 0; i < n; i++) {
        otk_list_set_selected (playlist_menu_list, selected[i], true);
    }
    if (focused) {
        otk_widget_set_focused (button_mode, true);
    }

    playlist_menu_update_buttons ();
    show_user_interface (NULL);

    ho_free (selected);
}


static void
playlist_menu_event_handler (void *p, oxine_event_t * event)
{
    if (!is_current_menu (show_menu_playlist)) {
        /* We cancel any thumbnail job that might be running. */
        cancel_event_job (&thumbnail_job);
        /* We cancel any fileinfo job that might be running. */
        cancel_event_job (&fileinfo_job);
        return;
    }

    switch (event->type) {
#ifdef HAVE_OSD_IMAGE
    case OXINE_EVENT_THUMBNAIL_UPDATE:
        playlist_menu_thumbnail_set ();
        break;
#endif
    case OXINE_EVENT_FILEINFO_UPDATE:
        playlist_menu_update_fileinfo ();
        break;
    case OXINE_EVENT_KEY:
        switch (event->source.key) {
        case OXINE_KEY_BACK:
        case OXINE_KEY_INSERT:
            show_menu_filelist (NULL);
            event->source.key = OXINE_KEY_NULL;
            break;
        case OXINE_KEY_SAVE:
            playlist_save_cb (NULL);
            event->source.key = OXINE_KEY_NULL;
            break;
        case OXINE_KEY_SIDEBAR:
            playlist_toggle_sidebar_cb (NULL);
            event->source.key = OXINE_KEY_NULL;
            break;
        case OXINE_KEY_RED:
            event->source.key = OXINE_KEY_SAVE;
            odk_oxine_event_send (oxine->odk, event);
            event->source.key = OXINE_KEY_NULL;
            break;
        case OXINE_KEY_GREEN:
            event->source.key = OXINE_KEY_INSERT;
            odk_oxine_event_send (oxine->odk, event);
            event->source.key = OXINE_KEY_NULL;
            break;
        case OXINE_KEY_YELLOW:
            event->source.key = OXINE_KEY_REMOVE;
            odk_oxine_event_send (oxine->odk, event);
            event->source.key = OXINE_KEY_NULL;
            break;
        case OXINE_KEY_BLUE:
            if (!otk_list_has_focus (playlist_menu_list)) {
                /* do nothing */
            }
            else if (ho_checkref (focused)) {
                playlist_lock (focused->parent_list);
                show_menu_playitem_info (focused);
                playlist_unlock (focused->parent_list);
            }
            event->source.key = OXINE_KEY_NULL;
            break;
        default:
            break;
        }
        break;
    default:
        break;
    }
}


static void
playlist_menu_show_list (int x, int y, int w, int h)
{
    playlist_t *playlist = oxine->rw_playlist;
    int a = (OTK_ALIGN_LEFT | OTK_ALIGN_VCENTER);

    int x0 = x;
    int y0 = y;
    int w0 = w - 4 * 46 + 6;
    playlist_menu_title = otk_label_new (oxine->otk, x0, y0, w0, 35, a,
                                         _("Current playlist..."));
    otk_widget_set_fontsize (playlist_menu_title, 32);

    int x1 = x0 + w0;
    int y1 = y;
    const char *m = playlist_get_playmode_string (playlist, true);
    button_mode = otk_text_button_new (oxine->otk, x1, y1, 40, 35, m,
                                       playlist_toggle_playmode_cb, NULL);
    otk_widget_set_alignment (button_mode, OTK_ALIGN_CENTER);
    x1 += 46;
    button_move_up = otk_vector_button_new (oxine->otk, x1, y1, 40, 35,
                                            OSD_VECTOR_ARROW_UP, 20, 20,
                                            playlist_move_up_selected_cb,
                                            NULL);
    x1 += 46;
    button_move_down = otk_vector_button_new (oxine->otk, x1, y1, 40, 35,
                                              OSD_VECTOR_ARROW_DOWN, 20, 20,
                                              playlist_move_down_selected_cb,
                                              NULL);
    x1 += 46;
    button_hide = otk_vector_button_new (oxine->otk, x1, y1, 40, 35,
                                         OSD_VECTOR_SIDEBAR, 20, 20,
                                         playlist_toggle_sidebar_cb, NULL);

    int x2 = x;
    int y2 = y + 40;
    int w2 = w;
    int h2 = h - 40;
    playlist_menu_list = otk_list_new (oxine->otk, x2, y2, w2, h2, 30, 33,
                                       true, true, OTK_LIST_SELECTION_SINGLE,
                                       oxine);
}


static void
playlist_menu_show_info (int x, int y, int w, int h)
{
    int x1 = x;
    int w1 = w;

#ifdef HAVE_OSD_IMAGE
    int x3 = x;
    int y3 = y;
    int w3 = h;
    int h3 = h;
    otk_border_new (oxine->otk, x3, y3, w3, h3);

    int a2 = (ODK_ALIGN_CENTER | ODK_ALIGN_VCENTER);
    int w2 = h - 10;
    int h2 = h - 10;
    int x2 = x + 5 + (w2 / 2);
    int y2 = y + 5 + (h2 / 2);

    playlist_menu_image = otk_image_new (oxine->otk, x2, y2, w2,
                                         h2, NULL, a2, false);

    x1 += w3 + 5;
    w1 -= w3 + 5;
#endif /* HAVE_OSD_IMAGE */

    playlist_menu_info = otk_infolist_new (oxine->otk, x1, y, w1, h, 26, 26);
    playlist_menu_info_width = w1 - 40;
}


static void
playlist_menu_show_sidebar (int x, int y, int w, int h)
{
    otk_widget_t *b = NULL;

    int x1 = x;
    int y1 = y;

    b = otk_text_button_new (oxine->otk, x1, y1, w, 35, _("Play"),
                             playlist_play_cb, oxine);
    playlist_menu_buttons[0] = b;
    y1 += 40;

    b = otk_text_button_new (oxine->otk, x1, y1, w, 35, _("Add"),
                             show_menu_filelist, oxine);
    playlist_menu_buttons[1] = b;
    y1 += 40;

    b = otk_text_button_new (oxine->otk, x1, y1, w, 35, _("Remove"),
                             playlist_selected_remove_cb, oxine);
    playlist_menu_buttons[2] = b;
    y1 += 40;

    b = otk_text_button_new (oxine->otk, x1, y1, w, 35, _("Clear"),
                             playlist_clear_cb, oxine);
    playlist_menu_buttons[3] = b;
    y1 += 40;

    b = otk_text_button_new (oxine->otk, x1, y1, w, 35, _("Save"),
                             playlist_save_cb, oxine);
    playlist_menu_buttons[4] = b;
    y1 += 50;

    b = otk_text_button_new (oxine->otk, x1, y1, w, 35, _("Add subtitle"),
                             playlist_subtitle_cb, oxine);
    playlist_menu_buttons[5] = b;
    y1 += 50;

    b = otk_text_button_new (oxine->otk, x1, y1, w, 35, _("Mainmenu"),
                             show_menu_main, oxine);
    playlist_menu_buttons[6] = b;
    y1 += 40;

    b = otk_text_button_new (oxine->otk, x1, y1, w, 35, _("Current title"),
                             show_menu_playback, oxine);
    playlist_menu_buttons[7] = b;
    y1 += 40;
}


void
show_menu_playlist (void *p)
{
    playlist_t *playlist = oxine->rw_playlist;

    /* We cancel any thumbnail job that might be running. */
    cancel_event_job (&thumbnail_job);
    /* We cancel any fileinfo job that might be running. */
    cancel_event_job (&fileinfo_job);

    if (playlist_menu_window) {
        otk_set_current_window (oxine->otk, playlist_menu_window);

        playlist_lock (playlist);
        playlist->top_position = otk_list_get_pos (playlist_menu_list);
        playlist->cur_position = otk_list_get_focus (playlist_menu_list);
        playlist_unlock (playlist);

        otk_window_destroy (playlist_menu_window);
    }

    odk_add_event_handler (oxine->odk, playlist_menu_event_handler,
                           oxine, EVENT_HANDLER_PRIORITY_NORMAL);

    playlist_menu_window = create_new_window (true, true);

    int x = 20;
    int y = 100;
    int w = odk_osd_get_width (oxine->odk) - 40;
    int h = odk_osd_get_height (oxine->odk) - 120;

    int x1 = x;
    int y1 = y;
    int w1 = w;
    int h1 = h - 190;

    int x2 = x;
    int y2 = y1 + h1 + 5;
    int w2 = w;
    int h2 = h - h1 - 5;

    int x3 = x;
    int y3 = y;
    int w3 = h2;
    int h3 = h1;

    if (oxine->is_sidebar_visible) {
        playlist_menu_show_sidebar (x3, y3, w3, h3);
        x1 += w3 + 5;
        w1 -= w3 + 5;
    }
    {
        playlist_menu_show_list (x1, y1, w1, h1);
        playlist_menu_show_info (x2, y2, w2, h2);
    }

    set_backto_menu (show_menu_playlist, NULL);
    set_current_menu (show_menu_playlist, NULL);

    playlist_menu_update_list ();
    playlist_menu_update_buttons ();

    show_user_interface (NULL);
    show_menu_background ("menu_playlist.png");
}


void
free_menu_playlist (void)
{
    ho_setref (focused, NULL);
}
