
/*
 * Copyright (C) 2004-2005 Maximilian Schwerin
 *
 * This file is part of oxine a free media player.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: meta_info.c 3922 2009-05-17 11:31:32Z mschwerin $
 *
 */
#include "config.h"

#include <xine.h>
#include <math.h>

#include "codeset.h"
#include "heap.h"
#include "i18n.h"
#include "logger.h"
#include "meta_info.h"
#include "mutex.h"
#include "oxine.h"
#include "utils.h"

extern oxine_t *oxine;

pthread_mutex_t meta_info_mutex;
pthread_mutexattr_t meta_info_mutex_attr;

static xine_stream_t *stream = NULL;

static struct {
    char *mrl;
    char *title;
    char *artist;
    char *genre;
    char *album;
    char *year;
    char *track;
    char *discid;
#ifdef XINE_META_INFO_COMPOSER
    char *composer;
#endif
} meta_info_cache;


char *
meta_info_get_from_stream (meta_info_t type, xine_stream_t * stream,
                           const char *mrl)
{
    char *meta = NULL;
    const char *xmeta = xine_get_meta_info (stream, type);

    if (xmeta) {
        /* Strip whitespace from the beginning of the string. */
        while ((xmeta[0] == ' ')
               || (xmeta[0] == '\t')
               || (xmeta[0] == '\n')
               || (xmeta[0] == '\r')) {
            xmeta = (char *) (xmeta + 1);
        }

        /* Convert '_' to ' '. */
        for (meta = (char *) xmeta; meta[0] != '\0'; meta++) {
            if (meta[0] == '_') {
                meta[0] = ' ';
            }
        }

        /* Convert the string to our local codeset. */
        recode_t *xr = recode_init ("UTF-8", USE_SYSTEM_ENCODING);
        meta = recode (xr, xmeta);
        recode_done (xr);
    }

    if (meta) {
        char *tmeta = ho_strdup (trim_whitespace (meta));
        ho_free (meta);
        meta = tmeta;

        /* Hack to destinguish between the language Latin and the genre Latin. */
        if ((type == META_INFO_GENRE) && (strcasecmp (meta, "Latin") == 0)) {
            ho_free (meta);
            meta = ho_strdup ("Latino");
        }
        /* Translate genres. */
        if (strlen (meta) && (type == META_INFO_GENRE)) {
            char *temp = ho_strdup (_(meta));
            ho_free (meta);
            meta = temp;
        }

        if ((type == META_INFO_TRACK_NUMBER) && (index (meta, '/'))) {
            char *i = index (meta, '/');
            int a = atoi (meta);
            int b = atoi (&i[1]);
            ho_free (meta);
            meta = ho_strdup_printf (_("%d of %d"), a, b);
        }
    }
    if (meta && (strlen (meta) == 0)) {
        ho_free (meta);
    }
    if (!meta && mrl_has_protocol (mrl, "cdda")) {
        int track;
        switch (type) {
        case META_INFO_TITLE:
            {
                meta = ho_strdup (_("Unknown Title"));
            }
            break;
        case META_INFO_ALBUM:
            {
                meta = ho_strdup (_("Unknown Album"));
            }
            break;
        case META_INFO_TRACK_NUMBER:
            if ((track = mrl_get_number (mrl)) > 0) {
                meta = ho_strdup_printf ("%d", track);
            }
            break;
        default:
            break;
        }
    }

    return meta;
}


static void
meta_info_free_cache ()
{
    ho_free (meta_info_cache.mrl);
    ho_free (meta_info_cache.title);
    ho_free (meta_info_cache.artist);
    ho_free (meta_info_cache.genre);
    ho_free (meta_info_cache.album);
    ho_free (meta_info_cache.year);
    ho_free (meta_info_cache.track);
    ho_free (meta_info_cache.discid);
#ifdef XINE_META_INFO_COMPOSER
    ho_free (meta_info_cache.composer);
#endif
}


static void
meta_info_fill_cache (xine_stream_t * stream, const char *mrl)
{
    meta_info_free_cache ();
    meta_info_cache.mrl = ho_strdup (mrl);
    meta_info_cache.album = meta_info_get_from_stream (META_INFO_ALBUM,
                                                       stream, mrl);
    meta_info_cache.title = meta_info_get_from_stream (META_INFO_TITLE,
                                                       stream, mrl);
    meta_info_cache.track = meta_info_get_from_stream (META_INFO_TRACK_NUMBER,
                                                       stream, mrl);
    meta_info_cache.discid = meta_info_get_from_stream (META_INFO_DISCID,
                                                        stream, mrl);
    if (!meta_info_cache.album && !meta_info_cache.title) {
        return;
    }
    meta_info_cache.artist = meta_info_get_from_stream (META_INFO_ARTIST,
                                                        stream, mrl);
    meta_info_cache.genre = meta_info_get_from_stream (META_INFO_GENRE,
                                                       stream, mrl);
    meta_info_cache.year = meta_info_get_from_stream (META_INFO_YEAR, stream,
                                                      mrl);
#ifdef XINE_META_INFO_COMPOSER
    meta_info_cache.composer = meta_info_get_from_stream (META_INFO_COMPOSER,
                                                          stream, mrl);
#endif
}


static char *
meta_info_get_from_cache (meta_info_t type, const char *mrl)
{
    char *meta = NULL;

    switch (type) {
    case META_INFO_TITLE:
        meta = meta_info_cache.title;
        break;
    case META_INFO_ARTIST:
        meta = meta_info_cache.artist;
        break;
    case META_INFO_GENRE:
        meta = meta_info_cache.genre;
        break;
    case META_INFO_ALBUM:
        meta = meta_info_cache.album;
        break;
    case META_INFO_YEAR:
        meta = meta_info_cache.year;
        break;
    case META_INFO_TRACK_NUMBER:
        meta = meta_info_cache.track;
        break;
    case META_INFO_DISCID:
        meta = meta_info_cache.discid;
        break;
#ifdef XINE_META_INFO_COMPOSER
    case META_INFO_COMPOSER:
        meta = meta_info_cache.composer;
        break;
#endif
    default:
        meta = NULL;
        break;
    }
    if (meta && strlen (meta)) {
        return ho_strdup (meta);
    }
    else {
        return NULL;
    }
}


char *
meta_info_get (meta_info_t type, const char *mrl)
{
    if (!mrl) {
        return NULL;
    }

    char *meta = NULL;
    char *encoded_mrl = url_encode (mrl);
    mutex_lock (&meta_info_mutex);
    if (mrl_is_television (mrl)) {
        /* do nothing */
    }
#ifdef HAVE_IMAGE_PLAYBACK
    else if (mrl_is_image (mrl)) {
        /* do nothing */
    }
#endif
    else if (mrls_are_equal (meta_info_cache.mrl, mrl)) {
        meta = meta_info_get_from_cache (type, mrl);
    }
    else if (xine_open (stream, mrl) == 1) {
        meta_info_fill_cache (stream, mrl);
        meta = meta_info_get_from_cache (type, mrl);
        xine_close (stream);
    }
    else if (xine_open (stream, encoded_mrl) == 1) {
        meta_info_fill_cache (stream, mrl);
        meta = meta_info_get_from_cache (type, mrl);
        xine_close (stream);
    }
    else {
        odk_log_stream_error (stream, __FILE__, __LINE__,
                              _("Could not open '%s': %s!"), mrl);
    }
    mutex_unlock (&meta_info_mutex);
    ho_free (encoded_mrl);

    return meta;
}


static int
meta_info_get_playback_length_from_stream (xine_stream_t * stream)
{
    int pos;
    int time;
    int length;
    if (xine_get_pos_length (stream, &pos, &time, &length) == 1) {
        return (length / 1000);
    }
    else {
        return 0;
    }
}


int
meta_info_get_playback_length (const char *mrl)
{
    if (!mrl) {
        return 0;
    }

    int playback_length = 0;
    char *encoded_mrl = url_encode (mrl);
    mutex_lock (&meta_info_mutex);
    if (mrl_is_television (mrl)) {
        /* Nothing to do, we return a playback length of zero. */
    }
#ifdef HAVE_IMAGE_PLAYBACK
    else if (mrl_is_image (mrl)) {
        if (config_get_bool ("image.slideshow.enable")) {
            playback_length = config_get_number ("image.slideshow.delay");
        }
    }
#endif
    else if (xine_open (stream, mrl) == 1) {
        playback_length = meta_info_get_playback_length_from_stream (stream);
        xine_close (stream);
    }
    else if (xine_open (stream, encoded_mrl) == 1) {
        playback_length = meta_info_get_playback_length_from_stream (stream);
        xine_close (stream);
    }
    else {
        odk_log_stream_error (stream, __FILE__, __LINE__,
                              _("Could not open '%s': %s!"), mrl);
    }
    mutex_unlock (&meta_info_mutex);
    ho_free (encoded_mrl);

    return playback_length;
}


char *
meta_info_get_playback_length_str (int playback_length)
{
    int d;
    int h;
    int m;
    int s;

    parse_time (playback_length, &d, &h, &m, &s);

    char *time = NULL;
    if (d == 1) {
        m = s > 0 ? m + 1 : m;
        time = ho_strdup_printf ("%d %s, %02d:%02d %s", d, _("day"),
                                 h, m, _("h"));
    }
    else if (d > 1) {
        m = s > 0 ? m + 1 : m;
        time = ho_strdup_printf ("%d %s, %02d:%02d %s", d, _("days"),
                                 h, m, _("hours"));
    }
    else if (h > 0) {
        m = s > 0 ? m + 1 : m;
        time = ho_strdup_printf ("%02d:%02d %s", h, m, _("hours"));
    }
    else {
        time = ho_strdup_printf ("%02d:%02d %s", m, s, _("minutes"));
    }

    return time;
}


void
meta_info_init (xine_t * xine)
{
    assert (xine);

    pthread_mutexattr_init (&meta_info_mutex_attr);
    pthread_mutexattr_settype (&meta_info_mutex_attr,
                               PTHREAD_MUTEX_ERRORCHECK);
    pthread_mutex_init (&meta_info_mutex, &meta_info_mutex_attr);

    stream = xine_stream_new (oxine->xine, NULL, NULL);
    if (!stream) {
        error (_("Could not create stream!"));
    }

    memset (&meta_info_cache, 0, sizeof (meta_info_cache));
}


void
meta_info_free (void)
{
    if (stream) {
        xine_dispose (stream);
        stream = NULL;
    }

    meta_info_free_cache ();

    pthread_mutex_destroy (&meta_info_mutex);
    pthread_mutexattr_destroy (&meta_info_mutex_attr);
}


#ifdef HAVE_LIBEXIF

#include <libexif/exif-data.h>


char *
exif_info_get (exif_info_t type, const char *mrl)
{
    int i = 0;
    int j = 0;
    char *exif = NULL;

    if (!mrl_is_image (mrl)) {
        return NULL;
    }

    char *filename = mrl_get_filename (mrl);
    ExifData *edata = exif_data_new_from_file (filename);
    if (!edata) {
        error (_("Could not find EXIF information in '%s'!"), mrl);
        goto out_free;
    }

    switch (type) {
    case EXIF_INFO_SORT:
    case EXIF_INFO_TIME:
    case EXIF_INFO_DATE:
        for (i = 0; i < EXIF_IFD_COUNT; i++) {
            ExifContent *content = edata->ifd[i];

            if (!content || !content->count)
                continue;

            for (j = 0; j < content->count; j++) {
                ExifEntry *e = content->entries[j];
                struct tm tm = { 0, };
                char *data;

                if (!content->entries[j])
                    continue;

                if ((e->tag != EXIF_TAG_DATE_TIME) &&
                    (e->tag != EXIF_TAG_DATE_TIME_ORIGINAL) &&
                    (e->tag != EXIF_TAG_DATE_TIME_DIGITIZED))
                    continue;

                if (e->data == NULL)
                    continue;

                if (strlen ((char *) e->data) < 10)
                    continue;

                data = ho_strdup ((char *) e->data);

                data[4] = data[7] = data[10] = '\0';

                tm.tm_year = atoi (data) - 1900;
                tm.tm_mon = atoi (data + 5) - 1;
                tm.tm_mday = atoi (data + 8);
                tm.tm_hour = 0;
                tm.tm_min = 0;
                tm.tm_sec = 0;
                tm.tm_isdst = -1;

                if (strlen ((char *) e->data) > 10) {
                    data[13] = data[16] = '\0';
                    tm.tm_hour = atoi (data + 11);
                    tm.tm_min = atoi (data + 14);
                    tm.tm_sec = atoi (data + 17);
                }
                ho_free (data);

                time_t time = mktime (&tm);
                struct tm *brokentime = localtime (&time);

                const char *format = NULL;
                if (type == EXIF_INFO_SORT) {
                    format = "%Y-%m-%d %T";
                }
                else if (type == EXIF_INFO_TIME) {
                    format = config_get_string ("gui.time_format");
                }
                else {
                    format = config_get_string ("gui.date_format");
                }

                char time_str[128];
                strftime (time_str, 128, format, brokentime);
                exif = ho_strdup (time_str);

                goto out_free;
            }
        }
        break;
    default:
        break;
    }

  out_free:
    exif_data_unref (edata);
    ho_free (filename);

    return exif;
}


#endif /* HAVE_LIBEXIF */
