
/*
 * Copyright (C) 2004-2005 Maximilian Schwerin
 *
 * This file is part of oxine a free media player.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: odk_event.c 3989 2009-09-03 20:29:48Z mschwerin $
 *
 */

#include "config.h"

#include <assert.h>
#include <errno.h>
#include <pthread.h>
#include <unistd.h>

#include "heap.h"
#include "logger.h"
#include "mutex.h"
#include "scheduler.h"
#include "utils.h"

#include "odk.h"
#include "odk_http.h"
#include "odk_netr.h"
#include "odk_joystick.h"
#include "odk_lirc.h"
#include "odk_private.h"

/**
 * This mutex is used to make sure the frontend only gets one event at the
 * time.
 */
static pthread_mutex_t frontend_mutex;
static pthread_mutexattr_t frontend_mutex_attr;


void
odk_run (odk_t * odk)
{
    debug (" event loop thread: 0x%X", (int) pthread_self ());
    odk->win->enter_event_loop (odk->win);
}


void
odk_exit (odk_t * odk)
{
    odk->win->leave_event_loop (odk->win);
}


/// Frees one event handler object.
static void
event_handlers_destroy (void *event_handler_p)
{
    odk_event_handler_t *eh = (odk_event_handler_t *) event_handler_p;
    ho_free (eh);
}


/// Returns TRUE if the event handlers must be swapped to be in order.
static bool
swap_handler (void *h1_p, void *h2_p)
{
    odk_event_handler_t *h1 = (odk_event_handler_t *) h1_p;
    odk_event_handler_t *h2 = (odk_event_handler_t *) h2_p;

    return (h1->priority < h2->priority);
}


void
odk_add_event_handler (odk_t * odk, event_handler_t eh, void *data,
                       odk_event_handler_priority_t priority)
{
    assert (eh);
    assert (mutex_trylock (&frontend_mutex) == EBUSY);

    odk_event_handler_t *cur = l_list_first (odk->event_handlers);
    while (cur) {
        if (cur->event_handler == eh) {
            return;
        }
        cur = l_list_next (odk->event_handlers, cur);
    }

    odk_event_handler_t *handler = ho_new (odk_event_handler_t);
    handler->event_handler = eh;
    handler->event_handler_data = data;
    handler->priority = priority;

    l_list_append (odk->event_handlers, handler);
    l_list_sort (odk->event_handlers, swap_handler);
}


void
odk_del_event_handler (odk_t * odk, event_handler_t eh)
{
    assert (eh);
    assert (mutex_trylock (&frontend_mutex) == EBUSY);

    odk_event_handler_t *cur = l_list_first (odk->event_handlers);
    while (cur) {
        if (cur->event_handler == eh) {
            l_list_remove (odk->event_handlers, cur);
            ho_free (cur);
            return;
        }
        cur = l_list_next (odk->event_handlers, cur);
    }
}


bool
odk_get_forward_events_to_xine (odk_t * odk)
{
    return odk->forward_events_to_xine;
}


void
odk_set_forward_events_to_xine (odk_t * odk, bool forward)
{
    odk->forward_events_to_xine = forward;
}


void
odk_xine_event_send (odk_t * odk, bool force, int type)
{
    if (!odk->forward_events_to_xine && !force)
        return;

    xine_event_t ev;
    ev.type = type;
    ev.data = NULL;
    ev.data_length = 0;
    xine_event_send (odk->main_stream, &ev);
}


void
_odk_oxine_event_send (odk_t * odk, oxine_event_t * event,
                       const char *file, int line)
{
    assert (odk);
    assert (event);
    assert (mutex_trylock (&frontend_mutex) == EBUSY);

    odk_event_handler_t *handler = l_list_first (odk->event_handlers);
    while (handler) {
        handler->event_handler (handler->event_handler_data, event);
        handler = l_list_next (odk->event_handlers, handler);
    }
}


/// Scales the mouse position to the current video size.
static void
scale_mouse_position (odk_t * odk, int *x, int *y)
{
    /* Scale position. */
    x11_rectangle_t rect;
    rect.x = *x;
    rect.y = *y;
    rect.w = 0;
    rect.h = 0;

    if (xine_port_send_gui_data (odk->win->video_port,
                                 XINE_GUI_SEND_TRANSLATE_GUI_TO_VIDEO,
                                 (void *) &rect) == -1) {
        return;
    }

    *x = rect.x;
    *y = rect.y;
}


/**
 * Handles events coming from the current video output window and the input
 * plugins (e.g. LIRC, joystick).
 */
static void
odk_event_handler (void *odk_p, oxine_event_t * event)
{
    odk_lock_frontend ();

    odk_t *odk = (odk_t *) odk_p;
    int event_repeat = 1;

    switch (event->type) {
    case OXINE_EVENT_OUTPUT_FORMAT_CHANGED:
        odk_osd_adapt_size (odk, NULL);
        break;
    case OXINE_EVENT_BUTTON:
    case OXINE_EVENT_MOTION:
        /* Send an event to the xine engine. */
        if (odk->forward_events_to_xine) {
            int vx = event->data.mouse.pos.x;
            int vy = event->data.mouse.pos.y;
            scale_mouse_position (odk, &vx, &vy);

            xine_input_data_t inp;
            inp.x = vx;
            inp.y = vy;
            inp.button = event->source.button;

            xine_event_t xine_event;
            xine_event.type = event->type;
            xine_event.data = &inp;
            xine_event.data_length = sizeof (inp);

            xine_event_send (odk->main_stream, &xine_event);
        }

        /* Scale x and y to video size. */
        if (!odk_osd_use_unscaled_osd (odk)) {
            int vx = event->data.mouse.pos.x;
            int vy = event->data.mouse.pos.y;
            scale_mouse_position (odk, &vx, &vy);

            event->data.mouse.pos.x = vx;
            event->data.mouse.pos.y = vy;
        }

        /* Scale x any y to OSD size. */
        {
            event->data.mouse.pos.x =
                (int) ((double) (event->data.mouse.pos.x - odk->osd.x_off) /
                       odk->osd.hscale);
            event->data.mouse.pos.y =
                (int) ((double) (event->data.mouse.pos.y - odk->osd.y_off) /
                       odk->osd.vscale);
        }

        break;
    case OXINE_EVENT_KEY:
        if (odk_current_is_dvd (odk)) {
            switch (event->source.key) {
            case OXINE_KEY_RED:
                event->source.key = OXINE_KEY_MENU1;
                break;
            case OXINE_KEY_GREEN:
                event->source.key = OXINE_KEY_MENU2;
                break;
            case OXINE_KEY_YELLOW:
                event->source.key = OXINE_KEY_MENU3;
                break;
            case OXINE_KEY_BLUE:
                event->source.key = OXINE_KEY_MENU4;
                break;
            default:
                break;
            }
        }
        else if (odk_current_is_television (odk)) {
            switch (event->source.key) {
            case OXINE_KEY_MENU1:
                event->source.key = OXINE_KEY_RED;
                break;
            case OXINE_KEY_MENU2:
                event->source.key = OXINE_KEY_GREEN;
                break;
            case OXINE_KEY_MENU3:
                event->source.key = OXINE_KEY_YELLOW;
                break;
            case OXINE_KEY_MENU4:
                event->source.key = OXINE_KEY_BLUE;
                break;
            case OXINE_KEY_MENU5:
                event->source.key = OXINE_KEY_MENU1;
                break;
            case OXINE_KEY_MENU6:
                event->source.key = OXINE_KEY_CHANNELS;
                break;
            case OXINE_KEY_MENU7:
                event->source.key = OXINE_KEY_SCHEDULE;
                break;
            case OXINE_KEY_MENU8:
                event->source.key = OXINE_KEY_RECORDINGS;
                break;
            default:
                break;
            }
        }
        if (odk_current_is_vdr (odk)) {
            switch (event->source.key) {
            case OXINE_KEY_PAGE_UP:
                event->source.key = OXINE_KEY_UP;
                event_repeat = 4;
                break;
            case OXINE_KEY_PAGE_DOWN:
                event->source.key = OXINE_KEY_DOWN;
                event_repeat = 4;
                break;
            default:
                break;
            }
        }
        /* Send an event to the xine engine. */
        if (odk->forward_events_to_xine
            && (event->source.key < OXINE_KEY_NULL)) {
            int i = 0;
            xine_event_t xine_event;

            xine_event.stream = odk->main_stream;
            xine_event.type = event->source.key;
            xine_event.data = NULL;
            xine_event.data_length = 0;

            for (i = 0; i < event_repeat; i++) {
                xine_event_send (odk->main_stream, &xine_event);
            }
        }
        break;
    default:
        break;
    }

    /* Send the event to the frontend. */
    odk_oxine_event_send (odk, event);

    odk_unlock_frontend ();
}


/**
 * Handles events coming from the event queues of the xine streams.
 */
static void
xine_event_handler (odk_t * odk, const xine_event_t * xine_event)
{
    odk_lock_frontend ();

    switch (xine_event->type) {
    case XINE_EVENT_FRAME_FORMAT_CHANGE:
        {
            xine_format_change_data_t *data =
                (xine_format_change_data_t *) xine_event->data;
            odk_osd_adapt_size (odk, data);
        }
        break;
    case XINE_EVENT_UI_PLAYBACK_FINISHED:
        if (false) {
            /* do nothing */
        }
        else if (odk->current_mode == ODK_MODE_NULL) {
            /* do nothing */
        }
        else if (odk->current_mode == ODK_MODE_LOGO) {
            /* do nothing */
        }
        /* If the source is the animation stream, we restart the stream. */
        else if (xine_event->stream == odk->animation_stream) {
            xine_play (odk->animation_stream, 0, 0);
        }
        /* If the source is not the main stream, we ignore the event. */
        else if (xine_event->stream != odk->main_stream) {
            /* do nothing */
        }
        /* If it's an image the playback is handled by the slideshow manager. */
        else if (odk_current_is_image (odk)) {
            /* do nothing */
        }
        /* If the current stream has alternatives, we tell the frontend to
         * play those. */
        else if (odk->current_has_alternatives) {
            odk->current_mode = ODK_MODE_NULL;
            oxine_event_t event;
            oxine_event_init (&event);
            event.type = OXINE_EVENT_PLAYBACK_FINISHED_WITH_ALTERNATIVES;
            odk_oxine_event_send (odk, &event);
        }
        /* Else we inform the frontend about the ending of the stream. */
        else {
            odk->current_mode = ODK_MODE_NULL;
            oxine_event_t event;
            oxine_event_init (&event);
            event.type = OXINE_EVENT_PLAYBACK_FINISHED;
            odk_oxine_event_send (odk, &event);
        }
        break;
    case XINE_EVENT_UI_CHANNELS_CHANGED:
    case XINE_EVENT_UI_SET_TITLE:
        /* We use this event to find out if the meta info has changed
         * without a new track starting (e.g. web-radio). */
        if (odk->current_mode == ODK_MODE_NULL) {
            /* do nothing */
        }
        else if (xine_event->stream != odk->main_stream) {
            /* do nothing */
        }
        else if (odk_current_is_dvd (odk)) {
            /* do nothing */
        }
        else {
            char *current_title = odk_get_meta_info (odk, META_INFO_TITLE);

            if (!current_title && odk->current_title) {
                oxine_event_t event;
                oxine_event_init (&event);
                event.type = OXINE_EVENT_GUI_REPAINT;
                odk_oxine_event_send (odk, &event);
            }
            else if (current_title && odk->current_title
                     && (strcmp (current_title, odk->current_title) != 0)) {
                ho_free (odk->current_title);
                odk->current_title = current_title;
                oxine_event_t event;
                oxine_event_init (&event);
                event.type = OXINE_EVENT_GUI_REPAINT;
                odk_oxine_event_send (odk, &event);
            }
            else if (current_title) {
                ho_free (current_title);
            }
        }
        break;
    case XINE_EVENT_MRL_REFERENCE_EXT:
        /* If we receive an alternative stream, we inform the frontend about
         * it. The frontend is then responsible for starting it. */
        if (xine_event->stream == odk->main_stream) {
            xine_mrl_reference_data_ext_t *p =
                (xine_mrl_reference_data_ext_t *) xine_event->data;

            debug ("Got alternative MRL [%d]: '%s'", p->alternative, p->mrl);
            odk->current_has_alternatives = true;

            oxine_event_t event;
            oxine_event_init (&event);
            event.type = OXINE_EVENT_ADD_ALTERNATIVE;
            event.data.alternative.pos = p->alternative;
            event.data.alternative.mrl = p->mrl;
            event.data.alternative.title = (p->mrl + strlen (p->mrl) + 1);
            odk_oxine_event_send (odk, &event);
        }
        break;
    case XINE_EVENT_UI_MESSAGE:
#ifdef DEBUG
        if (xine_event->stream == odk->main_stream) {
            xine_ui_message_data_t *msg =
                (xine_ui_message_data_t *) xine_event->data;
            // TODO: implement message reading
            debug ("Got message from xine-lib: '%s'", msg->messages);
        }
#endif
        break;
    case XINE_EVENT_PROGRESS:
        {
            xine_progress_data_t *d =
                (xine_progress_data_t *) xine_event->data;
            debug ("%s: %3d%%", d->description, d->percent);
        }
        break;
    default:
        break;
    }

    odk_unlock_frontend ();
}


static void *
xine_event_thread (void *data)
{
    odk_t *odk = (odk_t *) data;

    info (_("Successfully started xine event thread."));
    debug (" xine event thread: 0x%X", (int) pthread_self ());

    while (odk->xine_event_thread_run) {
        xine_event_t *event = NULL;

        if ((event = xine_event_get (odk->main_stream_queue)) != NULL) {
            xine_event_handler (odk, event);
            xine_event_free (event);
        }
        if ((event = xine_event_get (odk->back_stream_queue)) != NULL) {
            xine_event_handler (odk, event);
            xine_event_free (event);
        }
        if ((event = xine_event_get (odk->anim_stream_queue)) != NULL) {
            xine_event_handler (odk, event);
            xine_event_free (event);
        }

        usleep (1000);
    }

    return NULL;
}


bool
odk_start_input_plugin (odk_t * odk, const char *id)
{
    if (false) {
        /* do nothing */
    }
#ifdef HAVE_LIRC
    else if (strc_equals (id, "lirc")) {
        if (!odk->lirc) {
            odk->lirc = start_lirc (odk, odk_event_handler);
        }
        return (odk->lirc != NULL);
    }
#endif /* HAVE_LIRC */
#ifdef HAVE_JOYSTICK
    else if (strc_equals (id, "joystick")) {
        if (!odk->joystick) {
            odk->joystick = start_joystick (odk, odk_event_handler);
        }
        return (odk->joystick != NULL);
    }
#endif /* HAVE_JOYSTICK */
#ifdef HAVE_HTTP
    else if (strc_equals (id, "http")) {
        if (!odk->http) {
            odk->http = start_http (odk, odk_event_handler);
        }
        return (odk->http != NULL);
    }
#endif /* HAVE_HTTP */
#ifdef HAVE_NETR
    else if (strc_equals (id, "netr")) {
        if (!odk->netr) {
            odk->netr = start_netr (odk, odk_event_handler);
        }
        return (odk->netr != NULL);
    }
#endif /* HAVE_NETR */
    else {
        return odk->win->start_input_plugin (odk->win, id);
    }
}


bool
odk_stop_input_plugin (odk_t * odk, const char *id)
{
    if (false) {
        /* do nothing */
    }
#ifdef HAVE_LIRC
    else if (strc_equals (id, "lirc")) {
        if (odk->lirc) {
            stop_lirc (odk->lirc);
            odk->lirc = NULL;
        }
        return true;
    }
#endif /* HAVE_LIRC */
#ifdef HAVE_JOYSTICK
    else if (strc_equals (id, "joystick")) {
        if (odk->joystick) {
            stop_joystick (odk->joystick);
            odk->joystick = NULL;
        }
        return true;
    }
#endif /* HAVE_JOYSTICK */
#ifdef HAVE_HTTP
    else if (strc_equals (id, "http")) {
        if (odk->http) {
            stop_http (odk->http);
            odk->http = NULL;
        }
        return true;
    }
#endif /* HAVE_HTTP */
#ifdef HAVE_NETR
    else if (strc_equals (id, "netr")) {
        if (odk->netr) {
            stop_netr (odk->netr);
            odk->netr = NULL;
        }
        return true;
    }
#endif /* HAVE_NETR */
    else {
        return odk->win->stop_input_plugin (odk->win, id);
    }
}


void
odk_event_init (odk_t * odk)
{
#ifdef DEBUG
    assert (odk);
    assert (odk->win);
    assert (odk->main_stream);
    assert (odk->background_stream);
    assert (odk->animation_stream);
#endif

    pthread_mutexattr_init (&frontend_mutex_attr);
    pthread_mutexattr_settype (&frontend_mutex_attr,
                               PTHREAD_MUTEX_ERRORCHECK);
    pthread_mutex_init (&frontend_mutex, &frontend_mutex_attr);

    odk->event_handlers = l_list_new ();
    odk->forward_events_to_xine = false;
    odk->win->set_event_handler (odk->win, odk_event_handler, odk);

    odk->main_stream_queue = xine_event_new_queue (odk->main_stream);
    odk->back_stream_queue = xine_event_new_queue (odk->background_stream);
    odk->anim_stream_queue = xine_event_new_queue (odk->animation_stream);

    odk->xine_event_thread_run = true;
    if (pthread_create (&odk->xine_event_thread, NULL,
                        xine_event_thread, odk) != 0) {
        error (_("Could not create xine event thread: %s!"),
               strerror (errno));
        abort ();
    }
}


void
odk_event_free (odk_t * odk)
{
    odk_stop_input_plugin (odk, "http");
    odk_stop_input_plugin (odk, "netr");
    odk_stop_input_plugin (odk, "lirc");
    odk_stop_input_plugin (odk, "mouse");
    odk_stop_input_plugin (odk, "keyboard");
    odk_stop_input_plugin (odk, "joystick");

    void *ret = NULL;
    odk->xine_event_thread_run = false;
    pthread_join (odk->xine_event_thread, &ret);
    info (_("Successfully stopped xine event thread."));

    if (odk->main_stream_queue) {
        xine_event_dispose_queue (odk->main_stream_queue);
    }
    if (odk->back_stream_queue) {
        xine_event_dispose_queue (odk->back_stream_queue);
    }
    if (odk->anim_stream_queue) {
        xine_event_dispose_queue (odk->anim_stream_queue);
    }

    if (odk->event_handlers) {
        l_list_free (odk->event_handlers, event_handlers_destroy);
        odk->event_handlers = NULL;
    }

    pthread_mutex_destroy (&frontend_mutex);
    pthread_mutexattr_destroy (&frontend_mutex_attr);
}


void
_odk_lock_frontend (const char *file, int line)
{
#ifdef DEBUG
    _mutex_lock (&frontend_mutex, file, line);
#else
    mutex_lock (&frontend_mutex);
#endif
    //_log (LEVEL_DEBUG, __func__, file, line, "entered frontend");
}


void
_odk_unlock_frontend (const char *file, int line)
{
    //_log (LEVEL_DEBUG, __func__, file, line, "leaving frontend");
#ifdef DEBUG
    _mutex_unlock (&frontend_mutex, file, line);
#else
    mutex_unlock (&frontend_mutex);
#endif
}
