
/*
 * Copyright (C) 2007 Maximilian Schwerin
 *
 * This file is part of oxine a free media player.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: odk_http.c 3881 2009-05-04 11:36:14Z mschwerin $
 *
 */
#include "config.h"

#include <errno.h>
#include <pthread.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>

#ifdef HAVE_LIBX11
#include <X11/Xlib.h>
#endif

#include "filelist.h"
#include "heap.h"
#include "i18n.h"
#include "logger.h"
#include "oxine.h"
#include "playlist.h"
#include "types.h"
#include "utils.h"

#include "odk.h"
#include "odk_http.h"
#include "odk_keymap.h"
#include "odk_snapshot.h"

#define BUFFER_SIZE             2048
#define HTTP_PORT               3080
#define HTTP_QUEUE_LENGTH       5
#define CRLF                    "\r\n"

#define HTTP_VERSION            "HTTP/1.1"

#ifdef HAVE_HTTP
typedef struct {
    odk_input_t class;

    odk_t *odk;
    void (*event_handler) (void *data, oxine_event_t * ev);

    int server_port;
    int server_socket;
    int client_socket;
} http_input_t;


extern oxine_t *oxine;


static const char *
http_get_reason_phrase (int status_code)
{
    switch (status_code) {
    case 200:
        return "OK";
    case 303:
        return "See Other";
    case 400:
        return "Bad request";
    case 404:
        return "Not Found";
    case 500:
        return "Internal Server Error";
    case 501:
        return "Not Implemented";
    default:
        assert (false);
    }
}


static void
http_send_status (int client_socket, int status_code)
{
    server_printf (client_socket,
                   HTTP_VERSION " %d %s" CRLF,
                   status_code, http_get_reason_phrase (status_code));
}


static void
http_send_error (int client_socket, int status_code)
{
    http_send_status (client_socket, status_code);
    server_printf (client_socket,
                   "Content-Type: text/html; charset=UTF-8" CRLF);
    server_printf (client_socket, CRLF);
    server_printf (client_socket,
                   "<html>"
                   "<head><title>oxine - %s</title></head>"
                   "<body><strong>Error:</strong> %d - %s</body>"
                   "</html>",
                   http_get_reason_phrase (status_code),
                   status_code, http_get_reason_phrase (status_code));
}


static void
handle_connection (http_input_t * hi)
{
    int ret = 0;
    char *end;
    char buffer[BUFFER_SIZE];

    ret = read (hi->client_socket, buffer, BUFFER_SIZE);
    if (ret == -1) {
        error (_("Failed to read from socket: %s!"), strerror (errno));
        return;
    }
    buffer[ret] = '\0';

#ifdef DEBUG
    char *eol = NULL;
    char *start = buffer;
    while ((eol = strstr (start, CRLF))) {
        eol[0] = '\0';
        debug ("< %s", trim_whitespace (start));
        start = &eol[2];
    }
#endif

    char *cmd = strstr (buffer, "cmd=");
    if (cmd) {
        cmd = ho_strdup (&cmd[4]);
        end = index (cmd, ' ');
        if (end) {
            end[0] = '\0';
        }
        end = strstr (cmd, "&gui=");
        if (end) {
            end[0] = '\0';
        }
        end = strstr (cmd, "&prm=");
        if (end) {
            end[0] = '\0';
        }
    }

    char *prm = strstr (buffer, "prm=");
    if (prm) {
        prm = ho_strdup (&prm[4]);
        end = index (prm, ' ');
        if (end) {
            end[0] = '\0';
        }
        end = strstr (prm, "&cmd=");
        if (end) {
            end[0] = '\0';
        }
        end = strstr (prm, "&gui=");
        if (end) {
            end[0] = '\0';
        }
    }

    char *gui = strstr (buffer, "gui=");
    if (gui) {
        gui = ho_strdup (&gui[4]);
        end = index (gui, ' ');
        if (end) {
            end[0] = '\0';
        }
        end = strstr (gui, "&cmd=");
        if (end) {
            end[0] = '\0';
        }
        end = strstr (gui, "&prm=");
        if (end) {
            end[0] = '\0';
        }
    }

    char *cmd_decode = url_decode (cmd);
    char *prm_decode = url_decode (prm);

    info (_("     Command: %s"), cmd_decode);
    info (_("   Parameter: %s"), prm_decode);

    oxine_event_t event;
    oxine_event_init (&event);
    bool response_sent = false;

    if (!cmd_decode) {
        http_send_error (hi->client_socket, 400);
        response_sent = true;
    }
    else if (strcasecmp (cmd_decode, "get_snapshot") == 0) {
        odk_lock_frontend ();
        int size = 0;
        char *buffer = odk_create_snapshot (oxine->odk, &size);

        if (buffer) {
            http_send_status (hi->client_socket, 200);
            server_printf (hi->client_socket, "Content-Type: image/png" CRLF);
            server_printf (hi->client_socket, "Content-Length: %d" CRLF,
                           size);
            server_printf (hi->client_socket, CRLF);
            if (write (hi->client_socket, buffer, size) == -1) {
                error (_("Failed to write to socket: %s!"), strerror (errno));
            }
            ho_free (buffer);
        }
        else {
            http_send_error (hi->client_socket, 500);
        }

        response_sent = true;
        odk_unlock_frontend ();
    }
    else if (strcasecmp (cmd_decode, "add_mrl") == 0) {
        odk_lock_frontend ();
        char *mrl = mrl_from_filename (prm);

        filelist_t *filelist = filelist_get (FILELIST_TEMPORARY_MRL);
        filelist_lock (filelist);
        fileitem_t *fileitem = filelist_add_magic (filelist, NULL, mrl);
        if (fileitem) {
            playlist_add_fileitem (oxine->rw_playlist, fileitem);
            filelist_remove (filelist, fileitem);
        }
        filelist_unlock (filelist);

        ho_free (mrl);
        odk_unlock_frontend ();
    }
    else if (strcasecmp (cmd_decode, "play_mrl") == 0) {
        odk_lock_frontend ();
        char *mrl = mrl_from_filename (prm);

        filelist_t *filelist = filelist_get (FILELIST_TEMPORARY_MRL);
        filelist_lock (filelist);
        fileitem_t *fileitem = filelist_add_magic (filelist, NULL, mrl);
        if (fileitem) {
            playlist_clear (oxine->rw_playlist);
            playlist_add_fileitem (oxine->rw_playlist, fileitem);
            filelist_remove (filelist, fileitem);
            playlist_play_first (oxine->rw_playlist);
        }
        filelist_unlock (filelist);

        ho_free (mrl);
        odk_unlock_frontend ();
    }
    else if (strcasecmp (cmd_decode, "set_fullscreen") == 0) {
        odk_lock_frontend ();
        bool fullscreen = (strcasecmp (prm, "true") == 0);
        odk_window_set_fullscreen (oxine->odk, fullscreen);
        odk_unlock_frontend ();
    }
#ifdef HAVE_LIBX11
    else if (strcasecmp (cmd_decode, "set_geometry") == 0) {
        odk_lock_frontend ();
        int x = 0;
        int y = 0;
        int w = 0;
        int h = 0;
        XParseGeometry (prm, &x, &y, ((unsigned int *) &w),
                        ((unsigned int *) &h));
        odk_window_set_geometry (oxine->odk, x, y, w, h);
        odk_unlock_frontend ();
    }
#endif
    else if (odk_keymap_http2action (&event, cmd_decode)) {
        hi->event_handler (hi->odk, &event);
    }
    else {
        warn (_("Got unknown HTTP command: '%s'"), cmd);
        http_send_error (hi->client_socket, 501);
        response_sent = true;
    }

    if (!response_sent) {
        if (gui) {
            http_send_status (hi->client_socket, 303);
            server_printf (hi->client_socket, "Location: %s" CRLF, gui);
            server_printf (hi->client_socket, CRLF);
        }
        else {
            http_send_status (hi->client_socket, 200);
            server_printf (hi->client_socket, CRLF CRLF);
        }
    }

    ho_free (cmd);
    ho_free (prm);
    ho_free (gui);

    ho_free (cmd_decode);
    ho_free (prm_decode);
}


static void *
http_thread (void *this)
{
    http_input_t *hi = (http_input_t *) this;

    info (_("Successfully started the HTTP input plugin."));
    info (_("Listening for network connections on port %d."),
          hi->server_port);
    debug ("       HTTP thread: 0x%X", (int) pthread_self ());

    while (1) {
        struct sockaddr_in client_addr_in;
        struct sockaddr *client_addr = (struct sockaddr *) &client_addr_in;
        unsigned int client_addr_size = sizeof (client_addr_in);

        hi->client_socket = accept (hi->server_socket, client_addr,
                                    &client_addr_size);
        if (hi->client_socket == -1) {
            error (_("Failed to accept connection from client: %s!"),
                   strerror (errno));
        }
        else {
            handle_connection (hi);
            close (hi->client_socket);
            hi->client_socket = 0;
        }
    }

    return NULL;
}


odk_input_t *
start_http (odk_t * odk, event_handler_t event_handler)
{
    http_input_t *hi = ho_new (http_input_t);
    hi->odk = odk;
    hi->event_handler = event_handler;

    /* Initialize the server socket. */
    hi->server_port = HTTP_PORT;
    hi->server_socket = server_init (hi->server_port, HTTP_QUEUE_LENGTH);
    if (hi->server_socket == -1) {
        error (_("Failed to initialize HTTP input plugin!"));
        ho_free (hi);
        return NULL;
    }

    /* Create the thread to handle HTTP commands. */
    if (pthread_create (&hi->class.thread, NULL, http_thread, hi) != 0) {
        error (_("Could not create HTTP thread: %s!"), strerror (errno));
        ho_free (hi);
        return NULL;
    }

    return (odk_input_t *) hi;
}


void
stop_http (odk_input_t * oi)
{
    http_input_t *li = (http_input_t *) oi;
    pthread_cancel (oi->thread);
    if (li->client_socket > 0) {
        close (li->client_socket);
    }
    if (li->server_socket > 0) {
        close (li->server_socket);
    }
    ho_free (li);

    info (_("Successfully stopped the HTTP input plugin."));
}


#endif /* HAVE_HTTP */
