
/*
 * Copyright (C) 2008 Maximilian Schwerin
 *
 * This file is part of oxine a free media player.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: odk_netr.c 3881 2009-05-04 11:36:14Z mschwerin $
 *
 */
#include "config.h"

#include <errno.h>
#include <pthread.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>

#ifdef HAVE_LIBX11
#include <X11/Xlib.h>
#endif

#include "filelist.h"
#include "heap.h"
#include "i18n.h"
#include "logger.h"
#include "oxine.h"
#include "playlist.h"
#include "types.h"
#include "utils.h"

#include "odk.h"
#include "odk_netr.h"
#include "odk_keymap.h"
#include "odk_snapshot.h"

#define NETR_PORT               3023
#define NETR_QUEUE_LENGTH       5
#define BUFFER_SIZE             2048
#define CRLF                    "\r\n"

#ifdef HAVE_NETR
typedef struct {
    odk_input_t class;

    odk_t *odk;
    void (*event_handler) (void *data, oxine_event_t * ev);

    int server_port;
    int server_socket;
    int client_socket;
} netr_input_t;


static const char *input_plugins[] = {
    "keyboard",
    "mouse",
#ifdef HAVE_LIRC
    "lirc",
#endif /* HAVE_LIRC */
#ifdef HAVE_JOYSTICK
    "joystick",
#endif /* HAVE_JOYSTICK */
#ifdef HAVE_HTTP
    "http",
#endif /* HAVE_HTTP */
    NULL,
};


extern oxine_t *oxine;


static void
handle_connection (netr_input_t * hi)
{
    int i = 0;
    int ret = 0;
    char buffer[BUFFER_SIZE];
    bool keep_connection = true;

    oxine_event_t event;
    oxine_event_init (&event);

    server_printf (hi->client_socket,
                   "This is " PACKAGE_STRING " network remote control server!"
                   CRLF);

    while (keep_connection) {
        ret = readline (hi->client_socket, buffer, BUFFER_SIZE);
        if (ret == -1) {
            error (_("Failed to read from socket: %s!"), strerror (errno));
            return;
        }
        if (ret <= 1) {
            continue;
        }

#ifdef DEBUG
        char *eol = NULL;
        char *start = buffer;
        while ((eol = strstr (start, CRLF))) {
            eol[0] = '\0';
            debug ("< %s", trim_whitespace (start));
            start = &eol[2];
        }
#endif

        int argc = 0;
        char *argv[20];
        arguments_to_argv (buffer, argv, &argc);

        if (argc == 0) {
            /* do nothing */
        }
        else if (strcasecmp (argv[0], "help") == 0) {
            server_printf (hi->client_socket, CRLF);
            server_printf (hi->client_socket, "Remote commands:" CRLF);
            server_printf (hi->client_socket,
                           "    All LIRC commands are accepted." CRLF);
            server_printf (hi->client_socket, CRLF);
            server_printf (hi->client_socket, "Extra commands:" CRLF);
            server_printf (hi->client_socket, "    logout              - "
                           "Closes connection" CRLF);
            server_printf (hi->client_socket, "    list_input          - "
                           "Lists all available input plugins" CRLF);
            server_printf (hi->client_socket, "    start_input <input> - "
                           "Starts an input plugin (e.g. keyboard, LIRC)"
                           CRLF);
            server_printf (hi->client_socket, "    stop_input  <input> - "
                           "Stops an input plugin  (e.g. keyboard, LIRC)"
                           CRLF);
            server_printf (hi->client_socket, CRLF);
        }
        else if (buffer[0] == 0x04) {
            server_printf (hi->client_socket, "Closing connection" CRLF);
            keep_connection = false;
        }
        else if (strcasecmp (argv[0], "logout") == 0) {
            server_printf (hi->client_socket, "Closing connection" CRLF);
            keep_connection = false;
        }
#ifdef HAVE_LIBX11
        else if (strcasecmp (argv[0], "set_geometry") == 0) {
            odk_lock_frontend ();
            int x = 0;
            int y = 0;
            int w = 0;
            int h = 0;
            XParseGeometry (argv[1], &x, &y,
                            ((unsigned int *) &w),
                            ((unsigned int *) &h));
            odk_window_set_geometry (oxine->odk, x, y, w, h);
            odk_unlock_frontend ();
        }
#endif
        else if (strcasecmp (argv[0], "list_input") == 0) {
            int i = 0;
            for (i = 0; input_plugins[i]; i++) {
                server_printf (hi->client_socket, input_plugins[i]);
                server_printf (hi->client_socket, CRLF);
            }
        }
        else if (strcasecmp (argv[0], "start_input") == 0) {
            odk_lock_frontend ();
            const char *f1 = "Starting input plugin '%s'..." CRLF;
            const char *ef = "Starting input plugin '%s' failed!" CRLF;
            if (strcasecmp (argv[1], "all") == 0) {
                int i = 0;
                for (i = 0; input_plugins[i]; i++) {
                    server_printf (hi->client_socket, f1, input_plugins[i]);
                    if (!odk_start_input_plugin (hi->odk, input_plugins[i])) {
                        server_printf (hi->client_socket, ef,
                                       input_plugins[i]);
                    }
                    else {
                        server_printf (hi->client_socket, "Ok" CRLF);
                    }
                }
            }
            else if (strcasecmp (argv[1], "netr") == 0) {
                server_printf (hi->client_socket, "Ok" CRLF);
            }
            else if (odk_start_input_plugin (hi->odk, argv[1])) {
                server_printf (hi->client_socket, "Ok" CRLF);
            }
            else {
                server_printf (hi->client_socket, ef, argv[1]);
            }
            odk_unlock_frontend ();
        }
        else if (strcasecmp (argv[0], "stop_input") == 0) {
            odk_lock_frontend ();
            const char *f0 = "Stopping network remote is forbidden!" CRLF;
            const char *f1 = "Stopping input plugin '%s'..." CRLF;
            const char *ef = "input plugin '%s' failed!" CRLF;
            if (strcasecmp (argv[1], "all") == 0) {
                int i = 0;
                for (i = 0; input_plugins[i]; i++) {
                    server_printf (hi->client_socket, f1, input_plugins[i]);
                    if (!odk_stop_input_plugin (hi->odk, input_plugins[i])) {
                        server_printf (hi->client_socket, ef,
                                       input_plugins[i]);
                    }
                    else {
                        server_printf (hi->client_socket, "Ok" CRLF);
                    }
                }
            }
            else if (strcasecmp (argv[1], "netr") == 0) {
                server_printf (hi->client_socket, f0);
            }
            else if (odk_stop_input_plugin (hi->odk, argv[1])) {
                server_printf (hi->client_socket, "Ok" CRLF);
            }
            else {
                server_printf (hi->client_socket, ef, argv[1]);
            }
            odk_unlock_frontend ();
        }
        else if (strcasecmp (argv[0], "add_mrl") == 0) {
            char *mrl = mrl_from_filename (argv[1]);

            filelist_t *filelist = filelist_get (FILELIST_TEMPORARY_MRL);
            filelist_lock (filelist);
            fileitem_t *fileitem = filelist_add_magic (filelist, NULL, mrl);
            if (fileitem) {
                playlist_add_fileitem (oxine->rw_playlist, fileitem);
                filelist_remove (filelist, fileitem);
            }
            filelist_unlock (filelist);

            ho_free (mrl);
        }
        else if (strcasecmp (argv[0], "play_mrl") == 0) {
            odk_lock_frontend ();
            char *mrl = mrl_from_filename (argv[1]);

            filelist_t *filelist = filelist_get (FILELIST_TEMPORARY_MRL);
            filelist_lock (filelist);
            fileitem_t *fileitem = filelist_add_magic (filelist, NULL, mrl);
            if (fileitem) {
                playlist_clear (oxine->rw_playlist);
                playlist_add_fileitem (oxine->rw_playlist, fileitem);
                filelist_remove (filelist, fileitem);
                playlist_play_first (oxine->rw_playlist);
            }
            filelist_unlock (filelist);

            ho_free (mrl);
            odk_unlock_frontend ();
        }
        else if (odk_keymap_netr2action (&event, argv[0])) {
            hi->event_handler (hi->odk, &event);
            server_printf (hi->client_socket, "Ok" CRLF);
        }
        else {
            warn (_("Got unknown NETR command: '%s'"), argv[0]);
            server_printf (hi->client_socket,
                           "Unknown command: %s" CRLF, argv[0]);
        }

        for (i = 0; i < argc; i++) {
            ho_free (argv[i]);
        }
    }
}


static void *
netr_thread (void *this)
{
    netr_input_t *hi = (netr_input_t *) this;

    info (_("Successfully started the network input plugin."));
    info (_("Listening for network connections on port %d."),
          hi->server_port);
    debug ("    network thread: 0x%X", (int) pthread_self ());

    while (1) {
        struct sockaddr_in client_addr_in;
        struct sockaddr *client_addr = (struct sockaddr *) &client_addr_in;
        unsigned int client_addr_size = sizeof (client_addr_in);

        hi->client_socket = accept (hi->server_socket, client_addr,
                                    &client_addr_size);
        if (hi->client_socket == -1) {
            error (_("Failed to accept connection from client: %s!"),
                   strerror (errno));
        }
        else {
            handle_connection (hi);
            close (hi->client_socket);
            hi->client_socket = 0;
        }
    }

    return NULL;
}


odk_input_t *
start_netr (odk_t * odk, event_handler_t event_handler)
{
    netr_input_t *hi = ho_new (netr_input_t);
    hi->odk = odk;
    hi->event_handler = event_handler;

    /* Initialize the server socket. */
    hi->server_port = NETR_PORT;
    hi->server_socket = server_init (hi->server_port, NETR_QUEUE_LENGTH);
    if (hi->server_socket == -1) {
        error (_("Failed to initialize network input plugin!"));
        ho_free (hi);
        return NULL;
    }

    /* Create the thread to handle NETR commands. */
    if (pthread_create (&hi->class.thread, NULL, netr_thread, hi) != 0) {
        error (_("Could not create network thread: %s!"), strerror (errno));
        ho_free (hi);
        return NULL;
    }

    return (odk_input_t *) hi;
}


void
stop_netr (odk_input_t * oi)
{
    netr_input_t *li = (netr_input_t *) oi;
    pthread_cancel (oi->thread);
    if (li->client_socket > 0) {
        close (li->client_socket);
    }
    if (li->server_socket > 0) {
        close (li->server_socket);
    }
    ho_free (li);

    info (_("Successfully stopped the network input plugin."));
}


#endif /* HAVE_NETR */
