
/*
 * Copyright (C) 2003 by Dirk Meyer
 * Copyright (C) 2008 Maximilian Schwerin
 *
 * This file is part of oxine a free media player.
 * Parts of this file are part of xine, a unix video player.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: odk_post.c 3761 2009-04-23 09:55:54Z mschwerin $
 *
 * The code was taken from xine-ui/src/xitk/post.c at changed to work with fbxine
 * 
 * Modified for VDR xineliboutput plugin by Petri Hintukainen, 2006
 *   - runtime re-configuration (load/unload, enable/disable)
 *   - support for multiple streams 
 *   - support for mosaico post plugin (picture-in-picture)
 *
 */

#include "config.h"

#include <assert.h>
#include <stdio.h>
#include <xine.h>

#include "heap.h"
#include "i18n.h"
#include "logger.h"
#include "odk_private.h"
#include "odk_post.h"
#include "utils.h"


#define DEFAULT_DEINTERLACER "tvtime:method=LinearBlend,cheap_mode=1,pulldown=0,use_progressive_frame_flag=1"

#define fe_t post_plugins_t

typedef struct {
    xine_post_t *post;
    xine_post_api_t *api;
    xine_post_api_descr_t *descr;
    xine_post_api_parameter_t *param;
    char *param_data;

    int x;
    int y;

    int readonly;

    char **properties_names;
} post_object_t;


static int
__pplugin_retrieve_parameters (post_object_t * pobj)
{
    xine_post_in_t *input_api;

    if ((input_api = (xine_post_in_t *) xine_post_input (pobj->post,
                                                         "parameters"))) {
        xine_post_api_t *post_api;
        xine_post_api_descr_t *api_descr;
        xine_post_api_parameter_t *parm;
        int pnum = 0;

        post_api = (xine_post_api_t *) input_api->data;

        api_descr = post_api->get_param_descr ();

        parm = api_descr->parameter;
        pobj->param_data = ho_malloc (api_descr->struct_size);

        while (parm->type != POST_PARAM_TYPE_LAST) {

            post_api->get_parameters (pobj->post, pobj->param_data);

            if (!pnum)
                pobj->properties_names =
                    (char **) ho_malloc (sizeof (char *) * 2);
            else
                pobj->properties_names = (char **)
                    ho_realloc (pobj->properties_names,
                                sizeof (char *) * (pnum + 2));

            pobj->properties_names[pnum] = ho_strdup (parm->name);
            pobj->properties_names[pnum + 1] = NULL;
            pnum++;
            parm++;
        }

        pobj->api = post_api;
        pobj->descr = api_descr;
        pobj->param = api_descr->parameter;

        return 1;
    }

    return 0;
}

static void
_pplugin_update_parameter (post_object_t * pobj)
{
    pobj->api->set_parameters (pobj->post, pobj->param_data);
    pobj->api->get_parameters (pobj->post, pobj->param_data);
}

static void
__pplugin_update_parameters (xine_post_t * post, char *args)
{
    post_object_t pobj;
    char *p;

    memset (&pobj, 0, sizeof (post_object_t));
    pobj.post = post;

    if (__pplugin_retrieve_parameters (&pobj)) {
        int i;

        if (pobj.properties_names && args && *args) {
            char *param;

            while ((param = strsep (&args, ",")) != NULL) {

                p = param;

                while ((*p != '\0') && (*p != '='))
                    p++;

                if (p && strlen (p)) {
                    int param_num = 0;

                    *p++ = '\0';

                    while (pobj.properties_names[param_num]
                           && strcasecmp (pobj.properties_names[param_num],
                                          param))
                        param_num++;

                    if (pobj.properties_names[param_num]) {

                        pobj.param = pobj.descr->parameter;
                        pobj.param += param_num;
                        pobj.readonly = pobj.param->readonly;

                        switch (pobj.param->type) {
                        case POST_PARAM_TYPE_INT:
                            if (!pobj.readonly) {
                                if (pobj.param->enum_values) {
                                    char **values = pobj.param->enum_values;
                                    int i = 0;

                                    while (values[i]) {
                                        if (!strcasecmp (values[i], p)) {
                                            *(int *) (pobj.param_data +
                                                      pobj.param->offset) = i;
                                            break;
                                        }
                                        i++;
                                    }

                                    if (!values[i])
                                        *(int *) (pobj.param_data +
                                                  pobj.param->offset) =
                                            (int) strtol (p, &p, 10);
                                }
                                else {
                                    *(int *) (pobj.param_data +
                                              pobj.param->offset) =
                                        (int) strtol (p, &p, 10);
                                }
                                _pplugin_update_parameter (&pobj);
                            }
                            break;

                        case POST_PARAM_TYPE_DOUBLE:
                            if (!pobj.readonly) {
                                *(double *) (pobj.param_data +
                                             pobj.param->offset) =
                                    strtod (p, &p);
                                _pplugin_update_parameter (&pobj);
                            }
                            break;

                        case POST_PARAM_TYPE_CHAR:
                        case POST_PARAM_TYPE_STRING:
                            if (!pobj.readonly) {
                                if (pobj.param->type == POST_PARAM_TYPE_CHAR) {
                                    int maxlen =
                                        pobj.param->size / sizeof (char);

                                    snprintf ((char *) (pobj.param_data +
                                                        pobj.param->offset),
                                              maxlen, "%s", p);
                                    _pplugin_update_parameter (&pobj);
                                }
                                else
                                    fprintf (stderr,
                                             "parameter type POST_PARAM_TYPE_STRING not supported yet.\n");
                            }
                            break;

                        case POST_PARAM_TYPE_STRINGLIST:       /* unsupported */
                            if (!pobj.readonly)
                                fprintf (stderr,
                                         "parameter type POST_PARAM_TYPE_STRINGLIST not supported yet.\n");
                            break;

                        case POST_PARAM_TYPE_BOOL:
                            if (!pobj.readonly) {
                                *(int *) (pobj.param_data +
                                          pobj.param->offset) =
                                    ((int) strtol (p, &p, 10)) ? 1 : 0;
                                _pplugin_update_parameter (&pobj);
                            }
                            break;
                        }
                    }
                    else {
                        error ("Unknown post plugin parameter %s !", param);
                    }
                }
            }

            i = 0;

            while (pobj.properties_names[i]) {
                ho_free (pobj.properties_names[i]);
                i++;
            }

            ho_free (pobj.properties_names);
        }

        ho_free (pobj.param_data);
    }
}

/* -post <name>:option1=value1,option2=value2... */
static post_element_t **
pplugin_parse_and_load (fe_t * fe,
                        int plugin_type,
                        const char *pchain, int *post_elements_num)
{
    post_element_t **post_elements = NULL;
    char *post_chain;

    *post_elements_num = 0;

    if (pchain && strlen (pchain)) {
        char *p;

        post_chain = ho_strdup (pchain);

        while ((p = strsep (&post_chain, ";"))) {

            if (p && strlen (p)) {
                char *plugin,
                 *args = NULL;
                xine_post_t *post;

                while (*p == ' ')
                    p++;

                plugin = ho_strdup (p);

                if ((p = strchr (plugin, ':')))
                    *p++ = '\0';

                if (p && (strlen (p) > 1))
                    args = p;
#if 0
                post = xine_post_init (fe->xine, plugin, 0,
                                       &fe->audio_port, &fe->video_port);
#else
                if (plugin_type == XINE_POST_TYPE_VIDEO_COMPOSE) {
                    post = xine_post_init (fe->xine, plugin, 2,
                                           &fe->audio_port, &fe->video_port);
                }
                else
                    post = xine_post_init (fe->xine, plugin, 0,
                                           &fe->audio_port, &fe->video_port);
#endif

                if (post && plugin_type) {
                    if (post->type != plugin_type) {
                        xine_post_dispose (fe->xine, post);
                        post = NULL;
                    }
                }

                if (post) {

                    if (!(*post_elements_num))
                        post_elements = (post_element_t **)
                            ho_malloc (sizeof (post_element_t *) * 2);
                    else
                        post_elements = (post_element_t **)
                            ho_realloc (post_elements,
                                        sizeof (post_element_t *) *
                                        ((*post_elements_num) + 2));

                    post_elements[(*post_elements_num)] = (post_element_t *)
                        ho_malloc (sizeof (post_element_t));
                    post_elements[(*post_elements_num)]->post = post;
                    post_elements[(*post_elements_num)]->name =
                        ho_strdup (plugin);
#if 1
                    post_elements[(*post_elements_num)]->args =
                        args ? ho_strdup (args) : NULL;
                    post_elements[(*post_elements_num)]->enable = 0;
#endif
                    (*post_elements_num)++;
                    post_elements[(*post_elements_num)] = NULL;

                    __pplugin_update_parameters (post, args);
                }

                ho_free (plugin);
            }
        }
    }

    return post_elements;
}

static void
pplugin_parse_and_store_post (fe_t * fe, int plugin_type,
                              const char *post_chain)
{
    post_element_t ***_post_elements;
    int *_post_elements_num;
    post_element_t **posts = NULL;
    int num;

    switch (plugin_type) {
    case XINE_POST_TYPE_VIDEO_FILTER:
        _post_elements = &fe->post_video_elements;
        _post_elements_num = &fe->post_video_elements_num;
        break;
    case XINE_POST_TYPE_VIDEO_COMPOSE:
        _post_elements = &fe->post_pip_elements;
        _post_elements_num = &fe->post_pip_elements_num;
        break;
    case XINE_POST_TYPE_AUDIO_VISUALIZATION:
        _post_elements = &fe->post_vis_elements;
        _post_elements_num = &fe->post_vis_elements_num;
        break;
    default:
        _post_elements = &fe->post_audio_elements;
        _post_elements_num = &fe->post_audio_elements_num;
        break;
    }

    if ((posts = pplugin_parse_and_load (fe, plugin_type, post_chain, &num))) {
        if (*_post_elements_num) {
            int i;
            int ptot = *_post_elements_num + num;

            *_post_elements = (post_element_t **) ho_realloc (*_post_elements,
                                                              sizeof
                                                              (post_element_t
                                                               *)
                                                              * (ptot + 1));
            for (i = *_post_elements_num; i < ptot; i++)
                (*_post_elements)[i] = posts[i - *_post_elements_num];

            (*_post_elements)[i] = NULL;
            (*_post_elements_num) += num;
        }
        else {
            *_post_elements = posts;
            *_post_elements_num = num;
        }

#ifdef DEBUG
        {
            /* dump list of all loaded plugins */
            int ptot = *_post_elements_num + num;
            int i;
            char s[4096] = "";
            for (i = 0; i < ptot; i++)
                if ((*_post_elements)[i])
                    if (((*_post_elements)[i])->post) {
                        if (((*_post_elements)[i])->enable)
                            strcat (s, "*");
                        if (((*_post_elements)[i])->name)
                            strcat (s, ((*_post_elements)[i])->name);
                        else
                            strcat (s, "<no name!>");
                        strcat (s, " ");
                    }
            debug ("    loaded plugins (type %d.%d): %s",
                   (plugin_type >> 16), (plugin_type & 0xffff), s);
        }
#endif
    }
}


void
vpplugin_parse_and_store_post (fe_t * fe, const char *post_chain)
{
    pplugin_parse_and_store_post (fe, XINE_POST_TYPE_VIDEO_FILTER,
                                  post_chain);
    pplugin_parse_and_store_post (fe, XINE_POST_TYPE_VIDEO_COMPOSE,
                                  post_chain);
}


void
applugin_parse_and_store_post (fe_t * fe, const char *post_chain)
{
    pplugin_parse_and_store_post (fe, XINE_POST_TYPE_AUDIO_FILTER,
                                  post_chain);
    pplugin_parse_and_store_post (fe, XINE_POST_TYPE_AUDIO_VISUALIZATION,
                                  post_chain);
}


static void
_vpplugin_unwire (fe_t * fe)
{
    xine_post_out_t *vo_source;

    if (fe->slave_stream)
        vo_source = xine_get_video_source (fe->slave_stream);
    else
        vo_source = xine_get_video_source (fe->stream);

    (void) xine_post_wire_video_port (vo_source, fe->video_port);
}


static void
_applugin_unwire (fe_t * fe)
{
    xine_post_out_t *ao_source;

    if (fe->slave_stream)
        ao_source = xine_get_audio_source (fe->slave_stream);
    else
        ao_source = xine_get_audio_source (fe->stream);

    (void) xine_post_wire_audio_port (ao_source, fe->audio_port);
}


static void
_vpplugin_rewire_from_post_elements (fe_t * fe,
                                     post_element_t ** post_elements,
                                     int post_elements_num)
{
    if (post_elements_num) {
        xine_post_out_t *vo_source;
        int i = 0;

        for (i = (post_elements_num - 1); i >= 0; i--) {
            const char *const *outs =
                xine_post_list_outputs (post_elements[i]->post);
            const xine_post_out_t *vo_out =
                xine_post_output (post_elements[i]->post, (char *) *outs);
            if (i == (post_elements_num - 1)) {
                debug ("        wiring %10s[out] -> [in]video_out",
                       post_elements[i]->name);
                xine_post_wire_video_port ((xine_post_out_t *) vo_out,
                                           fe->video_port);
            }
            else {
                const xine_post_in_t *vo_in;
                int err;

                /* look for standard input names */
                vo_in = xine_post_input (post_elements[i + 1]->post, "video");
                if (!vo_in)
                    vo_in =
                        xine_post_input (post_elements[i + 1]->post,
                                         "video in");

                debug ("        wiring %10s[out] -> [in]%-10s ",
                       post_elements[i]->name, post_elements[i + 1]->name);
                err = xine_post_wire ((xine_post_out_t *) vo_out,
                                      (xine_post_in_t *) vo_in);
            }
        }

        if (fe->post_pip_enable &&
            !strcmp (post_elements[0]->name, "mosaico") && fe->pip_stream) {
            vo_source = xine_get_video_source (fe->pip_stream);
            debug ("        wiring %10s[out] -> [in1]%-10s ", "pip stream",
                   post_elements[0]->name);
            xine_post_wire_video_port (vo_source,
                                       post_elements[0]->post->
                                       video_input[1]);
        }

        if (fe->slave_stream)
            vo_source = xine_get_video_source (fe->slave_stream);
        else
            vo_source = xine_get_video_source (fe->stream);
        debug ("        wiring %10s[out] -> [in]%-10s", "stream",
               post_elements[0]->name);
        xine_post_wire_video_port (vo_source,
                                   post_elements[0]->post->video_input[0]);
    }
}


static void
_applugin_rewire_from_post_elements (fe_t * fe,
                                     post_element_t ** post_elements,
                                     int post_elements_num)
{
    if (post_elements_num) {
        xine_post_out_t *ao_source;
        int i = 0;

        for (i = (post_elements_num - 1); i >= 0; i--) {
            const char *const *outs =
                xine_post_list_outputs (post_elements[i]->post);
            const xine_post_out_t *ao_out =
                xine_post_output (post_elements[i]->post, (char *) *outs);

            if (i == (post_elements_num - 1)) {
                debug ("        wiring %10s[out] -> [in]audio_out",
                       post_elements[i]->name);
                xine_post_wire_audio_port ((xine_post_out_t *) ao_out,
                                           fe->audio_port);
            }
            else {
                const xine_post_in_t *ao_in;
                int err;

                /* look for standard input names */
                ao_in = xine_post_input (post_elements[i + 1]->post, "audio");
                if (!ao_in)
                    ao_in =
                        xine_post_input (post_elements[i + 1]->post,
                                         "audio in");

                debug ("        wiring %10s[out] -> [in]%-10s ",
                       post_elements[i]->name, post_elements[i + 1]->name);
                err =
                    xine_post_wire ((xine_post_out_t *) ao_out,
                                    (xine_post_in_t *) ao_in);
            }
        }

        if (fe->slave_stream)
            ao_source = xine_get_audio_source (fe->slave_stream);
        else
            ao_source = xine_get_audio_source (fe->stream);
        debug ("        wiring %10s[out] -> [in]%-10s", "stream",
               post_elements[0]->name);
        xine_post_wire_audio_port (ao_source,
                                   post_elements[0]->post->audio_input[0]);
    }
}

static post_element_t **
_pplugin_join_deinterlace_and_post_elements (fe_t * fe,
                                             int *post_elements_num)
{
    post_element_t **post_elements;
    int i = 0,
        j = 0,
        n = 0,
        p = 0;
    static const char *order[] = { "autocrop", "tvtime", NULL };

    *post_elements_num = 0;
    if (fe->post_video_enable)
        *post_elements_num += fe->post_video_elements_num;

    if (fe->post_pip_enable)
        *post_elements_num += fe->post_pip_elements_num;

    if (*post_elements_num == 0)
        return NULL;

    post_elements = (post_element_t **)
        ho_malloc (sizeof (post_element_t *) * (*post_elements_num));

    if (fe->post_pip_enable)
        for (i = 0; i < fe->post_pip_elements_num; i++) {
            if (fe->post_pip_elements[i]->enable)
                post_elements[i + j - n] = fe->post_pip_elements[i];
            else
                n++;
        }

    if (fe->post_video_enable)
        for (j = 0; j < fe->post_video_elements_num; j++) {
            if (fe->post_video_elements[j]->enable) {
                post_elements[i + j - n] = fe->post_video_elements[j];
            }
            else
                n++;
        }

    *post_elements_num -= n;
    if (*post_elements_num == 0) {
        ho_free (post_elements);
        return NULL;
    }

    /* in some special cases order is important. By default plugin order 
     * in post plugin chain is post plugin loading order.
     * But, we want:
     *
     *   1. autocrop       - less data to process for other plugins
     *                     - accepts only YV12
     *   2. deinterlace    - blur etc. makes deinterlacing difficult.
     *                     - upscales chroma (YV12->YUY2) in some modes.
     *   3. anything else
     *
     * So let's move those two to beginning ...
     */
    n = 0;
    while (order[p]) {
        for (i = 0; i < *post_elements_num; i++)
            if (!strcmp (post_elements[i]->name, order[p])) {
                if (i != n) {
                    post_element_t *tmp = post_elements[i];
                    for (j = i; j > n; j--)
                        post_elements[j] = post_elements[j - 1];
                    post_elements[n] = tmp;
                    debug ("      moved %s to post slot %d", order[p], n);
                }
                n++;
                break;
            }
        p++;
    }

    return post_elements;
}

static post_element_t **
_pplugin_join_visualization_and_post_elements (fe_t * fe,
                                               int *post_elements_num)
{
    post_element_t **post_elements;
    int i = 0,
        j = 0,
        n = 0;

    *post_elements_num = 0;
    if (fe->post_audio_enable)
        *post_elements_num += fe->post_audio_elements_num;

    if (fe->post_vis_enable)
        *post_elements_num += fe->post_vis_elements_num;

    if (*post_elements_num == 0)
        return NULL;

    post_elements = (post_element_t **)
        ho_malloc (sizeof (post_element_t *) * (*post_elements_num));

    if (fe->post_audio_enable)
        for (j = 0; j < fe->post_audio_elements_num; j++) {
            if (fe->post_audio_elements[j]->enable)
                post_elements[i + j - n] = fe->post_audio_elements[j];
            else
                n++;
        }

    if (fe->post_vis_enable)
        for (i = 0; i < fe->post_vis_elements_num; i++) {
            if (fe->post_vis_elements[i]->enable)
                post_elements[i + j - n] = fe->post_vis_elements[i];
            else
                n++;
        }

    *post_elements_num -= n;
    if (*post_elements_num == 0) {
        ho_free (post_elements);
        return NULL;
    }

    return post_elements;
}

static void
_vpplugin_rewire (fe_t * fe)
{
    static post_element_t **post_elements;
    int post_elements_num;

    post_elements =
        _pplugin_join_deinterlace_and_post_elements (fe, &post_elements_num);

    if (post_elements) {
        _vpplugin_rewire_from_post_elements (fe, post_elements,
                                             post_elements_num);

        ho_free (post_elements);
    }
}

static void
_applugin_rewire (fe_t * fe)
{
    static post_element_t **post_elements;
    int post_elements_num;

    post_elements =
        _pplugin_join_visualization_and_post_elements (fe,
                                                       &post_elements_num);

    if (post_elements) {
        _applugin_rewire_from_post_elements (fe, post_elements,
                                             post_elements_num);

        ho_free (post_elements);
    }
}

void
vpplugin_rewire_posts (fe_t * fe)
{
    /*TRACELINE; */
    _vpplugin_unwire (fe);
    _vpplugin_rewire (fe);
}

void
applugin_rewire_posts (fe_t * fe)
{
    /*TRACELINE; */
    _applugin_unwire (fe);
    _applugin_rewire (fe);
}

static int
_pplugin_enable_post (post_plugins_t * fe, const char *name,
                      const char *args,
                      post_element_t ** post_elements,
                      int post_elements_num, int *found)
{
    int i;
    int result = 0;

    for (i = 0; i < post_elements_num; i++)
        if (post_elements[i])
            if (!strcmp (post_elements[i]->name, name)) {
                if (post_elements[i]->enable == 0) {
                    post_elements[i]->enable = 1;
                    result = 1;
                }

                *found = 1;

                if (args && *args) {
                    if (post_elements[i]->enable != 2) {
                        char *tmp = ho_strdup (args);
                        __pplugin_update_parameters (post_elements[i]->post,
                                                     tmp);
                        ho_free (tmp);
                        if (post_elements[i]->args)
                            ho_free (post_elements[i]->args);
                        post_elements[i]->args = ho_strdup (args);
                    }
                    else {
                        debug
                            ("  * enable post %s, parameters fixed in command line.",
                             name);
                        debug ("      requested: %s", args ? : "none");
                        debug ("      using    : %s",
                               post_elements[i]->args ? : "none");
                    }
                }
            }

    return result;
}

static int
_vpplugin_enable_post (post_plugins_t * fe, const char *name,
                       const char *args, int *found)
{
    int result = 0;
    if (!*found)
        result =
            _pplugin_enable_post (fe, name, args, fe->post_video_elements,
                                  fe->post_video_elements_num, found);
    if (!*found)
        result = _pplugin_enable_post (fe, name, args, fe->post_pip_elements,
                                       fe->post_pip_elements_num, found);
    return result;
}

static int
_applugin_enable_post (post_plugins_t * fe, const char *name,
                       const char *args, int *found)
{
    int result = 0;
    if (!*found)
        result =
            _pplugin_enable_post (fe, name, args, fe->post_audio_elements,
                                  fe->post_audio_elements_num, found);
    if (!*found)
        result = _pplugin_enable_post (fe, name, args, fe->post_vis_elements,
                                       fe->post_vis_elements_num, found);
    return result;
}

static char *
_pp_name_ho_strdup (const char *initstr)
{
    char *name = ho_strdup (initstr),
        *pt;

    if (NULL != (pt = strchr (name, ':')))
        *pt = 0;

    return name;
}

static const char *
_pp_args (const char *initstr)
{
    char *pt = strchr (initstr, ':');
    if (pt && *(pt + 1))
        return pt + 1;
    return NULL;
}

int
vpplugin_enable_post (post_plugins_t * fe, const char *initstr, int *found)
{
    char *name = _pp_name_ho_strdup (initstr);
    const char *args = _pp_args (initstr);

    int result = _vpplugin_enable_post (fe, name, args, found);

    debug ("  * enable post %s --> %s, %s", name,
           *found ? "found" : "not found", result ? "enabled" : "no action");

    if (!*found) {
        debug ("  * loading post %s", initstr);
        vpplugin_parse_and_store_post (fe, initstr);
        result = _vpplugin_enable_post (fe, name, NULL, found);
        debug ("  * enable post %s --> %s, %s", name,
               *found ? "found" : "not found",
               result ? "enabled" : "no action");
    }

    if (result)
        _vpplugin_unwire (fe);

    ho_free (name);
    return result;
}

int
applugin_enable_post (post_plugins_t * fe, const char *initstr, int *found)
{
    const char *args = _pp_args (initstr);
    char *name = _pp_name_ho_strdup (initstr);

    int result = _applugin_enable_post (fe, name, args, found);

    debug ("  * enable post %s --> %s, %s", name,
           *found ? "found" : "not found", result ? "enabled" : "no action");

    if (!*found) {
        debug ("  * loading post %s", initstr);
        applugin_parse_and_store_post (fe, initstr);
        result = _applugin_enable_post (fe, name, NULL, found);
        debug ("  * enable post %s --> %s, %s", name,
               *found ? "found" : "not found",
               result ? "enabled" : "no action");
    }

    if (result)
        _applugin_unwire (fe);

    ho_free (name);
    return result;
}

static int
_pplugin_disable_post (post_plugins_t * fe, const char *name,
                       post_element_t ** post_elements, int post_elements_num)
{
    int i,
      result = 0;
    /*TRACELINE; */
    if (post_elements)
        for (i = 0; i < post_elements_num; i++)
            if (post_elements[i])
                if (!name || !strcmp (post_elements[i]->name, name))
                    if (post_elements[i]->enable == 1) {
                        post_elements[i]->enable = 0;
                        result = 1;
                    }
    return result;
}

int
vpplugin_disable_post (post_plugins_t * fe, const char *name)
{
    /*TRACELINE; */
    if (_pplugin_disable_post (fe, name, fe->post_video_elements,
                               fe->post_video_elements_num) ||
        _pplugin_disable_post (fe, name, fe->post_pip_elements,
                               fe->post_pip_elements_num)) {
        _vpplugin_unwire (fe);
        return 1;
    }
    return 0;
}

int
applugin_disable_post (post_plugins_t * fe, const char *name)
{
    /*TRACELINE; */
    if (_pplugin_disable_post (fe, name, fe->post_audio_elements,
                               fe->post_audio_elements_num) ||
        _pplugin_disable_post (fe, name, fe->post_vis_elements,
                               fe->post_vis_elements_num)) {
        _applugin_unwire (fe);
        return 1;
    }
    return 0;
}

static int
_pplugin_unload_post (post_plugins_t * fe, const char *name,
                      post_element_t *** post_elements,
                      int *post_elements_num)
{
    /* does not unwrire plugins ! */
    int i,
      j,
      result = 0;
    /*TRACELINE; */

    if (!*post_elements || !*post_elements_num)
        return 0;

    for (i = 0; i < *post_elements_num; i++) {
        if ((*post_elements)[i]) {
            if (!name || !strcmp ((*post_elements)[i]->name, name)) {

                if ((*post_elements)[i]->enable == 0 || !name) {
                    xine_post_dispose (fe->xine, (*post_elements)[i]->post);

                    ho_free ((*post_elements)[i]->name);

                    if ((*post_elements)[i]->args)
                        ho_free ((*post_elements)[i]->args);

                    ho_free ((*post_elements)[i]);

                    for (j = i; j < *post_elements_num - 1; j++)
                        (*post_elements)[j] = (*post_elements)[j + 1];

                    (*post_elements_num)--;
                    (*post_elements)[(*post_elements_num)] = 0;

                    result = 1;
                    i--;

                }
                else {
                    debug ("Unload %s failed: plugin enabled and in use",
                           (*post_elements)[i]->name);
                }
            }
        }
    }

    if (*post_elements_num <= 0) {
        if (*post_elements)
            ho_free (*post_elements);
        *post_elements = NULL;
    }

    return result;
}

int
vpplugin_unload_post (post_plugins_t * fe, const char *name)
{
    int result = vpplugin_disable_post (fe, name);

    /* unload already disabled plugins too (result=0) */
    _pplugin_unload_post (fe, name, &fe->post_video_elements,
                          &fe->post_video_elements_num);
    _pplugin_unload_post (fe, name, &fe->post_pip_elements,
                          &fe->post_pip_elements_num);

    /* result indicates only unwiring condition, not unload result */
    return result;
}

int
applugin_unload_post (post_plugins_t * fe, const char *name)
{
    int result = applugin_disable_post (fe, name);

    /* unload already disabled plugins too (result=0) */
    _pplugin_unload_post (fe, name, &fe->post_audio_elements,
                          &fe->post_audio_elements_num);
    _pplugin_unload_post (fe, name, &fe->post_vis_elements,
                          &fe->post_vis_elements_num);

    /* result indicates only unwiring condition, not unload result */
    return result;
}


/* ************************************************************************* */

/* ************************************************************************* */


#if 0
static char *
odk_post_plugin_type_get_name (int plugin_type)
{
    switch (plugin_type) {
    case XINE_POST_TYPE_VIDEO_FILTER:
        return _("video filter");
    case XINE_POST_TYPE_VIDEO_VISUALIZATION:
        return _("video visualization");
    case XINE_POST_TYPE_VIDEO_COMPOSE:
        return _("video compose");
    case XINE_POST_TYPE_AUDIO_FILTER:
        return _("audio filter");
    case XINE_POST_TYPE_AUDIO_VISUALIZATION:
        return _("audio visualization");
    default:
        break;
    }

    return _("unknown post plugin type");
}
#endif


void
odk_post_free (odk_t * odk)
{
    ho_free (odk->postplugins);
}


void
odk_post_init (odk_t * odk)
{
    int plugin_type;
    const char *const *plugin;

    post_plugins_t *posts = ho_new (post_plugins_t);
    memset (posts, 0, sizeof (post_plugins_t));
    posts->xine = odk->xine;
    posts->audio_port = odk->audio_port;
    posts->video_port = odk->win->video_port;
    posts->stream = odk->main_stream;
    odk->postplugins = posts;

    /* Print information about available post plugins. */
#if 0
    int plugin_types[] = {
        XINE_POST_TYPE_VIDEO_FILTER,
        XINE_POST_TYPE_VIDEO_VISUALIZATION,
        XINE_POST_TYPE_VIDEO_COMPOSE,
        XINE_POST_TYPE_AUDIO_FILTER,
        XINE_POST_TYPE_AUDIO_VISUALIZATION
    };
    int p = 0;
    for (p = 0; p < 5; p++) {
        plugin_type = plugin_types[p];
        debug ("post plugins (%s)",
               odk_post_plugin_type_get_name (plugin_types[p]));
        if ((plugin = xine_list_post_plugins_typed (odk->xine, plugin_type))) {
            int i = 0;
            for (i = 0; plugin[i]; i++) {
                debug ("  plugin[%d] = '%s'", i, plugin[i]);
            }
        }
    }
#endif

    /* Audio visualization post plugin */
    plugin_type = XINE_POST_TYPE_AUDIO_VISUALIZATION;
    if ((plugin = xine_list_post_plugins_typed (odk->xine, plugin_type))) {
        int i = 0;
        for (i = 0; plugin[i]; i++);

        char **plugins = (char **) ho_malloc (sizeof (char **) * (i + 1));
        for (i = 0; plugin[i]; i++) {
            plugins[i] = ho_strdup (plugin[i]);
        }
        plugins[i] = NULL;

        config_register_enum ("visual_anim.post_plugin", 0, plugins,
                              _("post plugin type"));

        for (i = 0; plugins[i]; i++) {
            ho_free (plugins[i]);
        }
        ho_free (plugins);
    }

    /* Video deinterlacer post plugin */
#if 0
    {
        xine_config_register_string (odk->xine, "gui.deinterlace_plugin",
                                     DEFAULT_DEINTERLACER,
                                     _("Deinterlace plugin."),
                                     _("Plugin (with optional parameters) to "
                                       "use when deinterlace is used (plugin "
                                       "separator is ';')."), 10, NULL, NULL);
    }
#endif

    odk->novideo_post = NULL;
}


static void
pplugin_audio_unwire (odk_t * odk)
{
    xine_audio_port_t *audio_port = odk->audio_port;
    xine_post_out_t *out = xine_get_audio_source (odk->main_stream);
    xine_post_wire_audio_port (out, audio_port);
}


#if 0
static void
pplugin_video_unwire (odk_t * odk)
{
    xine_post_out_t *out = xine_get_video_source (odk->stream);
    xine_post_wire_video_port (out, odk->win->video_port);
}
#endif


void
odk_post_audio_rewire (odk_t * odk)
{
    xine_cfg_entry_t centry;
    xine_config_lookup_entry (odk->xine, "visual_anim.post_plugin", &centry);
    const char *plugin_name = centry.enum_values[centry.num_value];

    xine_audio_port_t *audio_port = odk->audio_port;
    xine_video_port_t *video_port = odk->win->video_port;

    xine_post_t *plugin = xine_post_init (odk->xine, plugin_name, 0,
                                          &audio_port, &video_port);

    if (!plugin) {
        error (_("Failed to initialize post plugin '%s'!"), plugin_name);
        return;
    }

    xine_post_out_t *output = xine_get_audio_source (odk->main_stream);
    if (!output) {
        xine_post_dispose (odk->xine, plugin);
        error (_("Failed to retrieve audio output from current stream!"));
        return;
    }
    xine_post_in_t *input = xine_post_input (plugin, "audio in");
    if (!input) {
        xine_post_dispose (odk->xine, plugin);
        error (_("Failed to retrieve post plugin input '%s'!"));
        return;
    }
    if (!xine_post_wire (output, input)) {
        xine_post_dispose (odk->xine, plugin);
        error (_("Failed to rewire post plugin!"));
        return;
    }

    odk->novideo_post = plugin;
}


void
odk_post_audio_unwire (odk_t * odk)
{
    assert (odk->novideo_post);

    pplugin_audio_unwire (odk);
    xine_post_dispose (odk->xine, odk->novideo_post);

    odk->novideo_post = NULL;
}
