
/*
 * Copyright (C) 2002-2003 Stefan Holst
 * Copyright (C) 2004-2005 Maximilian Schwerin
 *
 * This file is part of oxine a free media player.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: otk_list.c 4022 2010-01-12 14:20:53Z mschwerin $
 */

#include "config.h"

#include <ctype.h>

#include "heap.h"
#include "i18n.h"
#include "list.h"
#include "logger.h"
#include "odk.h"
#include "odk_xk.h"
#include "otk.h"
#include "oxine.h"
#include "utils.h"
#include "xmlparser.h"

extern oxine_t *oxine;

typedef struct otk_listentry_s otk_listentry_t;

typedef enum {
    OTK_LIST_TYPE_NORMAL,
    OTK_LIST_TYPE_INFO,
    OTK_LIST_TYPE_MENU,
    OTK_LIST_TYPE_SELECTOR,
} otk_list_type_t;


typedef struct {
    int x;
    int w;
    int alignment;
    int fontsize;
    char *font;
    char *content;
} otk_list_column_t;


typedef struct {
    otk_widget_t widget;

    /// The scrollbar associated with this list.
    otk_widget_t *scrollbar;
    otk_widget_t *button_up;
    otk_widget_t *button_down;

    /// The entries of this list (otk_listentry_t).
    l_list_t *entries;

    /// The last listentry that was selected alone.
    otk_listentry_t *last_single_selected;

    /// The last listentry to have the focus in this list.
    otk_listentry_t *last_with_focus;

    /// Is it necessary to adapt the list entries.
    bool need_adaptation;

    /// What type of list is this.
    otk_list_type_t type;

    /// Number of selected entries.
    int num_selected;
    /// Number of visible entries.
    int num_visible;
    /// Number of the entry at the top of the visible list.
    int first_visible;

    /// How may the user select an item.
    otk_list_selection_t allow_select;

    int entry_x;
    int entry_y;
    int entry_height;
    int entry_width;
    int entry_spacing;

    /// Userdefined data that is sent in every callback.
    void *cb_data;

    /**
     * Callback that is called whenever the visible entry changes. This is
     * only used by the selector widget.
     */
    otk_int_set_cb_t entry_changed_cb;

    /// User-defined data for the above callback.
    void *entry_changed_cb_data;
} otk_list_t;


struct otk_listentry_s {
    otk_widget_t widget;

    otk_list_t *list;

    bool is_first;
    bool is_last;
    bool is_selected;

    /// Columns of this entry.
    l_list_t *columns;

    /// Index character of this entry.
    char index_char;

    /// Time this item was last clicked.
    time_t doubleclicktime;

    /**
     * This callback is called whenever the list receives an activation event.
     * Activation events are OXINE_KEY_ACTIVATE or a double-click
     * (OXINE_MOUSE_BUTTON_LEFT twice).
     */
    otk_cb_t activate_cb;

    /// User-defined data for the above callback.
    void *activate_cb_data;

    /**
     * This callback is called whenever the list receives a selection event.
     * Selection events are OXINE_KEY_SELECT or a single-click
     * (OXINE_MOUSE_BUTTON_LEFT once).
     */
    otk_cb_t select_cb;

    /// User-defined data for the above callback.
    void *select_cb_data;

    /**
     * This callback is called whenever the list receives a remove event. The
     * remove event is OXINE_KEY_REMOVE. Apart from calling the callback
     * nothing is done. The user is responsible for removing the entry.
     */
    otk_cb_t remove_cb;

    /// User-defined data for the above callback.
    void *remove_cb_data;

    /**
     * This callback is called whenever the list receives a copy event. The
     * copy event is OXINE_KEY_COPY. Apart from calling the callback nothing
     * is done. The user is responsible for removing the entry.
     */
    otk_cb_t copy_cb;

    /// User-defined data for the above callback.
    void *copy_cb_data;

    /**
     * This callback is called whenever the list receives a cut event. The
     * cut event is OXINE_KEY_CUT. Apart from calling the callback nothing
     * is done. The user is responsible for removing the entry.
     */
    otk_cb_t cut_cb;

    /// User-defined data for the above callback.
    void *cut_cb_data;

    /**
     * This callback is called whenever the list receives a paste event. The
     * paste event is OXINE_KEY_PASTE. Apart from calling the callback nothing
     * is done. The user is responsible for removing the entry.
     */
    otk_cb_t paste_cb;

    /// User-defined data for the above callback.
    void *paste_cb_data;
};


static void listentries_adapt (otk_list_t * list);


static void
list_page_down (void *data)
{
    otk_list_t *list = (otk_list_t *) data;

    list->first_visible += list->num_visible;
    list->need_adaptation = true;
    otk_draw_needy (list->widget.otk);

    if (list->entry_changed_cb) {
        list->entry_changed_cb (list->entry_changed_cb_data,
                                list->first_visible);
    }
}


static void
list_page_up (void *data)
{
    otk_list_t *list = (otk_list_t *) data;

    list->first_visible -= list->num_visible;
    list->need_adaptation = true;
    otk_draw_needy (list->widget.otk);

    if (list->entry_changed_cb) {
        list->entry_changed_cb (list->entry_changed_cb_data,
                                list->first_visible);
    }
}


static void
list_scroll_down (void *data)
{
    otk_list_t *list = (otk_list_t *) data;

    list->first_visible++;
    list->need_adaptation = true;
    otk_draw_needy (list->widget.otk);

    if (list->entry_changed_cb) {
        list->entry_changed_cb (list->entry_changed_cb_data,
                                list->first_visible);
    }
}


static void
list_scroll_up (void *data)
{
    otk_list_t *list = (otk_list_t *) data;

    list->first_visible--;
    list->need_adaptation = true;
    otk_draw_needy (list->widget.otk);

    if (list->entry_changed_cb) {
        list->entry_changed_cb (list->entry_changed_cb_data,
                                list->first_visible);
    }
}


static int
list_set_position (void *data, int position)
{
    otk_list_t *list = (otk_list_t *) data;

    list->first_visible = position * l_list_length (list->entries) / 100;
    list->need_adaptation = true;
    otk_draw_needy (list->widget.otk);

    return position;
}


static void
listentry_column_free (void *p)
{
    otk_list_column_t *c = p;

    ho_free (c->font);
    ho_free (c->content);
    ho_free (c);
}


static void
listentry_destroy (otk_widget_t * this)
{
    otk_listentry_t *listentry = (otk_listentry_t *) this;

    if (!otk_widget_is_correct (this, OTK_WIDGET_LISTENTRY)) {
        return;
    }

    l_list_free (listentry->columns, listentry_column_free);
    ho_free (listentry);
}


static void
listentry_draw (otk_widget_t * this)
{
    otk_listentry_t *entry = (otk_listentry_t *) this;

    if (!otk_widget_is_correct (this, OTK_WIDGET_LISTENTRY)) {
        return;
    }

    otk_list_t *list = entry->list;
    assert (!list->need_adaptation);

    if (this->is_focused) {
        list->last_with_focus = entry;
    }

    int palette = 0;
    if (!this->is_enabled) {
        palette = otk_get_palette (this->otk, OTK_PALETTE_BUTTON_DISABLED);
    }
    else if (this->is_focused) {
        palette = otk_get_palette (this->otk, OTK_PALETTE_BUTTON_FOCUSED);
    }
    else {
        palette = otk_get_palette (this->otk, OTK_PALETTE_BUTTON);
    }

    bool draw_bg = false;
    draw_bg |= (list->type == OTK_LIST_TYPE_MENU);
    draw_bg |= this->is_focused;
    draw_bg |= !this->is_enabled;
    draw_bg |= entry->is_selected;

    if (draw_bg) {
        int background_color = palette + OSD_TEXT_PALETTE_BACKGROUND;
        odk_draw_rect (this->odk, this->x, this->y, this->w,
                       this->h, 5, background_color, true);
    }
    else {
        int background_color = palette + OSD_TEXT_PALETTE_TRANSPARENT;
        odk_draw_rect (this->odk, this->x, this->y, this->w,
                       this->h, 5, background_color, true);
    }

    otk_list_column_t *column = l_list_first (entry->columns);
    while (column) {
        int x = column->x;
        int y = this->y + this->h / 2;
        int w = column->w;
        int a = column->alignment;

        int s = column->fontsize;
        char *f = column->font ? column->font : entry->widget.font;

        char *t = NULL;
        char *c = column->content;

#if 0
        odk_draw_rect (this->odk, x, this->y, w, this->h, 0,
                       palette + OSD_TEXT_PALETTE_FOREGROUND, false);
#endif

        if (str_equals (c, "*--*")) {
            int y = this->y + this->h / 2;
            int w = this->w / 2;
            int x = this->x + (w / 2);

            odk_draw_line (this->odk, x, y, x + w, y, 1,
                           palette + OSD_TEXT_PALETTE_FOREGROUND);
        }
        else {
            if (a & OTK_ALIGN_RIGHT) {
                x += w;
            }
            else if (a & OTK_ALIGN_CENTER) {
                x += w / 2;
            }

            odk_osd_set_font (this->odk, f, s);
            t = otk_trunc_text_to_width (this->otk, c, w);
            odk_draw_text (this->odk, x, y, t, a, palette);
            ho_free (t);
        }

        column = l_list_next (entry->columns, column);
    }
}


static void
listentry_select (otk_listentry_t * listentry, int kbd_modifier)
{
    otk_list_t *list = listentry->list;

    listentry->widget.need_repaint = true;
    listentry->doubleclicktime = time (NULL);
    int was_selected = listentry->is_selected;

    /* We only care about Ctrl and Shift */
    kbd_modifier &= ControlMask | ShiftMask;

    if (list->allow_select == OTK_LIST_SELECTION_SINGLE) {
        otk_list_set_selected_none ((otk_widget_t *) list);

        if (kbd_modifier & ControlMask) {
            listentry->is_selected = !was_selected;
            list->num_selected = was_selected ? 0 : 1;
        }
        else {
            listentry->is_selected = true;
            list->num_selected = 1;
        }
        list->last_single_selected = listentry;
    }

    else if (list->allow_select == OTK_LIST_SELECTION_MULTIPLE) {
        if (!config_get_bool ("misc.list_select_ctrl_shift")
            || (kbd_modifier & ControlMask)) {
            listentry->is_selected = !was_selected;
            list->num_selected += was_selected ? -1 : 1;
        }

        else if (kbd_modifier & ShiftMask) {
            bool selected;

            listentry->is_selected = true;

            selected = false;
            otk_listentry_t *entry = l_list_first (list->entries);
            while (entry && (entry != listentry)) {
                if (entry == list->last_single_selected)
                    selected = true;
                entry->is_selected = selected;
                entry = l_list_next (list->entries, entry);
            }

            selected = false;
            entry = l_list_last (list->entries);
            while (entry && (entry != listentry)) {
                if (entry == list->last_single_selected)
                    selected = true;
                entry->is_selected = selected;
                entry = l_list_prev (list->entries, entry);
            }

            list->num_selected = 0;
            entry = l_list_first (list->entries);
            while (entry) {
                if (entry->is_selected)
                    list->num_selected++;
                entry = l_list_next (list->entries, entry);
            }
        }

        else {
            otk_list_set_selected_none ((otk_widget_t *) list);
            listentry->is_selected = true;
            list->num_selected = 1;
            list->last_single_selected = listentry;
        }
    }

    if (listentry->select_cb) {
        listentry->select_cb (listentry->select_cb_data);
    }

    otk_draw_needy (list->widget.otk);
}


static void
listentry_activate (otk_listentry_t * listentry)
{
    listentry->doubleclicktime = 0;

    otk_list_set_selected_none ((otk_widget_t *) listentry->list);

    if (listentry->activate_cb) {
        listentry->activate_cb (listentry->activate_cb_data);
    }
}

static otk_widget_t *
_widget_by_index_char (otk_list_t * list, int c, bool search)
{
    c = toupper (c);

    bool good = false;
    good |= (('A' <= c) && (c <= 'Z'));
    good |= (('0' <= c) && (c <= '9'));
    if (!good) {
        return NULL;
    }

    int position = 0;
    otk_widget_t *w = l_list_first (list->entries);
    while (w) {
        otk_listentry_t *entry = (otk_listentry_t *) w;
        if (w->is_enabled && (entry->index_char == c)) {
            list->first_visible = position - list->num_visible / 2;
            list->need_adaptation = true;
            list->last_with_focus = entry;
            return w;
        }
        w = l_list_next (list->entries, w);
        position++;
    }

    if (!search) {
        return NULL;
    }

    int d = 1;
    for (d = 1; (c - d >= 'A') || (c + d <= 'Z'); d++) {
        if ((w = _widget_by_index_char (list, c - d, false))) {
            return w;
        }
        if ((w = _widget_by_index_char (list, c + d, false))) {
            return w;
        }
    }

    return NULL;
}


static otk_widget_t *
widget_by_index_char (otk_list_t * list, int c)
{
    return _widget_by_index_char (list, c, true);
}


otk_widget_t *
otk_list_focus_by_index_char (otk_widget_t * this, int c)
{
    otk_list_t *list = (otk_list_t *) this;

    if (!otk_widget_is_correct (this, OTK_WIDGET_LIST)) {
        return NULL;
    }

    return widget_by_index_char (list, c);
}

static otk_widget_t *
widget_by_number (otk_list_t * list, otk_listentry_t * cur, char *cs)
{
    otk_widget_t *new_widget = NULL;

    int i = 0;
    int c = 0;
    int s = strlen (cs);

    for (c = 0; cs[c + 1] != '\0'; c++) {
        if (cur->index_char == cs[c]) {
            break;
        }
    }

    debug ("%d %c", c, cs[c]);

    for (i = (c + 1) % s; (i != c) && !new_widget; i = (i + 1) % s) {
        debug ("%d %c", i, cs[i]);
        new_widget = _widget_by_index_char (list, cs[i], false);
    }

    if (!new_widget) {
        new_widget = (otk_widget_t *)cur;
    }

    return new_widget;
}


static void
listentry_key_handler (otk_widget_t * this, oxine_event_t * ev)
{
    otk_listentry_t *entry = (otk_listentry_t *) this;
    otk_list_t *list = entry ? entry->list : NULL;

    if (!otk_widget_is_correct (this, OTK_WIDGET_LISTENTRY)) {
        return;
    }

    oxine_key_id_t saved_key = ev->source.key;
    ev->source.key = OXINE_KEY_NULL;

    otk_widget_t *new_widget = NULL;

    switch (saved_key) {
    case OXINE_KEY_FIRST:
        list->first_visible = 0;
        listentries_adapt (list);

        new_widget = l_list_first (list->entries);
        while (new_widget
               && !(new_widget->is_visible && new_widget->is_enabled)) {
            new_widget = l_list_next (list->entries, new_widget);
        }

        if (list->entry_changed_cb) {
            list->entry_changed_cb (list->entry_changed_cb_data,
                                    list->first_visible);
        }
        break;
    case OXINE_KEY_LAST:
        list->first_visible = l_list_length (list->entries);
        listentries_adapt (list);

        new_widget = l_list_last (list->entries);
        while (new_widget
               && !(new_widget->is_visible && new_widget->is_enabled)) {
            new_widget = l_list_prev (list->entries, new_widget);
        }

        if (list->entry_changed_cb) {
            list->entry_changed_cb (list->entry_changed_cb_data,
                                    list->first_visible);
        }
        break;
    case OXINE_KEY_PAGE_UP:
        list->first_visible -= list->num_visible;
        listentries_adapt (list);

        new_widget = this;
        while (!((otk_listentry_t *) new_widget)->is_first) {
            new_widget = l_list_prev (list->entries, new_widget);
        }

        while (new_widget
               && !(new_widget->is_visible && new_widget->is_enabled)) {
            new_widget = l_list_next (list->entries, new_widget);
        }

        if (list->entry_changed_cb) {
            list->entry_changed_cb (list->entry_changed_cb_data,
                                    list->first_visible);
        }
        break;
    case OXINE_KEY_PAGE_DOWN:
        list->first_visible += list->num_visible;
        listentries_adapt (list);

        new_widget = this;
        while (!((otk_listentry_t *) new_widget)->is_last) {
            new_widget = l_list_next (list->entries, new_widget);
        }

        while (new_widget
               && !(new_widget->is_visible && new_widget->is_enabled)) {
            new_widget = l_list_prev (list->entries, new_widget);
        }

        if (list->entry_changed_cb) {
            list->entry_changed_cb (list->entry_changed_cb_data,
                                    list->first_visible);
        }
        break;
    case OXINE_KEY_UP:
        /* Search backwards in the list until we find an entry, that is
         * visible and enabled. */
        {
            otk_listentry_t *new_entry = entry;
            do {
                if (new_entry->is_first) {
                    list->first_visible--;
                    listentries_adapt (list);
                }

                new_entry = l_list_prev (list->entries, new_entry);
            } while (new_entry && (!(new_entry->widget.is_visible)
                                   || !(new_entry->widget.is_enabled)));
            new_widget = (otk_widget_t *) new_entry;

            if (new_entry && (new_entry != entry) && list->entry_changed_cb) {
                list->entry_changed_cb (list->entry_changed_cb_data,
                                        list->first_visible);
            }
        }
        break;
    case OXINE_KEY_DOWN:
        /* Search forewards in the list until we find an entry, that is
         * visible and enabled. */
        {
            otk_listentry_t *new_entry = entry;
            do {
                if (new_entry->is_last) {
                    list->first_visible++;
                    listentries_adapt (list);
                }

                new_entry = l_list_next (list->entries, new_entry);
            } while (new_entry && (!(new_entry->widget.is_visible)
                                   || !(new_entry->widget.is_enabled)));
            new_widget = (otk_widget_t *) new_entry;

            if (new_entry && (new_entry != entry) && list->entry_changed_cb) {
                list->entry_changed_cb (list->entry_changed_cb_data,
                                        list->first_visible);
            }
        }
        break;
    case OXINE_KEY_SELECT:
        listentry_select (entry, ev->data.keyboard.modifier);
        break;
    case OXINE_KEY_SELECT_ALL:
        otk_list_set_selected_all (&list->widget);
        break;
    case OXINE_KEY_ACTIVATE:
        listentry_activate (entry);
        break;
    case OXINE_KEY_REMOVE:
        if (entry->remove_cb) {
            entry->remove_cb (entry->remove_cb_data);
        }
        break;
    case OXINE_KEY_COPY:
        if (entry->copy_cb) {
            entry->copy_cb (entry->copy_cb_data);
        }
        break;
    case OXINE_KEY_CUT:
        if (entry->cut_cb) {
            entry->cut_cb (entry->cut_cb_data);
        }
        break;
    case OXINE_KEY_PASTE:
        if (entry->paste_cb) {
            entry->paste_cb (entry->paste_cb_data);
        }
        break;
    case OXINE_KEY_INPUT:
        if (!(ev->data.keyboard.modifier & ControlMask)
            && !(ev->data.keyboard.modifier & Mod1Mask)) {
            int key = toupper (ev->data.keyboard.key);
            new_widget = widget_by_index_char (list, key);
        }
        break;
    case OXINE_KEY_2:
        new_widget = widget_by_number (list, entry, "ABC");
        break;
    case OXINE_KEY_3:
        new_widget = widget_by_number (list, entry, "DEF");
        break;
    case OXINE_KEY_4:
        new_widget = widget_by_number (list, entry, "GHI");
        break;
    case OXINE_KEY_5:
        new_widget = widget_by_number (list, entry, "JKL");
        break;
    case OXINE_KEY_6:
        new_widget = widget_by_number (list, entry, "MNO");
        break;
    case OXINE_KEY_7:
        new_widget = widget_by_number (list, entry, "PQRS");
        break;
    case OXINE_KEY_8:
        new_widget = widget_by_number (list, entry, "TUV");
        break;
    case OXINE_KEY_9:
        new_widget = widget_by_number (list, entry, "WXYZ");
        break;
    default:
        /* If we did not use (consume) this event here, we restore it for
         * others to use. */
        ev->source.key = saved_key;
        break;
    }

    if (list->need_adaptation) {
        listentries_adapt (list);
    }
    if (new_widget) {
        otk_widget_set_focused (new_widget, true);
        otk_draw_needy (this->otk);
    }
}


static void
listentry_button_handler (otk_widget_t * this, oxine_event_t * ev)
{
    otk_listentry_t *entry = (otk_listentry_t *) this;

    if (!otk_widget_is_correct (this, OTK_WIDGET_LISTENTRY)) {
        return;
    }

    otk_list_t *list = entry->list;
    oxine_button_id_t saved_button = ev->source.button;
    ev->source.button = OXINE_MOUSE_BUTTON_NULL;

    switch (saved_button) {
    case OXINE_MOUSE_BUTTON_LEFT:
        if (time (NULL) - 1 < entry->doubleclicktime) {
            listentry_activate (entry);
        }
        else {
            listentry_select (entry, ev->data.mouse.button.modifier);
        }
        break;
    case OXINE_MOUSE_SCROLLWHEEL_UP:
        list->first_visible--;
        list->need_adaptation = true;
        otk_draw_needy (this->otk);

        if (list->entry_changed_cb) {
            list->entry_changed_cb (list->entry_changed_cb_data,
                                    list->first_visible);
        }
        break;
    case OXINE_MOUSE_SCROLLWHEEL_DOWN:
        list->first_visible++;
        list->need_adaptation = true;
        otk_draw_needy (this->otk);

        if (list->entry_changed_cb) {
            list->entry_changed_cb (list->entry_changed_cb_data,
                                    list->first_visible);
        }
        break;
    default:
        /* If we did not use this event we restore the event. */
        ev->source.button = saved_button;
        break;
    }
}


otk_widget_t *
otk_listentry_new (otk_widget_t * this, const char *text,
                   otk_cb_t activate_cb, void *activate_cb_data,
                   otk_cb_t select_cb, void *select_cb_data,
                   otk_cb_t remove_cb, void *remove_cb_data,
                   otk_cb_t copy_cb, void *copy_cb_data,
                   otk_cb_t cut_cb, void *cut_cb_data,
                   otk_cb_t paste_cb, void *paste_cb_data)
{
    otk_list_t *list = (otk_list_t *) this;

    if (!otk_widget_is_correct (this, OTK_WIDGET_LIST)) {
        return NULL;
    }

    otk_listentry_t *entry = ho_new (otk_listentry_t);

    otk_widget_constructor ((otk_widget_t *) entry, this->otk,
                            OTK_WIDGET_LISTENTRY, list->entry_x, 0,
                            list->entry_width, list->entry_height);

    entry->widget.selectable = OTK_SELECTABLE_MOUSE;

    entry->widget.draw = listentry_draw;
    entry->widget.destroy = listentry_destroy;

    entry->widget.button_handler = listentry_button_handler;
    entry->widget.key_handler = listentry_key_handler;

    ho_free (entry->widget.font);
    entry->widget.font = ho_strdup (list->widget.font);
    entry->widget.fontsize = list->widget.fontsize;
    entry->widget.alignment = list->widget.alignment;
    entry->widget.alignment &= ~OTK_ALIGN_TOP;
    entry->widget.alignment &= ~OTK_ALIGN_BOTTOM;
    entry->widget.alignment |= OTK_ALIGN_VCENTER;

    entry->list = list;
    entry->is_first = false;
    entry->is_last = false;
    entry->is_selected = false;
    entry->activate_cb_data = activate_cb_data;
    entry->activate_cb = activate_cb;
    entry->select_cb_data = select_cb_data;
    entry->select_cb = select_cb;
    entry->remove_cb_data = remove_cb_data;
    entry->remove_cb = remove_cb;
    entry->copy_cb_data = copy_cb_data;
    entry->copy_cb = copy_cb;
    entry->cut_cb_data = cut_cb_data;
    entry->cut_cb = cut_cb;
    entry->paste_cb_data = paste_cb_data;
    entry->paste_cb = paste_cb;

    if (!text) {
        entry->index_char = '\0';
    }
    /* Mmmh, this is a bit hacky and actually does not belong in a
     * library but is the easiest way to do this. */
    else if (str_equals (text, _("Play all titles in this folder "
                                 "and all subfolders..."))) {
        entry->index_char = '\0';
    }
    else if (str_equals (text, _("Play all titles in this folder..."))) {
        entry->index_char = '\0';
    }
    else if (strc_starts_with (text, "[] ")) {
        entry->index_char = toupper (text[3]);
    }
    else {
        entry->index_char = toupper (text[0]);
    }

    int c = 0;
    int x = list->entry_x + 5;
    int w = list->entry_width - 10;

    entry->columns = l_list_new ();
    if (strstr (text, "<tr>")) {
        xml_node_t *nline;
        xml_node_t *ncolumn;
        xml_parser_init (text, strlen (text), XML_PARSER_CASE_INSENSITIVE);
        assert (xml_parser_build_tree (&nline) == XML_PARSER_OK);

        ncolumn = nline->child;
        while (ncolumn) {
            if (strcasecmp (ncolumn->name, "td") == 0) {
                otk_list_column_t *column = ho_new (otk_list_column_t);
                l_list_append (entry->columns, column);

                column->alignment = OTK_ALIGN_LEFT;
                column->content = ho_strdup (ncolumn->data);
                column->font = NULL;
                column->fontsize = entry->widget.fontsize;

                const char *pw = xml_parser_get_property (ncolumn, "width");
                if (pw) {
                    column->w = atoi (pw);
                }
                else {
                    column->w = 0;
                    c++;
                }

                const char *pa = xml_parser_get_property (ncolumn, "align");
                if (pa && (strcasecmp (pa, "right") == 0)) {
                    column->alignment = OTK_ALIGN_RIGHT;
                }
                else if (pa && (strcasecmp (pa, "center") == 0)) {
                    column->alignment = OTK_ALIGN_CENTER;
                }

                const char *pf = xml_parser_get_property (ncolumn, "font");
                if (pf) {
                    column->font = ho_strdup (pf);
                }

                const char *ps =
                    xml_parser_get_property (ncolumn, "fontsize");
                if (ps) {
                    column->fontsize = atoi (ps);
                }

                w -= column->w;
            }
            ncolumn = ncolumn->next;
        }
        xml_parser_free_tree (nline);
    }

    /* If no column was found we assume that there is only one column
     * containing the complete text. */
    if (l_list_length (entry->columns) == 0) {
        otk_list_column_t *column = ho_new (otk_list_column_t);
        l_list_append (entry->columns, column);
        column->alignment = entry->widget.alignment;
        column->content = ho_strdup (text);
        column->font = NULL;
        column->fontsize = entry->widget.fontsize;
        column->w = 0;
        c++;
    }

    /* Split the remaining width evenly over the columns with automatic
     * width and update x-position and alignment. */
    otk_list_column_t *column = l_list_first (entry->columns);
    while (column) {
        column->x = x;
        column->alignment |= OTK_ALIGN_VCENTER;
        if (column->w == 0) {
            column->w = w / c;
        }
        x += column->w;
        column = l_list_next (entry->columns, column);
    }

    l_list_append (list->entries, entry);
    list->need_adaptation = true;
    list->widget.need_repaint = true;

    return (otk_widget_t *) entry;
}


void
otk_list_set_selected_none (otk_widget_t * this)
{
    otk_list_t *list = (otk_list_t *) this;

    if (!otk_widget_is_correct (this, OTK_WIDGET_LIST)) {
        return;
    }

    otk_listentry_t *entry = l_list_first (list->entries);
    while (entry) {
        entry->widget.need_repaint = entry->is_selected;
        entry->is_selected = false;
        entry = l_list_next (list->entries, entry);
    }
    list->num_selected = 0;

    otk_draw_needy (this->otk);
}


void
otk_list_set_selected_all (otk_widget_t * this)
{
    otk_list_t *list = (otk_list_t *) this;

    if (!otk_widget_is_correct (this, OTK_WIDGET_LIST)) {
        return;
    }
    if (list->allow_select != OTK_LIST_SELECTION_MULTIPLE) {
        return;
    }

    otk_listentry_t *entry = l_list_first (list->entries);
    while (entry) {
        entry->widget.need_repaint = !entry->is_selected;
        entry->is_selected = true;
        entry = l_list_next (list->entries, entry);
    }
    list->num_selected = l_list_length (list->entries);

    entry = l_list_first (list->entries);
    while (entry) {
        if (entry->select_cb) {
            entry->select_cb (entry->select_cb_data);
        }
        entry = l_list_next (list->entries, entry);
    }

    otk_draw_needy (this->otk);
}


int
otk_list_get_selected_count (otk_widget_t * this)
{
    otk_list_t *list = (otk_list_t *) this;

    if (!otk_widget_is_correct (this, OTK_WIDGET_LIST)) {
        return 0;
    }

    return list->num_selected;
}


int
otk_list_get_selected_position (otk_widget_t * this)
{
    otk_list_t *list = (otk_list_t *) this;

    if (!otk_widget_is_correct (this, OTK_WIDGET_LIST)) {
        return -1;
    }
    if (list->num_selected != 1) {
        return -1;
    }

    int p = 0;
    otk_listentry_t *entry = l_list_first (list->entries);
    while (entry) {
        if (entry->is_selected) {
            return p;
        }
        entry = l_list_next (list->entries, entry);
        p++;
    }

    return -1;
}


int *
otk_list_get_selected_positions (otk_widget_t * this, int *num_selected)
{
    otk_list_t *list = (otk_list_t *) this;

    if (num_selected) {
        *num_selected = 0;
    }
    if (!otk_widget_is_correct (this, OTK_WIDGET_LIST)) {
        return NULL;
    }
    if (list->num_selected == 0) {
        return NULL;
    }
    if (num_selected) {
        *num_selected = list->num_selected;
    }

    int p = 0;
    int i = 0;
    int size = list->num_selected * sizeof (int);
    int *selected_entries = ho_malloc (size);
    memset (selected_entries, 0, size);

    otk_listentry_t *entry = l_list_first (list->entries);
    while (entry) {
        if (entry->is_selected) {
            selected_entries[i++] = p;
        }
        entry = l_list_next (list->entries, entry);
        p++;
    }

    return selected_entries;
}


void *
otk_list_get_selected_value (otk_widget_t * this)
{
    otk_list_t *list = (otk_list_t *) this;

    if (!otk_widget_is_correct (this, OTK_WIDGET_LIST)) {
        return NULL;
    }
    if (list->num_selected != 1) {
        return NULL;
    }

    otk_listentry_t *entry = l_list_first (list->entries);
    while (entry) {
        if (entry->is_selected) {
            return entry->select_cb_data;
        }
        entry = l_list_next (list->entries, entry);
    }

    return NULL;
}


void **
otk_list_get_selected_values (otk_widget_t * this, int *num_selected)
{
    otk_list_t *list = (otk_list_t *) this;

    if (num_selected) {
        *num_selected = 0;
    }
    if (!otk_widget_is_correct (this, OTK_WIDGET_LIST)) {
        return NULL;
    }
    if (list->num_selected == 0) {
        return NULL;
    }
    if (num_selected) {
        *num_selected = list->num_selected;
    }

    int i = 0;
    int size = list->num_selected * sizeof (void *);
    void **selected_entries = ho_malloc (size);
    memset (selected_entries, 0, size);

    otk_listentry_t *entry = l_list_first (list->entries);
    while (entry) {
        if (entry->is_selected) {
            selected_entries[i++] = entry->select_cb_data;
        }
        entry = l_list_next (list->entries, entry);
    }

    return selected_entries;
}


otk_widget_t *
otk_listentry_get_last_with_focus (otk_widget_t * this)
{
    otk_list_t *list = (otk_list_t *) this;

    if (!otk_widget_is_correct (this, OTK_WIDGET_LIST)) {
        return NULL;
    }
    else if (list->last_with_focus) {
        return (otk_widget_t *) list->last_with_focus;
    }
    else {
        otk_listentry_t *entry = l_list_first (list->entries);
        while (entry) {
            if (entry->widget.is_visible && entry->widget.is_enabled) {
                return (otk_widget_t *) entry;
            }
            entry = l_list_next (list->entries, entry);
        }
        return NULL;
    }
}


otk_widget_t *
otk_listentry_get_list (otk_widget_t * this)
{
    otk_listentry_t *entry = (otk_listentry_t *) this;
    otk_list_t *list = entry ? entry->list : NULL;

    if (!otk_widget_is_correct (this, OTK_WIDGET_LISTENTRY)) {
        return NULL;
    }

    return &list->widget;
}


static void
listentries_adapt (otk_list_t * list)
{
    list->need_adaptation = false;
    list->widget.need_repaint = true;

    if (l_list_length (list->entries) == 0) {
        if (list->scrollbar) {
            otk_scrollbar_set (list->scrollbar, 0, 100);
        }
        if (list->type == OTK_LIST_TYPE_SELECTOR) {
            otk_widget_set_enabled (list->button_up, false);
            otk_widget_set_enabled (list->button_down, false);
        }
        return;
    }

    int num_entries = l_list_length (list->entries);
    if ((num_entries - list->first_visible) < list->num_visible) {
        list->first_visible = num_entries - list->num_visible;
    }
    if (list->first_visible < 0) {
        list->first_visible = 0;
    }

    int first_visible = list->first_visible;
    int last_visible = list->first_visible + list->num_visible - 1;
    if (last_visible >= num_entries) {
        last_visible = num_entries - 1;
    }

    int first_y = list->entry_y;
    int entry_s = list->entry_spacing;
    if ((list->type == OTK_LIST_TYPE_MENU)
        && (num_entries < list->num_visible)) {
        first_y += ((list->num_visible * entry_s)
                    - (num_entries * entry_s)) / 2;
    }

    int pos = 0;
    otk_listentry_t *entry = l_list_first (list->entries);
    while (entry && (pos < first_visible)) {
        entry->is_first = false;
        entry->is_last = false;
        entry->widget.is_visible = false;
        entry = l_list_next (list->entries, entry);
        pos += 1;
    }
    while (entry && (pos <= last_visible)) {
        int y = first_y + entry_s * (pos - first_visible);
        bool is_first = (pos == first_visible);
        bool is_last = (pos == last_visible);
        entry->widget.need_repaint |= (entry->is_first != is_first);
        entry->widget.need_repaint |= (entry->is_last != is_last);
        entry->widget.need_repaint |= (entry->widget.y != y);
        entry->widget.need_repaint |= !entry->widget.is_visible;
        entry->is_first = is_first;
        entry->is_last = is_last;
        entry->widget.y = y;
        entry->widget.is_visible = true;
        entry = l_list_next (list->entries, entry);
        pos += 1;
    }
    while (entry) {
        entry->is_first = false;
        entry->is_last = false;
        entry->widget.is_visible = false;
        entry = l_list_next (list->entries, entry);
        pos += 1;
    }

    if (list->scrollbar) {
        double cnt = (double) num_entries;
        double beg = (double) list->first_visible;
        double num = (double) list->num_visible;
        double pos = (beg * 100.0) / cnt;
        double len = (num * 100.0) / cnt;
        /* Just to make sure rounding errors don't bite us! */
        if ((num_entries - list->num_visible) == list->first_visible) {
            pos = 100.0 - len;
        }
        otk_scrollbar_set (list->scrollbar, pos, len);
    }

    if (list->type == OTK_LIST_TYPE_SELECTOR) {
        otk_widget_t *last = l_list_last (list->entries);
        otk_widget_t *first = l_list_first (list->entries);
        otk_widget_set_enabled (list->button_down, !last->is_visible);
        otk_widget_set_enabled (list->button_up, !first->is_visible);
    }
}


void
otk_list_adapt (otk_widget_t * this)
{
    otk_list_t *list = (otk_list_t *) this;

    if (!otk_widget_is_correct (this, OTK_WIDGET_LIST)) {
        return;
    }
    if (!list->need_adaptation) {
        return;
    }

    listentries_adapt (list);
}


static void
entries_destroy (void *data)
{
    otk_widget_t *this = (otk_widget_t *) data;

    if (!otk_widget_is_correct (this, OTK_WIDGET_LISTENTRY)) {
        return;
    }

    otk_widget_destructor (this);
}


static void
list_destroy (otk_widget_t * this)
{
    otk_list_t *list = (otk_list_t *) this;

    if (!otk_widget_is_correct (this, OTK_WIDGET_LIST)) {
        return;
    }

    l_list_free (list->entries, entries_destroy);
    ho_free (list);
}


static void
list_draw (otk_widget_t * this)
{
    otk_list_t *list = (otk_list_t *) this;

    if (!otk_widget_is_correct (this, OTK_WIDGET_LIST)) {
        return;
    }


    /* Fill all parts of the list that don't contain entries with
     * transparency. */
    int length = l_list_length (list->entries);
    if (length < list->num_visible) {
        int palette = otk_get_palette (this->otk, OTK_PALETTE_BUTTON);
        int color = palette + OSD_TEXT_PALETTE_TRANSPARENT;
        int x = list->entry_x;
        int w = list->entry_width;
        otk_listentry_t *entry = l_list_first (list->entries);
        while (entry) {
            if (entry->is_first) {
                int y = list->entry_y;
                int h = entry->widget.y - list->entry_y;
                odk_draw_rect (this->odk, x, y, w, h, 0, color, true);
            }
            if (entry->is_last) {
                int y = entry->widget.y + entry->widget.h;
                int h = (list->num_visible * list->entry_spacing) - (y - list->entry_y);
                odk_draw_rect (this->odk, x, y, w, h, 0, color, true);
            }
            entry = l_list_next (list->entries, entry);
        }
        if (length == 0) {
            int y = list->entry_y;
            int h = (list->num_visible * list->entry_spacing);
            odk_draw_rect (this->odk, x, y, w, h, 0, color, true);
        }
    }

    if (list->type != OTK_LIST_TYPE_MENU) {
        return;
    }
    if (length < list->num_visible) {
        return;
    }

    int first_visible = list->first_visible;
    {
        int x1 = this->x + (this->w / 2);
        int y1 = list->entry_y - 30;
        int x2 = x1 - 20;
        int y2 = y1 + 20;
        int x3 = x1 + 20;
        int y3 = y1 + 20;

        int palette = otk_get_palette (this->otk, OTK_PALETTE_BUTTON);
        int color = palette + OSD_TEXT_PALETTE_BACKGROUND;
        if (first_visible == 0) {
            color = palette;
        }

        odk_draw_triangle (this->odk, x1, y1, x2, y2, x3, y3, color, true);
    }

    int last_visible = list->first_visible + list->num_visible;
    {
        int x1 = this->x + (this->w / 2);
        int y1 = list->entry_y;
        y1 += ((list->num_visible - 1) * list->entry_spacing);
        y1 += list->entry_height + 30;
        int x2 = x1 - 20;
        int y2 = y1 - 20;
        int x3 = x1 + 20;
        int y3 = y1 - 20;

        int palette = otk_get_palette (this->otk, OTK_PALETTE_BUTTON);
        int color = palette + OSD_TEXT_PALETTE_BACKGROUND;
        if (last_visible == length) {
            color = palette;
        }

        odk_draw_triangle (this->odk, x1, y1, x2, y2, x3, y3, color, true);
    }
}


otk_widget_t *
otk_list_new (otk_t * otk, int x, int y, int w, int h,
              int entry_height, int entry_spacing,
              bool show_border, bool show_scrollbar,
              otk_list_selection_t allow_select, void *list_cb_data)
{
    otk_list_t *list = ho_new (otk_list_t);

    otk_widget_constructor ((otk_widget_t *) list, otk,
                            OTK_WIDGET_LIST, x, y, w, h);

    list->widget.selectable = OTK_SELECTABLE_NONE;

    list->widget.draw = list_draw;
    list->widget.destroy = list_destroy;

    list->type = OTK_LIST_TYPE_NORMAL;
    list->allow_select = allow_select;

    list->entry_x = list->widget.x;
    list->entry_y = list->widget.y;
    list->entry_width = list->widget.w;
    list->entry_height = entry_height;
    list->entry_spacing = entry_spacing;

    list->entries = l_list_new ();
    list->num_selected = 0;
    list->first_visible = 0;

    list->num_visible = list->widget.h - entry_height;
    list->num_visible -= show_border ? 10 : 0;
    list->num_visible /= entry_spacing;
    list->num_visible += 1;

    list->cb_data = list_cb_data;
    list->entry_changed_cb = NULL;
    list->entry_changed_cb_data = NULL;

    list->scrollbar = NULL;
    list->button_up = NULL;
    list->button_down = NULL;

    if (show_scrollbar && (list->num_visible > 1)) {
        list->entry_width -= 35;

        int x1 = x + w - 35;
        int y1 = y + 5;
        int w1 = 30;
        int h1 = h - 10;

        list->scrollbar = otk_scrollbar_new (otk, x1, y1, w1, h1,
                                             list_set_position,
                                             list_scroll_up, list_scroll_down,
                                             list_page_up, list_page_down,
                                             list);

        otk_scrollbar_set (list->scrollbar, 0, 100);
    }

    else if (show_scrollbar && (list->num_visible == 1)) {
        list->entry_width -= 70;

        int x1 = x + w - 35;
        int y1 = y + 5;
        int x2 = x + w - 70;
        int y2 = y + 5;

        int bw = 30;
        int bh = 30;
        int vw = 20;
        int vh = (vw / 2);

        list->button_up = otk_vector_button_new (otk, x1, y1, bw, bh,
                                                 OSD_VECTOR_ARROW_UP_SIMPLE,
                                                 vw, vh, list_scroll_up,
                                                 list);
        list->button_up->selectable = OTK_SELECTABLE_MOUSE;

        list->button_down = otk_vector_button_new (otk, x2, y2, bw, bh,
                                                   OSD_VECTOR_ARROW_DOWN_SIMPLE,
                                                   vw, vh, list_scroll_down,
                                                   list);
        list->button_down->selectable = OTK_SELECTABLE_MOUSE;
    }

    if (show_border) {
        otk_border_new (otk, x, y, w, h);

        list->entry_width -= 10;
        list->entry_x += 5;
        list->entry_y += 5;
    }

    return (otk_widget_t *) list;
}


int
otk_list_get_length (otk_widget_t * this)
{
    otk_list_t *list = (otk_list_t *) this;

    if (!otk_widget_is_correct (this, OTK_WIDGET_LIST)) {
        return 0;
    }

    return l_list_length (list->entries);
}


int
otk_list_get_visible_count (otk_widget_t * this)
{
    otk_list_t *list = (otk_list_t *) this;

    if (!otk_widget_is_correct (this, OTK_WIDGET_LIST)) {
        return 0;
    }

    return list->num_visible;
}


bool
otk_list_has_focus (otk_widget_t * this)
{
    otk_list_t *list = (otk_list_t *) this;

    if (!otk_widget_is_correct (this, OTK_WIDGET_LIST)) {
        return false;
    }

    otk_widget_t *window = otk_get_current_window (this->otk);
    if (!window) {
        return false;
    }

    otk_widget_t *focused = otk_window_focus_pointer_get (window);
    if (!focused) {
        return false;
    }

    int i = 0;
    otk_listentry_t *current = l_list_first (list->entries);
    while (current) {
        if (((otk_widget_t *) current) == focused) {
            return true;
        }
        current = l_list_next (list->entries, current);
        i++;
    }

    return false;
}


int
otk_list_get_focus (otk_widget_t * this)
{
    otk_list_t *list = (otk_list_t *) this;

    if (!otk_widget_is_correct (this, OTK_WIDGET_LIST)) {
        return 0;
    }

    otk_widget_t *window = otk_get_current_window (this->otk);
    if (!window) {
        return 0;
    }

    otk_widget_t *focused = otk_window_focus_pointer_get (window);
    if (!focused) {
        return 0;
    }

    int i = 0;
    otk_listentry_t *current = l_list_first (list->entries);
    while (current) {
        if (((otk_widget_t *) current) == focused) {
            return i;
        }
        current = l_list_next (list->entries, current);
        i++;
    }

    return 0;
}


int
otk_list_set_focus (otk_widget_t * this, int pos)
{
    otk_list_t *list = (otk_list_t *) this;

    if (!otk_widget_is_correct (this, OTK_WIDGET_LIST)) {
        return pos;
    }

    if (pos > l_list_length (list->entries) - 1) {
        pos = l_list_length (list->entries) - 1;
    }

    int i = 0;
    otk_widget_t *current = l_list_first (list->entries);
    while (current) {
        if (i == pos) {
            while (current && !current->is_enabled) {
                current = l_list_next (list->entries, current);
            }
            if (current) {
                otk_widget_set_focused (current, true);
            }
            break;
        }
        current = l_list_next (list->entries, current);
        i++;
    }

    return pos;
}


int
otk_list_set_pos (otk_widget_t * this, int pos)
{
    otk_list_t *list = (otk_list_t *) this;

    if (!otk_widget_is_correct (this, OTK_WIDGET_LIST)) {
        return pos;
    }

    list->first_visible = pos;
    list->need_adaptation = true;

    return pos;
}


int
otk_list_get_pos (otk_widget_t * this)
{
    otk_list_t *list = (otk_list_t *) this;

    if (!otk_widget_is_correct (this, OTK_WIDGET_LIST)) {
        return -1;
    }

    return list->first_visible;
}


void
otk_list_clear (otk_widget_t * this)
{
    otk_list_t *list = (otk_list_t *) this;

    if (!otk_widget_is_correct (this, OTK_WIDGET_LIST)) {
        return;
    }

    l_list_clear (list->entries, entries_destroy);
    list->last_single_selected = NULL;
    list->last_with_focus = NULL;
    list->num_selected = 0;
    list->first_visible = 0;
    list->need_adaptation = true;
    list->widget.need_repaint = true;
}


void
otk_list_set_selected (otk_widget_t * this, int pos, bool selected)
{
    otk_list_t *list = (otk_list_t *) this;

    if (!otk_widget_is_correct (this, OTK_WIDGET_LIST)) {
        return;
    }

    int i = 0;
    otk_listentry_t *current = l_list_first (list->entries);
    while (current) {
        if ((i == pos)
            && (current->widget.is_enabled)
            && (current->widget.is_visible)
            && (current->widget.selectable)) {

            if (!current->is_selected && selected)
                list->num_selected++;
            if (current->is_selected && !selected)
                list->num_selected--;

            current->is_selected = selected;
            return;
        }
        current = l_list_next (list->entries, current);
        i++;
    }
}


void
otk_list_set_allow_select (otk_widget_t * this,
                           otk_list_selection_t allow_select)
{
    otk_list_t *list = (otk_list_t *) this;

    if (!otk_widget_is_correct (this, OTK_WIDGET_LIST)) {
        return;
    }

    list->allow_select = allow_select;

    switch (list->allow_select) {
    case OTK_LIST_SELECTION_NONE:
    case OTK_LIST_SELECTION_SINGLE:
        otk_list_set_selected_none (this);
        break;
    default:
        break;
    }
}


otk_widget_t *
otk_selector_new (otk_t * otk, int x, int y, int w,
                  otk_int_set_cb_t cb, void *cb_data)
{
    otk_list_t *list =
        (otk_list_t *) otk_list_new (otk, x, y, w, 40, 30, 30, true, true,
                                     OTK_LIST_SELECTION_NONE, NULL);

    list->entry_changed_cb = cb;
    list->entry_changed_cb_data = cb_data;

    list->type = OTK_LIST_TYPE_SELECTOR;

    return (otk_widget_t *) list;
}


otk_widget_t *
otk_selectoritem_new (otk_widget_t * selector, const char *label,
                      otk_cb_t activate_cb, void *activate_cb_data)
{
    otk_widget_t *selectoritem =
        otk_listentry_new (selector, label, activate_cb, activate_cb_data,
                           NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
                           NULL, NULL);

    return selectoritem;
}


otk_widget_t *
otk_menulist_new (otk_t * otk, int x, int y, int w, int h,
                  int entry_height, int entry_spacing, void *menu_cb_data)
{
    otk_list_t *list =
        (otk_list_t *) otk_list_new (otk, x, y + 30, w, h - 60, entry_height,
                                     entry_spacing, false, false,
                                     OTK_LIST_SELECTION_NONE, menu_cb_data);

    list->type = OTK_LIST_TYPE_MENU;

    int off = (h - 60);
    off -= ((list->num_visible - 1) * entry_spacing);
    off -= entry_height;
    list->entry_y += (off / 2);

#if 0
    otk_border_new (otk, x, y, w, h);
#endif

    return (otk_widget_t *) list;
}


otk_widget_t *
otk_menuitem_new (otk_widget_t * menu, const char *text,
                  otk_cb_t activate_cb, void *activate_cb_data)
{
    otk_widget_t *menuitem =
        otk_listentry_new (menu, text, activate_cb, activate_cb_data,
                           activate_cb, activate_cb_data, NULL, NULL,
                           NULL, NULL, NULL, NULL, NULL, NULL);

    return menuitem;
}


otk_widget_t *
otk_infolist_new (otk_t * otk, int x, int y, int w, int h,
                  int entry_height, int entry_spacing)
{
    otk_list_t *list =
        (otk_list_t *) otk_list_new (otk, x, y, w, h, entry_height,
                                     entry_spacing, true, false,
                                     OTK_LIST_SELECTION_NONE, NULL);

    list->type = OTK_LIST_TYPE_INFO;

    list->entry_width -= 35;

    int x1 = x + w - 35;
    int y1 = y + 5;
    int w1 = 30;
    int h1 = h - 10;

    list->scrollbar = otk_scrollbar_new (otk, x1, y1, w1, h1,
                                         list_set_position,
                                         list_scroll_up, list_scroll_down,
                                         NULL, NULL, list);

    otk_scrollbar_set (list->scrollbar, 0, 100);

    return (otk_widget_t *) list;
}


otk_widget_t *
otk_infoitem_new (otk_widget_t * info, const char *label)
{
    otk_widget_t *item = NULL;
    otk_list_t *list = (otk_list_t *) info;

    if (!otk_widget_is_correct (info, OTK_WIDGET_LIST)) {
        return NULL;
    }
    if (list->type != OTK_LIST_TYPE_INFO) {
        return NULL;
    }

    if (strlen (label) == 0) {
        item = otk_listentry_new (info, label, NULL, NULL, NULL, NULL, NULL,
                                  NULL, NULL, NULL, NULL, NULL, NULL, NULL);
        return item;
    }

    if (strstr (label, "<tr>")) {
        item = otk_listentry_new (info, label, NULL, NULL, NULL, NULL, NULL,
                                  NULL, NULL, NULL, NULL, NULL, NULL, NULL);
        return item;
    }

    char *t = ho_strdup (label);
    char *nl = index (t, '\n');
    if (nl) {
        nl[0] = '\0';
        item = otk_infoitem_new (info, t);
        item = otk_infoitem_new (info, &nl[1]);
        ho_free (t);
        return item;
    }
    ho_free (t);

    int l = 0;
    int w = 0;
    int h = 0;
    odk_osd_set_font (info->odk, info->font, info->fontsize);
    odk_get_text_size (info->odk, label, &w, &h);
    if (w > list->entry_width - 10) {
        do {
            char *t = ho_strdup (&label[l]);
            char *c = index (t, ' ');
            char *d = c;
            if (c) {
                c[0] = '\0';
            }
            odk_get_text_size (info->odk, t, &w, &h);
            while (c && (w < list->entry_width - 10)) {
                d = c;
                c[0] = ' ';
                c = index (c + 1, ' ');
                if (!c) {
                    break;
                }
                c[0] = '\0';
                odk_get_text_size (info->odk, t, &w, &h);
            }
            if (c) {
                c[0] = ' ';
                d[0] = '\0';
            }
            item = otk_listentry_new (info, t, NULL, NULL, NULL, NULL, NULL,
                                      NULL, NULL, NULL, NULL, NULL, NULL,
                                      NULL);
            l += strlen (t) + 1;
            ho_free (t);
        } while (l < strlen (label));
    }
    else {
        item = otk_listentry_new (info, label, NULL, NULL, NULL, NULL, NULL,
                                  NULL, NULL, NULL, NULL, NULL, NULL, NULL);
    }

    return item;
}


otk_widget_t *
otk_list_separator_new (otk_widget_t * info)
{
    return otk_listentry_new (info, "*--*", NULL, NULL, NULL, NULL,
                              NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
}
