
/*
 * Copyright (C) 2002-2003 Stefan Holst
 * Copyright (C) 2004-2005 Maximilian Schwerin
 *
 * This file is part of oxine a free media player.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: otk_widget.c 3790 2009-04-25 20:07:09Z mschwerin $
 */
#include "config.h"

#include <assert.h>

#include "heap.h"
#include "logger.h"
#include "otk.h"
#include "utils.h"


const char *
otk_get_default_fonttype (void)
{
    static const char *default_fonttype = NULL;
    if (!default_fonttype) {
        default_fonttype = config_get_string ("gui.fonttype");
    }
    if (!default_fonttype) {
        default_fonttype = OTK_DEFAULT_FONTTYPE;
    }
    return default_fonttype;
}


int
otk_get_default_fontsize (void)
{
    static int default_fontsize = -1;
    if (default_fontsize == -1) {
        default_fontsize = config_get_number ("gui.fontsize");
    }
    if (default_fontsize == -1) {
        default_fontsize = OTK_DEFAULT_FONTSIZE;
    }
    return default_fontsize;
}


void
otk_widget_destructor (otk_widget_t * widget)
{
    if (!widget) {
        return;
    }

    otk_t *otk = widget->otk;
    otk_widget_t *window = otk_get_current_window (otk);

    /* Inform the user that this widget will be destructed. */
    if (widget->destruct_cb) {
        widget->destruct_cb (widget->destruct_cb_data);
    }

    /* Remove the widget from the current window. */
    otk_window_remove (window, widget);

    /* Free the font. */
    ho_free (widget->font);

    /* And call the widgets own destroy function. */
    widget->destroy (widget);
}


void
otk_widget_constructor (otk_widget_t * widget, otk_t * otk,
                        otk_widget_type_t type, int x, int y, int w, int h)
{
    if (!widget) {
        return;
    }

    memset (widget, 0, sizeof (otk_widget_t));

    otk_widget_t *window = otk_get_current_window (otk);

    widget->type = type;

    widget->otk = otk;
    widget->odk = otk_get_odk (otk);

    widget->x = ((otk_widget_t *) window)->x + x;
    widget->y = ((otk_widget_t *) window)->y + y;
    widget->w = w;
    widget->h = h;

    widget->selectable = OTK_SELECTABLE_MOUSE | OTK_SELECTABLE_KEY;

    widget->is_focused = false;
    widget->is_enabled = true;
    widget->is_visible = true;

    widget->need_update = false;
    widget->need_repaint = true;

    widget->font = ho_strdup (otk_get_default_fonttype ());
    widget->fontsize = otk_get_default_fontsize ();
    widget->alignment = ODK_ALIGN_LEFT | ODK_ALIGN_VCENTER;

    otk_window_append (window, widget);
}


void
otk_widget_draw (otk_widget_t * widget)
{
    assert (widget);
    if (widget->draw) {
        widget->draw (widget);
    }
    widget->need_repaint = false;
}


void
otk_widget_hide (otk_widget_t * widget)
{
    assert (widget);
    if (widget->hide) {
        widget->hide (widget);
    }
    widget->need_repaint = false;
}


void
otk_widget_set_focused (otk_widget_t * widget, bool focused)
{
    if (!widget) {
        return;
    }

    otk_widget_t *window = otk_get_current_window (widget->otk);
    otk_widget_t *focused_widget = otk_window_focus_pointer_get (window);

    if (focused) {
        if (!widget->is_visible) {
            return;
        }
        if (!widget->is_enabled) {
            return;
        }
        if (widget->selectable == OTK_SELECTABLE_NONE) {
            return;
        }
        if (focused_widget) {
            otk_widget_set_focused (focused_widget, false);
        }

        widget->is_focused = true;
        otk_window_focus_pointer_set (window, widget);

        if (widget->focus_enter_cb) {
            widget->focus_enter_cb (widget->focus_enter_cb_data);
        }
    }

    else {
        if (focused_widget != widget) {
            return;
        }

        widget->is_focused = false;
        otk_window_focus_pointer_set (window, NULL);

        if (widget->focus_leave_cb) {
            widget->focus_leave_cb (widget->focus_leave_cb_data);
        }
    }

    widget->need_repaint = true;
}


const char *
otk_widget_get_type (otk_widget_t * widget)
{
    if (!widget) {
        return "null";
    }

    switch (widget->type) {
    case OTK_WIDGET_WINDOW:
        return "window";
    case OTK_WIDGET_BUTTON:
        return "button";
    case OTK_WIDGET_LABEL:
        return "label";
    case OTK_WIDGET_LIST:
        return "list";
    case OTK_WIDGET_LISTENTRY:
        return "listentry";
    case OTK_WIDGET_SLIDER:
        return "slider";
    case OTK_WIDGET_SCROLLBAR:
        return "scrollbar";
    case OTK_WIDGET_CHECKBOX:
        return "checkbox";
    case OTK_WIDGET_BORDER:
        return "border";
    case OTK_WIDGET_EDITBOX:
        return "editbox";
    case OTK_WIDGET_IMAGE:
        return "image";
    default:
        break;
    }

    return "unknown";
}
