
/*
 * Copyright (C) 2002-2003 Stefan Holst
 * Copyright (C) 2004-2005 Maximilian Schwerin
 *
 * This file is part of oxine a free media player.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: otk_widget.h 3790 2009-04-25 20:07:09Z mschwerin $
 *
 */

#ifndef HAVE_OTK_WIDGET_H
#define HAVE_OTK_WIDGET_H

#include "config.h"
#include "heap.h"

#define OTK_ALIGN_LEFT              (0x001)
#define OTK_ALIGN_CENTER            (0x002)
#define OTK_ALIGN_RIGHT             (0x004)
#define OTK_ALIGN_TOP               (0x008)
#define OTK_ALIGN_VCENTER           (0x010)
#define OTK_ALIGN_BOTTOM            (0x020)

#define OTK_SELECTABLE_NONE         (0x000)
#define OTK_SELECTABLE_KEY          (0x001)
#define OTK_SELECTABLE_MOUSE        (0x010)

#define OTK_DEFAULT_FONTTYPE        "sans"
#define OTK_DEFAULT_FONTSIZE        20

typedef enum {
    OTK_WIDGET_WINDOW,
    OTK_WIDGET_BUTTON,
    OTK_WIDGET_LABEL,
    OTK_WIDGET_LIST,
    OTK_WIDGET_LISTENTRY,
    OTK_WIDGET_SLIDER,
    OTK_WIDGET_SCROLLBAR,
    OTK_WIDGET_CHECKBOX,
    OTK_WIDGET_BORDER,
    OTK_WIDGET_EDITBOX,
    OTK_WIDGET_IMAGE,
} otk_widget_type_t;


/// The datastructure of a widget.
struct otk_widget_s {
    otk_widget_type_t type;

    otk_t *otk;
    odk_t *odk;

    int x;
    int y;
    int w;
    int h;

    int selectable;

    bool is_focused;
    bool is_enabled;
    bool is_visible;

    bool need_update;
    bool need_repaint;

    char *font;
    int fontsize;
    int alignment;

    int foreground_color;
    int background_color;

    void (*draw) (otk_widget_t * this);
    void (*hide) (otk_widget_t * this);
    void (*destroy) (otk_widget_t * this);
    void (*update) (otk_widget_t * this);

    void (*key_handler) (otk_widget_t * this, oxine_event_t * event);
    void (*button_handler) (otk_widget_t * this, oxine_event_t * event);
    void (*motion_handler) (otk_widget_t * this, oxine_event_t * event);

    /// Call this callback when this item gets the focus.
    void *focus_enter_cb_data;
    otk_cb_t focus_enter_cb;
    /// Call this callback when this item looses the focus.
    void *focus_leave_cb_data;
    otk_cb_t focus_leave_cb;
    /// Call this callback just before the widget ist destructed.
    void *destruct_cb_data;
    otk_cb_t destruct_cb;
};

void otk_widget_constructor (otk_widget_t * widget, otk_t * otk,
                             otk_widget_type_t type,
                             int x, int y, int w, int h);

void otk_widget_destructor (otk_widget_t * widget);

void otk_widget_draw (otk_widget_t * widget);

void otk_widget_hide (otk_widget_t * widget);

void otk_widget_set_focused (otk_widget_t * widget, bool focused);


static inline void
otk_widget_set_destruct_callback (otk_widget_t * widget,
                                  otk_cb_t destruct_cb,
                                  void *destruct_cb_data)
{
    if (!widget) {
        return;
    }

    widget->destruct_cb = destruct_cb;
    widget->destruct_cb_data = destruct_cb_data;
}


static inline void
otk_widget_set_focus_callbacks (otk_widget_t * widget,
                                otk_cb_t enter_cb, void *enter_cb_data,
                                otk_cb_t leave_cb, void *leave_cb_data)
{
    if (!widget) {
        return;
    }

    widget->focus_enter_cb = enter_cb;
    widget->focus_enter_cb_data = enter_cb_data;
    widget->focus_leave_cb = leave_cb;
    widget->focus_leave_cb_data = leave_cb_data;
}


const char *otk_get_default_fonttype (void);

int otk_get_default_fontsize (void);

static inline void
otk_widget_set_fonttype (otk_widget_t * widget, const char *font)
{
    if (!widget) {
        return;
    }
    if (!font) {
        return;
    }

    ho_free (widget->font);
    widget->font = ho_strdup (font);
    widget->need_repaint = true;
}


static inline void
otk_widget_set_fontsize (otk_widget_t * widget, int fontsize)
{
    if (!widget) {
        return;
    }

    /* If the default font size has been changed by the user, we scale all
     * fonts to reflect this change. */
    double nsize = (double) fontsize;
    double dsize = (double) OTK_DEFAULT_FONTSIZE;
    double csize = (double) otk_get_default_fontsize ();
    double ssize = (nsize / dsize) * csize;

    widget->fontsize = (int) ssize;
    widget->need_repaint = true;
}


static inline void
otk_widget_set_alignment (otk_widget_t * widget, int alignment)
{
    if (!widget) {
        return;
    }

    widget->alignment = alignment;
    widget->need_repaint = true;
}


static inline void
otk_widget_set_enabled (otk_widget_t * widget, bool enabled)
{
    if (!widget) {
        return;
    }

    widget->is_enabled = enabled;
    widget->need_repaint = true;
}


static inline void
otk_widget_set_visible (otk_widget_t * widget, bool visible)
{
    if (!widget) {
        return;
    }

    widget->is_visible = visible;
    widget->need_repaint = true;
}


static inline void
otk_widget_set_selectable (otk_widget_t * widget, int selectable)
{
    if (!widget) {
        return;
    }

    widget->selectable = selectable;
    widget->need_repaint = true;
}


static inline void
otk_widget_set_updated (otk_widget_t * widget, bool update)
{
    if (!widget) {
        return;
    }

    widget->need_update = update;
}


static inline bool
otk_widget_is_enabled (otk_widget_t * widget)
{
    if (!widget) {
        return false;
    }
    return widget->is_enabled;
}


static inline bool
otk_widget_is_focused (otk_widget_t * widget)
{
    if (!widget) {
        return false;
    }
    return widget->is_focused;
}


/**
 * Checks if the expected type and the real type of a widget are the same.
 */
static inline bool
otk_widget_is_correct (otk_widget_t * widget, otk_widget_type_t expected)
{
    if (!widget) {
        return false;
    }
    if (widget->type != expected) {
        return false;
    }

    return true;
}

/**
 * Returns a string identifying the type of the widget.
 */
const char *otk_widget_get_type (otk_widget_t * widget);

#endif /* HAVE_OTK_WIDGET_H */
