
/*
 * Copyright (C) 2004-2005 Maximilian Schwerin
 *
 * This file is part of oxine a free media player.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: playlist.c 4035 2010-10-20 11:10:14Z mschwerin $
 *
 */
#include "config.h"

#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>

#include "environment.h"
#include "filelist.h"
#include "heap.h"
#include "i18n.h"
#include "logger.h"
#include "mutex.h"
#include "oxine.h"
#include "playlist.h"
#include "playlist_m3u.h"
#include "playlist_pls.h"
#include "playlist_xml.h"
#include "utils.h"

#define INFO_LIST_FORMAT        "<tr>" \
                                "<td width=150 align=right>%s</td>" \
                                "<td width=10></td><td>%s</td>" \
                                "</tr>"

extern oxine_t *oxine;

static playitem_length_cb_t playitem_length_cb = NULL;


static bool
playlist_is_locked (playlist_t * playlist)
{
#ifdef DEBUG
    return (!playlist || playlist->mutex.__data.__count > 0);
#else
    return true;
#endif
}


/**
 * Frees one item of the info lists.
 */
static void
fileinfo_free_cb (void *p)
{
    ho_free (p);
}


static void
playitem_free (void *p)
{
    playitem_t *playitem = (playitem_t *) p;

    assert (playitem);
    //debug ("%s (%x)", __func__, playitem);

    if (playitem->info_list) {
        l_list_free (playitem->info_list, fileinfo_free_cb);
        playitem->info_list = NULL;
    }
    if (playitem->alternative_list) {
        ho_badref (playitem->alternative_list);
        ho_setref (playitem->alternative_list, NULL);
    }

    ho_free (playitem->title);
    ho_free (playitem->mrl);
    ho_free (playitem->subtitle_mrl);
    ho_free (playitem->thumbnail_mrl);

    ho_free (playitem);
}


/**
 * Callback used by l_list_free and l_list_clear to free a playitem.
 */
static void
playitem_free_cb (void *p)
{
    ho_badref (p);
    ho_decref (p);
}


static void
playlist_free (void *p)
{
    playlist_t *playlist = (playlist_t *) p;

    assert (playlist);
    assert (playlist->list);
    assert (!playlist_is_locked (playlist));
    //debug ("%s (%x)", __func__, playlist);

    if (playlist->info_list) {
        l_list_free (playlist->info_list, fileinfo_free_cb);
        playlist->info_list = NULL;
    }
    {
        l_list_free (playlist->list, playitem_free_cb);
        playlist->list = NULL;
    }

    pthread_mutex_destroy (&playlist->mutex);
    pthread_mutexattr_destroy (&playlist->mutex_attr);

    ho_free (playlist->mrl);
    ho_free (playlist);
}


playlist_t *
playlist_new (const char *mrl, playlist_change_cb_t cb)
{
    playlist_t *playlist = ho_initref (ho_new (playlist_t));
    playlist->r.free = playlist_free;

    playlist->list = l_list_new ();

    pthread_mutexattr_init (&playlist->mutex_attr);
    pthread_mutexattr_settype (&playlist->mutex_attr,
                               PTHREAD_MUTEX_RECURSIVE);
    pthread_mutex_init (&playlist->mutex, &playlist->mutex_attr);

    playlist->mrl = mrl ? ho_strdup (mrl) : NULL;
    playlist->current = NULL;
    playlist->playmode = PLAYBACK_MODE_NORMAL;
    playlist->playstatus = PLAYBACK_STATUS_STOPPED;
    playlist->top_position = -1;
    playlist->cur_position = -1;
    playlist->change_cb = cb;
    playlist->call_change_cb = true;
    playlist->playback_length = 0;

    return playlist;
}


/**
 * Calls the playlist-changed callback if allowed. If the playlist has an info
 * list this is freed.
 */
static void
playlist_changed (playlist_t * playlist)
{
    if (playlist->info_list) {
        l_list_free (playlist->info_list, fileinfo_free_cb);
        playlist->info_list = NULL;
    }
    if (playlist->call_change_cb && playlist->change_cb) {
        playlist->change_cb (playlist);
    }
}


void
playlist_clear (playlist_t * playlist)
{
    assert (ho_checkref (playlist));

    playlist_lock (playlist);

    l_list_clear (playlist->list, playitem_free_cb);
    playlist->current = NULL;
    playlist->playmode = PLAYBACK_MODE_NORMAL;
    playlist->playstatus = PLAYBACK_STATUS_STOPPED;
    playlist->top_position = -1;
    playlist->cur_position = -1;
    playlist->playback_length = 0;

    playlist_changed (playlist);
    playlist_unlock (playlist);
}


static char *
try_subtitles (const char *mrl)
{
    const char *endings[] =
        { "sub", "srt", "asc", "smi", "ssa", "txt", NULL };

    /* Allocate enough memory to hold the MRL plus a suffix of four
     * characters plus the terminating \0. */
    int length = strlen (mrl) + 5;
    char *subtitle = ho_malloc (length);

    /* Try to find a subtitle file that is named like the MRL with 
     * an appended suffix. */
    int i = 0;
    for (i = 0; endings[i]; i++) {
        snprintf (subtitle, length, "%s.%s", mrl, endings[i]);

        if (file_exists (subtitle)) {
            return subtitle;
        }
    }

    ho_free (subtitle);
    return NULL;
}


static char *
guess_subtitle (const char *mrl)
{
    char *subtitle = NULL;

    if (strc_starts_with (mrl, "file:")) {
        return guess_subtitle (mrl + 5);
    }

    if ((subtitle = try_subtitles (mrl)) == NULL) {
        char *mrl_no_suffix = ho_strdup (mrl);
        char *suffix = NULL;
        if ((suffix = rindex (mrl_no_suffix, '.')) != NULL)
            suffix[0] = '\0';
        subtitle = try_subtitles (mrl_no_suffix);
        ho_free (mrl_no_suffix);
    }
#ifdef DEBUG
    if (subtitle) {
        char *mrl_bn = get_basename (mrl);
        char *sub_bn = get_basename (subtitle);
        debug ("found subtitle file");
        debug ("         mrl: %s", mrl_bn);
        debug ("    subtitle: %s", sub_bn);
        ho_free (mrl_bn);
        ho_free (sub_bn);
    }
#endif

    return subtitle;
}


playitem_t *
playlist_insert (playlist_t * playlist, int position,
                 const char *title, const char *mrl)
{
    playitem_t *playitem = NULL;

    assert (ho_checkref (playlist));
    assert (title);
    assert (mrl);

    playlist_lock (playlist);
    if (!mrl_is_valid (mrl)) {
        char *new_mrl = mrl_from_filename (mrl);
        playitem = playlist_insert (playlist, position, title, new_mrl);
        ho_free (new_mrl);
    }
    else {
        playitem = ho_initref (ho_new (playitem_t));
        playitem->r.free = playitem_free;

        playitem->parent_list = playlist;
        playitem->title = ho_strdup (title);
        playitem->mrl = ho_strdup (mrl);

        if (mrl_is_networked (playitem->mrl)) {
            /* don't query the playback length of network streams */
        }
        else if (playitem_length_cb) {
            playitem->playback_length = playitem_length_cb (mrl);
            playlist->playback_length += playitem->playback_length;
        }

        if (config_get_bool ("subtitles.autoload")) {
            playitem->subtitle_mrl = guess_subtitle (mrl);
        }

        l_list_insert (playlist->list, position, playitem);

        playlist_changed (playlist);
    }
    playlist_unlock (playlist);

    return playitem;
}


playitem_t *
playitem_insert_alternative (playitem_t * playitem,
                             const char *mrl, const char *title, int position)
{
    playitem_t *alternative = NULL;

    assert (ho_checkref (playitem));
    assert (ho_checkref (playitem->parent_list));
    assert (playlist_is_locked (playitem->parent_list));
    assert (mrl);

    playlist_lock (playitem->parent_list);

    if (!playitem->alternative_list) {
        playitem->alternative_list = playlist_new (NULL, NULL);
    }

    title = title ? title : playitem->title;
    alternative = playlist_insert (playitem->alternative_list,
                                   position, title, mrl);

    const char *thumbnail_mrl = playitem_get_cover_image (playitem);
    if (alternative && thumbnail_mrl) {
        ho_free (alternative->thumbnail_mrl);
        alternative->thumbnail_mrl = ho_strdup (thumbnail_mrl);
    }

    playlist_unlock (playitem->parent_list);

    return alternative;
}


playitem_t *
playitem_append_alternative (playitem_t * playitem,
                             const char *mrl, const char *title)
{
    playitem_t *alternative = NULL;

    assert (ho_checkref (playitem));
    assert (ho_checkref (playitem->parent_list));
    assert (playlist_is_locked (playitem->parent_list));
    assert (mrl);

    playlist_lock (playitem->parent_list);

    if (!playitem->alternative_list) {
        playitem->alternative_list = playlist_new (NULL, NULL);
    }

    int length = playlist_length (playitem->alternative_list);
    alternative = playitem_insert_alternative (playitem, mrl, title, length);

    playlist_unlock (playitem->parent_list);

    return alternative;
}


playitem_t *
playlist_add (playlist_t * playlist, const char *title, const char *mrl)
{
    assert (ho_checkref (playlist));
    assert (title);
    assert (mrl);

    if (mrl_is_directory (mrl) && !mrl_is_directory_dvd (mrl)) {
        playitem_t *playitem = NULL;
        filelist_t *filelist = filelist_get (FILELIST_TEMPORARY_MRL);
        filelist_lock (filelist);
        fileitem_t *fileitem = filelist_add_magic (filelist, NULL, mrl);
        if (fileitem) {
            playitem = playlist_add_fileitem (playlist, fileitem);
            filelist_remove (filelist, fileitem);
        }
        filelist_unlock (filelist);
        return playitem;
    }

    int length = l_list_length (playlist->list);

    return playlist_insert (playlist, length, title, mrl);
}


playitem_t *
playlist_add_autoplay (playlist_t * playlist,
                       const char *title, const char *plugin_id)
{
    playitem_t *playitem = NULL;

    playlist_lock (playlist);
    playlist->call_change_cb = false;

    if (strc_starts_with (plugin_id, "dvb")) {
        int default_channel = config_get_number ("media.dvb.last_channel");
        const char *dvb_channels = get_file_dvb_channels ();

        FILE *f = fopen (dvb_channels, "r");
        if (f) {
            int i = 1;
            char line[256];
            char title[64];

            while (!feof (f)) {
                if (!fgets (line, 256, f)) {
                    break;
                }
                sscanf (line, "%[^:]", title);

                char *mrl = mrl_build ("dvb", "/", i);
                debug (mrl);
                playitem_t *item = playlist_add (playlist, title, mrl);
                ho_free (mrl);

                if ((default_channel == i) || !playitem) {
                    playitem = item;
                }

                i += 1;
            }

            fclose (f);
        }
        else {
            error (_("Could not open '%s': %s!"), dvb_channels,
                   strerror (errno));
        }
    }
    else {
        int i = 0;
        int c = 0;
        char **mrls = xine_get_autoplay_mrls (oxine->xine, plugin_id, &c);
        if (c > 0) {
            playitem = playlist_add (playlist, title, mrls[0]);
            debug ("%s", mrls[0]);
        }
        for (i = 1; i < c; i++) {
            playlist_add (playlist, title, mrls[i]);
            debug ("%s", mrls[i]);
        }
    }

    playlist->call_change_cb = true;
    playlist_changed (playlist);
    playlist_unlock (playlist);

    if (!playitem) {
        error (_("Failed to retrieve autoplay MRL for plugin '%s'!"),
               plugin_id);
    }

    return playitem;
}


playitem_t *
playlist_add_fileitem (playlist_t * playlist, fileitem_t * fileitem)
{
    playitem_t *playitem = NULL;

    assert (ho_checkref (playlist));
    assert (ho_checkref (fileitem));

    playlist_lock (playlist);
    playlist->call_change_cb = false;

    filelist_expand (fileitem);

    switch (fileitem->type) {
    case FILE_TYPE_REGULAR:
        {
            const char *thumbnail_mrl = fileitem_get_cover_image (fileitem);
            playitem = playlist_add (playlist, fileitem->title,
                                     fileitem->mrl);
            if (thumbnail_mrl && playitem) {
                playitem->thumbnail_mrl = ho_strdup (thumbnail_mrl);
            }
        }
        if (playitem && (filelist_length (fileitem->child_list) > 0)) {
            filelist_lock (fileitem->child_list);
            filelist_t *sl = fileitem->child_list;
            fileitem_t *si = filelist_first (sl);
            while (si) {
                playitem_append_alternative (playitem, si->mrl, si->title);
                si = filelist_next (sl, si);
            }
            filelist_unlock (fileitem->child_list);
        }
        if (playitem && fileitem->info_list
            && (l_list_length (fileitem->info_list) > 0)) {
            playitem->info_list = l_list_new ();
            char *i = l_list_first (fileitem->info_list);
            while (i) {
                l_list_append (playitem->info_list, ho_strdup (i));
                i = l_list_next (fileitem->info_list, i);
            }
        }
        break;
    case FILE_TYPE_FOLDER:
    case FILE_TYPE_MEDIAMARKS:
        if (filelist_length (fileitem->child_list) > 0) {
            filelist_lock (fileitem->child_list);
            filelist_t *sl = fileitem->child_list;
            fileitem_t *si = filelist_first (sl);
            while (si) {
                playitem = playlist_add_fileitem (playlist, si);
                si = filelist_next (sl, si);
            }
            filelist_unlock (fileitem->child_list);
        }
        break;
    case FILE_TYPE_PLAYLIST:
        if (filelist_length (fileitem->child_list) > 0) {
            filelist_lock (fileitem->child_list);
            filelist_t *sl = fileitem->child_list;
            fileitem_t *si = filelist_first (sl);
            while (si) {
                playitem = playlist_add_fileitem (playlist, si);
                si = filelist_next (sl, si);
            }
            filelist_unlock (fileitem->child_list);
        }
        else {
            playlist_load (playlist, fileitem->mrl);
        }
        break;
    default:
        error (_("Unknown filetype: %d!"), fileitem->type);
        break;
    }

    playlist->call_change_cb = true;
    playlist_changed (playlist);
    playlist_unlock (playlist);

    return playitem;
}


void
playlist_move_up (playlist_t * playlist, playitem_t * playitem)
{
    assert (ho_checkref (playlist));
    assert (ho_checkref (playitem));

    playlist_lock (playlist);

    l_list_move_up (playlist->list, playitem);

    playlist_changed (playlist);
    playlist_unlock (playlist);
}


void
playlist_move_down (playlist_t * playlist, playitem_t * playitem)
{
    assert (ho_checkref (playlist));
    assert (ho_checkref (playitem));

    playlist_lock (playlist);

    l_list_move_down (playlist->list, playitem);

    playlist_changed (playlist);
    playlist_unlock (playlist);
}


void
playlist_remove (playlist_t * playlist, playitem_t * playitem)
{
    playlist_lock (playlist);

    assert (ho_checkref (playlist));
    assert (ho_checkref (playitem));
    assert (playitem->parent_list == playlist);

    if (playlist->current == playitem) {
        playlist->current = NULL;
    }

    playlist->playback_length -= playitem->playback_length;

    l_list_remove (playlist->list, playitem);
    ho_badref (playitem);
    ho_decref (playitem);

    playlist_changed (playlist);
    playlist_unlock (playlist);
}


static playitem_t *
playlist_random_next (playlist_t * playlist)
{
    assert (ho_checkref (playlist));

    if (!l_list_length (playlist->list))
        return NULL;

    int num_played = 0;
    playitem_t *item = l_list_first (playlist->list);
    while (item) {
        if (item->played) {
            num_played++;
        }
        item = l_list_next (playlist->list, item);
    }
    if (num_played == l_list_length (playlist->list)) {
        return NULL;
    }

    int found = 0;
    while (!found) {
        int next =
            (int) ((double) l_list_length (playlist->list) *
                   ((double) random () / (double) RAND_MAX));
        playitem_t *item = l_list_first (playlist->list);
        while (item && next > 0) {
            next -= 1;
            item = l_list_next (playlist->list, item);
        }
        if (next == 0) {
            if (!item->played) {
                return item;
            }
        }
    }

    return NULL;
}


playitem_t *
playlist_get_first (playlist_t * playlist)
{
    assert (ho_checkref (playlist));

    playlist_lock (playlist);

    playitem_t *item = NULL;
    switch (playlist->playmode) {
    case PLAYBACK_MODE_NORMAL:
    case PLAYBACK_MODE_REPEAT:
        item = l_list_first (playlist->list);
        break;
    case PLAYBACK_MODE_RANDOM:
        srandom ((unsigned int) time (NULL));
        item = l_list_first (playlist->list);
        while (item) {
            item->played = false;
            item = l_list_next (playlist->list, item);
        }
        item = playlist_random_next (playlist);
        break;
    }

    playlist_set_current (playlist, item);

    playlist_unlock (playlist);

    return playlist->current;
}


playitem_t *
playlist_get_prev (playlist_t * playlist)
{
    assert (ho_checkref (playlist));

    playlist_lock (playlist);

    if (!playlist->current)
        return NULL;

    playitem_t *item = NULL;
    switch (playlist->playmode) {
    case PLAYBACK_MODE_NORMAL:
        item = l_list_prev (playlist->list, playlist->current);
        break;
    case PLAYBACK_MODE_REPEAT:
        item = l_list_prev (playlist->list, playlist->current);
        if (!item) {
            item = l_list_last (playlist->list);
        }
        break;
    case PLAYBACK_MODE_RANDOM:
        item = playlist_random_next (playlist);
        break;
    }

    playlist_set_current (playlist, item);

    playlist_unlock (playlist);

    return playlist->current;
}


static playitem_t *
playlist_get_next_to_play (playlist_t * playlist)
{
    if (!playlist->current) {
        return NULL;
    }

    playitem_t *item = NULL;
    switch (playlist->playmode) {
    case PLAYBACK_MODE_NORMAL:
        item = l_list_next (playlist->list, playlist->current);
        break;
    case PLAYBACK_MODE_REPEAT:
        item = l_list_next (playlist->list, playlist->current);
        if (!item) {
            item = l_list_first (playlist->list);
        }
        break;
    case PLAYBACK_MODE_RANDOM:
        item = playlist_random_next (playlist);
        break;
    }

    return item;
}


playitem_t *
playlist_get_next (playlist_t * playlist)
{
    assert (ho_checkref (playlist));

    playlist_lock (playlist);

    playlist_set_current (playlist, playlist_get_next_to_play (playlist));

    playlist_unlock (playlist);

    return playlist->current;
}


playitem_t *
playlist_set_current (playlist_t * playlist, playitem_t * current)
{
    assert (ho_checkref (playlist));

    playlist_lock (playlist);

    if (playlist->current != current) {
        playlist->current = current;

        if (playlist->current) {
            playlist->current->played = true;
        }

        playlist_changed (playlist);
    }

    playlist_unlock (playlist);

    return playlist->current;
}


playitem_t *
playlist_get_current (playlist_t * playlist)
{
    assert (ho_checkref (playlist));

    return playlist->current;
}


bool
playlist_current_is_last (playlist_t * playlist)
{
    assert (ho_checkref (playlist));

    playlist_lock (playlist);

    bool result = (playlist_get_next_to_play (playlist) == NULL);

    playlist_unlock (playlist);

    return result;
}


playitem_t *
playlist_set_current_pos (playlist_t * playlist, int pos)
{
    assert (ho_checkref (playlist));

    playlist_lock (playlist);

    playitem_t *cur = NULL;
    if ((pos >= 0) && (pos < playlist_length (playlist))) {
        int i = 0;
        cur = l_list_first (playlist->list);
        while (cur) {
            if (i == pos) {
                break;
            }
            cur = l_list_next (playlist->list, cur);
            i++;
        }
    }

    playlist_set_current (playlist, cur);

    playlist_unlock (playlist);

    return playlist->current;
}


int
playlist_get_current_pos (playlist_t * playlist)
{
    assert (ho_checkref (playlist));

    playlist_lock (playlist);

    int p = -1;
    int i = 0;
    playitem_t *cur = l_list_first (playlist->list);
    while (cur) {
        if (cur == playlist->current) {
            p = i;
            break;
        }
        cur = l_list_next (playlist->list, cur);
        i++;
    }

    playlist_unlock (playlist);

    return p;
}


playitem_t *
playlist_first (playlist_t * playlist)
{
    assert (ho_checkref (playlist));

    return l_list_first (playlist->list);
}


playitem_t *
playlist_next (playlist_t * playlist, playitem_t * current)
{
    assert (ho_checkref (playlist));
    assert (current);

    return l_list_next (playlist->list, current);
}


playitem_t *
playlist_prev (playlist_t * playlist, playitem_t * current)
{
    assert (ho_checkref (playlist));
    assert (current);

    return l_list_prev (playlist->list, current);
}


int
playlist_length (playlist_t * playlist)
{
    assert (ho_checkref (playlist));

    return l_list_length (playlist->list);
}


playback_mode_t
playlist_get_playmode (playlist_t * playlist)
{
    assert (ho_checkref (playlist));

    return playlist->playmode;
}


const char *
playlist_get_playmode_string (playlist_t * playlist, bool s)
{
    const char *mode = NULL;

    assert (ho_checkref (playlist));

    switch (playlist->playmode) {
    case PLAYBACK_MODE_NORMAL:
        mode = s ? _("N") : _("Normal Mode");
        break;
    case PLAYBACK_MODE_REPEAT:
        mode = s ? _("L") : _("Loop Mode");
        break;
    case PLAYBACK_MODE_RANDOM:
        mode = s ? _("S") : _("Shuffle Mode");
        break;
    }

    return mode;
}


void
playlist_set_playmode (playlist_t * playlist, playback_mode_t playmode)
{
    assert (ho_checkref (playlist));

    playlist_lock (playlist);

    playlist->playmode = playmode;

    playlist_changed (playlist);
    playlist_unlock (playlist);
}


playback_status_t
playlist_get_playstatus (playlist_t * playlist)
{
    assert (ho_checkref (playlist));

    return playlist->playstatus;
}


void
playlist_set_playstatus (playlist_t * playlist, playback_status_t playstatus)
{
    assert (ho_checkref (playlist));

    playlist_lock (playlist);

    playlist->playstatus = playstatus;

    playlist_changed (playlist);
    playlist_unlock (playlist);
}


bool
playlist_load (playlist_t * playlist, const char *mrl)
{
    assert (ho_checkref (playlist));

    playlist_lock (playlist);
    playlist->call_change_cb = false;

    bool res = false;
    const char *lmrl = mrl ? mrl : playlist->mrl;
    if (mrl_is_playlist_oxp (lmrl)) {
        res = playlist_xml_load (playlist, lmrl);
    }
    else if (mrl_is_playlist_m3u (lmrl)) {
        res = playlist_m3u_load (playlist, lmrl);
    }
    else if (mrl_is_playlist_pls (lmrl)) {
        res = playlist_pls_load (playlist, lmrl);
    }

    playlist->call_change_cb = true;
    playlist_changed (playlist);
    playlist_unlock (playlist);

    return res;
}


bool
playlist_save (playlist_t * playlist, const char *mrl)
{
    assert (ho_checkref (playlist));

    playlist_lock (playlist);

    bool res = false;
    const char *lmrl = mrl ? mrl : playlist->mrl;
    if (mrl_is_playlist_oxp (lmrl)) {
        res = playlist_xml_save (playlist, lmrl);
    }
    else if (mrl_is_playlist_m3u (lmrl)) {
        res = playlist_m3u_save (playlist, lmrl);
    }
    else if (mrl_is_playlist_pls (lmrl)) {
        res = playlist_pls_save (playlist, lmrl);
    }

    playlist_unlock (playlist);

    return res;
}


const char *
playitem_get_cover_image (playitem_t * playitem)
{
    assert (ho_checkref (playitem));
    assert (ho_checkref (playitem->parent_list));
    assert (playlist_is_locked (playitem->parent_list));

    const char *mrl = NULL;
    if (is_cover_image_valid (playitem->thumbnail_mrl)) {
        mrl = playitem->thumbnail_mrl;
    }
    else if (mrl_is_image (playitem->mrl)) {
        ho_free (playitem->thumbnail_mrl);
        playitem->thumbnail_mrl = ho_strdup (playitem->mrl);
        mrl = playitem->thumbnail_mrl;
    }
    else {
        ho_free (playitem->thumbnail_mrl);
        playitem->thumbnail_mrl = get_cover_image_from_mrl (playitem->mrl);
        mrl = playitem->thumbnail_mrl;
    }

    return mrl;
}


void
playlist_set_length_cb (playitem_length_cb_t cb)
{
    playitem_length_cb = cb;
}


int
playlist_get_playback_length (playlist_t * playlist)
{
    assert (ho_checkref (playlist));

    return playlist->playback_length;
}


l_list_t *
playlist_get_info_list (playlist_t * playlist)
{
    assert (ho_checkref (playlist));
    assert (playlist_is_locked (playlist));

    if (playlist->info_list) {
        return playlist->info_list;
    }

    playlist->info_list = l_list_new ();

    char *t = NULL;
    {
        int count = playlist_length (playlist);
        int length = playlist->playback_length;
        char *time = meta_info_get_playback_length_str (length);

        if (length && time) {
            char *f = _("The current playlist contains %d %s with a total "
                        "playback time of %s.");
            char *e = (count == 1) ? _("entry") : _("entries");
            t = ho_strdup_printf (f, count, e, time);
        }
        else if (count == 0) {
            char *f = _("The current playlist contains no entries.");
            t = ho_strdup (f);
        }
        else {
            char *f = _("The current playlist contains %d %s.");
            char *e = (count == 1) ? _("entry") : _("entries");
            t = ho_strdup_printf (f, count, e);
        }

        ho_free (time);
    }

    char *m = NULL;
    {
        switch (playlist->playmode) {
        case PLAYBACK_MODE_NORMAL:
            m = _("The titles in this playlist will be "
                  "played in normal order.");
            break;
        case PLAYBACK_MODE_REPEAT:
            m = _("The titles in this playlist will be "
                  "played repeatedly.");
            break;
        case PLAYBACK_MODE_RANDOM:
            m = _("The titles in this playlist will be "
                  "played in random order.");
            break;
        }
    }

    char *i = ho_strdup_printf ("%s %s", t ? t : "", m ? m : "");
    l_list_append (playlist->info_list, i);
    ho_free (t);

    return playlist->info_list;
}


static void
playitem_info_list_append (l_list_t * info_list, const char *title,
                           const char *value)
{
    char *t = xml_escape_string (value, XML_ESCAPE_NO_QUOTE);
    char *i = ho_strdup_printf (INFO_LIST_FORMAT, title, t);
    l_list_append (info_list, i);
    ho_free (t);
}


l_list_t *
playitem_get_info_list (playitem_t * playitem)
{
    assert (ho_checkref (playitem));
    assert (ho_checkref (playitem->parent_list));
    assert (playlist_is_locked (playitem->parent_list));

    if (playitem->info_list) {
        return playitem->info_list;
    }

    playitem->info_list = l_list_new ();
    {
        if (false) {
            /* do nothing */
        }
        else if (mrl_is_networked (playitem->mrl) && playitem->title) {
            playitem_info_list_append (playitem->info_list, _("Title:"),
                                       playitem->title);
        }
#ifdef HAVE_LIBEXIF
        else if (mrl_is_image (playitem->mrl)) {
            char *info = NULL;
            char *date = exif_info_get (EXIF_INFO_DATE, playitem->mrl);
            char *time = exif_info_get (EXIF_INFO_TIME, playitem->mrl);
            if (date && time) {
                info = ho_strdup_printf (_("The selected image was "
                                           "taken on %s at %s."), date, time);
            }
            else if (date) {
                info = ho_strdup_printf (_("The selected image was "
                                           "taken on %s."), date);
            }
            else if (time) {
                info = ho_strdup_printf (_("The selected image was "
                                           "taken at %s."), time);
            }
            if (info) {
                l_list_append (playitem->info_list, info);
            }

            ho_free (date);
            ho_free (time);
        }
#endif /* HAVE_LIBEXIF */
        else {
            char *mrl = playitem->mrl;
            time_t now = time (NULL);
            struct tm *tm_now = localtime (&now);
            int current_year = tm_now->tm_year + 1900;

            char *title = meta_info_get (META_INFO_TITLE, mrl);
            if (title && (strlen (title) > 0)) {
                playitem_info_list_append (playitem->info_list, _("Title:"),
                                           title);
            }
            else if (playitem->title) {
                playitem_info_list_append (playitem->info_list, _("Title:"),
                                           playitem->title);
            }
            ho_free (title);

            char *album = meta_info_get (META_INFO_ALBUM, mrl);
            if (album && (strlen (album) > 0)) {
                playitem_info_list_append (playitem->info_list, _("Album:"),
                                           album);
            }
            ho_free (album);

            char *artist = meta_info_get (META_INFO_ARTIST, mrl);
            if (artist && (strlen (artist) > 0)) {
                playitem_info_list_append (playitem->info_list, _("Artist:"),
                                           artist);
            }
            ho_free (artist);

            char *genre = meta_info_get (META_INFO_GENRE, mrl);
            if (genre && (strlen (genre) > 0)) {
                playitem_info_list_append (playitem->info_list, _("Genre:"),
                                           genre);
            }
            ho_free (genre);

            char *year = meta_info_get (META_INFO_YEAR, mrl);
            if (year && (atoi (year) > 0) && (atoi (year) <= current_year)) {
                playitem_info_list_append (playitem->info_list, _("Year:"),
                                           year);
            }
            ho_free (year);

            char *track = meta_info_get (META_INFO_TRACK_NUMBER, mrl);
            if (track && (strlen (track) > 0)) {
                playitem_info_list_append (playitem->info_list, _("Track:"),
                                           track);
            }
            ho_free (track);

            int length = playitem->playback_length;
            char *time = meta_info_get_playback_length_str (length);
            if (length && time) {
                playitem_info_list_append (playitem->info_list, _("Length:"),
                                           time);
            }
            ho_free (time);

            char *quality = mrl_get_quality (mrl);
            if (quality && (strlen (quality) > 0)) {
                playitem_info_list_append (playitem->info_list, _("Quality:"),
                                           _(quality));
            }
            ho_free (quality);
        }
    }

    return playitem->info_list;
}
