
/*
 * Copyright (C) 2004-2005 Maximilian Schwerin
 *
 * This file is part of oxine a free media player.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: playlist_xml.c 3806 2009-04-27 15:13:17Z mschwerin $
 *
 */
#include "config.h"

#include <errno.h>
#include <stdio.h>
#include <string.h>

#include "codeset.h"
#include "environment.h"
#include "heap.h"
#include "i18n.h"
#include "logger.h"
#include "playlist_xml.h"
#include "utils.h"
#include "xmlparser.h"

#define OXINE_PL_XML "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
#define OXINE_PL_DTD "<!DOCTYPE oxine_playlist SYSTEM \"http://oxine.sf.net/DTD/playlist.dtd\">\n\n"


static void
playlist_print_mrl (FILE * f, recode_t * xr, const char *node_name,
                    const char *mrl)
{
    char *esc_mrl = xml_escape (mrl, xr, XML_ESCAPE_NO_QUOTE);
    if (esc_mrl) {
        fprintf (f, "    <%s href=\"%s\" />\n", node_name, esc_mrl);
    }
    ho_free (esc_mrl);
}


static void
playlist_write (FILE * f, recode_t * xr, playlist_t * playlist)
{
    int cur_pos = playlist_get_current_pos (playlist);
    char current[5] = "none";
    if (playlist_get_current_pos (playlist) >= 0) {
        snprintf (current, 5, "%d", cur_pos);
    }

    char *playmode = NULL;
    switch (playlist->playmode) {
    case PLAYBACK_MODE_NORMAL:
        playmode = "normal";
        break;
    case PLAYBACK_MODE_REPEAT:
        playmode = "repeat";
        break;
    case PLAYBACK_MODE_RANDOM:
        playmode = "random";
        break;
    default:
        break;
    }

    char *playstatus = NULL;
    switch (playlist->playstatus) {
    case PLAYBACK_STATUS_STOPPED:
        playstatus = "stopped";
        break;
    case PLAYBACK_STATUS_PAUSED:
        playstatus = "paused";
        break;
    case PLAYBACK_STATUS_PLAYING:
        playstatus = "playing";
        break;
    default:
        break;
    }

    fprintf (f, "<oxine_playlist current=\"%s\" "
             "playmode=\"%s\" playstatus=\"%s\">\n",
             current, playmode, playstatus);

    playitem_t *playitem = playlist_first (playlist);
    while (playitem) {
        fprintf (f, "  <playitem played=\"%s\" start_time=\"%d\">\n",
                 playitem->played ? "true" : "false", playitem->start_time);

        char *esc_str0 = xml_escape (playitem->title,
                                     xr, XML_ESCAPE_NO_QUOTE);

        fprintf (f, "    <title>%s</title>\n", esc_str0);
        playlist_print_mrl (f, xr, "mrl", playitem->mrl);
        playlist_print_mrl (f, xr, "subtitle_mrl", playitem->subtitle_mrl);
        playlist_print_mrl (f, xr, "thumbnail_mrl", playitem->thumbnail_mrl);

        if (playitem->alternative_list) {
            playlist_write (f, xr, playitem->alternative_list);
        }

        ho_free (esc_str0);

        fprintf (f, "  </playitem>\n");
        playitem = playlist_next (playlist, playitem);
    }

    fprintf (f, "</oxine_playlist>\n");
}


bool
playlist_xml_save (playlist_t * playlist, const char *mrl)
{
    FILE *f = fopen (mrl, "w");
    if (f == NULL) {
        error (_("Could not open '%s': %s!"), mrl, strerror (errno));
        return false;
    }

    recode_t *xr = recode_init (USE_SYSTEM_ENCODING, "UTF-8");

    fprintf (f, OXINE_PL_XML);
    fprintf (f, OXINE_PL_DTD);

    playlist_write (f, xr, playlist);

    fclose (f);

    recode_done (xr);

    return true;
}

static void read_playlist (const char *xml_mrl, playlist_t * playlist,
                           xml_node_t * node);

static void
read_playitem (const char *xml_mrl, playlist_t * playlist, xml_node_t * node)
{
    bool played = false;
    int start_time = 0;
    char *mrl = NULL;
    char *title = NULL;
    char *subtitle_mrl = NULL;
    char *thumbnail_mrl = NULL;
    playlist_t *alternative_list = NULL;

    recode_t *xr = recode_init ("UTF-8", USE_SYSTEM_ENCODING);

    const char *str_played = xml_parser_get_property (node, "played");
    if (str_played && (strcasecmp (str_played, "true") == 0)) {
        played = true;
    }

    const char *str_start_time = xml_parser_get_property (node, "start_time");
    if (str_start_time) {
        start_time = atoi (str_start_time);
    }

    node = node->child;
    while (node) {
        if (strcasecmp (node->name, "title") == 0) {
            title = recode (xr, node->data);
        }
        else if (strcasecmp (node->name, "mrl") == 0) {
            const char *href = xml_parser_get_property (node, "href");
            mrl = recode (xr, href);
        }
        else if (strcasecmp (node->name, "subtitle_mrl") == 0) {
            const char *href = xml_parser_get_property (node, "href");
            subtitle_mrl = recode (xr, href);
        }
        else if (strcasecmp (node->name, "thumbnail_mrl") == 0) {
            const char *href = xml_parser_get_property (node, "href");
            thumbnail_mrl = recode (xr, href);
        }
        else if (strcasecmp (node->name, "oxine_playlist") == 0) {
            alternative_list = playlist_new (NULL, NULL);
            read_playlist (xml_mrl, alternative_list, node->child);
        }
        else {
            error (_("Unknown name '%s' for node!"), node->name);
        }

        node = node->next;
    }

    if (mrl) {
        playitem_t *playitem = playlist_add (playlist, title, mrl);
        playitem->played = played;
        playitem->start_time = start_time;
        if (subtitle_mrl) {
            ho_free (playitem->subtitle_mrl);
            playitem->subtitle_mrl = subtitle_mrl;
        }
        if (thumbnail_mrl) {
            ho_free (playitem->thumbnail_mrl);
            playitem->thumbnail_mrl = thumbnail_mrl;
        }
        if (alternative_list) {
            ho_setref (playitem->alternative_list, alternative_list);
        }
    }
    else {
        error (_("Found an entry that did not contain a MRL!"));

        if (subtitle_mrl) {
            ho_free (subtitle_mrl);
        }
        if (thumbnail_mrl) {
            ho_free (thumbnail_mrl);
        }
        if (alternative_list) {
            ho_decref (alternative_list);
        }
    }

    ho_free (mrl);
    ho_free (title);

    recode_done (xr);
}


static void
read_playlist (const char *xml_mrl, playlist_t * playlist, xml_node_t * node)
{
    while (node) {
        if (!strcasecmp (node->name, "playitem")) {
            read_playitem (xml_mrl, playlist, node);
        }
        else {
            error (_("Unknown name '%s' for node!"), node->name);
        }

        node = node->next;
    }
}


bool
playlist_xml_load (playlist_t * playlist, const char *mrl)
{
    bool success = false;
    int size;
    xml_node_t *node;

    char *buffer = read_entire_file (mrl, &size);
    if (!buffer) {
        goto buffer_free;
    }

    xml_parser_init (buffer, size, XML_PARSER_CASE_INSENSITIVE);
    if (xml_parser_build_tree (&node) != XML_PARSER_OK) {
        error (_("Parsing '%s' failed!"), mrl);
        goto parser_free;
    }

    if (strcasecmp (node->name, "oxine_playlist")) {
        error (_("Root node of '%s' must be '%s'!"), mrl, "oxine_playlist");
        goto parser_free;
    }

    playlist_change_cb_t change_cb = playlist->change_cb;
    playlist->change_cb = NULL;

    read_playlist (mrl, playlist, node->child);

    const char *str_current = xml_parser_get_property (node, "current");
    if (str_current && (strcasecmp (str_current, "none") != 0)) {
        playlist_set_current_pos (playlist, atoi (str_current));
    }

    const char *str_mode = xml_parser_get_property (node, "playmode");
    if (str_mode) {
        if (strcasecmp (str_mode, "normal") == 0) {
            playlist_set_playmode (playlist, PLAYBACK_MODE_NORMAL);
        }
        else if (strcasecmp (str_mode, "repeat") == 0) {
            playlist_set_playmode (playlist, PLAYBACK_MODE_REPEAT);
        }
        else if (strcasecmp (str_mode, "random") == 0) {
            playlist_set_playmode (playlist, PLAYBACK_MODE_RANDOM);
        }
    }

    const char *str_status = xml_parser_get_property (node, "playstatus");
    if (str_status) {
        if (strcasecmp (str_status, "stopped") == 0) {
            playlist_set_playstatus (playlist, PLAYBACK_STATUS_STOPPED);
        }
        else if (strcasecmp (str_status, "paused") == 0) {
            playlist_set_playstatus (playlist, PLAYBACK_STATUS_PAUSED);
        }
        else if (strcasecmp (str_status, "playing") == 0) {
            playlist_set_playstatus (playlist, PLAYBACK_STATUS_PLAYING);
        }
    }

    playlist->change_cb = change_cb;

    success = true;
  parser_free:
    xml_parser_free_tree (node);
  buffer_free:
    ho_free (buffer);

    return success;
}
