
/*
 * Copyright (C) 2002-2003 Stefan Holst
 * Copyright (C) 2004-2005 Maximilian Schwerin
 *
 * This file is part of oxine a free media player.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: scheduler.c 3913 2009-05-14 11:44:54Z mschwerin $
 *
 */
#include "config.h"

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>

#include "i18n.h"
#include "heap.h"
#include "list.h"
#include "logger.h"
#include "mutex.h"
#include "scheduler.h"

typedef struct {
    time_t start_time;
    l_list_t *jobs;
    pthread_t scheduler_thread;
    pthread_mutex_t jobs_change;
    pthread_mutex_t job_execute;
    pthread_mutex_t wait_mutex;
    pthread_cond_t reorganize;
    int job_id;
    int keep_going;
} ox_scheduler_t;

typedef struct {
    struct timespec ts;                                         /* when to do */
    scheduler_cb_t cb;
    void *cb_data;
    int id;
    int priority;
} job_t;

static ox_scheduler_t *ox_scheduler = NULL;

static void *
scheduler_thread (void *data)
{
    int ret;

    info (_("Successfully started scheduler thread."));
    debug ("  scheduler thread: 0x%X", (int) pthread_self ());

    if (!ox_scheduler) {
        pthread_exit (NULL);
        return NULL;
    }

    mutex_lock (&ox_scheduler->wait_mutex);
    while (ox_scheduler->keep_going) {
        job_t *next_job;
        struct timespec ts;

        mutex_lock (&ox_scheduler->jobs_change);
        if ((next_job = l_list_last (ox_scheduler->jobs))) {
            memcpy (&ts, &next_job->ts, sizeof (struct timespec));
        }
        mutex_unlock (&ox_scheduler->jobs_change);

        if (!next_job) {
            ret = pthread_cond_wait (&ox_scheduler->reorganize,
                                     &ox_scheduler->wait_mutex);
            continue;
        }
        else {
            ret = pthread_cond_timedwait (&ox_scheduler->reorganize,
                                          &ox_scheduler->wait_mutex, &ts);
        }

        if (ret == ETIMEDOUT) {
            scheduler_cb_t cb = NULL;
            void *cb_data = NULL;

            mutex_lock (&ox_scheduler->job_execute);

            mutex_lock (&ox_scheduler->jobs_change);
            job_t *job = l_list_last (ox_scheduler->jobs);
            if (job == next_job) {
                cb = job->cb;
                cb_data = job->cb_data;
                l_list_remove (ox_scheduler->jobs, job);
                ho_free (job);
            }
            mutex_unlock (&ox_scheduler->jobs_change);

            if (cb) {
                cb (cb_data);
            }

            mutex_unlock (&ox_scheduler->job_execute);
        }
    }
    mutex_unlock (&ox_scheduler->wait_mutex);
    pthread_exit (NULL);

    return NULL;
}


void
start_scheduler (void)
{
    struct timeval tv;
    struct timezone tz;

    ox_scheduler = ho_new (ox_scheduler_t);

    gettimeofday (&tv, &tz);
    ox_scheduler->start_time = tv.tv_sec;

    ox_scheduler->jobs = l_list_new ();
    pthread_mutex_init (&ox_scheduler->wait_mutex, NULL);
    pthread_mutex_init (&ox_scheduler->jobs_change, NULL);
    pthread_mutex_init (&ox_scheduler->job_execute, NULL);
    pthread_cond_init (&ox_scheduler->reorganize, NULL);
    ox_scheduler->job_id = 0;
    ox_scheduler->keep_going = 1;

    if (pthread_create (&ox_scheduler->scheduler_thread, NULL,
                        scheduler_thread, ox_scheduler) != 0) {
        fatal (_("Could not create scheduler thread: %s!"), strerror (errno));
        abort ();
    }
}


static void
job_free_cb (void *j1)
{
    ho_free (j1);
}


void
stop_scheduler (void)
{
    void *ret = NULL;

    if (!ox_scheduler) {
        return;
    }

    ox_scheduler->keep_going = 0;
    pthread_cond_signal (&ox_scheduler->reorganize);

    pthread_join (ox_scheduler->scheduler_thread, ret);

    l_list_free (ox_scheduler->jobs, job_free_cb);

    pthread_mutex_destroy (&ox_scheduler->wait_mutex);
    pthread_mutex_destroy (&ox_scheduler->jobs_change);
    pthread_cond_destroy (&ox_scheduler->reorganize);

    ho_free (ox_scheduler);
    ox_scheduler = NULL;

    info (_("Successfully stopped scheduler thread."));
}


static bool
job_swap_cb (void *p1, void *p2)
{
    job_t *j1 = (job_t *) p1;
    job_t *j2 = (job_t *) p2;

    return j1->priority < j2->priority;
}


int
schedule_job (int delay_ms, scheduler_cb_t cb, void *cb_data)
{
    struct timeval tv;
    struct timezone tz;

    if (!ox_scheduler) {
        return -1;
    }

    mutex_lock (&ox_scheduler->jobs_change);
    {
        job_t *job = ho_new (job_t);

        gettimeofday (&tv, &tz);

        time_t secs = tv.tv_sec;
        time_t msec = tv.tv_usec / 1000;

        secs = (delay_ms / 1000) + secs;
        msec = (delay_ms % 1000) + msec;

        secs = (msec / 1000) + secs;
        msec = (msec % 1000);

        job->ts.tv_sec = secs;
        job->ts.tv_nsec = msec * 1000 * 1000;

        job->cb = cb;
        job->cb_data = cb_data;
        ox_scheduler->job_id = (ox_scheduler->job_id + 1) % 10000;
        job->id = ox_scheduler->job_id;

        job->priority = (secs - ox_scheduler->start_time) * 1000 + msec;

        l_list_append (ox_scheduler->jobs, job);
        l_list_sort (ox_scheduler->jobs, job_swap_cb);
    }
    mutex_unlock (&ox_scheduler->jobs_change);

    pthread_cond_signal (&ox_scheduler->reorganize);

    return ox_scheduler->job_id;
}


void *
cancel_job (int * job_id)
{
    void *cb_data = NULL;

    if (!ox_scheduler) {
        return NULL;
    }
    if (!job_id) {
        return NULL;
    }
    if (*job_id < 0) {
        return NULL;
    }

    mutex_lock (&ox_scheduler->jobs_change);
    {
        job_t *job = l_list_first (ox_scheduler->jobs);
        while (job) {
            if (job->id == *job_id) {
                l_list_remove (ox_scheduler->jobs, job);
                cb_data = job->cb_data;
                ho_free (job);
                break;
            }
            job = l_list_next (ox_scheduler->jobs, job);
        }
        *job_id = -1;
    }
    mutex_unlock (&ox_scheduler->jobs_change);

    pthread_cond_signal (&ox_scheduler->reorganize);

    return cb_data;
}


void
lock_job_mutex (void)
{
    if (!ox_scheduler) {
        return;
    }
    mutex_lock (&ox_scheduler->job_execute);
}


void
unlock_job_mutex (void)
{
    if (!ox_scheduler) {
        return;
    }
    mutex_unlock (&ox_scheduler->job_execute);
}
