
/*
 * Copyright (C) 2004-2005 Maximilian Schwerin
 *
 * This file is part of oxine a free media player.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: utils.c 3675 2008-12-19 11:10:29Z mschwerin $
 *
 */
#include "config.h"

#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <sys/types.h>
#include <sys/wait.h>

#include "heap.h"
#include "logger.h"
#include "utils.h"


char *
replace_chars (char *str, char old, char new)
{
    int i = 0;
    int l = strlen (str);

    for (i = 0; i < l; i++) {
        if (str[i] == old) {
            str[i] = new;
        }
    }

    return str;
}


static void
replace_special_char (char *str, char *src, char dst)
{
    char *d;
    if (!str || !src) {
        return;
    }

    while ((d = strstr (str, src))) {
        int i = 0;
        int l = strlen (d);
        for (i = 1; i < l; i++) {
            d[i] = d[i + 1];
        }
        d[0] = dst;
    }
}


char *
replace_special_chars (char *str)
{
    if (!str) {
        return NULL;
    }

    replace_special_char (str, "ä", 'a');
    replace_special_char (str, "Ä", 'a');
    replace_special_char (str, "ö", 'o');
    replace_special_char (str, "Ö", 'o');
    replace_special_char (str, "ü", 'u');
    replace_special_char (str, "Ü", 'u');
    replace_special_char (str, "ß", 's');

    return str;
}


bool
swap_strings (const char *s1, const char *s2)
{
    if (!s1 || !s2) {
        return false;
    }

    char *n1 = ho_strdup (s1);
    char *n2 = ho_strdup (s2);

    n1 = replace_special_chars (n1);
    n2 = replace_special_chars (n2);

    bool swap = (strcasecmp (n1, n2) > 0);

    ho_free (n1);
    ho_free (n2);

    return swap;
}

char *
trim_whitespace (char *str)
{
    if (!str) {
        return NULL;
    }

    char *start = str;
    while (isspace (start[0])) {
        ++start;
    }

    char *end = start + strlen (start);
    while ((end > start) && isspace (end[-1])) {
        --end;
    }
    *end = '\0';

    memmove (str, start, (end - start) + 1);

    return str;
}


char *
conv_lowercase (char *str)
{
    if (!str) {
        return NULL;
    }

    int i = 0;
    int l = strlen (str);
    for (; i < l; i++)
        str[i] = tolower (str[i]);

    return str;
}


bool
execute_shell (const char *command, int retval)
{
    pid_t pid = 0;

    if ((pid = start_process (command, 0, NULL)) != 0) {
        return (wait_process (pid) == retval) ? true : false;
    }
    else {
        return false;
    }
}


void
arguments_to_argv (const char *arguments, char **argv, int *_argc)
{
    if (!arguments || (strlen (arguments) == 0)) {
        return;
    }

    int i = 0;
    int c = 0;
    int argc = *_argc;
    char *trimmed = ho_strdup (arguments);
    trim_whitespace (trimmed);

    argv[argc] = ho_strdup (trimmed);
    for (i = 0; i < strlen (trimmed); i++) {
        if (trimmed[i] == '\\') {
            argv[argc][c++] = trimmed[++i];
        }
        else if (trimmed[i] != ' ') {
            argv[argc][c++] = trimmed[i];
        }
        else {
            argv[argc][c] = '\0';
            if (strlen (argv[argc]) > 0) {
                argv[++argc] = ho_strdup (&trimmed[i + 1]);
            }
            c = 0;
        }
    }
    argv[argc][c] = '\0';
    argv[++argc] = NULL;

    ho_free (trimmed);

    *_argc = argc;
}


pid_t
start_process_argv (char **argv, int readfd, int *pipefd)
{
    int _pipefd[2] = { 0, 0 };
    if (pipefd && (pipe (_pipefd) != 0)) {
        error (_("Failed to create pipe: %s!"), strerror (errno));
    }

#if 0
    int i = 0;
    for (i = 0; argv[i]; i++) {
        debug ("argv[%02d] %s", i, argv[i]);
    }
#endif

    pid_t pid = fork ();
    if (pid == 0) {
        int i = 0;
        for (i = sysconf (_SC_OPEN_MAX); i > 2; i--) {
            if (i != _pipefd[1]) {
                close (i);
            }
        }

        if (pipefd) {
            dup2 (_pipefd[1], readfd);
            close (_pipefd[1]);
        }

        execvp (argv[0], argv);

        pid = 0;
        error (_("Execution of '%s' failed: %s!"), argv[0], strerror (errno));

        abort ();
    }
    else if (pid < 0) {
        pid = 0;
        error (_("Execution of '%s' failed: %s!"), argv[0], strerror (errno));
    }

    if (pipefd) {
        close (_pipefd[1]);
        *pipefd = _pipefd[0];
    }

    return pid;
}


pid_t
start_process (const char *command, int readfd, int *pipefd)
{
    if (!command || (strlen (command) == 0)) {
        return 0;
    }

    debug ("Running '%s'...", command);

    int argc = 0;
    char *argv[256];
    arguments_to_argv (command, argv, &argc);

    pid_t pid = start_process_argv (argv, readfd, pipefd);

    int i = 0;
    for (i = 0; argv[i] != NULL; i++) {
        ho_free (argv[i]);
    }

    return pid;
}


int
wait_process (pid_t pid)
{
    int status;

    waitpid (pid, &status, 0);

    return WIFEXITED (status) ? WEXITSTATUS (status) : -1;
}


bool
has_process_terminated (pid_t pid)
{
    int status;

    return (waitpid (pid, &status, WNOHANG) != 0);
}


char *
xml_escape (const char *str, recode_t * xr, xml_escape_quote_t quote_type)
{
    if (!str) {
        return NULL;
    }

    char *str_encoded = recode (xr, str);
    if (!str_encoded) {
        return ho_strdup (str);
    }

    char *str_escaped = xml_escape_string (str_encoded, quote_type);
    if (!str_escaped) {
        return str_encoded;
    }

    ho_free (str_encoded);
    return str_escaped;
}


#define IS_ALNUM(ch) (\
        ( (ch >= 'a') && (ch <= 'z') ) || \
        ( (ch >= 'A') && (ch <= 'Z') ) || \
        ( (ch >= '0') && (ch <= '9') ) || \
          (ch == '.') || \
          (ch == '-') || \
          (ch == '_') || \
          (ch == '~') || \
          (ch == '/') || \
          (ch == ':')    \
        )


char *
url_decode (const char *str)
{
    int i;
    int j = 0;
    int len;
    char *tmp;
    char hex[3];
    hex[2] = 0;

    if (!str) {
        return NULL;
    }

    len = strlen (str);
    tmp = (char *) ho_malloc (sizeof (char) * (len + 1));

    for (i = 0; i < len; i++, j++) {
        if (str[i] != '%') {
            tmp[j] = str[i];
        }
        else if (IS_ALNUM (str[i + 1])
                 && IS_ALNUM (str[i + 2])
                 && (i < (len - 2))) {
            hex[0] = str[i + 1];
            hex[1] = str[i + 2];
            tmp[j] = strtol (hex, NULL, 16);

            i += 2;
        }
        else {
            tmp[j] = '%';
        }
    }
    tmp[j] = 0;

    return tmp;

}


char *
url_encode (const char *str)
{
    int i;
    int j = 0;
    int len;
    char *tmp;

    if (!str) {
        return NULL;
    }

#if 0
    for (i = '!'; i <= 'z'; i++) {
        debug ("%c %s", (char) i, IS_ALNUM (i) ? "true" : "false");
    }
#endif

    len = strlen (str);
    tmp = (char *) ho_malloc ((sizeof (char) * 3 * len) + 1);

    for (i = 0; i < len; i++) {
        if (IS_ALNUM (str[i])) {
            tmp[j] = str[i];
        }
        else {
            snprintf (&tmp[j], 4, "%%%02X\n", (unsigned char) str[i]);
            j += 2;
        }
        j++;
    }
    tmp[j] = 0;

    return tmp;
}


const char *
nindex (const char *s, int c, int n)
{
    int i = 0;
    const char *ni = s;
    const char *si = s;
    for (i = 0; i < n; i++) {
        ni = index (si, c);
        si = ni ? ni + 1 : si;
    }
    return ni;
}


void
parse_time (int sec_in, int *d, int *h, int *m, int *s)
{
    *d = sec_in / 86400;
    sec_in -= *d * 86400;
    *h = sec_in / 3600;
    sec_in -= *h * 3600;
    *m = sec_in / 60;
    sec_in -= *m * 60;
    *s = sec_in;
}
