
/*
 * Copyright (C) 2004-2005 Maximilian Schwerin
 *
 * This file is part of oxine a free media player.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: utils.h 3997 2009-09-05 18:44:02Z mschwerin $
 *
 */

#ifndef HAVE_UTILS_H
#define HAVE_UTILS_H

#include "config.h"

#include <string.h>

#include "utils_config.h"
#include "utils_covers.h"
#include "utils_gui.h"
#include "utils_io.h"
#include "utils_lastfm.h"
#include "utils_mrl.h"
#include "utils_network.h"
#include "utils_playlist.h"

#include "codeset.h"
#include "types.h"
#include "xmlparser.h"


/**
 * Returns TRUE if s1 should be sorted after s2.
 */
bool swap_strings (const char *s1, const char *s2);


/**
 * Replaces all characters org with the character rpl. The passed string is
 * changed and returned.
 */
char *replace_chars (char *str, char org, char rpl);


/**
 * Replaces special characters with 'normal' character. The passed string is
 * changed and returned.
 */
char *replace_special_chars (char *str);


/**
 * Trims whitespace off the start and the end of a string. The passed string
 * is changed and returned.
 */
char *trim_whitespace (char *str);


/**
 * Converts a string to lowercase. The passed string is changed and returned.
 */
char *conv_lowercase (char *string);


/**
 * Returns TRUE if s1 contains s2.
 */
static inline bool
str_contains (const char *s1, const char *s2)
{
    return strstr (s1, s2);
}


/**
 * Returns TRUE if s1 contains s2.
 */
static inline bool
strc_contains (const char *s1, const char *s2)
{
    char *sl1 = ho_strdup (s1);
    char *sl2 = ho_strdup (s2);
    conv_lowercase (sl1);
    conv_lowercase (sl2);
    bool result = strstr (sl1, sl2);
    ho_free (sl1);
    ho_free (sl2);
    return result;
}


/**
 * Returns TRUE if s1 equals s2.
 */
static inline bool
str_equals (const char *s1, const char *s2)
{
    return (s1 && s2 && (strcmp (s1, s2) == 0));
}


/**
 * Returns TRUE if s1 equals s2.
 */
static inline bool
strc_equals (const char *s1, const char *s2)
{
    return (s1 && s2 && (strcasecmp (s1, s2) == 0));
}


/**
 * Returns TRUE if s1 starts with s2.
 */
static inline bool
str_starts_with (const char *s1, const char *s2)
{
    return (s1 && s2 && (strlen (s1) >= strlen (s2)) && (strncmp (s1, s2, strlen (s2)) == 0));
}


/**
 * Returns TRUE if s1 starts with s2.
 */
static inline bool
strc_starts_with (const char *s1, const char *s2)
{
    return (s1 && s2 && (strlen (s1) >= strlen (s2)) && (strncasecmp (s1, s2, strlen (s2)) == 0));
}


/**
 * Returns TRUE if s1 ends with s2.
 */
static inline bool
str_ends_with (const char *s1, const char *s2)
{
    return (s1 && s2 && (strlen (s1) >= strlen (s2)) && (strncmp (&s1[strlen (s1) - strlen (s2)], s2, strlen (s2)) == 0));
}


/**
 * Returns TRUE if s1 ends with s2.
 */
static inline bool
strc_ends_with (const char *s1, const char *s2)
{
    return (s1 && s2 && (strlen (s1) >= strlen (s2)) && (strncasecmp (&s1[strlen (s1) - strlen (s2)], s2, strlen (s2)) == 0));
}


/**
 * Executes a shell command.
 *
 * @param command               The command to execute.
 * @param retval                The expected return value.
 *
 * @return                      If the exit value returned by the process
 *                              equals the expected return value provided
 *                              by the caller TRUE is returned.
 */
bool execute_shell (const char *command, int retval);


/**
 * Splits a list of arguments into an argument vector.
 *
 * @param arguments             The arguments in one long string.
 * @param argv                  A pointer to the argument vector.
 * @param argc                  A pointer to the number of arguments in the
 *                              argument vector.
 */
void arguments_to_argv (const char *arguments, char **argv, int *argc);


/**
 * Starts a new process.
 *
 * @param command               The command to execute.
 * @param readfd                Filedescriptor of the stream to pipe to pipefd.
 * @param pipefd                Filedescriptor of our end of the pipe.
 *
 * @return                      On success the ID of new process is
 *                              returned. If an error occurs 0 is returned.
 */
pid_t start_process (const char *command, int readfd, int *pipefd);


/**
 * Starts a new process.
 *
 * @param argv                  The argument vector of the command to execute.
 * @param readfd                Filedescriptor of the stream to pipe to pipefd.
 * @param pipefd                Filedescriptor of our end of the pipe.
 *
 * @return                      On success the ID of new process is
 *                              returned. If an error occurs 0 is returned.
 */
pid_t start_process_argv (char **argv, int readfd, int *pipefd);


/**
 * Waits for the process with the given ID to terminate.
 *
 * @param pid                   The process ID of the process to wait for.
 *
 * @return                      The exit status of the process.
 */
int wait_process (pid_t pid);


/**
 * Returns TRUE if a process has been terminated.
 *
 * @param pid                   The process ID of the process to wait for.
 */
bool has_process_terminated (pid_t pid);


/**
 * Escapes and recodes a string to be written to an XML file.
 *
 * @param str                   The original string.
 * @param xr                    The recoding datestructure.
 * @param quote_type            The type of quotes the string 
 *                              will be enclosed in.
 */
char *xml_escape (const char *str, recode_t * xr,
                  xml_escape_quote_t quote_type);


/**
 * Decode a URL-encoded string.
 *
 * The returned string must be freed.
 */
char *url_decode (const char *str);


/**
 * URL-encode a string.
 *
 * The returned string must be freed.
 */
char *url_encode (const char *str);


/**
 * Returns a pointer to the n-th occurence of
 * the character c in the string s.
 */
const char *nindex (const char *s, int c, int n);


/**
 * Splits a time given in seconds into days, hours, minutes and seconds.
 */
void parse_time (int sec_in, int *d, int *h, int *m, int *s);


#define MEASURE_TIME_START                          \
    struct timeval ts;                              \
    struct timeval te;                              \
    struct timezone tz;                             \
    time_t duration = 0;                            \
    gettimeofday (&ts, &tz);

#define MEASURE_TIME_STOP                           \
    gettimeofday (&te, &tz);                        \
    duration += (te.tv_sec - ts.tv_sec) * 1000;     \
    duration += (te.tv_usec - ts.tv_usec) / 1000;   \
    debug ("duration %d ms", duration);

#endif /* HAVE_UTILS_H */
