
/*
 * Copyright (C) 2008 Maximilian Schwerin
 *
 * This file is part of oxine a free media player.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: utils_covers.c 4017 2010-01-12 12:22:06Z mschwerin $
 *
 */

#include "config.h"

#include <errno.h>
#include <dirent.h>
#include <unistd.h>

#include <sys/types.h>

#include "desktop_entry.h"
#include "download.h"
#include "environment.h"
#include "filelist.h"
#include "heap.h"
#include "meta_info.h"
#include "logger.h"
#include "utils.h"


char *
get_cover_image_from_mrl (const char *mrl)
{
    if (!mrl) {
        return NULL;
    }

    /* We try to determine the thumbnail by protocol. */
    if (mrl_has_protocol (mrl, "file")) {
        /* do nothing */
    }
    else if (mrl_has_protocol (mrl, "vdr")) {
        return ho_strdup_printf ("%s/icons/logo_television.png",
                                 get_dir_oxine_skin ());
    }
    else if (mrl_has_protocol (mrl, "dvd")) {
        return ho_strdup_printf ("%s/icons/logo_device_dvd.png",
                                 get_dir_oxine_skin ());
    }
    else if (mrl_has_protocol (mrl, "vcd")) {
        return ho_strdup_printf ("%s/icons/logo_device_vcd.png",
                                 get_dir_oxine_skin ());
    }
    else if (mrl_has_protocol (mrl, "cdda")) {
        return ho_strdup_printf ("%s/icons/logo_device_cdda.png",
                                 get_dir_oxine_skin ());
    }

    /* We try to determine the thumbnail by location. */
    if (mrls_are_equal (mrl, FILELIST_HOME_MRL)) {
        return ho_strdup_printf ("%s/icons/logo_folder_home.png",
                                 get_dir_oxine_skin ());
    }
    else if (mrls_are_equal (mrl, FILELIST_DESKTOP_MRL)) {
        return ho_strdup_printf ("%s/icons/logo_folder_desktop.png",
                                 get_dir_oxine_skin ());
    }
    else if (mrls_are_equal (mrl, FILELIST_ROOT_MRL)) {
        return ho_strdup_printf ("%s/icons/logo_device_disc.png",
                                 get_dir_oxine_skin ());
    }
    else if (mrls_are_equal (mrl, FILELIST_MEDIA_MRL)) {
        return ho_strdup_printf ("%s/icons/logo_folder_music.png",
                                 get_dir_oxine_skin ());
    }
    else if (mrls_are_equal (mrl, FILELIST_MUSIC_MRL)) {
        return ho_strdup_printf ("%s/icons/logo_folder_music.png",
                                 get_dir_oxine_skin ());
    }
    else if (mrls_are_equal (mrl, FILELIST_XDG_MUSIC_MRL)) {
        return ho_strdup_printf ("%s/icons/logo_folder_music.png",
                                 get_dir_oxine_skin ());
    }
    else if (mrls_are_equal (mrl, FILELIST_VIDEO_MRL)) {
        return ho_strdup_printf ("%s/icons/logo_folder_video.png",
                                 get_dir_oxine_skin ());
    }
    else if (mrls_are_equal (mrl, FILELIST_XDG_VIDEOS_MRL)) {
        return ho_strdup_printf ("%s/icons/logo_folder_video.png",
                                 get_dir_oxine_skin ());
    }
    else if (mrls_are_equal (mrl, FILELIST_IMAGE_MRL)) {
        return ho_strdup_printf ("%s/icons/logo_folder_image.png",
                                 get_dir_oxine_skin ());
    }
    else if (mrls_are_equal (mrl, FILELIST_XDG_PICTURES_MRL)) {
        return ho_strdup_printf ("%s/icons/logo_folder_image.png",
                                 get_dir_oxine_skin ());
    }
    else if (mrls_are_equal (mrl, FILELIST_FAVORITES_MOST_MRL)) {
        return ho_strdup_printf ("%s/icons/logo_folder_favorites.png",
                                 get_dir_oxine_skin ());
    }
    else if (mrls_are_equal (mrl, FILELIST_FAVORITES_RECENT_MRL)) {
        return ho_strdup_printf ("%s/icons/logo_folder_favorites.png",
                                 get_dir_oxine_skin ());
    }
    else if (mrls_are_equal (mrl, FILELIST_PLAYLISTS_MRL)) {
        return ho_strdup_printf ("%s/icons/logo_folder_playlists.png",
                                 get_dir_oxine_skin ());
    }
    else if (mrls_are_equal (mrl, OXINE_AUDIODIR)) {
        return ho_strdup_printf ("%s/icons/logo_folder_mediamarks.png",
                                 get_dir_oxine_skin ());
    }
    else if (mrls_are_equal (mrl, OXINE_VIDEODIR)) {
        return ho_strdup_printf ("%s/icons/logo_folder_mediamarks.png",
                                 get_dir_oxine_skin ());
    }
    else if (mrls_are_equal (mrl, OXINE_IMAGEDIR)) {
        return ho_strdup_printf ("%s/icons/logo_folder_mediamarks.png",
                                 get_dir_oxine_skin ());
    }

    char *image_mrl = NULL;
    char *cover_mrl = NULL;
    char *directory = NULL;
    if (!mrl_is_directory (mrl)) {
        char *filename = mrl_get_filename (mrl);
        directory = get_dirname (filename);
        ho_free (filename);
    }
    else {
        char *filename = mrl_get_filename (mrl);
        directory = ho_strdup (filename);
        ho_free (filename);
    }

    /* First we try to retrieve the thumbnail from a .directory file. */
    if (!cover_mrl) {
        char *filename = ho_strdup_printf ("%s/.directory", directory);
        if (file_exists (filename)) {
            cover_mrl = get_localestring_from_desktop_entry (filename,
                                                             "Desktop Entry",
                                                             "Icon");
        }
        ho_free (filename);

        if (file_exists (cover_mrl) && mrl_is_image (cover_mrl)) {
            /* Make sure we're using a valid MRL from now on. */
            char *mrl = mrl_from_filename (cover_mrl);
            ho_free (cover_mrl);
            cover_mrl = mrl;
        }
        else {
            ho_free (cover_mrl);
        }
    }

    /* Next we try to find an image (JPG, PNG, GIF) with title or folder as
     * basename. */
    if (!cover_mrl) {
        DIR *dirp;
        struct dirent *entp;

        if ((dirp = opendir (directory))) {
            while (!cover_mrl && (entp = readdir (dirp))) {
                if (entp->d_name[0] == '.') {
                    continue;
                }
                if (!mrl_is_image (entp->d_name)) {
                    continue;
                }

                char *filename = ho_strdup_printf ("%s/%s", directory,
                                                   entp->d_name);

                if (!image_mrl) {
                    image_mrl = mrl_from_filename (filename);
                }
                if (strc_starts_with (entp->d_name, "title")
                    || strc_starts_with (entp->d_name, "cover")
                    || strc_starts_with (entp->d_name, "front")
                    || strc_starts_with (entp->d_name, "folder")) {
                    cover_mrl = mrl_from_filename (filename);
                }

                ho_free (filename);
            }
            closedir (dirp);
        }
    }

    /* Next we try to get an image from Last.fm */
#ifdef HAVE_LASTFM
    if (!cover_mrl && mrl_is_audio (mrl)) {
        char *artist = meta_info_get (META_INFO_ARTIST, mrl);
        char *album = meta_info_get (META_INFO_ALBUM, mrl);
        {
            cover_mrl = get_cover_image_from_lastfm (artist, album);
        }
        if (!cover_mrl) {
            cover_mrl = get_cover_image_from_lastfm (artist, NULL);
        }
        ho_free (artist);
        ho_free (album);
    }
#endif

    /* Next we search for any image and use the first we find. */
    if (!cover_mrl && image_mrl) {
        cover_mrl = ho_strdup (image_mrl);
    }

    if (!cover_mrl && mrl_is_audio (mrl) && mrl_is_networked (mrl)) {
        cover_mrl = ho_strdup_printf ("%s/icons/logo_internet_music.png",
                                      get_dir_oxine_skin ());
    }
    if (!cover_mrl && mrl_is_audio (mrl)) {
        cover_mrl = ho_strdup_printf ("%s/icons/logo_file_music.png",
                                      get_dir_oxine_skin ());
    }

    if (!cover_mrl && mrl_is_video (mrl) && mrl_is_networked (mrl)) {
        cover_mrl = ho_strdup_printf ("%s/icons/logo_internet_video.png",
                                      get_dir_oxine_skin ());
    }
    if (!cover_mrl && mrl_is_video (mrl)) {
        cover_mrl = ho_strdup_printf ("%s/icons/logo_file_video.png",
                                      get_dir_oxine_skin ());
    }

    if (!cover_mrl && mrl_is_playlist (mrl) && mrl_is_networked (mrl)) {
        cover_mrl = ho_strdup_printf ("%s/icons/logo_internet_playlist.png",
                                      get_dir_oxine_skin ());
    }
    if (!cover_mrl && mrl_is_playlist (mrl)) {
        cover_mrl = ho_strdup_printf ("%s/icons/logo_file_playlist.png",
                                      get_dir_oxine_skin ());
    }

    ho_free (directory);
    ho_free (image_mrl);

    return cover_mrl;
}


bool
is_cover_image_valid (const char *mrl)
{
    if (!mrl) {
        return false;
    }
    else if (!mrl_is_local (mrl)) {
        return true;
    }
    else if (file_exists (mrl)) {
        return true;
    }
    else {
        return false;
    }
}
