
/*
 * Copyright (C) 2004-2005 Maximilian Schwerin
 *
 * This file is part of oxine a free media player.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: utils_gui.c 3960 2009-07-12 11:28:09Z mschwerin $
 *
 */
#include "config.h"

#include <assert.h>
#include <stdarg.h>
#include <stdio.h>

#include "environment.h"
#include "download.h"
#include "i18n.h"
#include "heap.h"
#include "logger.h"
#include "otk.h"
#include "odk.h"
#include "oxine.h"
#include "scheduler.h"
#include "utils.h"

#include "menu_base.h"
#include "menu_playback.h"
#include "menu_playlist.h"

extern oxine_t *oxine;

static int hide_gui_job_id = -1;


void
schedule_hide_gui_job (int timeout)
{
    cancel_hide_gui_job ();
    oxine_event_id_t id = OXINE_EVENT_GUI_HIDE;
    hide_gui_job_id = schedule_simple_event_job (timeout, id);
}


void
cancel_hide_gui_job (void)
{
    cancel_event_job (&hide_gui_job_id);
    hide_gui_job_id = -1;
}


static void
event_job (void *p)
{
    oxine_event_t *event = (oxine_event_t *) p;
    odk_oxine_event_send_synchronized (oxine->odk, event);
    ho_free (event);
}


int
schedule_event_job (int timeout, oxine_event_t * event)
{
    int job = schedule_job (timeout, event_job, event);
    if (job <= -1) {
        ho_free (event);
    }
    return job;
}


int
schedule_simple_event_job (int timeout, oxine_event_id_t id)
{
    oxine_event_t * event = oxine_event_new ();
    event->type = id;
    return schedule_event_job (timeout, event);
}


void
cancel_event_job (int * job)
{
    oxine_event_t * event = (oxine_event_t *) cancel_job (job);
    if (event) {
        ho_free (event);
    }
}


void
show_user_interface (void *data_p)
{
    cancel_hide_gui_job ();

    otk_draw (oxine->otk);

    oxine->user_interface_is_visible = true;
    odk_set_forward_events_to_xine (oxine->odk, false);
}


void
hide_user_interface (void *data_p)
{
    cancel_hide_gui_job ();

    otk_clear (oxine->otk);

    oxine->user_interface_is_visible = false;
    oxine->playback_controls_are_visible = false;
    oxine->stream_parameter_is_visible = false;

    if (odk_current_is_dvd (oxine->odk)) {
        odk_set_forward_events_to_xine (oxine->odk, true);
    }
#ifdef HAVE_VDR
    else if (odk_current_is_vdr (oxine->odk)) {
        odk_set_forward_events_to_xine (oxine->odk, true);
    }
#endif
    else {
        odk_set_forward_events_to_xine (oxine->odk, false);
    }

    set_current_menu (hide_user_interface, NULL);
}


static void
set_time_string (void *data_p, otk_widget_t * widget)
{
    char time_str[128];

    static char format[128] = { '\0' };
    if (strlen (format) == 0) {
        const char *cformat = config_get_string ("gui.clock_format");
        if (!cformat || (strlen (cformat) == 0)) {
            strncpy (format, _("%a, %d. %b, %H:%M"), 128);
        }
        else {
            strncpy (format, cformat, 128);
        }
    }

    time_t current = time (NULL);
    struct tm *brokentime = localtime (&current);
    strftime (time_str, 128, format, brokentime);
    otk_label_set_text (widget, time_str);
}


static void
show_clock (void)
{
    int a = OTK_ALIGN_RIGHT | OTK_ALIGN_VCENTER;
    int x = 20;
    int w = odk_osd_get_width (oxine->odk) - 40;

    otk_widget_t *time = otk_label_new (oxine->otk, x, 20, w, 40, a, "");
    set_time_string (oxine, time);
    otk_label_set_update_cb (time, set_time_string, oxine);
    otk_widget_set_updated (time, true);
}


static void
window_iconify_cb (void *p)
{
    oxine_event_t event;
    oxine_event_init (&event);
    event.type = OXINE_EVENT_KEY;
    event.source.key = OXINE_KEY_WINDOW_ICONIFY;
    odk_oxine_event_send (oxine->odk, &event);
}


static void
show_iconify (void)
{
    int x = odk_osd_get_width (oxine->odk) - 81;

    otk_invisible_button_new (oxine->otk, x, 0, 80, 20,
                              window_iconify_cb, NULL);
}


otk_widget_t *
create_new_window (bool persistent, bool with_clock)
{
    otk_widget_t *w = otk_window_new (oxine->otk, 0, 0,
                                      odk_osd_get_width (oxine->odk),
                                      odk_osd_get_height (oxine->odk));

    if (persistent) {
        otk_window_keep (w, true);
    }
    if (with_clock) {
        show_clock ();
    }
    {
        show_iconify ();
    }

    return w;
}


void
show_menu_background (const char *background)
{
    const char *current = odk_current_get_mrl (oxine->odk);
    char *mrl = NULL;

    if (background && !file_exists (background)) {
        char *a = ho_strdup_printf ("%s/backgrounds/%s",
                                    get_dir_oxine_skin (), background);
        mrl = mrl_from_filename (a);
        ho_free (a);
    }
    else {
        mrl = mrl_from_filename (background);
    }

    if (!file_exists (mrl)) {
        ho_free (mrl);
    }

    if (!mrl) {
        char *a = ho_strdup_printf ("%s/backgrounds/menu_main.png",
                                    get_dir_oxine_skin (), background);
        mrl = mrl_from_filename (a);
        ho_free (a);
    }

    if (odk_current_is_playback_mode (oxine->odk)) {
        odk_play_background_stream (oxine->odk, mrl);
        goto out_free;
    }

    if (str_equals (mrl, current)) {
        goto out_free;
    }

    if (mrl && odk_play_stream (oxine->odk, "Background",
                                mrl, NULL, 0, ODK_MODE_LOGO)) {
        goto out_free;
    }

    error (_("Could not open or play menu background image!"));
    fatal (_("Please make sure that your version of xine-lib "
             "supports playback of PNG files."));
    abort ();

  out_free:
    ho_free (mrl);
}


#define STREAM_INFO_POS_X (20)
#define STREAM_INFO_POS_Y (20)
#define STREAM_INFO_WIDTH (odk_osd_get_width (oxine->odk) - 40)
#define STREAM_INFO_ALIGN (OTK_ALIGN_LEFT | OTK_ALIGN_TOP)


static int
position_get (void *data_p)
{
    return odk_get_stream_param (oxine->odk, ODK_PARAM_POSITION);
}


static int
volume_get (void *data_p)
{
    return odk_get_stream_param (oxine->odk, ODK_PARAM_AO_VOLUME);
}


bool
show_stream_parameter (odk_stream_param_t param)
{
    /* If we're in the playback menu and the parameter that was changed is
     * PARAM_AUDIO_MUTE, we have to repaint the menu to update the mute/
     * no-mute icon. */
    if (is_current_menu (show_menu_playback)
        && (param == ODK_PARAM_AO_VOLUME_MUTE)) {
        show_menu_current ();
        return false;
    }

    /* If we're in the playlist menu and the parameter that was changed is
     * PARAM_SPEED, we have to repaint the menu to update the played/ paused
     * icon. */
    if (is_current_menu (show_menu_playlist)
        && (param == ODK_PARAM_SPEED)) {
        show_menu_current ();
        return false;
    }

    /* If we're in any menu (meaning we're not showing no OSD and not showing
     * playback info), we return. */
    if (!is_current_menu (hide_user_interface)
        && !is_current_menu (show_playback_audio_info)
        && !is_current_menu (show_playback_video_info)) {
        return false;
    }

    /* If we're not supposed to show the volume slider we return. */
    if ((param == ODK_PARAM_AO_VOLUME)
        && odk_current_has_video (oxine->odk)
        && !config_get_bool ("misc.show_volume_when_playing_videos")) {
        return true;
    }

    /* If we're not in one of the info menus we have to create a new window to
     * paint on. */
    if (is_current_menu (hide_user_interface)) {
        create_new_window (false, false);
    }

    if ((param == ODK_PARAM_POSITION) || (param == ODK_PARAM_SPEED)) {
        int y = odk_osd_get_height (oxine->odk) - 65;
        int w = odk_osd_get_width (oxine->odk) - 70;
        otk_widget_t *s = otk_slider_new (oxine->otk, 50, y, w, 40, 0, 20,
                                          false, 0, 0, OTK_SLIDER_HORIZONTAL,
                                          NULL, 0, false, false, 0, 5, 0, 100,
                                          position_get, oxine, NULL, NULL);
        otk_widget_set_updated (s, true);
        otk_widget_set_enabled (s, false);

        char *media_pplay_label = NULL;
        if (odk_get_stream_param (oxine->odk, ODK_PARAM_SPEED) ==
            ODK_SPEED_PAUSE) {
            media_pplay_label = "<";
        }
        else {
            media_pplay_label = ">";
        }
        otk_widget_t *b = otk_text_button_new (oxine->otk, 20, y, 40, 40,
                                               media_pplay_label, NULL, NULL);
        otk_widget_set_selectable (b, OTK_SELECTABLE_NONE);
        otk_widget_set_fonttype (b, "cetus");
        otk_widget_set_fontsize (b, 20);
        otk_widget_set_alignment (b, OTK_ALIGN_CENTER);
    }

    else if ((param == ODK_PARAM_AO_VOLUME)
             || (param == ODK_PARAM_AO_VOLUME_MUTE)) {
        int y = odk_osd_get_height (oxine->odk) - 65;
        int w = odk_osd_get_width (oxine->odk) - 70;
        otk_widget_t *s = otk_slider_new (oxine->otk, 50, y, w, 40, 0, 20,
                                          false, 0, 0, OTK_SLIDER_HORIZONTAL,
                                          NULL, 0, false, false, 0, 5, 0, 100,
                                          volume_get, oxine, NULL, NULL);
        otk_widget_set_updated (s, true);
        otk_widget_set_enabled (s, false);

        odk_osd_vector_type_t type = OSD_VECTOR_VOLUME;
        if (odk_get_stream_param (oxine->odk, ODK_PARAM_AO_VOLUME_MUTE)) {
            type = OSD_VECTOR_VOLUME_MUTE;
        }
        otk_widget_t *b = otk_vector_button_new (oxine->otk, 20, y, 40, 40,
                                                 type, 30, 30, NULL, NULL);
        otk_widget_set_selectable (b, OTK_SELECTABLE_NONE);
    }

    else {
        char msg[256];
        char *param_str = odk_get_stream_param_as_string (oxine->odk, param);

        if (param == ODK_PARAM_SPEED) {
            snprintf (msg, 256, "%s", param_str);
        }

        else {
            char *param_name = odk_get_stream_param_name (oxine->odk, param);
            snprintf (msg, 256, "%s: %s", param_name, param_str);
            ho_free (param_name);
        }

        ho_free (param_str);

        otk_label_new (oxine->otk, STREAM_INFO_POS_X, STREAM_INFO_POS_Y,
                       STREAM_INFO_WIDTH, 30, STREAM_INFO_ALIGN, msg);
    }

    otk_draw (oxine->otk);
    oxine->stream_parameter_is_visible = true;

    return true;
}


void
show_download_progress_info (void *p)
{
    create_new_window (false, true);

    int w = 600;
    int h = 140;
    int x = (odk_osd_get_width (oxine->odk) - w) / 2;
    int y = (odk_osd_get_height (oxine->odk) - h) / 2;

    otk_border_new (oxine->otk, x, y, w, h);

    otk_label_new (oxine->otk, x, y, 560, 50,
                   OTK_ALIGN_CENTER | OTK_ALIGN_BOTTOM, _("Please wait..."));

    otk_widget_t *slider =
        otk_slider_new (oxine->otk, x + 20, y + 70, 560, 40, 0, 20, false, 0,
                        0,
                        OTK_SLIDER_HORIZONTAL, "%3.f %%", 1.0, false, false,
                        0, 1, 0, 100, download_get_percent, NULL, NULL, NULL);
    otk_widget_set_updated (slider, true);

    set_current_menu (show_download_progress_info, NULL);

    show_user_interface (NULL);
    show_menu_background (NULL);
}
