
/*
 * Copyright (C) 2008 Maximilian Schwerin
 *
 * This file is part of oxine a free media player.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: utils_lastfm.c 3894 2009-05-05 12:22:01Z mschwerin $
 *
 */

#include "config.h"

#include <errno.h>
#include <dirent.h>
#include <unistd.h>

#include <sys/types.h>

#include "download.h"
#include "environment.h"
#include "heap.h"
#include "logger.h"
#include "utils.h"

#ifdef HAVE_LASTFM

#define LASTFM_ALBUM_QUERY      "http://ws.audioscrobbler.com/2.0/" \
                                "album/info.xml" \
                                "?artist=%s" \
                                "&album=%s"

#define LASTFM_ARTIST_QUERY     "http://ws.audioscrobbler.com/2.0/" \
                                "artist/info.xml" \
                                "?artist=%s"

static xml_node_t *
lastfm_get_info (const char *query_url, const char *rootname)
{
    xml_node_t *n_root = NULL;

    /* Run the query and download the answer to a buffer. */
    download_t *d = download_new (NULL);
    if (!network_download (query_url, d)) {
        error (_("Failed to download cover information!"));
        goto download_free;
    }

    /* Parse the XML answer to our query. */
    xml_parser_init (d->buffer, d->size, XML_PARSER_CASE_INSENSITIVE);
    if (xml_parser_build_tree (&n_root) != XML_PARSER_OK) {
        error (_("Parsing '%s' failed!"), query_url);
        goto download_free;
    }
    if (strcasecmp (n_root->name, rootname) != 0) {
        error (_("Root node of '%s' must be '%s'!"), query_url, rootname);
        xml_parser_free_tree (n_root);
        n_root = NULL;
    }

  download_free:
    download_free (d, true);

    return n_root;
}


static xml_node_t *
lastfm_get_artist_info (const char *artist)
{
    assert (artist && strlen (artist));

    debug ("Searching for '%s' on Last.fm...", artist);

    char *enc_artist = url_encode (artist);
    char *url = ho_strdup_printf (LASTFM_ARTIST_QUERY, enc_artist);
    xml_node_t *n = lastfm_get_info (url, "artist");
    ho_free (enc_artist);
    ho_free (url);

    return n;
}


static xml_node_t *
lastfm_get_album_info (const char *artist, const char *album)
{
    assert (artist && strlen (artist));
    assert (album && strlen (album));

    debug ("Searching for '%s' by '%s' on Last.fm...", album, artist);

    char *enc_artist = url_encode (artist);
    char *enc_album = url_encode (album);
    char *url = ho_strdup_printf (LASTFM_ALBUM_QUERY, enc_artist, enc_album);
    xml_node_t *n = lastfm_get_info (url, "album");
    ho_free (enc_artist);
    ho_free (enc_album);
    ho_free (url);

    return n;
}


/**
 * If we failed to retrieve information, we try again every 5 days.
 */
static bool
lastfm_retry (const char *filename)
{
    FILE *f = fopen (filename, "r");
    if (!f) {
        error (_("Could not open '%s': %s!"), filename, strerror (errno));
        return true;
    }
    time_t lasttry = 0;
    if (fscanf (f, "%ld", (long*) &lasttry) != 1) {
        error (_("Parsing '%s' failed!"), filename);
        lasttry = 0;
    }
    fclose (f);
    return (lasttry < (time (NULL) - (3600 * 24 * 5)));
}


char *
get_cover_image_from_lastfm (const char *artist, const char *album)
{
    artist = (artist && strlen (artist)) ? artist : NULL;
    album = (album && strlen (album)) ? album : NULL;
    if (!artist) {
        return NULL;
    }
    if (strc_equals (artist, "Various Artists")) {
        return NULL;
    }
    if (strc_equals (artist, _("Various Artists"))) {
        return NULL;
    }

    char *cache_dir = ho_strdup_printf ("%s/lastfm/%s/%s",
                                        get_dir_oxine_cache (),
                                        artist, (album ? album : ""));
    char *nohit_mrl = NULL;
    char *cover_mrl = NULL;
    if (!mkdir_recursive (cache_dir, 0700)) {
        goto finished_free;
    }

    /* Check if we already tried and failed. */
    nohit_mrl = ho_strdup_printf ("%s/nohit", cache_dir);
    if (file_exists (nohit_mrl) && !lastfm_retry (nohit_mrl)) {
        goto finished_free;
    }

    /* Check if the cover is already in our cache. */
    cover_mrl = ho_strdup_printf ("%s/cover.jpg", cache_dir);
    if (file_exists (cover_mrl)) {
        goto finished_free;
    }

    /* We need to URL encode our artist and album strings for the query. */
    xml_node_t *n_root = NULL;
    if (album) {
        n_root = lastfm_get_album_info (artist, album);
    }
    else {
        n_root = lastfm_get_artist_info (artist);
    }
    if (!n_root) {
        ho_free (cover_mrl);
        goto finished_free;
    }

    const char *image_small = NULL;
    const char *image_medium = NULL;
    const char *image_large = NULL;
    const char *image_extralarge = NULL;

    xml_node_t *n_image = xml_parser_get_node_by_name (n_root->child, "image");
    while (n_image) {
        const char *s = xml_parser_get_property (n_image, "size");
        if (strc_equals (s, "small")) {
            image_small = n_image->data;
        }
        else if (strc_equals (s, "medium")) {
            image_medium = n_image->data;
        }
        else if (strc_equals (s, "large")) {
            image_large = n_image->data;
        }
        else if (strc_equals (s, "extralarge")) {
            image_extralarge = n_image->data;
        }
        n_image = xml_parser_get_node_by_name (n_image->next, "image");
    }

    if (image_extralarge && strlen (image_extralarge)) {
        char *tmp = NULL;
        if ((tmp = download_to_cache (image_extralarge, cover_mrl, false))) {
            ho_free (tmp);
            goto parser_free;
        }
    }
    if (image_large && strlen (image_large)) {
        char *tmp = NULL;
        if ((tmp = download_to_cache (image_large, cover_mrl, false))) {
            ho_free (tmp);
            goto parser_free;
        }
    }
    if (image_medium && strlen (image_medium)) {
        char *tmp = NULL;
        if ((tmp = download_to_cache (image_medium, cover_mrl, false))) {
            ho_free (tmp);
            goto parser_free;
        }
    }
    if (image_small && strlen (image_small)) {
        char *tmp = NULL;
        if ((tmp = download_to_cache (image_small, cover_mrl, false))) {
            ho_free (tmp);
            goto parser_free;
        }
    }
    {
        ho_free (cover_mrl);
    }

  parser_free:
    xml_parser_free_tree (n_root);

  finished_free:
    if (!cover_mrl) {
        FILE *f = fopen (nohit_mrl, "w");
        if (f) {
            fprintf (f, "%ld", (long) time (NULL));
            fclose (f);
        }
    }

    ho_free (nohit_mrl);
    ho_free (cache_dir);

    return cover_mrl;
}


#ifdef LASTFM_DOWNLOADER

oxine_t *oxine;

int
main (int argc, char **argv)
{
    oxine = ho_new (oxine);
    if (!file_exists (argv[3])) {
        char *filename = get_cover_image_from_lastfm (argv[1], argv[2]);
        if (filename) {
            copyfile (filename, argv[3]);
        }
        ho_free (filename);
    }
    return 0;
}

void
_odk_oxine_event_send (odk_t * a, oxine_event_t * b, const char *f, int l)
{
}

#endif /* LASTFM_DOWNLOADER */

#endif /* HAVE_LASTFM */
