
/*
 * Copyright (C) 2004-2005 Maximilian Schwerin
 *
 * This file is part of oxine a free media player.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: utils_mrl.c 4005 2009-09-26 12:57:04Z mschwerin $
 *
 */

#include "config.h"

#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <libgen.h>
#include <mntent.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <sys/ioctl.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/types.h>

#include "download.h"
#include "environment.h"
#include "heap.h"
#include "i18n.h"
#include "list.h"
#include "logger.h"
#include "utils.h"


static char *inputs[] = {
    "file",
    "dvb",
    "vdr",
    "xvdr",
    "v4l",
    "dvd",
    "vcd",
    "mms",
    "ftp",
    "http",
    "rtsp",
    "cdda",
    "computer",
    NULL
};


/**
 * See manpage of realpath.
 *
 * @return                      Returns the resolved path or NULL if an
 *                              error occurred. The returned string must be
 *                              freed.
 */
static char *
file_get_realpath (const char *filename)
{
    assert (filename);
    char temp[MAXPATHLEN];
    char *absolute = NULL;
    if (realpath (filename, temp)) {
        absolute = ho_strdup (temp);
    }
    else if (errno == ENOENT) {
        absolute = ho_strdup (filename);
    }
    else {
        error (_("Unable to resolve '%s': %s."), filename, strerror (errno));
        absolute = ho_strdup (filename);
    }
    return absolute;
}


static char *
file_get_cleanpath (const char *filename)
{
    assert (filename);
    int i = 0;
    int o = 0;
    char *clean = ho_malloc (strlen (filename) + 1);
    while (i < strlen (filename)) {
        if ((filename[i] == '/') && (filename[i + 1] == '/')) {
            i += 1;
        }
        clean[o++] = filename[i++];
    }
    return clean;
}


int
mrl_get_number (const char *mrl)
{
    int number = -1;
    char *numstr = rindex (mrl, '/');
    if (numstr) {
        numstr += 1;
        number = atoi (numstr);
        number = (number != 0) ? number : -1;
    }
    return number;
}


bool
mrl_has_protocol (const char *mrl, const char *protocol)
{
    int len = strlen (protocol);
    return strc_starts_with (mrl, protocol) && (mrl[len] == ':');
}


char *
mrl_get_protocol (const char *mrl)
{
    int i = 0;
    const char *protocol = "file";

    for (i = 0; inputs[i]; i++) {
        if (mrl_has_protocol (mrl, inputs[i])) {
            protocol = inputs[i];
            break;
        }
    }

    return ho_strdup (protocol);
}


char *
mrl_attach_query (const char *mrl, const char *key, const char *val)
{
    char *new_mrl = NULL;
    const char *query = index (mrl, '?');

    /* After encoding a string containing special characters, the encoded
     * string contains the '%' character. This is a problem for printf in
     * some cases so we replace it with something printf does not care
     * about and that does not appear in an url-encoded string. */
    char *enc_key = url_encode (key);
    replace_chars (enc_key, '%', '|');
    char *enc_val = url_encode (val);
    replace_chars (enc_val, '%', '|');
    if (query) {
        new_mrl = ho_strdup_printf ("%s&%s=%s", mrl, enc_key, enc_val);
    }
    else {
        new_mrl = ho_strdup_printf ("%s?%s=%s", mrl, enc_key, enc_val);
    }
    replace_chars (new_mrl, '|', '%');
    ho_free (enc_key);
    ho_free (enc_val);

    return new_mrl;
}


char *
mrl_without_query (const char *mrl)
{
    char *m = ho_strdup (mrl);
    char *q = index (m, '?');
    if (q) {
        q[0] = '?';
    }
    return m;
}


char *
mrl_get_query (const char *mrl)
{
    const char *query = index (mrl, '?');
    return ho_strdup (query);
}


char *
query_get_parameter (const char *query, const char *key)
{
    char *s;
    char *v;

    if (!query) {
        return NULL;
    }

    s = ho_strdup_printf ("%s=", key);
    v = strstr (query, s);
    ho_free (s);
    if (!v) {
        return NULL;
    }

    v = url_decode (&v[strlen (key) + 1]);
    s = index (v, '&');
    if (s) {
        s[0] = '\0';
    }

    return v;
}


char *
mrl_get_query_parameter (const char *mrl, const char *key)
{
    char *q;
    char *s;
    char *v;

    q = index (mrl, '?');
    if (!q) {
        return NULL;
    }

    s = ho_strdup_printf ("%s=", key);
    v = strstr (q, s);
    ho_free (s);
    if (!v) {
        return NULL;
    }

    v = url_decode (&v[strlen (key) + 1]);
    s = index (v, '&');
    if (s) {
        s[0] = '\0';
    }

    return v;
}


char *
mrl_get_filename (const char *mrl)
{
    if (!mrl) {
        return NULL;
    }

    const char *temp = mrl;

    int i = 0;
    for (i = 0; inputs[i]; i++) {
        if (mrl_has_protocol (temp, inputs[i])) {
            temp = (char *) (temp + strlen (inputs[i]) + 2);
            break;
        }
    }

    while ((temp[0] == '/') && (temp[1] == '/')) {
        temp += 1;
    }

    char *query = index (temp, '?');
    if (query) {
        query[0] = '\0';
    }

    char *fname = url_decode (temp);
    if (query) {
        query[0] = '?';
    }

    return fname;
}


char *
mrl_build (const char *input, const char *absolute, int tracknumber)
{
    assert (input);
    assert (absolute);

    /* We save this here as we change the tracknumber in some cases. */
    bool with_tracknumber = (tracknumber > 0);

    /* For some reason the VCD plugin wants a MRL to look like this vcd:// */
    const char *separator = "";
    if (strc_equals (input, "vcd")) {
        separator = "/";
    }
    /* For some reason the xine-lib developers have the channel numbers
     * start with 0 for DVB. */
    if (strc_equals (input, "dvb")) {
        tracknumber -= 1;
    }

    if (with_tracknumber) {
        return ho_strdup_printf ("%s:/%s%s/%d", input, separator,
                                 absolute, tracknumber);
    }
    else {
        return ho_strdup_printf ("%s:/%s%s", input, separator, absolute);
    }
}


char *
mrl_from_filename (const char *filename)
{
    char *mrl = NULL;
    char *cleaned = NULL;
    char *encoded = NULL;

    if (!filename) {
        return NULL;
    }
    if (mrl_is_valid (filename)) {
        return ho_strdup (filename);
    }
    cleaned = file_get_cleanpath (filename);
    if (!cleaned) {
        goto out_free;
    }
    encoded = url_encode (cleaned);
    if (!encoded) {
        goto out_free;
    }
    mrl = mrl_build ("file", encoded, 0);
    if (!mrl) {
        goto out_free;
    }

  out_free:
    ho_free (encoded);

    return mrl;
}


bool
mrl_is_valid (const char *mrl)
{
    int i = 0;
    bool valid = false;

    if (!mrl) {
        return false;
    }

    for (i = 0; inputs[i]; i++) {
        if (mrl_has_protocol (mrl, inputs[i])) {
            valid = true;
            break;
        }
    }

    return valid;
}


bool
mrl_is_device (const char *mrl)
{
    if (!mrl_is_local (mrl)) {
        return false;
    }

    char *filename = mrl_get_filename (mrl);
    bool res = strc_starts_with (filename, "/dev/");
    ho_free (filename);

    return res;
}


bool
mrl_is_local (const char *mrl)
{
    if (!mrl) {
        return false;
    }
    else if (mrl_is_valid (mrl)) {
        return mrl_has_protocol (mrl, "file");
    }
    else {
        return true;
    }
}


static bool
protocols_are_equal (const char *m1, const char *m2)
{
    int i = 0;

    for (i = 0; inputs[i]; i++) {
        if (mrl_has_protocol (m1, inputs[i])
            && mrl_has_protocol (m2, inputs[i])) {
            return true;
        }
    }

    return false;
}


bool
mrls_are_equal (const char *m1, const char *m2)
{
    if (!m1 || !m2) {
        return (m1 == m2) ? true : false;
    }
    else if (protocols_are_equal (m1, m2)) {
        char *mm1 = mrl_get_filename (m1);
        char *mm2 = mrl_get_filename (m2);

        bool equal = (strcmp (mm1, mm2) == 0) ? true : false;

        ho_free (mm1);
        ho_free (mm2);

        return equal;
    }
    else {
        return false;
    }
}


bool
mrl_has_extension (const char *filename, const char *extension)
{
    const char *ext = get_extension (filename);

    return ext && (strcasecmp (extension, ext) == 0);
}


const char *
get_extension (const char *filename)
{
    const char *ext;

    if (!filename) {
        return NULL;
    }

    ext = strrchr (filename, '.');

    if (!ext || (ext == filename)) {
        return NULL;
    }

    return (ext + 1);
}


char *
file_resolve_softlink (const char *mrl)
{
    char *absolute = NULL;
    char *filename = mrl_get_filename (mrl);
    if (filename) {
        absolute = file_get_realpath (filename);
        ho_free (filename);
    }
    return absolute;
}


char *
mrl_resolve_softlink (const char *mrl)
{
    char *absolute_mrl = NULL;
    char *absolute = file_resolve_softlink (mrl);
    if (absolute) {
        absolute_mrl = mrl_from_filename (absolute);
        ho_free (absolute);
    }
    return absolute_mrl;
}


char *
read_entire_file (const char *mrl, int *size)
{
    char *buffer = NULL;

    if (!mrl) {
        return NULL;
    }

    if (is_downloadable (mrl)) {
#ifdef HAVE_LIBCURL
        download_t *download = download_new (NULL);

        if (download && network_download (mrl, download)) {
            buffer = download->buffer;
            *size = download->size;
        }

        download_free (download, false);
#endif /* HAVE_LIBCURL */
    }
    else {
        struct stat filestat;
        char *filename = mrl_get_filename (mrl);

        if (access (filename, R_OK) != 0) {
            error (_("Could not open '%s': %s!"), filename, strerror (errno));
            ho_free (filename);
            return NULL;
        }

        if (stat (filename, &filestat) == -1) {
            error (_("Unable to get status for file '%s': %s."), filename,
                   strerror (errno));
            ho_free (filename);
            return NULL;
        }

        if ((*size = filestat.st_size) == 0) {
            error (_("File '%s' is empty!"), filename);
            ho_free (filename);
            return NULL;
        }

        int fd;
        if ((fd = open (filename, O_RDONLY)) == -1) {
            error (_("Could not open '%s': %s!"), filename, strerror (errno));
            ho_free (filename);
            return NULL;
        }

        if ((buffer = (char *) ho_malloc (*size + 1)) == NULL) {
            close (fd);
            ho_free (filename);
            return NULL;
        }

        int bytes_read;
        if ((bytes_read = read (fd, buffer, *size)) != *size) {
            error (_("Could not read '%s': %s!"), filename, strerror (errno));
            *size = bytes_read;
        }
        buffer[*size] = '\0';

        close (fd);
        ho_free (filename);
    }

    return buffer;
}


char *
get_basename (const char *mrl)
{
    if (!mrl) {
        return NULL;
    }

    char *mrl_copy = ho_strdup (mrl);
    char *mrl_base = ho_strdup (basename (mrl_copy));

    ho_free (mrl_copy);

    return mrl_base;
}


char *
get_dirname (const char *mrl)
{
    if (!mrl) {
        return NULL;
    }

    char *mrl_copy = ho_strdup (mrl);
    char *mrl_dirn = ho_strdup (dirname (mrl_copy));

    ho_free (mrl_copy);

    return mrl_dirn;
}


bool
file_exists (const char *mrl)
{
    bool exists = true;
    struct stat filestat;

    if (!mrl_is_local (mrl)) {
        return false;
    }

    char *filename = mrl_get_filename (mrl);
    if (stat (filename, &filestat) == -1) {
        switch (errno) {
        case EACCES:
        case ENOENT:
        case ENOTDIR:
            exists = false;
            break;
        default:
            error (_("Unable to get status for file '%s': %s."),
                   filename, strerror (errno));
            exists = false;
            break;
        }
    }
    ho_free (filename);

    return exists;
}


bool
mkdir_recursive (const char *mrl, mode_t mode)
{
    if (!mrl) {
        return false;
    }
    if (file_exists (mrl)) {
        return true;
    }

    char *filename = mrl_get_filename (mrl);
    if (mkdir (filename, mode) == -1) {
        char *dirn = get_dirname (filename);
        if (!mkdir_recursive (dirn, mode)) {
            ho_free (dirn);
            ho_free (filename);
            return false;
        }
        ho_free (dirn);

        if (mkdir (filename, mode) == -1) {
            error (_("Could not create '%s': %s!"), filename,
                   strerror (errno));
            ho_free (filename);
            return false;
        }
    }

    ho_free (filename);
    return true;
}


#define UNACCEPTABLE_SHELL(a) ((a == ' ') \
                              || (a == '"') \
                              || (a == '&') \
                              || (a == '(') \
                              || (a == ')') \
                              || (a == '\''))

char *
filename_escape_for_shell (const char *filename)
{
    if (!filename) {
        return NULL;
    }

    int unacceptable = 0;
    const char *p;
    for (p = filename; *p != '\0'; p++) {
        char c = (char) *p;
        if (UNACCEPTABLE_SHELL (c)) {
            unacceptable++;
        }
    }

    char *result = ho_malloc (strlen (filename) + unacceptable + 1);

    char *q = result;
    for (p = filename; *p != '\0'; p++) {
        char c = (char) *p;
        if (UNACCEPTABLE_SHELL (c)) {
            *q++ = '\\';
        }
        *q++ = *p;
    }
    *q = '\0';

#if 0
    debug ("%s -> %s", filename, result);
#endif

    return result;
}


bool
mrl_is_playlist_m3u (const char *mrl)
{
    /* This is the simple way to detect mime types. We could also use
     * gnomevfs-info -s <filename> */
    return mrl_has_extension (mrl, "m3u");
}


bool
mrl_is_playlist_pls (const char *mrl)
{
    /* Mimetype: audio/x-scpls */
    return mrl_has_extension (mrl, "pls");
}


bool
mrl_is_playlist_oxp (const char *mrl)
{
    return mrl_has_extension (mrl, "oxp");
}


bool
mrl_is_mediamarks_oxm (const char *mrl)
{
    return mrl_has_extension (mrl, "oxm");
}


bool
mrl_is_directory (const char *mrl)
{
    bool is_dir = false;
    struct stat filestat;

    if (!mrl_is_local (mrl)) {
        return false;
    }

    char *filename = mrl_get_filename (mrl);
    if (filename && (stat (filename, &filestat) == 0)) {
        is_dir = S_ISDIR (filestat.st_mode) ? true : false;
    }
    ho_free (filename);

    return is_dir;
}


/// Tests for the existance of a subentry of a directory.
static bool
subentry_exists (const char *directory, const char *subentry)
{
    char *mrl = ho_strdup_printf ("%s/%s", directory, subentry);
    bool exists = file_exists (mrl);
    ho_free (mrl);

    return exists;
}


bool
mrl_is_directory_dvd (const char *mrl)
{
    return (mrl_is_directory (mrl)
            && subentry_exists (mrl, "VIDEO_TS/VIDEO_TS.IFO"));
}


bool
mrl_is_directory_vcd (const char *mrl)
{
    return (mrl_is_directory (mrl)
            && (subentry_exists (mrl, "VCD/INFO.VCD")
                || subentry_exists (mrl, "SVCD/INFO.VCD")
                || subentry_exists (mrl, "SVCD/INFO.SVD")));
}


bool
mrl_is_iso_image (const char *mrl)
{
    return (mrl_has_extension (mrl, "iso")
            || mrl_has_extension (mrl, "img")
            || mrl_has_extension (mrl, "bin")
            || mrl_has_extension (mrl, "cue")
            || mrl_has_extension (mrl, "toc")
            || mrl_has_extension (mrl, "nrg")
            || mrl_has_extension (mrl, "mdf")
            || mrl_has_extension (mrl, "mds")
            || mrl_has_extension (mrl, "ccd"));
}


bool
mrl_is_television (const char *mrl)
{
    return (mrl_has_protocol (mrl, "v4l")
            || mrl_has_protocol (mrl, "dvb")
            || mrl_has_protocol (mrl, "vdr")
            || mrl_has_protocol (mrl, "xvdr"));
}


bool
mrl_is_networked (const char *mrl)
{
    return (mrl_has_protocol (mrl, "http")
            || mrl_has_protocol (mrl, "rtsp")
            || mrl_has_protocol (mrl, "mms")
            || mrl_has_protocol (mrl, "ftp"));
}


bool
mrl_is_hidden (const char *mrl)
{
    if (!mrl) {
        return false;
    }

    char *base = get_basename (mrl);
    bool hidden = strc_starts_with (base, ".");
    ho_free (base);

    return hidden;
}


bool
mrl_get_filesize (const char *mrl, off_t * filesize)
{
    bool success = false;
    struct stat filestat;

    if (!mrl_is_local (mrl)) {
        return -1;
    }

    char *filename = mrl_get_filename (mrl);
    if (stat (filename, &filestat) == -1) {
        error (_("Unable to get status for file '%s': %s."),
               filename, strerror (errno));
    }
    else {
        *filesize = filestat.st_size;
        success = true;
    }
    ho_free (filename);

    return success;
}


bool
mrl_get_mtime (const char *mrl, time_t * mtime)
{
    bool success = false;
    struct stat filestat;

    if (!mrl_is_local (mrl)) {
        return 0;
    }

    char *filename = mrl_get_filename (mrl);
    if (stat (filename, &filestat) == -1) {
        *mtime = time (NULL);
        error (_("Unable to get status for file '%s': %s."),
               filename, strerror (errno));
    }
    else {
        *mtime = filestat.st_mtime;
        success = true;
    }
    ho_free (filename);

    return success;
}


char *
mrl_get_cache_dir (const char *mrl)
{
    if (mrl[strlen (mrl) - 1] != '/') {
        return ho_strdup_printf ("%s/%s/", get_dir_oxine_cache (), mrl);
    }
    else {
        return ho_strdup_printf ("%s/%s", get_dir_oxine_cache (), mrl);
    }
}


char *
mrl_get_quality (const char *mrl)
{
    const char *ext = get_extension (mrl);

    if (!ext) {
        return NULL;
    }

    bool lossy = false;
    lossy |= (strcasecmp (ext, "mp3") == 0);
    lossy |= (strcasecmp (ext, "ogg") == 0);

    bool lossless = false;
    lossless |= (strcasecmp (ext, "wav") == 0);
    lossless |= (strcasecmp (ext, "flac") == 0);

    if (lossy) {
        return ho_strdup_printf ("%s (%s)", _("Lossy"), ext);
    }
    if (lossless) {
        return ho_strdup_printf ("%s (%s)", _("Lossless"), ext);
    }

    return NULL;
}


int
copyfile (const char *old_path, const char *new_path)
{
    char buffer[1024];

    int rlen = 0;
    int wlen = 0;

    int f0 = open (old_path, O_RDONLY);
    if (f0 == -1) {
        return -1;
    }
    int f1 = creat (new_path, 0666);
    if (f1 == -1) {
        close (f0);
        return -1;
    }

    while (1) {
        {
            if ((rlen = read (f0, buffer, 1024)) == -1) {
                close (f0);
                close (f1);
                return -1;
            }
            else if (rlen == 0) {
                break;
            }
        }

        while (rlen > 0) {
            if ((wlen = write (f1, buffer, rlen)) == -1) {
                close (f0);
                close (f1);
                return -1;
            }
            rlen -= wlen;
        }
    }

    close (f0);
    close (f1);

    return 0;
}


bool
mrl_is_newer (const char *m1, const char *m2)
{
    time_t mt1;
    time_t mt2;
    mrl_get_mtime (m1, &mt1);
    mrl_get_mtime (m2, &mt2);
    return (mt2 < mt1);
}
