
/*
 * Copyright (C) 2004-2005 Maximilian Schwerin
 *
 * This file is part of oxine a free media player.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: utils_mrl.h 3967 2009-08-23 13:45:38Z mschwerin $
 *
 */

#ifndef HAVE_UTILS_MRL_H
#define HAVE_UTILS_MRL_H

#include "types.h"


/**
 * Returns the number in a filename (e.g. cdda://1). If the MRL does not
 * contain a number -1 is returned.
 */
int mrl_get_number (const char *mrl);


/**
 * Returns the filename contained in the passed MRL.
 *
 * The returned string must be freed.
 */
char *mrl_get_filename (const char *mrl);


/**
 * Returns TRUE if the MRL has the given protocol.
 *
 */
bool mrl_has_protocol (const char *mrl, const char *protocol);


/**
 * Returns the protocol part of the MRL.
 *
 * The returned string must be freed.
 */
char *mrl_get_protocol (const char *mrl);


/**
 * Returns the MRL without any query string.
 *
 * The returned string must be freed.
 */
char *mrl_without_query (const char *mrl);


/**
 * Attaches one query parameter to an existing MRL.
 *
 * The returned string must be freed.
 */
char *mrl_attach_query (const char *mrl, const char *key, const char *value);


/**
 * Returns the query part of the MRL.
 *
 * The returned string must be freed.
 */
char *mrl_get_query (const char *mrl);


/**
 * Returns the value of the query key of the query.
 *
 * The returned string must be freed.
 */
char *query_get_parameter (const char *query, const char *key);


/**
 * Returns the value of the query key of the MRL.
 *
 * The returned string must be freed.
 */
char *mrl_get_query_parameter (const char *mrl, const char *key);


/**
 * Converts a regular filename to an MRL. If the filename is already in MRL
 * form an unchanged copy is returned.
 *
 * The returned string must be freed.
 *
 * @param absolute              An absolute pathname.
 */
char *mrl_from_filename (const char *absolute);


/**
 * Builds a MRL from the given input name, the absolute pathname and an
 * optional track number.
 *
 * The returned string must be freed.
 *
 * @param input                 The name of an input plugin.
 * @param absolute              An absolute pathname.
 * @param tracknumber           The number of a track. If this number is
 *                              smaller or equal to zero it is not used.
 */
char *mrl_build (const char *input, const char *absolute, int tracknumber);


/**
 * Returns TRUE if the MRL points to a device file.
 */
bool mrl_is_device (const char *mrl);


/**
 * Returns TRUE if the MRL points to a file on a local volume.
 */
bool mrl_is_local (const char *mrl);


/**
 * Returns TRUE if the passed string is a valid MRL.
 */
bool mrl_is_valid (const char *mrl);


/**
 * Returns TRUE if both MRLs point to the same place.
 */
bool mrls_are_equal (const char *m1, const char *m2);


/**
 * Resolves a softlink. The returned string is a a valid filename on a
 * filesystem.
 *
 * The returned string must be freed.
 */
char *file_resolve_softlink (const char *mrl);


/**
 * Resolves a softlink. If the passed MRL is a softlink the target is
 * returned. If it is not a softlink the MRL itself is returned. The
 * returned string is an MRL.
 *
 * The returned string must be freed.
 */
char *mrl_resolve_softlink (const char *mrl);


/**
 * Returns the directory of the passed MRL.
 *
 * The returned string must be freed.
 */
char *get_dirname (const char *mrl);


/**
 * Returns the base name of the passed MRL.
 *
 * The returned string must be freed.
 */
char *get_basename (const char *mrl);


/**
 * Returns the extension of the passed MRL.
 */
const char *get_extension (const char *mrl);


/**
 * Returns the size of a file.
 *
 * @return                      On success TRUE is returned.
 */
bool mrl_get_filesize (const char *mrl, off_t * filesize);


/**
 * Returns the time this file was last modified.
 *
 * @return                      On success TRUE is returned.
 */
bool mrl_get_mtime (const char *mrl, time_t * mtime);


/**
 * Returns TRUE if the file m1 was modified more recently than the file m2.
 */
bool mrl_is_newer (const char *m1, const char *m2);


/**
 * Returns TRUE if the passed filename matches the passed extension.
 */
bool mrl_has_extension (const char *filename, const char *extension);


/**
 * Recursively creates a directory.
 */
bool mkdir_recursive (const char *mrl, mode_t mode);


/**
 * Returns TRUE if the passed MRL exists.
 */
bool file_exists (const char *mrl);


/**
 * Escapes a filename to be used in a URI. The returned string must be
 * freed.
 */
char *filename_escape_to_uri (const char *filename);


/**
 * Escapes a filename to be used in a shell. The returned string must be
 * freed.
 */
char *filename_escape_for_shell (const char *filename);


/**
 * Read the complete file into a buffer which is returned. The returned buffer
 * must be freed.
 */
char *read_entire_file (const char *mrl, int *size);


/**
 * Returns TRUE if the MRL is an oxine playlist.
 */
bool mrl_is_playlist_oxp (const char *mrl);


/**
 * Returns TRUE if the MRL is an oxine mediamarks file.
 */
bool mrl_is_mediamarks_oxm (const char *mrl);


/**
 * Returns TRUE if the MRL is an M3U playlist.
 */
bool mrl_is_playlist_m3u (const char *mrl);


/**
 * Returns TRUE if the MRL is an PLS playlist.
 */
bool mrl_is_playlist_pls (const char *mrl);


/**
 * Returns TRUE if the MRL is a directory.
 */
bool mrl_is_directory (const char *mrl);


/**
 * Returns TRUE if the MRL is DVD copied to disc.
 */
bool mrl_is_directory_dvd (const char *mrl);


/**
 * Returns TRUE if the MRL is VCD copied to disc.
 */
bool mrl_is_directory_vcd (const char *mrl);


/**
 * Returns TRUE if the MRL is an ISO image.
 */
bool mrl_is_iso_image (const char *mrl);


/**
 * Returns TRUE if the MRL is a television MRL.
 */
bool mrl_is_television (const char *mrl);


/**
 * Returns TRUE if the MRL is a network MRL.
 */
bool mrl_is_networked (const char *mrl);


/**
 * Returns TRUE if the MRL is a hidden file.
 */
bool mrl_is_hidden (const char *mrl);


/**
 * Returns the cache directory for the MRL.
 */
char *mrl_get_cache_dir (const char *mrl);


/**
 * Returns the quality (lossy or lossless) for the MRL.
 */
char *mrl_get_quality (const char *mrl);


/**
 * Copies the file from old_path to new_path.
 */
int copyfile (const char *old_path, const char *new_path);


#endif /* HAVE_UTILS_MRL_H */
