
/*
 * Copyright (C) 2004-2005 Maximilian Schwerin
 *
 * This file is part of oxine a free media player.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: utils_playlist.c 3600 2008-12-08 22:28:25Z mschwerin $
 *
 */
#include "config.h"

#include "i18n.h"
#include "heap.h"
#include "logger.h"
#include "oxine.h"
#include "odk.h"
#include "playlist.h"
#include "utils.h"

#include "menu_base.h"
#include "menu_dialog.h"
#include "menu_main.h"

extern oxine_t *oxine;


static uint32_t
dvd_get_stream_info (xine_stream_t * stream, const char *mrl, int key)
{
    if (xine_open (stream, mrl) != 1) {
        return -1;
    }

    uint32_t value = xine_get_stream_info (stream, key);

    xine_close (stream);

    return value;
}


static void
dvd_find_main_title (playitem_t * playitem)
{
    int found_length = 0;
    char *found_mrl = NULL;

    xine_stream_t *stream = xine_stream_new (oxine->xine, NULL, NULL);
    if (!stream) {
        return;
    }

    int i = 0;
    int count = dvd_get_stream_info (stream, playitem->mrl,
                                     XINE_STREAM_INFO_DVD_TITLE_COUNT);
    for (i = 1; i <= count; i++) {
        char *smrl = ho_strdup_printf ("%s/%d", playitem->mrl, i);
        int length = dvd_get_stream_info (stream, smrl,
                                          XINE_STREAM_INFO_DVD_CHAPTER_COUNT);

        if (length > found_length) {
            found_length = length;
            ho_free (found_mrl);
            found_mrl = smrl;
        }
        else {
            ho_free (smrl);
        }
    }

    xine_dispose (stream);

    if (found_mrl) {
        ho_free (playitem->mrl);
        playitem->mrl = found_mrl;
    }
}


void
playlist_rw_change_cb (playlist_t * playlist)
{
    assert (playlist == oxine->rw_playlist);

    if (odk_current_is_playback_mode (oxine->odk)
        && (oxine->current_playlist == oxine->rw_playlist)) {
        const char *mrl = odk_current_get_mrl (oxine->odk);

        playitem_t *playitem = playlist_first (playlist);
        while (playitem) {
            playitem->start_time = 0;
            playitem = playlist_next (playlist, playitem);
        }

        playitem_t *current = playlist_get_current (playlist);
        if (current && str_equals (mrl, current->mrl)) {
            odk_get_pos_length (oxine->odk, NULL, &current->start_time, NULL);
        }
    }

    playlist_save (playlist, NULL);
}


typedef struct {
    bool (*error_cb) (int start_time);
    otk_cb_t backto_menu;
    otk_cb_t backto_menu_data;
    otk_cb_t current_menu;
    otk_cb_t current_menu_data;
    playitem_t *failed;
} playlist_try_next_t;


static void
playlist_try_next_yes (void * p)
{
    playlist_try_next_t *d = (playlist_try_next_t*) p;

    assert (d);
    assert (d->backto_menu);
    assert (d->current_menu);
    assert (d->failed);

    set_backto_menu (d->backto_menu, d->backto_menu_data);
    set_current_menu (d->current_menu, d->current_menu_data);
    d->error_cb (0);

    playlist_remove (oxine->current_playlist, d->failed);

    ho_free (d);
}


static void
playlist_try_next_no (void * p)
{
    playlist_try_next_t *d = (playlist_try_next_t*) p;

    playback_ended_menu_cb (NULL);

    ho_free (d);
}


static bool
playitem_play_stream (playitem_t * playitem)
{
    return odk_play_stream (oxine->odk, playitem->title,
                            playitem->mrl, playitem->subtitle_mrl,
                            playitem->start_time, ODK_MODE_NORMAL);
}


static bool
playitem_play (playitem_t * playitem, int start_time,
               bool (*error_cb) (int start_time))
{
    if (!playitem) {
        playback_ended_menu_cb (oxine);
        return true;
    }

    if (mrl_has_protocol (playitem->mrl, "http")
        || mrl_has_protocol (playitem->mrl, "rtsp")
        || mrl_has_protocol (playitem->mrl, "cdda")
        || mrl_has_protocol (playitem->mrl, "dvd")
        || mrl_has_protocol (playitem->mrl, "vcd")
        || mrl_has_protocol (playitem->mrl, "dvb")
        || mrl_has_protocol (playitem->mrl, "v4l")) {
        /* This is only necessary when we start oxine and start playing a
         * video at once. If we call please_wait in this situation we
         * first load the background and a very short time after that
         * start playing the video. That would not look good! */
        if (odk_current_is_logo_mode (oxine->odk)
            || odk_current_is_playback_mode (oxine->odk)) {
            otk_cb_t current_menu = get_current_menu ();
            void *current_menu_data = get_current_menu_data ();
            please_wait (NULL);
            set_current_menu (current_menu, current_menu_data);
        }
    }

    if (mrl_has_protocol (playitem->mrl, "dvd")
        && config_get_bool ("misc.try_to_find_main_dvd_title")) {
        dvd_find_main_title (playitem);
    }

    if (start_time < 0) {
        start_time += (playitem->playback_length * 1000);
    }
    if (start_time > 0) {
        playitem->start_time = start_time;
    }

    bool success = false;
    if (playitem->alternative_list) {
        playlist_t *altlist = playitem->alternative_list;
        playitem_t *altitem = playlist_first (altlist);
        while (!success && altitem) {
            altitem->start_time = playitem->start_time;
            success = playitem_play_stream (altitem);

            if (!success) {
                playlist_remove (altlist, altitem);
                altitem = playlist_first (altlist); 
            }
        }
    }
    else {
        success = playitem_play_stream (playitem);
    }

    if (!success) {
        char *msg = NULL;
        otk_cb_t cb0 = playback_ended_menu_cb;
        void * dt0 = NULL;
        otk_cb_t cb1 = playback_ended_menu_cb;
        void * dt1 = NULL;
        message_dialog_type_t type = DIALOG_OK_CANCEL;

        if (mrl_has_protocol (playitem->mrl, "cdda")) {
            if (mrl_is_device (playitem->mrl)) {
                msg = ho_strdup_printf ("%s\n%s",
                                        _("Could not play Audio CD!"),
                                        _("Please insert an Audio CD!"));
                cb0 = play_cdda_cb;
                type = DIALOG_RETRY_CANCEL;
            }
            else {
                msg = ho_strdup (_("Could not play Audio CD!"));
            }
        }
        else if (mrl_has_protocol (playitem->mrl, "vcd")) {
            if (mrl_is_device (playitem->mrl)) {
                msg = ho_strdup_printf ("%s\n%s",
                                        _("Could not play Video CD!"),
                                        _("Please insert a Video CD!"));
                cb0 = play_vcd_cb;
                type = DIALOG_RETRY_CANCEL;
            }
            else {
                msg = ho_strdup (_("Could not play Video CD!"));
            }
        }
        else if (mrl_has_protocol (playitem->mrl, "dvd")) {
            if (mrl_is_device (playitem->mrl)) {
                msg = ho_strdup_printf ("%s\n%s",
                                        _("Could not play DVD!"),
                                        _("Please insert a DVD!"));
                cb0 = play_dvd_cb;
                type = DIALOG_RETRY_CANCEL;
            }
            else {
                msg = ho_strdup (_("Could not play DVD!"));
            }
        }
        else if (mrl_has_protocol (playitem->mrl, "v4l")
                 || mrl_has_protocol (playitem->mrl, "dvb")
                 || mrl_has_protocol (playitem->mrl, "vdr")
                 || mrl_has_protocol (playitem->mrl, "xvdr")) {
            msg = ho_strdup (_("Failed to start television playback!"));
            cb0 = play_tv_cb;
            type = DIALOG_RETRY_CANCEL;
        }
        else {
            playlist_try_next_t *d = ho_new (playlist_try_next_t);
            d->error_cb = error_cb;
            d->backto_menu = get_backto_menu ();
            d->backto_menu_data = get_backto_menu_data ();
            d->current_menu = get_current_menu ();
            d->current_menu_data = get_current_menu_data ();
            d->failed = playitem;

            const char *s = _("Try to play next in list?");
            if (error_cb == playlist_play_prev) {
                s = _("Try to play previous in list?");
            }
            
            msg = ho_strdup_printf ("%s\n%s", _("Could not play title!"), s);

            type = DIALOG_YES_NO;
            cb0 = playlist_try_next_yes;
            dt0 = d;
            cb1 = playlist_try_next_no;
            dt1 = d;
        }

        set_backto_menu (playback_ended_menu_cb, NULL);
        show_message_dialog (cb0, dt0 , cb1, dt1 , type, NULL, msg);
        ho_free (msg);

        return false;
    }
    else {
        return true;
    }
}


bool
playlist_play_first (playlist_t * playlist)
{
    bool result = false;

    assert (playlist);

    set_backto_menu (hide_user_interface, NULL);
    set_current_menu (hide_user_interface, NULL);
    oxine->current_playlist = playlist;

    playlist_lock (oxine->current_playlist);
    playitem_t *playitem = playlist_get_first (oxine->current_playlist);
    result = playitem_play (playitem, 0, playlist_play_next);
    playlist_unlock (oxine->current_playlist);

    return result;
}


bool
playlist_play_item (playlist_t * playlist, playitem_t * playitem)
{
    bool result = false;

    assert (playlist);

    set_backto_menu (hide_user_interface, NULL);
    set_current_menu (hide_user_interface, NULL);
    oxine->current_playlist = playlist;

    playlist_lock (oxine->current_playlist);
    playlist_set_current (oxine->current_playlist, playitem);
    result = playitem_play (playitem, 0, playlist_play_next);
    playlist_unlock (oxine->current_playlist);

    return result;
}


bool
playlist_play_next (int start_time)
{
    bool result = false;

    playlist_lock (oxine->current_playlist);
    playitem_t *playitem = playlist_get_next (oxine->current_playlist);
    result = playitem_play (playitem, start_time, playlist_play_next);
    playlist_unlock (oxine->current_playlist);

    return result;
}


bool
playlist_play_prev (int start_time)
{
    bool result = false;

    playlist_lock (oxine->current_playlist);
    playitem_t *playitem = playlist_get_prev (oxine->current_playlist);
    result = playitem_play (playitem, start_time, playlist_play_prev);
    playlist_unlock (oxine->current_playlist);

    return result;
}


bool
play_autoplay (const char *title, const char *pid)
{
    playitem_t *item = NULL;
    playlist_clear (oxine->ro_playlist);

    if (!(item = playlist_add_autoplay (oxine->ro_playlist, title, pid))) {
        return false;
    }
    else if (!playlist_play_item (oxine->ro_playlist, item)) {
        return false;
    }
    else {
        return true;
    }
}
