
/*
 * Copyright (C) 2006 Maximilian Schwerin
 *
 * This file is part of oxine a free media player.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: vdr.c 4038 2011-05-06 19:02:45Z mschwerin $
 *
 */

#include "config.h"

#define _GNU_SOURCE

#include <errno.h>
#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>

#include "codeset.h"
#include "environment.h"
#include "heap.h"
#include "i18n.h"
#include "lang.h"
#include "list.h"
#include "logger.h"
#include "meta_info.h"
#include "oxine.h"
#include "utils.h"
#include "vdr.h"

#ifdef HAVE_VDR

//#define DEBUG_COMM

#define TIMERS_FLAG_ACTIVE      0x0001
#define TIMERS_FLAG_INSTANT     0x0002
#define TIMERS_FLAG_VPS         0x0004
#define TIMERS_FLAG_RECORDING   0x0008

#define BUFSIZE                 1024 * 8
#define CMDSIZE                 64

extern oxine_t *oxine;

typedef struct {
    int stream_content;
    int stream_type;
    char *language;
    char *description;
} stream_info_t;

typedef struct {
    int id;
    int channel_id;
    int timer_id;

    time_t time_beg;
    time_t time_end;

    char *title;
    char *description_short;
    char *description_long;

    bool is_active;
    bool is_cut;
    bool is_recording;
    bool is_scheduled;

    l_list_t *stream_info;
} epg_entry_t;

typedef struct {
    char *title;
    epg_entry_t epg_entry;
} recording_cache_entry_t;

static char *codeset_vdr = NULL;

static int num_cache_entries = 0;

static recording_cache_entry_t *cache_entries = NULL;


static void
stream_info_free (void *p)
{
    stream_info_t *s = (stream_info_t *) p;
    ho_free (s->language);
    ho_free (s->description);
    ho_free (s);
}


static bool
stream_info_swap (void *p1, void *p2)
{
    stream_info_t *i1 = (stream_info_t *) p1;
    stream_info_t *i2 = (stream_info_t *) p2;

    return (i1->stream_type > i2->stream_type);
}


static bool
send_command (int sd, const char *command)
{
#ifdef DEBUG_COMM
    debug ("->VDR: %s", command);
#endif

    if (send (sd, command, strlen (command), 0) == -1) {
        error (_("Could not send command to server: %s!"), strerror (errno));
        return false;
    }
    if (send (sd, "\n", 1, 0) == -1) {
        error (_("Could not send command to server: %s!"), strerror (errno));
        return false;
    }

    return true;
}


static bool
recv_answer (int sd, char *buf, int buf_size, const char *expected_code)
{
    int len = 0;

    if ((len = readline (sd, buf, buf_size)) == -1) {
        error (_("Could not read from server: %s!"), strerror (errno));
        return false;
    }
    else if (len == 1) {
        return recv_answer (sd, buf, buf_size, expected_code);
    }

    trim_whitespace (buf);
#ifdef DEBUG_COMM
    debug ("<-VDR: %s", buf);
#endif

    if (strlen (buf) < strlen (expected_code)) {
        error (_("Command to VDR failed: %s!"), &buf[0]);
        return false;
    }
    else if (!strc_starts_with (buf, expected_code)) {
        error (_("Command to VDR failed: %s!"), &buf[4]);
        return false;
    }

    return true;
}


static bool
simple_command (int sd, const char *command, const char *expected_code)
{
    if (!send_command (sd, command)) {
        return false;
    }

    char buf[BUFSIZE];
    if (!recv_answer (sd, buf, BUFSIZE, expected_code)) {
        return false;
    }

    return true;
}


static void
svdrp_disconnect (int sd)
{
    simple_command (sd, "QUIT", "221");
    close (sd);
}


static int
svdrp_connect (void)
{
    char buf[BUFSIZE];
    int sd;
    struct hostent *hp;
    struct sockaddr_in pin;

    const int port = config_get_number ("television.vdr.svdrp.port");
    const char *host = config_get_string ("television.vdr.svdrp.host");

    /* Go find out about the desired host machine. */
    if ((hp = gethostbyname (host)) == 0) {
        error (_("Could not resolve '%s': %s!"), host, strerror (errno));
        return 0;
    }

    /* Fill in the socket structure with host information. */
    memset (&pin, 0, sizeof (pin));
    pin.sin_family = AF_INET;
    pin.sin_addr.s_addr = ((struct in_addr *) (hp->h_addr))->s_addr;
    pin.sin_port = htons (port);

    /* Grab an internet domain socket. */
    if ((sd = socket (AF_INET, SOCK_STREAM, 0)) == -1) {
        error (_("Could not create socket: %s!"), strerror (errno));
        return 0;
    }

    /* Connect to port on host. */
    if (connect (sd, (struct sockaddr *) &pin, sizeof (pin)) == -1) {
        error (_("Could not connect to '%s': %s!"), host, strerror (errno));
        close (sd);
        return 0;
    }

    /* Wait for server to say hello. */
    if (!recv_answer (sd, buf, BUFSIZE, "220")) {
        close (sd);
        return 0;
    }

    if (codeset_vdr != NULL) {
        /* nichts tun */
    }
    else if (config_get_bool ("television.vdr.charset.override")) {
        const char *codeset = config_get_string ("television.vdr.charset");
        codeset_vdr = ho_strdup (codeset);
        debug ("Using user provided charset %s for VDR.", codeset_vdr);
    }
    else {
        int v0;
        int v1;
        int v2;

        if (sscanf (buf, "220 %*s SVDRP VideoDiskRecorder "
                    "%d.%d.%d", &v0, &v1, &v2) != 3) {
            error (_("Failed to detect VDR version: %s!"), strerror (errno));
            codeset_vdr = ho_strdup ("ISO-8859-1");
        }
        else if ((v0 >= 1) && (v1 >= 6)) {
            const char *codeset = get_system_encoding ();
            codeset_vdr = ho_strdup (codeset);
        }
        else {
            codeset_vdr = ho_strdup ("ISO-8859-1");
        }

        debug ("Detected VDR version %d.%d.%d. Using charset %s.",
               v0, v1, v2, codeset_vdr);
    }

    return sd;
}


static bool
vdr_rec_swap_cb (void *p1, void *p2)
{
    fileitem_t *f1 = (fileitem_t *) p1;
    fileitem_t *f2 = (fileitem_t *) p2;

    return (f1->vdr.time_beg < f2->vdr.time_beg);
}


static bool
vdr_tim_swap_cb (void *p1, void *p2)
{
    fileitem_t *f1 = (fileitem_t *) p1;
    fileitem_t *f2 = (fileitem_t *) p2;

    return (f1->vdr.time_beg > f2->vdr.time_beg);
}


/**
 * Parses one line of an EPG entry.
 *
 * ======
 * E 435 1240219800 1800 4E 1C
 * T Unsere erste gemeinsame Wohnung
 * D Junge Paare auf ihrem Weg zur ersten gemeinsamen Wohnung
 * X 2 03 deu deutsch
 * X 2 05 deu dolby digital
 * e
 * ======
 *
 * @return                      Returns the first character of the line.
 *                              Returns -1 on error.
 */
static int
vdr_parse_epg_entry_line (const char *line, recode_t * xr,
                          epg_entry_t * epg_entry)
{
    stream_info_t *stream_info;
    char language[BUFSIZE] = { '\0' };
    char description[BUFSIZE] = { '\0' };

    if (line[0] == 'e') {                                       // End of an EPG entry.
        return line[0];
    }
    if (line[0] == 'c') {                                       // End of a channel.
        return line[0];
    }
    if (line[1] != ' ') {
        error ("Bad EPG entry line: %s!", line);
        return -1;
    }

    switch (line[0]) {
    case 'C':
        break;
    case 'E':
        if (sscanf (line, "E %d %ld %ld", &epg_entry->id,
                    &epg_entry->time_beg, &epg_entry->time_end) != 3) {
            debug (line);
            error (_("Failed to parse EPG entry line: %s!"),
                   strerror (errno));
            return -1;
        }
        else {
            epg_entry->time_end += epg_entry->time_beg;
        }
        break;
    case 'T':
        {
            epg_entry->title = recode (xr, &line[2]);
        }
        break;
    case 'S':
        {
            epg_entry->description_short = recode (xr, &line[2]);
        }
        break;
    case 'D':
        {
            epg_entry->description_long = recode (xr, &line[2]);
            replace_chars (epg_entry->description_long, '|', '\n');
        }
        break;
    case 'X':
        if (!epg_entry->stream_info) {
            epg_entry->stream_info = l_list_new ();
        }
        {
            stream_info = ho_new (stream_info_t);
        }
        if (sscanf (line, "X %d %d %s %s",
                    &stream_info->stream_type, &stream_info->stream_content,
                    (char *) &language, (char *) &description) < 3) {
            debug (line);
            warn (_("Failed to parse EPG entry line: %s!"), strerror (errno));
            ho_free (stream_info);
        }
        else {
            stream_info->language = ho_strdup (language);
            stream_info->description = recode (xr, description);
            l_list_append (epg_entry->stream_info, stream_info);
        }
        break;
    }

    return line[0];
}


/**
 * Parses EPG entries.
 *
 * @return                      Returns 0 after successfully parsing any
 *                              entry of the list but the last.
 *                              Returns 1 after successfully parsing the
 *                              last entry of the list.
 *                              Returns -1 if an error occurs.
 */
static int
vdr_parse_epg_entry (int sd, recode_t * xr, epg_entry_t * epg_entry)
{
    memset (epg_entry, 0, sizeof (epg_entry_t));
    epg_entry->is_active = true;

    char buf[BUFSIZE];
    while (1) {
        if (!recv_answer (sd, buf, BUFSIZE, "215")) {
            return -1;
        }
        else if (strc_starts_with (buf, "215 ")) {
            return 1;
        }

        int res = vdr_parse_epg_entry_line (&buf[4], xr, epg_entry);

        if (res == -1) {
            return -1;
        }
        else if (res == 'e') {
            return 0;
        }
    }
}


/**
 * Parses timer entries.
 *
 * @return                      Returns 0 after successfully parsing any
 *                              entry of the list but the last.
 *                              Returns 1 after successfully parsing the
 *                              last entry of the list.
 *                              Returns -1 if an error occurs.
 */
static int
vdr_parse_timer_entry (int sd, recode_t * xr, epg_entry_t * epg_entry)
{
    memset (epg_entry, 0, sizeof (epg_entry_t));
    epg_entry->is_active = true;

    char buf[BUFSIZE];
    if (!recv_answer (sd, buf, BUFSIZE, "250")) {
        return -1;
    }

    bool is_last = strc_starts_with (buf, "250 ");

    const char *data = &buf[4];

    epg_entry->timer_id = atoi (data);

    const char *flags = nindex (data, ' ', 1);
    if (flags) {
        epg_entry->is_active = (atoi (&flags[1]) & TIMERS_FLAG_ACTIVE);
        epg_entry->is_recording = (atoi (&flags[1]) & TIMERS_FLAG_RECORDING);
    }

    const char *cid = nindex (data, ':', 1);
    if (cid) {
        epg_entry->channel_id = atoi (&cid[1]);
    }

    const char *epgid = strstr (data, "<epgid>");
    if (epgid) {
        epg_entry->id = atoi (&epgid[7]);
    }

    const char *eventid = strstr (data, "<eventid>");
    if (eventid) {
        epg_entry->id = atoi (&eventid[9]);
    }

    struct tm ts;
    char *ts_str = (char *) nindex (data, ':', 2);
    if (ts_str) {
        memset (&ts, 0, sizeof (struct tm));
        strptime (&ts_str[1], "%Y-%m-%d:%H%M", &ts);
        mktime (&ts);
        strptime (&ts_str[1], "%Y-%m-%d:%H%M", &ts);
    }
    else {
        return -1;
    }

    struct tm te;
    char *te_str = (char *) nindex (data, ':', 4);
    if (te_str) {
        memcpy (&te, &ts, sizeof (struct tm));
        strptime (&te_str[1], "%H%M", &te);
    }
    else {
        return -1;
    }

    /* If the end time is smaller than the starting time there is probably
     * a change of day in between. To fix this we add one to the day and
     * normalize the end time using mktime. */
    if ((te.tm_min + te.tm_hour * 60) < (ts.tm_min + ts.tm_hour * 60)) {
        te.tm_mday += 1;
    }

    epg_entry->time_beg = mktime (&ts);
    epg_entry->time_end = mktime (&te);

    char *n_end = (char *) nindex (data, ':', 8);
    if (n_end) {
        n_end[0] = '\0';
    }
    else {
        return -1;
    }

    char *n_beg = (char *) nindex (data, ':', 7);
    if (n_beg) {
        n_beg = trim_whitespace (&n_beg[1]);

        char *dp = index (n_beg, '|');
        while (dp) {
            dp[0] = ':';
            dp = index (n_beg, '|');
        }
    }
    else {
        return -1;
    }

    epg_entry->title = recode (xr, n_beg);

    return is_last ? 1 : 0;
}


static void
vdr_epg_entry_to_fileitem (fileitem_t * fileitem,
                           epg_entry_t * epg_entry, bool epg_entry_free,
                           bool update_fileitem, bool update_infolist)
{
    const char *title_gui_format =
        "<tr>"
        "<td width=40>%s,</td>"
        "<td width=130>%s</td>"
        "<td width=140>%s-%s</td>"
        "<td>%s</td>"
        "<td width=130>%s</td>"
        "</tr>";
    const char *info_line_format =
        "<tr>"
        "<td width=150>%s</td>"
        "<td>%s</td>"
        "<td width=150>%s</td>"
        "</tr>";

    if (update_fileitem) {
        struct tm ts;
        struct tm te;
        localtime_r (&epg_entry->time_beg, &ts);
        localtime_r (&epg_entry->time_end, &te);

        char *wday = ho_strdup_strftime ("%a", &ts);
        char *date = ho_strdup_strftime ("%d.%m.%Y", &ts);
        char *tbeg = ho_strdup_strftime ("%H:%M", &ts);
        char *tend = ho_strdup_strftime ("%H:%M", &te);

        const char *title_gui_comment = "";
        if (!epg_entry->is_active) {
            title_gui_comment = _("deactivated");
        }
        else if (epg_entry->is_cut) {
            title_gui_comment = _("cut");
        }
        else if (epg_entry->is_recording) {
            title_gui_comment = _("recording");
        }
        else if (epg_entry->is_scheduled) {
            title_gui_comment = _("scheduled");
        }

        ho_free (fileitem->title);
        fileitem->title = ho_strdup (epg_entry->title);
        ho_free (fileitem->title_gui);
        fileitem->title_gui = ho_strdup_printf (title_gui_format, wday,
                                                date, tbeg, tend,
                                                epg_entry->title,
                                                title_gui_comment);
        ho_free (fileitem->sort_by);
        fileitem->sort_by = ho_strdup_strftime ("%F %H:%M", &ts);
        fileitem->vdr.epg_id = epg_entry->id;
        fileitem->vdr.channel_id = epg_entry->channel_id;
        fileitem->vdr.time_beg = epg_entry->time_beg;
        fileitem->vdr.time_end = epg_entry->time_end;
        fileitem->vdr.is_active = epg_entry->is_active;
        fileitem->vdr.is_recording = epg_entry->is_recording;
        fileitem->vdr.is_scheduled = epg_entry->is_scheduled;

        ho_free (wday);
        ho_free (date);
        ho_free (tbeg);
        ho_free (tend);
    }
    if (update_infolist) {
        fileitem_free_info_list (fileitem);
        fileitem->info_list = l_list_new ();
        if (epg_entry->title) {
            char *title = ho_strdup (epg_entry->title);
            l_list_append (fileitem->info_list, title);
        }
        if (epg_entry->description_short) {
            char *description = ho_strdup (epg_entry->description_short);
            l_list_append (fileitem->info_list, description);
        }
        if (epg_entry->description_long) {
            char *description = ho_strdup (epg_entry->description_long);
            l_list_append (fileitem->info_list, ho_strdup (""));
            l_list_append (fileitem->info_list, description);
        }
        if (epg_entry->stream_info && l_list_length (epg_entry->stream_info)) {
            bool had_audio = false;
            bool had_video = false;
            l_list_append (fileitem->info_list, ho_strdup ("*--*"));
            l_list_sort (epg_entry->stream_info, stream_info_swap);
            stream_info_t *si = l_list_first (epg_entry->stream_info);
            while (si) {
                const char *lang = get_language_from_iso639_1 (si->language);
                char *type = NULL;
                char *desc = si->description;

                /**
                 * ETSI EN 300 468
                 */
                switch (si->stream_type) {
                case 0x01:
                    type = had_video ? "" : _("Video tracks:");
                    had_video = true;
                    lang = "";
                    switch (si->stream_content) {
                    case 0x01:
                    case 0x05:
                        desc = _("4:3");
                        break;
                    case 0x02:
                    case 0x03:
                    case 0x06:
                    case 0x07:
                        desc = _("16:9 (widescreen)");
                        break;
                    case 0x04:
                    case 0x08:
                        desc = _("> 16:9");
                        break;
                    case 0x09:
                    case 0x0D:
                        desc = _("high definition, 4:3");
                        break;
                    case 0x0A:
                    case 0x0B:
                    case 0x0E:
                    case 0x0F:
                        desc = _("high definition, 16:9 (widescreen)");
                        break;
                    case 0x0C:
                    case 0x10:
                        desc = _("high definition, > 16:9");
                        break;
                    }
                    break;
                case 0x02:
                    type = had_audio ? "" : _("Audio tracks:");
                    had_audio = true;
                    switch (si->stream_content) {
                    case 0x01:
                        desc = _("single mono channel");
                        break;
                    case 0x02:
                        desc = _("dual mono channel");
                        break;
                    case 0x03:
                        desc = _("stereo");
                        break;
                    case 0x04:
                        desc = _("multi-lingual, multi-channel");
                        break;
                    case 0x05:
                        desc = _("surround sound");
                        break;
                    case 0x40:
                        desc = _("audio description for "
                                 "the visually impaired");
                        break;
                    case 0x41:
                        desc = _("audio for the hard of hearing");
                        break;
                    }
                    break;
                }
                if (type != NULL) {
                    l_list_append (fileitem->info_list,
                                   ho_strdup_printf (info_line_format,
                                                     type, desc, lang));
                }
                si = l_list_next (epg_entry->stream_info, si);
            }
        }
    }
    if (epg_entry_free) {
        ho_free (epg_entry->title);
        ho_free (epg_entry->description_short);
        ho_free (epg_entry->description_long);
    }
    if (epg_entry_free && epg_entry->stream_info) {
        l_list_free (epg_entry->stream_info, stream_info_free);
    }
}


bool
vdr_is_recording (void)
{
    bool result = false;

    filelist_t *timers = filelist_get (FILELIST_VDR_TIM_MRL);
    if (timers) {
        filelist_lock (timers);
        fileitem_t *timer = filelist_first (timers);
        while (timer && !result) {
            result |= timer->vdr.is_recording;
            timer = filelist_next (timers, timer);
        }
        filelist_unlock (timers);
    }

    return result;
}


bool
vdr_recording_play (fileitem_t * fileitem)
{
    if (!strc_starts_with (fileitem->mrl, FILELIST_VDR_REC_MRL)) {
        error (_("This is not a valid VDR recordings MRL!"));
        return false;
    }

    int num = mrl_get_number (fileitem->mrl);
    debug ("starting VDR recording %d", num);

    int sd = svdrp_connect ();
    if (!sd) {
        return false;
    }

    bool result = false;
    char command[CMDSIZE];
    snprintf (command, CMDSIZE, "PLAY %d", num);
    if (!simple_command (sd, command, "250")) {
        goto error;
    }

    /* Make sure recordings is showing. */
    if (!simple_command (sd, "HITK Recordings", "250")) {
        goto error;
    }

    /* Hide the menu. */
    if (!simple_command (sd, "HITK Menu", "250")) {
        goto error;
    }

    /* Show the position. */
    if (!simple_command (sd, "HITK Ok", "250")) {
        goto error;
    }

    result = true;
  error:
    svdrp_disconnect (sd);

    return result;
}


bool
vdr_recording_remove (fileitem_t * fileitem)
{
    if (!strc_starts_with (fileitem->mrl, FILELIST_VDR_REC_MRL)) {
        error (_("This is not a valid MRL of a VDR recording!"));
        return false;
    }

    int num = mrl_get_number (fileitem->mrl);
    debug ("removing VDR recording %d", num);

    int sd = svdrp_connect ();
    if (!sd) {
        return false;
    }

    bool result = false;
    char command[CMDSIZE];
    snprintf (command, CMDSIZE, "DELR %d", num);
    if (!simple_command (sd, command, "250")) {
        goto error;
    }

    result = true;
  error:
    svdrp_disconnect (sd);

    return result;
}


static bool
vdr_recording_read (int sd, recode_t * xr, fileitem_t * fileitem)
{
    recording_cache_entry_t *ce = NULL;
    time_t time_now = time (NULL);

    int i = 0;
    for (i = 0; i < num_cache_entries; i++) {
        ce = &cache_entries[i];
        if (str_equals (ce->title, fileitem->title)) {
            goto out_good;
        }
    }

    num_cache_entries++;
    size_t size = sizeof (recording_cache_entry_t) * num_cache_entries;
    cache_entries = ho_realloc (cache_entries, size);
    ce = &cache_entries[num_cache_entries - 1];
    ce->title = ho_strdup (fileitem->title);

    char command[CMDSIZE];
    snprintf (command, CMDSIZE, "LSTR %d", mrl_get_number (fileitem->mrl));
    if (!send_command (sd, command)) {
        return false;
    }

    if (vdr_parse_epg_entry (sd, xr, &ce->epg_entry) == -1) {
        return false;
    }

  out_good:
    ce->epg_entry.is_recording = (ce->epg_entry.time_end >= time_now);
    ce->epg_entry.is_cut = (fileitem->title[16] == '%');
    vdr_epg_entry_to_fileitem (fileitem, &ce->epg_entry, false, true, true);

    return true;
}


bool
vdr_recordings_read (filelist_t * filelist)
{
    int sd = svdrp_connect ();
    if (!sd) {
        ho_free (filelist->error);
        filelist->error = ho_strdup (_("Connection failure..."));
        return false;
    }

    bool result = false;
    recode_t *xr = recode_init (codeset_vdr, USE_SYSTEM_ENCODING);

    bool cut[256];
    memset (cut, 0, sizeof (bool) * 256);

    if (!send_command (sd, "LSTR")) {
        goto error;
    }

    char buf[BUFSIZE];
    do {
        if (!recv_answer (sd, buf, BUFSIZE, "250")) {
            goto error;
        }

        int num = atoi (&buf[4]);
        char *mrl = mrl_build ("vdr", "/recordings", num);

        char *title = (char *) nindex (&buf[4], ' ', 1);
        title = trim_whitespace (title);
        title = recode (xr, title);

        filelist_add (filelist, title, mrl, FILE_TYPE_REGULAR);

        ho_free (mrl);
        ho_free (title);
    } while (!strc_starts_with (buf, "250 "));

    fileitem_t *item = filelist_first (filelist);
    while (item) {
        if (!vdr_recording_read (sd, xr, item)) {
            goto error;
        }
        item = filelist_next (filelist, item);
    }

    filelist_sort (filelist, vdr_rec_swap_cb);

    result = true;
  error:
    svdrp_disconnect (sd);
    recode_done (xr);

    return result;
}


static bool
_vdr_channels_read (filelist_t * filelist)
{
    int sd = svdrp_connect ();
    if (!sd) {
        ho_free (filelist->error);
        filelist->error = ho_strdup (_("Connection failure..."));
        return false;
    }

    bool result = false;
    recode_t *xr = recode_init (codeset_vdr, USE_SYSTEM_ENCODING);

    if (!send_command (sd, "LSTC")) {
        goto error;
    }

    char buf[BUFSIZE];
    do {
        if (!recv_answer (sd, buf, BUFSIZE, "250")) {
            goto error;
        }

        int num = atoi (buf + 4);

        char *title = recode (xr, index (buf + 4, ' ') + 1);
        char *e = index (title, ';');
        if (e) {
            e[0] = '\0';
        }
        char *f = index (title, ',');
        if (f) {
            f[0] = '\0';
        }

        char *mrl = mrl_build ("vdr", "/channels", num);
        fileitem_t *fileitem = filelist_add (filelist, title, mrl,
                                             FILE_TYPE_FOLDER);
        ho_free (fileitem->sort_by);
        fileitem->sort_by = ho_strdup_printf ("%04d", num);
        ho_free (title);
        ho_free (mrl);
    } while (!strc_starts_with (buf, "250 "));

    filelist_sort (filelist, NULL);

    result = true;
  error:
    svdrp_disconnect (sd);
    recode_done (xr);

    return result;
}


static bool
_vdr_epg_read (filelist_t * filelist)
{
    if (!strc_starts_with (filelist->mrl, FILELIST_VDR_CHA_MRL)) {
        error (_("This is not a valid VDR channel MRL!"));
        return false;
    }

    int channel_id = mrl_get_number (filelist->mrl);

    fileitem_t *fileitem = filelist_add (filelist, _("Watch this channel..."),
                                         filelist->mrl, FILE_TYPE_REGULAR);
    ho_free (fileitem->sort_by);
    fileitem->sort_by = ho_strdup ("1970-01-01 00:00");

    filelist_t *timers = filelist_get (FILELIST_VDR_TIM_MRL);

    int sd = svdrp_connect ();
    if (!sd) {
        ho_free (filelist->error);
        filelist->error = ho_strdup (_("Connection failure..."));
        return false;
    }

    bool result = false;
    recode_t *xr = recode_init (codeset_vdr, USE_SYSTEM_ENCODING);

    char command[CMDSIZE];
    snprintf (command, CMDSIZE, "LSTE %d", channel_id);
    if (!send_command (sd, command)) {
        goto error;
    }

    epg_entry_t epg_entry;
    while (vdr_parse_epg_entry (sd, xr, &epg_entry) == 0) {
        char *mrl = mrl_build ("vdr", "/epg", epg_entry.id);
        epg_entry.channel_id = channel_id;

        filelist_lock (timers);
        fileitem_t *timer = filelist_first (timers);
        while (timer) {
            if (timer->vdr.channel_id != channel_id) {
                /* ignore it */
            }
            else if (timer->vdr.epg_id == epg_entry.id) {
                epg_entry.is_scheduled = true;
                epg_entry.is_recording = timer->vdr.is_recording;
            }
            else if (timer->vdr.time_beg < epg_entry.time_beg - 30 * 60) {
                /* ignore it */
            }
            else if (timer->vdr.time_beg > epg_entry.time_beg + 30 * 60) {
                /* ignore it */
            }
            else if (timer->vdr.time_end < epg_entry.time_end - 30 * 60) {
                /* ignore it */
            }
            else if (timer->vdr.time_end > epg_entry.time_end + 30 * 60) {
                /* ignore it */
            }
            else if (str_equals (timer->title, epg_entry.title)) {
                epg_entry.is_scheduled = true;
                epg_entry.is_recording = timer->vdr.is_recording;
            }

            timer = filelist_next (timers, timer);
        }
        filelist_unlock (timers);

        fileitem_t *fileitem = filelist_add (filelist, epg_entry.title,
                                             mrl, FILE_TYPE_REGULAR);
        vdr_epg_entry_to_fileitem (fileitem, &epg_entry, true, true, true);
        ho_free (mrl);
    }

    filelist_sort (filelist, NULL);

    result = true;
  error:
    svdrp_disconnect (sd);
    recode_done (xr);

    return result;
}


bool
vdr_channels_read (filelist_t * filelist)
{
    if (strc_equals (filelist->mrl, FILELIST_VDR_CHA_MRL)) {
        return _vdr_channels_read (filelist);
    }
    else if (strc_starts_with (filelist->mrl, FILELIST_VDR_CHA_MRL "?")) {
        return _vdr_channels_read (filelist);
    }
    else {
        return _vdr_epg_read (filelist);
    }
}


bool
vdr_channel_play (fileitem_t * fileitem)
{
    if (!strc_starts_with (fileitem->mrl, FILELIST_VDR_CHA_MRL)) {
        error (_("This is not a valid VDR channel MRL!"));
        return false;
    }

    int channel = mrl_get_number (fileitem->mrl);
    debug ("switching to VDR channel %d", channel);

    int sd = svdrp_connect ();
    if (!sd) {
        return false;
    }

    bool result = false;
    if (!simple_command (sd, "HITK STOP", "250")) {
        goto error;
    }

    char command[CMDSIZE];
    snprintf (command, CMDSIZE, "CHAN %d", channel);
    if (!simple_command (sd, command, "250")) {
        goto error;
    }

    result = true;
  error:
    svdrp_disconnect (sd);

    return result;
}


bool
vdr_timer_read_info (fileitem_t * fileitem)
{
    if (!strc_starts_with (fileitem->mrl, FILELIST_VDR_TIM_MRL)) {
        error (_("This is not a valid MRL of a VDR timer!"));
        return false;
    }

    int sd = svdrp_connect ();
    if (!sd) {
        return false;
    }

    recode_t *xr = recode_init (codeset_vdr, USE_SYSTEM_ENCODING);

    time_t time_beg = fileitem->vdr.time_beg;
    /// @bug: For some reason we need to add an hour here?!
    time_beg += 3600;
    time_t time_end = fileitem->vdr.time_end;
    /// @bug: For some reason we need to add an hour here?!
    time_end += 3600;

    bool found = false;
    char command[CMDSIZE];
    {
        snprintf (command, CMDSIZE, "LSTE %d at %ld",
                  fileitem->vdr.channel_id, time_beg);
        if (!send_command (sd, command)) {
            goto error;
        }

        epg_entry_t epg_entry;
        while (vdr_parse_epg_entry (sd, xr, &epg_entry) == 0) {
            if (epg_entry.id == fileitem->vdr.epg_id) {
                vdr_epg_entry_to_fileitem (fileitem, &epg_entry,
                                           true, false, true);
                found = true;
            }
        }
    }
    if (!found) {
        snprintf (command, CMDSIZE, "LSTE %d", fileitem->vdr.channel_id);
        if (!send_command (sd, command)) {
            goto error;
        }

        epg_entry_t epg_entry;
        while (vdr_parse_epg_entry (sd, xr, &epg_entry) == 0) {
            if (epg_entry.id == fileitem->vdr.epg_id) {
                vdr_epg_entry_to_fileitem (fileitem, &epg_entry,
                                           true, false, true);
                found = true;
            }
        }
    }
    if (!found) {
        char *i = _("Could not find entry in the electronic program guide!");
        fileitem_free_info_list (fileitem);
        fileitem->info_list = l_list_new ();
        l_list_append (fileitem->info_list, ho_strdup (i));
    }

  error:
    svdrp_disconnect (sd);
    recode_done (xr);

    return true;
}


bool
vdr_timer_remove (fileitem_t * fileitem)
{
    if (!strc_starts_with (fileitem->mrl, FILELIST_VDR_TIM_MRL)) {
        error (_("This is not a valid MRL of a VDR timer!"));
        return false;
    }

    int num = mrl_get_number (fileitem->mrl);
    debug ("removing VDR timer %d", num);

    int sd = svdrp_connect ();
    if (!sd) {
        return false;
    }

    bool result = false;
    char command[CMDSIZE];
    snprintf (command, CMDSIZE, "DELT %d", num);
    if (!simple_command (sd, command, "250")) {
        goto error;
    }

    result = true;
  error:
    svdrp_disconnect (sd);

    return result;
}


bool
vdr_timer_set_active (fileitem_t * fileitem, bool activate)
{
    if (!strc_starts_with (fileitem->mrl, FILELIST_VDR_TIM_MRL)) {
        error (_("This is not a valid MRL of a VDR timer!"));
        return false;
    }

    int num = mrl_get_number (fileitem->mrl);
    debug ("setting VDR timer %d %s", num, activate ? "active" : "inactive");

    int sd = svdrp_connect ();
    if (!sd) {
        return false;
    }

    bool result = false;
    char command[CMDSIZE];
    snprintf (command, CMDSIZE, "MODT %d %s", num, activate ? "on" : "off");
    if (!simple_command (sd, command, "250")) {
        goto error;
    }

    result = true;
  error:
    svdrp_disconnect (sd);

    return result;
}


bool
vdr_timer_create (fileitem_t * fileitem)
{
    if (!strc_starts_with (fileitem->mrl, FILELIST_VDR_EPG_MRL)) {
        error (_("This is not a valid MRL of a VDR EPG entry!"));
        return false;
    }

    int sd = svdrp_connect ();
    if (!sd) {
        return false;
    }

    recode_t *xr = recode_init (USE_SYSTEM_ENCODING, codeset_vdr);
    char *title = recode (xr, fileitem->title);
    int active = 1;
    int priority = config_get_number ("television.vdr.timer.priority");
    int lifetime = config_get_number ("television.vdr.timer.lifetime");
    time_t time_beg = config_get_number ("television.vdr.timer.buffer_beg");
    time_beg = time_beg * 60;
    time_beg = fileitem->vdr.time_beg - time_beg;
    time_t time_end = config_get_number ("television.vdr.timer.buffer_end");
    time_end = time_end * 60;
    time_end = fileitem->vdr.time_end + time_end;

    struct tm ts;
    struct tm te;
    localtime_r (&time_beg, &ts);
    localtime_r (&time_end, &te);

    bool result = false;
    char *f = "NEWT %d:%d:%d:%02d%02d:%02d%02d:%d:%d:%s:"
        "<oxine><epgid>%d</epgid></oxine>";
    char *c = ho_strdup_printf (f, active, fileitem->vdr.channel_id,
                                ts.tm_mday, ts.tm_hour, ts.tm_min,
                                te.tm_hour, te.tm_min, priority,
                                lifetime, title, fileitem->vdr.epg_id);
    if (!simple_command (sd, c, "250")) {
        goto error;
    }

    result = true;
  error:
    svdrp_disconnect (sd);
    recode_done (xr);
    ho_free (title);

    return result;
}


bool
vdr_timers_read (filelist_t * filelist)
{
    const bool timers_all = config_get_bool ("television.show."
                                             "vfolder.timers_all");

    int sd = svdrp_connect ();
    if (!sd) {
        ho_free (filelist->error);
        filelist->error = ho_strdup (_("Connection failure..."));
        return false;
    }

    bool result = false;
    recode_t *xr = recode_init (codeset_vdr, USE_SYSTEM_ENCODING);

    if (!send_command (sd, "LSTT")) {
        goto error;
    }

    epg_entry_t epg_entry;
    int r = 0;
    while ((r == 0) && (r = vdr_parse_timer_entry (sd, xr, &epg_entry)) >= 0) {
        if (!epg_entry.is_active && !timers_all) {
            ho_free (epg_entry.title);
            ho_free (epg_entry.description_short);
            ho_free (epg_entry.description_long);
        }
        else {
            char *mrl = mrl_build ("vdr", "/timers", epg_entry.timer_id);
            fileitem_t *fileitem = filelist_add (filelist, epg_entry.title,
                                                 mrl, FILE_TYPE_REGULAR);
            vdr_epg_entry_to_fileitem (fileitem, &epg_entry, true, true, false);
            ho_free (mrl);
        }
    }
    if (r >= 0) {
        filelist_sort (filelist, vdr_tim_swap_cb);
        result = true;
    }
  error:
    svdrp_disconnect (sd);
    recode_done (xr);

    return result;
}


void
vdr_register_config (void)
{
    int num_mrls = 0;
    int num_plugins = 0;
    char *vdr_plugins[3];

    if (xine_get_autoplay_mrls (oxine->xine, "xvdr", &num_mrls)) {
        vdr_plugins[num_plugins++] = "xvdr - vdr-plugin-xineliboutput";
    }
    if (xine_get_autoplay_mrls (oxine->xine, "vdr", &num_mrls)) {
        vdr_plugins[num_plugins++] = "vdr - xine-plugin";
    }
    {
        vdr_plugins[num_plugins++] = _("try all available plugins");
        vdr_plugins[num_plugins++] = NULL;
    }

    config_register_enum ("television.vdr.plugin", 0, vdr_plugins,
                          _("input plugin used for playing VDR"));
    config_register_number ("television.vdr.svdrp.port", 2001,
                            _("port used for connecting to VDR"));
    config_register_string ("television.vdr.svdrp.host", "localhost",
                            _("host used for connecting to VDR"));
    config_register_bool ("television.show.vfolder.recordings", true,
                          _("show television recordings"));
    config_register_bool ("television.show.vfolder.timers", true,
                          _("show television timers"));
    config_register_bool ("television.show.vfolder.channels", true,
                          _("show television channels"));
    config_register_bool ("television.show.vfolder.timers_all", false,
                          _("also show deactivated television timers"));

    config_register_number ("television.vdr.timer.priority", 99,
                            _("VDR timer priority"));
    config_register_number ("television.vdr.timer.lifetime", 99,
                            _("VDR timer lifetime"));
    config_register_number ("television.vdr.timer.buffer_beg", 15,
                            _("VDR timer pre start time buffer (min)"));
    config_register_number ("television.vdr.timer.buffer_end", 15,
                            _("VDR timer post end time buffer (min)"));

    config_register_bool ("television.vdr.charset.override", false,
                          _("override VDR charset"));
    config_register_string ("television.vdr.charset", get_system_encoding (),
                            _("charset for VDR"));
}


bool
vdr_play (void)
{
    const char *id = config_get_enum ("television.vdr.plugin");

    if (false) {
        /* do nothing */
    }
    else if (strc_starts_with (id, "xvdr")
             && play_autoplay (_("Television"), "XVDR")) {
        return true;
    }
    else if (strc_starts_with (id, "vdr")
             && play_autoplay (_("Television"), "VDR")) {
        return true;
    }
    else if (play_autoplay (_("Television"), "XVDR")) {
        return true;
    }
    else if (play_autoplay (_("Television"), "VDR")) {
        return true;
    }
    else {
        return false;
    }
}


void
vdr_static_free (void)
{
    recording_cache_entry_t *ce = NULL;

    int i = 0;
    for (i = 0; i < num_cache_entries; i++) {
        ce = &cache_entries[i];
        ho_free (ce->title);
        ho_free (ce->epg_entry.title);
        ho_free (ce->epg_entry.description_short);
        ho_free (ce->epg_entry.description_long);
    }

    ho_free (cache_entries);
    ho_free (codeset_vdr);
}


bool
vdr_append_disc_stat (l_list_t * info_list)
{
    const char *f0 =
        "<tr>"
        "<td width=150 align=right>%s</td>"
        "<td width=10></td>"
        "<td width=120 align=right>%.2f GB</td>"
        "</tr>";
    const char *f1 =
        "<tr>"
        "<td width=150 align=right>%s</td>"
        "<td width=10></td>"
        "<td width=120 align=right>%.2f GB</td>"
        "<td width=10></td>"
        "<td>(%d %%)</td>"
        "</tr>";

    int sd = svdrp_connect ();
    if (!sd) {
        return false;
    }

    bool result = false;
    if (!send_command (sd, "STAT DISK")) {
        goto error;
    }
    char buf[BUFSIZE];
    if (!recv_answer (sd, buf, BUFSIZE, "250")) {
        goto error;
    }
    int total;
    int free;
    int used;
    int percent;
    if (sscanf (buf, "250 %dMB %dMB %d%%", &total, &free, &percent) != 3) {
        goto error;
    }
    used = total - free;
    l_list_append (info_list, ho_strdup ("*--*"));
    l_list_append (info_list,
                   ho_strdup_printf (f0, _("Disc size:"),
                                     (double) total / 1024.0));
    percent = 100 - percent;
    l_list_append (info_list,
                   ho_strdup_printf (f1, _("Free disc:"),
                                     (double) free / 1024.0, percent));
    percent = 100 - percent;
    l_list_append (info_list,
                   ho_strdup_printf (f1, _("Used disc:"),
                                     (double) used / 1024.0, percent));
    result = true;
  error:
    svdrp_disconnect (sd);
    return result;
}


#endif /* HAVE_VDR */
