
/*
 * Copyright (C) 2002-2003,2007 the xine project
 *
 * This file is part of xine, a free video player.
 * This file is part of oxine a free media player.
 *
 * The xine-lib XML parser is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * The xine-lib XML parser is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with the Gnome Library; see the file COPYING.LIB.  If not,
 * write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * $Id: xmlparser.c 3868 2009-04-30 19:21:37Z mschwerin $
 *
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <ctype.h>

#define lprintf(...)

#include "config.h"
#include "heap.h"
#include "logger.h"
#include "utils.h"
#include "xmllexer.h"
#include "xmlparser.h"


#define TOKEN_SIZE              4 * 1024
#define DATA_SIZE               4 * 1024
#define MAX_RECURSION           10

typedef enum {
    PARSER_STATE_IDLE       = 0,
    PARSER_STATE_NODE_OPEN  = 1,
    PARSER_STATE_NODE       = 2,
    PARSER_STATE_NODE_CLOSE = 3,
    PARSER_STATE_FINAL      = 4,
    PARSER_STATE_PROP_NAME  = 5,
    PARSER_STATE_PROP_VALUE = 6,
    PARSER_STATE_COMMENT    = 7,
    PARSER_STATE_TI         = 8,
    PARSER_STATE_DOCTYPE    = 9,
    PARSER_STATE_CDATA      = 10,
} parser_state_t;

/* private global variables */
static int xml_parser_mode;

/* private functions */

static char *
strtoupper (char *str)
{
    int i = 0;

    while (str[i] != '\0') {
        str[i] = (char) toupper ((int) str[i]);
        i++;
    }
    return str;
}

static xml_node_t *
new_xml_node (void)
{
    xml_node_t *new_node;

    new_node = (xml_node_t *) ho_malloc (sizeof (xml_node_t));
    new_node->name = NULL;
    new_node->data = NULL;
    new_node->cdata = NULL;
    new_node->props = NULL;
    new_node->child = NULL;
    new_node->next = NULL;
    return new_node;
}

static void
free_xml_node (xml_node_t * node)
{
    ho_free (node->name);
    ho_free (node->data);
    ho_free (node->cdata);
    ho_free (node);
}

static xml_property_t *
new_xml_property (void)
{
    xml_property_t *new_property;

    new_property = (xml_property_t *) ho_malloc (sizeof (xml_property_t));
    new_property->name = NULL;
    new_property->value = NULL;
    new_property->next = NULL;
    return new_property;
}

static void
free_xml_property (xml_property_t * property)
{
    ho_free (property->name);
    ho_free (property->value);
    ho_free (property);
}

void
xml_parser_init (const char *buf, int size, xml_parser_mode_t mode)
{
    lexer_init (buf, size);
    xml_parser_mode = mode;
}

static void
xml_parser_free_props (xml_property_t * current_property)
{
    if (current_property) {
        if (!current_property->next) {
            free_xml_property (current_property);
        }
        else {
            xml_parser_free_props (current_property->next);
            free_xml_property (current_property);
        }
    }
}

static void
xml_parser_free_tree_rec (xml_node_t * current_node, int free_next)
{
    lprintf ("xml_parser_free_tree_rec: %s\n", current_node->name);

    if (current_node) {
        /* properties */
        if (current_node->props) {
            xml_parser_free_props (current_node->props);
        }

        /* child nodes */
        if (current_node->child) {
            lprintf ("xml_parser_free_tree_rec: child\n");
            xml_parser_free_tree_rec (current_node->child, 1);
        }

        /* next nodes */
        if (free_next) {
            xml_node_t *next_node = current_node->next;
            xml_node_t *next_next_node;

            while (next_node) {
                next_next_node = next_node->next;
                lprintf ("xml_parser_free_tree_rec: next\n");
                xml_parser_free_tree_rec (next_node, 0);
                next_node = next_next_node;
            }
        }

        free_xml_node (current_node);
    }
}

void
xml_parser_free_tree (xml_node_t * current_node)
{
    lprintf ("xml_parser_free_tree\n");
    xml_parser_free_tree_rec (current_node, 1);
}

static int xml_parser_get_node (xml_node_t * current_node,
                                char *root_name, int rec);

static int
_xml_parser_get_node (char **token_buffer, int *token_buffer_size,
                      char **pname_buffer, int *pname_buffer_size,
                      char **nname_buffer, int *nname_buffer_size,
                      xml_node_t * current_node, char *root_name, int rec)
{
    char *tok = *token_buffer;
    char *property_name = *pname_buffer;
    char *node_name = *nname_buffer;
    parser_state_t state = PARSER_STATE_IDLE;
    int lexer_res = 0;
    int parse_res = 0;
    int bypass_get_token = 0;
    xml_node_t *subtree = NULL;
    xml_node_t *current_subtree = NULL;
    xml_property_t *current_property = NULL;
    xml_property_t *properties = NULL;

    if (rec >= MAX_RECURSION) {
        error ("The maximum recursion depth "
               "%d has been reached.", MAX_RECURSION);
        return -1;
    }

    while (1) {
        if (!bypass_get_token) {
            lexer_res = lexer_get_token_d (token_buffer,
                                           token_buffer_size, 0);
            if (lexer_res == T_ERROR) {
                error ("Lexer error.");
                return -1;
            }
            tok = *token_buffer;
        }

        bypass_get_token = 0;
        lprintf ("info: %d - %d : '%s'\n", state, lexer_res, tok);

        switch (state) {
        case PARSER_STATE_IDLE:
            switch (lexer_res) {
            case (T_EOL):
            case (T_SEPAR):
                /* do nothing */
                break;
            case (T_EOF):
                /* normal end */
                return 0;
                break;
            case (T_M_START_1):
                state = PARSER_STATE_NODE_OPEN;
                break;
            case (T_M_START_2):
                state = PARSER_STATE_NODE_CLOSE;
                break;
            case (T_C_START):
                state = PARSER_STATE_COMMENT;
                break;
            case (T_TI_START):
                state = PARSER_STATE_TI;
                break;
            case (T_DOCTYPE_START):
                state = PARSER_STATE_DOCTYPE;
                break;
            case (T_CDATA_START):
                state = PARSER_STATE_CDATA;
                break;
            case (T_DATA):
                /* current data */
                ho_free (current_node->data);
                current_node->data = lexer_decode_entities (tok);
                lprintf ("info: node data : %s\n", current_node->data);
                state = PARSER_STATE_IDLE;
                break;
            default:
                error ("Found unexpected token '%s' (state %d).", tok, state);
                return -1;
                break;
            }
            break;

        case PARSER_STATE_NODE_OPEN:
            switch (lexer_res) {
            case (T_IDENT):
                properties = NULL;
                current_property = NULL;

                /* save node name */
                if (xml_parser_mode == XML_PARSER_CASE_INSENSITIVE) {
                    strtoupper (tok);
                }
                /* make sure the buffer for the node name is big enough */
                if (*token_buffer_size > *nname_buffer_size) {
                    *nname_buffer_size = *token_buffer_size;
                    *nname_buffer = ho_realloc (*nname_buffer,
                                                *nname_buffer_size);
                    node_name = *nname_buffer;
                }
                strcpy (node_name, tok);
                state = PARSER_STATE_NODE;
                lprintf ("info: current node name \"%s\"\n", node_name);
                break;
            default:
                error ("Found unexpected token '%s' (state %d).", tok, state);
                return -1;
                break;
            }
            break;

        case PARSER_STATE_NODE:
            switch (lexer_res) {
            case (T_EOL):
            case (T_SEPAR):
                /* nothing */
                break;
            case (T_M_STOP_1):
                /* new subtree */
                subtree = new_xml_node ();

                /* set node name */
                subtree->name = ho_strdup (node_name);

                /* set node propertys */
                subtree->props = properties;
                lprintf ("info: rec %d new subtree %s\n", rec, node_name);
                parse_res = xml_parser_get_node (subtree, node_name, rec + 1);
                if (parse_res != 0) {
                    return parse_res;
                }
                if (current_subtree == NULL) {
                    current_node->child = subtree;
                    current_subtree = subtree;
                }
                else {
                    current_subtree->next = subtree;
                    current_subtree = subtree;
                }
                state = PARSER_STATE_IDLE;
                break;
            case (T_M_STOP_2):
                /* new leaf */
                /* new subtree */
                subtree = new_xml_node ();

                /* set node name */
                subtree->name = ho_strdup (node_name);

                /* set node propertys */
                subtree->props = properties;

                lprintf ("info: rec %d new subtree %s\n", rec, node_name);

                if (current_subtree == NULL) {
                    current_node->child = subtree;
                    current_subtree = subtree;
                }
                else {
                    current_subtree->next = subtree;
                    current_subtree = subtree;
                }
                state = PARSER_STATE_IDLE;
                break;
            case (T_IDENT):
                /* save property name */
                if (xml_parser_mode == XML_PARSER_CASE_INSENSITIVE) {
                    strtoupper (tok);
                }
                /* make sure the buffer for the property name is big enough */
                if (*token_buffer_size > *pname_buffer_size) {
                    *pname_buffer_size = *token_buffer_size;
                    *pname_buffer = ho_realloc (*pname_buffer,
                                                *pname_buffer_size);
                    property_name = *pname_buffer;
                }
                strcpy (property_name, tok);
                state = PARSER_STATE_PROP_NAME;
                lprintf ("info: current property name \"%s\"\n",
                         property_name);
                break;
            default:
                error ("Found unexpected token '%s' (state %d).", tok, state);
                return -1;
                break;
            }
            break;

        case PARSER_STATE_NODE_CLOSE:
            switch (lexer_res) {
            case (T_IDENT):
                /* must be equal to root_name */
                if (xml_parser_mode == XML_PARSER_CASE_INSENSITIVE) {
                    strtoupper (tok);
                }
                if (strcmp (tok, root_name) == 0) {
                    state = PARSER_STATE_FINAL;
                }
                else {
                    error ("Found unexpected token '%s'. "
                           "Expected token '%s' (state %d).",
                           tok, root_name, state);
                    return -1;
                }
                break;
            default:
                error ("Found unexpected token '%s' (state %d).", tok, state);
                return -1;
                break;
            }
            break;

        case PARSER_STATE_FINAL:
            /* > expected */
            switch (lexer_res) {
            case (T_M_STOP_1):
                return 0;
                break;
            default:
                error ("Found unexpected token '%s' (state %d).", tok, state);
                return -1;
                break;
            }
            break;

        case PARSER_STATE_PROP_NAME:
            /* = or > or ident or separator expected */
            switch (lexer_res) {
            case (T_EOL):
            case (T_SEPAR):
                /* do nothing */
                break;
            case (T_EQUAL):
                state = PARSER_STATE_PROP_VALUE;
                break;
            case (T_IDENT):
                /* jump to state 2 without fetching a new token */
                bypass_get_token = 1;
                state = PARSER_STATE_NODE;
                break;
            case (T_M_STOP_1):
                /* add a new property without value */
                if (current_property == NULL) {
                    properties = new_xml_property ();
                    current_property = properties;
                }
                else {
                    current_property->next = new_xml_property ();
                    current_property = current_property->next;
                }
                current_property->name = ho_strdup (property_name);
                lprintf ("info: new property %s\n", current_property->name);
                /* jump to state 2 without fetching a new token */
                bypass_get_token = 1;
                state = PARSER_STATE_NODE;
                break;
            default:
                error ("Found unexpected token '%s' (state %d).", tok, state);
                return -1;
                break;
            }
            break;

        case PARSER_STATE_PROP_VALUE:
            /* string, ident or separator expected */
            switch (lexer_res) {
            case (T_EOL):
            case (T_SEPAR):
                /* do nothing */
                break;
            case (T_STRING):
            case (T_IDENT):
                /* add a new property */
                if (current_property == NULL) {
                    properties = new_xml_property ();
                    current_property = properties;
                }
                else {
                    current_property->next = new_xml_property ();
                    current_property = current_property->next;
                }
                current_property->name = ho_strdup (property_name);
                current_property->value = lexer_decode_entities (tok);
                lprintf ("info: new property %s=%s\n",
                         current_property->name, current_property->value);
                state = PARSER_STATE_NODE;
                break;
            default:
                error ("Found unexpected token '%s' (state %d).", tok, state);
                return -1;
                break;
            }
            break;

        case PARSER_STATE_COMMENT:
            /* --> expected */
            switch (lexer_res) {
            case (T_C_STOP):
                state = PARSER_STATE_IDLE;
                break;
            default:
                state = PARSER_STATE_COMMENT;
                break;
            }
            break;

        case PARSER_STATE_TI:
            /* ?> expected */
            switch (lexer_res) {
            case (T_TI_STOP):
                state = PARSER_STATE_IDLE;
                break;
            default:
                state = PARSER_STATE_TI;
                break;
            }
            break;

        case PARSER_STATE_DOCTYPE:
            /* > expected */
            switch (lexer_res) {
            case (T_DOCTYPE_STOP):
            case (T_M_STOP_1):
                state = PARSER_STATE_IDLE;
                break;
            default:
                state = PARSER_STATE_DOCTYPE;
                break;
            }
            break;

        case PARSER_STATE_CDATA:
            /* ]]> expected */
            switch (lexer_res) {
            case (T_CDATA_STOP):
                /* current cdata */
                ho_free (current_node->cdata);
                current_node->cdata = ho_strdup (tok);
                lprintf ("info: node cdata : %s\n", current_node->cdata);
                state = PARSER_STATE_IDLE;
                break;
            default:
                state = PARSER_STATE_CDATA;
                break;
            }
            break;

        default:
            error ("We are in an unknown parser state (state=%d).", state);
            return -1;
        }
    }
}


static int
xml_parser_get_node (xml_node_t * current_node, char *root_name, int rec)
{
    int res = 0;
    int token_buffer_size = TOKEN_SIZE;
    int pname_buffer_size = TOKEN_SIZE;
    int nname_buffer_size = TOKEN_SIZE;
    char *token_buffer = ho_malloc (token_buffer_size);
    char *pname_buffer = ho_malloc (pname_buffer_size);
    char *nname_buffer = ho_malloc (nname_buffer_size);

    res = _xml_parser_get_node (&token_buffer, &token_buffer_size,
                                &pname_buffer, &pname_buffer_size,
                                &nname_buffer, &nname_buffer_size,
                                current_node, root_name, rec);

    ho_free (token_buffer);
    ho_free (pname_buffer);
    ho_free (nname_buffer);

    return res;
}


int
xml_parser_build_tree (xml_node_t ** root_node)
{
    xml_node_t *tmp_node;
    int res;

    tmp_node = new_xml_node ();
    res = xml_parser_get_node (tmp_node, "", 0);
    if ((tmp_node->child) && (!tmp_node->child->next)) {
        *root_node = tmp_node->child;
        free_xml_node (tmp_node);
        res = 0;
    }
    else {
        *root_node = NULL;
        xml_parser_free_tree (tmp_node);
        res = -1;
    }

    return res;
}

const char *
xml_parser_get_property (const xml_node_t * node, const char *name)
{
    xml_property_t *prop;

    prop = node->props;
    while (prop) {
        lprintf ("looking for %s in %s\n", name, prop->name);

        if (!strcasecmp (prop->name, name)) {
            lprintf ("found it. value=%s\n", prop->value);
            return prop->value;
        }

        prop = prop->next;
    }

    return NULL;
}

int
xml_parser_get_property_int (const xml_node_t * node, const char *name,
                             int def_value)
{

    const char *v;
    int ret;

    v = xml_parser_get_property (node, name);

    if (!v)
        return def_value;

    if (sscanf (v, "%d", &ret) != 1)
        return def_value;
    else
        return ret;
}

int
xml_parser_get_property_bool (const xml_node_t * node, const char *name,
                              int def_value)
{

    const char *v;

    v = xml_parser_get_property (node, name);

    if (!v)
        return def_value;

    return !strcasecmp (v, "true");
}

static int
xml_escape_string_internal (char *buf, const char *s,
                            xml_escape_quote_t quote_type)
{
    int c;
    int length = 0;
    int sl = buf ? 8 : 0;
    /* calculate max required buffer size */
    while ((c = *s++ & 0xFF))
        switch (c) {
        case '"':
            if (quote_type != XML_ESCAPE_DOUBLE_QUOTE)
                goto literal;
            length += snprintf (buf + length, sl, "&quot;");
            break;
        case '\'':
            if (quote_type != XML_ESCAPE_SINGLE_QUOTE)
                goto literal;
            length += snprintf (buf + length, sl, "&apos;");
            break;
        case '&':
            length += snprintf (buf + length, sl, "&amp;");
            break;
        case '<':
            length += snprintf (buf + length, sl, "&lt;");
            break;
        case '>':
            length += snprintf (buf + length, sl, "&gt;");
            break;
        case 127:
            length += snprintf (buf + length, sl, "&#127;");
            break;
        case '\t':
        case '\n':
          literal:if (buf)
                buf[length] = c;
            ++length;
            break;
        default:
            if (c >= ' ')
                goto literal;
            length += snprintf (buf + length, sl, "&#%d;", c);
            break;
        }
    if (buf)
        buf[length] = 0;
    return length + 1;
}

char *
xml_escape_string (const char *s, xml_escape_quote_t quote_type)
{
    char *buf = ho_malloc (xml_escape_string_internal (NULL, s, quote_type));
    return buf ? (xml_escape_string_internal (buf, s, quote_type), buf) : NULL;
}

static void
xml_parser_dump_node (const xml_node_t * node, int indent)
{
    xml_property_t *p;
    xml_node_t *n;
    size_t l = strlen (node->name);

    printf ("%*s<%s", indent, "", node->name);

    p = node->props;
    while (p) {
        char *value = xml_escape_string (p->value, XML_ESCAPE_SINGLE_QUOTE);
        printf (" %s='%s'", p->name, value);
        ho_free (value);
        p = p->next;
        if (p) {
            printf ("\n%*s", indent + 1 + l, "");
        }
    }

    if (node->child) {
        printf (">\n");
        n = node->child;
        while (n) {
            xml_parser_dump_node (n, indent + 2);
            n = n->next;
        }
        printf ("%*s</%s>\n", indent, "", node->name);
    }
    else if (node->cdata) {
        printf ("><![CDATA[");
        int i = 0;
        while (node->cdata[i]) {
            if (node->cdata[i] == '\r') {
                putchar('\n');
                i++;
            }
            else {
                putchar(node->cdata[i++]);
            }
        }
        printf ("]]></%s>\n", node->name);
    }
    else if (node->data) {
        char *data = xml_escape_string (node->data, XML_ESCAPE_NO_QUOTE);
        printf (">");
        printf ("%s", data);
        printf ("</%s>\n", node->name);
        ho_free (data);
    }
    else {
        printf (" />\n");
    }
}

void
xml_parser_dump_tree (const xml_node_t * node)
{
    xml_parser_dump_node (node, 0);
}


xml_node_t *
xml_parser_get_node_by_name (xml_node_t * current_node, const char *nodename)
{
    xml_node_t *node = current_node;
    while (node) {
        if (strcasecmp (node->name, nodename) == 0) {
            return node;
        }
        node = node->next;
    }

    return NULL;
}


#ifdef XMLPARSER_TEST

oxine_t *oxine;

int
main (int argc, char **argv)
{
    oxine = ho_new (oxine_t);

    int size = 0;
    char *buffer = read_entire_file (argv[1], &size);

    /* Parse the XML answer to our query. */
    xml_node_t *n_root;
    xml_parser_init (buffer, size, XML_PARSER_CASE_SENSITIVE);
    if (xml_parser_build_tree (&n_root) != XML_PARSER_OK) {
        error (_("Parsing '%s' failed!"), argv[1]);
        goto buffer_free;
    }

    xml_parser_dump_tree (n_root);
    xml_parser_free_tree (n_root);
  buffer_free:
    ho_free (buffer);
    ho_free (oxine);

    return 0;
}

void
_odk_oxine_event_send (odk_t * a, oxine_event_t * b, const char *f, int l)
{
}

#endif
